using System.Collections.Generic;
using Ato.CD.Inbound.Shared;
using VaTS;
using Ato.EN.IntegrationServices.CodeGenerationCTR;
using Ato.EN.IntegrationServices.CodeGenerationIEE;

using DataContracts;

namespace Ato.CD.Inbound.CTR202102
{
    internal class CrossFormValidatorIEE : ICrossFormValidator
    {
        private CTR2021 ParentDocument { get; }

        private IEE2018 ChildDocument { get; }

        internal CrossFormValidatorIEE(CTR2021 report, BusinessDocument childDocument)
        {
            ParentDocument = report;
           ChildDocument = (IEE2018)childDocument.ConsumedReport;
        }

        public IEnumerable<ProcessMessageDocument> ValidateCrossFormRules()
        {
            var response = new List<ProcessMessageDocument>();
            VRATOCTR428234(response);
            VRATOCTR428240(response);
            VRATOGEN410115(response);
            VRATOGEN410119(response);
            VRATOGEN402010(response);
            VRATOIEE410051(response);
            VRATOIEE410070(response);
            VRATOIEE410226(response);
            VRATOIEE410223(response);
            VRATOIEE410225(response);

            return response;
        }
      #region  CTR-IEE

        #region VR.ATO.CTR.428234

         /*           
                    IF (RP:entity.identifier.TFN <> NULLORBLANK) AND (COUNT(SCHEDULE = "IEE") > 0) AND (RP:entity.identifier.TFN = ANY OCCURRENCE OF ([IEE50]))
                       RETURN VALIDATION MESSAGE
                    ENDIF                    
                    [IEE50] = IEE:RP:pyid.xx.xx:Identifiers.TaxFileNumber.Identifier IN TUPLE(iee.0003.lodge.req.xx.xx:FamilyTrust)
        */

        protected void VRATOCTR428234(List<ProcessMessageDocument> response)
        {
            ProcessMessageDocument processMessage;
                  
            bool assertion = !string.IsNullOrWhiteSpace(ParentDocument.RPIdentifierTFN) && ParentDocument.RPIdentifierTFN == ChildDocument.IEE50;
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.CTR.428234",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Main entity cannot interpose itself",
                    LongDescription = @"An Entity cannot interpose itself, therefore the family trust tax file number quoted in Interposed Entity Election cannot be the same as the entity making the election",
                    Location = ParentDocument.RPOccurrenceIndex > 1 ? "/xbrli:xbrl/xbrli:context[" + ParentDocument.RPOccurrenceIndex.ToString()+ "]/xbrli:entity/xbrli:identifier" : $"/xbrli:xbrl/xbrli:context/xbrli:entity/xbrli:identifier",                 
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.CTR.428234" } },
                };

                processMessage.Parameters.Add(new ProcessMessageParameter { Name = "IEE50", Value = ChildDocument.IEE50.ToString() });
                response.Add(processMessage);
            }
        }
      
        #endregion // VR.ATO.CTR.428234


        #region VR.ATO.CTR.428240

        /*  VR.ATO.CTR.428240
              
                    IF (CTR:RP:pyin.xx.xx:Report.TargetFinancial.Year <> NULL) AND (COUNT(SCHEDULE = "IEE") > 0) AND (ANY OCCURRENCE OF ([IEE116] = "R") AND ([IEE82] <> CTR:RP:pyin.xx.xx:Report.TargetFinancial.Year))
                       RETURN VALIDATION MESSAGE
                    ENDIF
                    
                    [IEE82] = IEE:RP:rvctc3.xx.xx:Elections.Revocation.Year
                    [IEE116] = IEE:RP:rvctc3.xx.xx:Elections.InterposedEntityElectionOrRevocation.Code
         */
      protected void VRATOCTR428240(List<ProcessMessageDocument> response)
        
            {
            ProcessMessageDocument processMessage;
            
       

            bool assertion = (ParentDocument.CTR2.HasValue  && !string.IsNullOrWhiteSpace(ChildDocument.IEE116) && ChildDocument.IEE116.ToUpper() == "R" && ChildDocument.IEE82.HasValue && ChildDocument.IEE82.GetValueOrDefault() != ParentDocument.CTR2.GetValueOrDefault());
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.CTR.428240",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Revocation date of the Interposed Entity Election is invalid",
                    LongDescription = @"The income year for which the details of the Interposed Entity Election are being revoked must equal the income year of the main return being lodged",
                    Location = "/xbrli:xbrl/tns:Report.TargetFinancial.Year",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.CTR.428240" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter { Name = "CTR2", Value = ParentDocument.CTR2.GetValueOrDefault().ToString() });
                processMessage.Parameters.Add(new ProcessMessageParameter { Name = "IEE116", Value = ChildDocument.IEE116.ToString() });
                processMessage.Parameters.Add(new ProcessMessageParameter { Name = "IEE82", Value = ChildDocument.IEE82.ToString() });
                response.Add(processMessage);
              
            }
        }

        #endregion // VR.ATO.CTR.428240
      #endregion
      #region  IEE-CTR
        #region VR.ATO.GEN.410115

            /*  VR.ATO.GEN.410115
                    WHERE PARENT RETURN EXISTS 
                    IF (IEE:RP:entity.identifier.TFN <> NULLORBLANK) AND (IEE:RP:entity.identifier.TFN <> PARENT RETURN:RP:entity.identifier.TFN)
                           RETURN VALIDATION MESSAGE
                    ENDIF
          */
        protected void VRATOGEN410115(List<ProcessMessageDocument> response)
        {
            ProcessMessageDocument processMessage;
            bool assertion = !string.IsNullOrWhiteSpace(ChildDocument.RPIdentifierTFN) && 
                              ChildDocument.RPIdentifierTFN != ParentDocument.RPIdentifierTFN;

            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.410115",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"TFN on Schedule must match TFN on main form",                    
                    Location = ChildDocument.RPOccurrenceIndex > 1 ? "/xbrli:xbrl/xbrli:context[" + ChildDocument.RPOccurrenceIndex.ToString() + "]/xbrli:entity/xbrli:identifier" : $"/xbrli:xbrl/xbrli:context/xbrli:entity/xbrli:identifier",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.410115" } },
                };
                response.Add(processMessage);
            }
        }
        #endregion // VR.ATO.GEN.410115
        #region VR.ATO.GEN.410119

        /*  VR.ATO.GEN.410119
                    WHERE PARENT RETURN EXISTS
                    IF (IEE:INT:entity.identifier.ABN <> NULLORBLANK) AND (IEE:INT:entity.identifier.ABN <> PARENT RETURN:INT:entity.identifier.ABN)
                          RETURN VALIDATION MESSAGE
                    ENDIF
                  */
        protected void VRATOGEN410119(List<ProcessMessageDocument> response)
        {
            ProcessMessageDocument processMessage;
            bool assertion = !string.IsNullOrWhiteSpace(ChildDocument.INTIdentifierABN) && 
                             !ChildDocument.INTIdentifierABN.Equals(ParentDocument.INTIdentifierABN);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.410119",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"ABN of the Intermediary on the schedule must match the ABN on the main form",
                   
                    Location = ChildDocument.INTOccurrenceIndex > 1 ? "/xbrli:xbrl/xbrli:context[" + ChildDocument.INTOccurrenceIndex.ToString() + "]/xbrli:entity/xbrli:identifier" : $"/xbrli:xbrl/xbrli:context/xbrli:entity/xbrli:identifier",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.410119" } },
                };
                response.Add(processMessage);
            }
        }
        #endregion // VR.ATO.GEN.410119

        #region VR.ATO.GEN.402010

        /*  VR.ATO.GEN.402010
        
                    IF (RP:pyid.xx.xx:Identifiers.AustralianBusinessNumber.Identifier <> NULLORBLANK) AND 
                    (PARENT RETURN:RP:pyid.xx.xx:Identifiers.AustralianBusinessNumber.Identifier <> NULLORBLANK) AND 
                    (RP:pyid.xx.xx:Identifiers.AustralianBusinessNumber.Identifier <> PARENT RETURN:RP:pyid.xx.xx:Identifiers.AustralianBusinessNumber.Identifier)
                    RETURN VALIDATION MESSAGE
                    ENDIF
                    */
        protected void VRATOGEN402010(List<ProcessMessageDocument> response)
        {
            ProcessMessageDocument processMessage;
               bool  assertion = !string.IsNullOrWhiteSpace(ChildDocument.IEE162) &&
               !ChildDocument.IEE162.Equals(ParentDocument.CTR11);        
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.402010",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Your supplied ABN does not match the ABN supplied on the form it was submitted with",
                    LongDescription = @"Your supplied ABN does not match the ABN supplied on the parent form it was submitted with",
                    Location = "/xbrli:xbrl/tns:Identifiers.AustralianBusinessNumber.Identifier",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.402010" } },
                };  
                response.Add(processMessage);
            }
        }
        #endregion // VR.ATO.GEN.402010

        #region VR.ATO.IEE.410051

        /*  VR.ATO.IEE.410051            
                    WHERE PARENT RETURN EXISTS
                    IF ([IEE49] <> NULL) AND ([IEE45] <> PARENT RETURN:Report.TargetFinancial.Year) 
                        RETURN VALIDATION MESSAGE
                    ENDIF
                    [IEE45] = IEE:RP:rvctc3.02.00:Elections.InterposedEntityElectionStatus.Year
                    [IEE49] = IEE:RP:rvctc3.02.00:Elections.Commencement.Date
                   */
        protected void VRATOIEE410051(List<ProcessMessageDocument> response)
        {
            ProcessMessageDocument processMessage;  
            bool assertion = ChildDocument.IEE49.HasValue && ChildDocument.IEE45 != ParentDocument.CTR2;
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.IEE.410235",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Election Commencement Time must not be present if Income Year Specified is different to the parent return target financial year",
                    Location = "/xbrli:xbrl/tns:Elections.Commencement.Date",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.IEE.410051" } },
                };
                response.Add(processMessage);
            }
        }
        #endregion // VR.ATO.IEE.410051

        #region VR.ATO.IEE.410070

        /*  VR.ATO.IEE.410070   
                    WHERE PARENT RETURN EXISTS
                    IF ([IEE116] = "R") AND ([IEE82] <> PARENT RETURN:Report.TargetFinancial.Year) 
                        RETURN VALIDATION MESSAGE
                    ENDIF
                    
                    [IEE82] = IEE:RP:rvctc3.02.00:Elections.Revocation.Year
                    [IEE116] = IEE:RP:rvctc3.02.02:Elections.InterposedEntityElectionOrRevocation.Code
                    */
        protected void VRATOIEE410070(List<ProcessMessageDocument> response)
        {
            ProcessMessageDocument processMessage;  
            bool assertion = ChildDocument.IEE116.ToUpper().Equals("R") && ChildDocument.IEE82.HasValue &&
                             ChildDocument.IEE82 != ParentDocument.CTR2;
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.IEE.410070",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"The income year of the Interposed Entity Revocation must equal the income year of the main return",
                    Location = "/xbrli:xbrl/tns:Elections.Revocation.Year",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.IEE.410070" } },
                };

                response.Add(processMessage);
            }
        }
        #endregion // VR.ATO.IEE.410070

        #region VR.ATO.IEE.410226

        /*  VR.ATO.IEE.410226
            
                    WHERE PARENT RETURN EXISTS
                    IF ([IEE4] <> NULLORBLANK) AND ([IEE4] <> (PARENT RETURN:RP:pyde.xx.xx:OrganisationNameDetails.OrganisationalName.Text 
                    IN TUPLE (xbrli\orgname2.xx.xx:OrganisationNameDetails) WHERE ((TUPLE EXPLICIT pyde.xx.xx:OrganisationNameDetails.OrganisationalNameType.Code = "MN") 
                    AND (TUPLE EXPLICIT pyde.xx.xx:OrganisationNameDetails.Currency.Code = "C"))))
                             RETURN VALIDATION MESSAGE
                    ENDIF              
                    */
        protected void VRATOIEE410226(List<ProcessMessageDocument> response)
        {
            ProcessMessageDocument processMessage;
         
            bool assertion = !string.IsNullOrWhiteSpace(ChildDocument.IEE4) &&
                             !ChildDocument.IEE4.Equals(ParentDocument.CTR9);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.402063",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Name on schedule does not match parent form",
                    LongDescription = @"Name on schedule does not match name on parent form.",
                    Location = ChildDocument.OrganisationNameDetailsCollection_MN_C_RPOccurrenceIndex > 1 ? "/xbrli:xbrl/tns:OrganisationNameDetails[" + ChildDocument.OrganisationNameDetailsCollection_MN_C_RPOccurrenceIndex + "]/tns:OrganisationNameDetails.OrganisationalName.Text"
                    : "/xbrli:xbrl/tns:OrganisationNameDetails/tns:OrganisationNameDetails.OrganisationalName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.IEE.410226" } },
                };

                response.Add(processMessage);
            }
        }
        #endregion // VR.ATO.IEE.410226

        #region VR.ATO.IEE.410223

        /*  VR.ATO.IEE.410223
            
                    WHERE PARENT RETURN EXISTS
                    IF ([IEE5] <> NULLORBLANK) AND ([IEE5] <> (PARENT RETURN:RP:pyde.xx.xx:AddressDetails.Line1.Text WHERE 
                    ((TUPLE ELEMENT EXPLICIT pyde.xx.xx:AddressDetails.Usage.Code = "POS") AND (TUPLE ELEMENT EXPLICIT pyde.xx.xx:AddressDetails.Currency.Code = "C"))
                    IN TUPLE(xbrli\address2.xx.xx:AddressDetails)))
                       RETURN VALIDATION MESSAGE
                    ENDIF
                  */
        protected void VRATOIEE410223(List<ProcessMessageDocument> response)
        {
            ProcessMessageDocument processMessage;
            bool assertion = !string.IsNullOrWhiteSpace(ChildDocument.IEE5) &&
                            !ChildDocument.IEE5.Equals(ParentDocument.CTR15);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.IEE.410223",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Interposed Entity Address must be blank or be the same as on main form",
                    LongDescription = @"When the IEE is lodged as a schedule, the Address of the interposed entity must be either blank or must equal the address on the main form",
                    Location = ChildDocument.AddressDetailsCollection_POS_C_RPOccurrenceIndex > 1 ? "/xbrli:xbrl/tns:AddressDetails[" + ChildDocument.AddressDetailsCollection_POS_C_RPOccurrenceIndex + "]/tns:AddressDetails.Line1.Text"
                    : "/xbrli:xbrl/tns:AddressDetails/tns:AddressDetails.Line1.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.IEE.410223" } },
                };
                response.Add(processMessage);
            }
        }
        #endregion // VR.ATO.IEE.410223

        #region VR.ATO.IEE.410225

        /*  VR.ATO.IEE.410225
                    WHERE PARENT RETURN = (“CTR”, “PTR”, “TRT”)
                    IF ([IEE10] <> NULLORBLANK) AND ([IEE10] <> (PARENT RETURN:RP:pyde.xx.xx:OrganisationNameDetails.OrganisationalName.Text) 
                    IN TUPLE (xbrli\orgname2.xx.xx:OrganisationNameDetails) WHERE ((TUPLE EXPLICIT pyde.xx.xx:OrganisationNameDetails.OrganisationalNameType.Code = "MN") 
                    AND (TUPLE EXPLICIT pyde.xx.xx:OrganisationNameDetails.Currency.Code = "P"))) 
                      RETURN VALIDATION MESSAGE
                    ENDIF
                    
            [IEE10] = IEE:RP:pyde.02.00:OrganisationNameDetails.OrganisationalName.Text WHERE (pyde.02.00:OrganisationNameDetails.OrganisationalNameType.Code = "MN" AND pyde.02.00:OrganisationNameDetails.Currency.Code = "P") IN TUPLE(iee.0003.lodge.req.02.00:OrganisationNameDetails)
         */
        protected void VRATOIEE410225(List<ProcessMessageDocument> response)
        {
            ProcessMessageDocument processMessage;
             bool assertion = !string.IsNullOrWhiteSpace(ChildDocument.IEE10) &&
                             !ChildDocument.IEE10.Equals(ParentDocument.CTR12);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.402063",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Name on schedule does not match parent form",
                    LongDescription = @"Name on schedule does not match name on parent form.",
                  
                    Location = ChildDocument.OrganisationNameDetailsCollection_MN_P_RPOccurrenceIndex > 1 ? "/xbrli:xbrl/tns:OrganisationNameDetails[" + ChildDocument.OrganisationNameDetailsCollection_MN_P_RPOccurrenceIndex + "]/tns:OrganisationNameDetails.OrganisationalName.Text"
                    : "/xbrli:xbrl/tns:OrganisationNameDetails/tns:OrganisationNameDetails.OrganisationalName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.IEE.410225" } },
                };
                response.Add(processMessage);
            }
        }
        #endregion // VR.ATO.IEE.410225

        #endregion
    }
}
