using System;
using System.Collections.Generic;
using System.Linq;
using System.Xml;
using System.Text.RegularExpressions;
using Au.Gov.Sbr.Xbrl.Document;
using DataContracts;
using Au.Gov.Sbr.Xbrl.Metadata;
using System.Runtime.CompilerServices;

namespace Ato.EN.IntegrationServices.CodeGenerationFITR
{

    public partial class FITR2021Validator
    {
        /// <summary>
        /// The max parameter name length is restricted by the SBR1 and SBR2 schemas - the lowest common denominator of SBR 1 is used as the default
        /// </summary>
        private int _maxParameterNameLength;

        /// <summary>
        /// The max parameter value length is restricted by the SBR1 and SBR2 schemas
        /// </summary>
        private int _maxParameterValueLength;

        /// <summary>
        /// The SBR1 and ebms3 schemas do not allow parameter names or values to be the empty string
        /// </summary>
        private string _emptyParameterValue;

        /// <summary>
        /// Initializes a new instance of the <see cref="FITR2021Validator" /> class.
        /// </summary>
        /// <param name="maxParameterNameLength">Maximum length of the parameter name.</param>
        /// <param name="maxParameterValueLength">Maximum length of the parameter value.</param>
        /// <param name="emptyParameterValue">This value will be used in place of any parameter values that result in a null or empty value.</param>
        public FITR2021Validator(int maxParameterNameLength = 20, int maxParameterValueLength = 4096, string emptyParameterValue = "EMPTY")
        {
            _maxParameterNameLength = maxParameterNameLength;
            _maxParameterValueLength = maxParameterValueLength;
            _emptyParameterValue = emptyParameterValue;
        }

        #region Functions
        private static IEnumerable<string> Union(IEnumerable<string> list1, IEnumerable<string> list2)
        {
            IEnumerable<string> response;

            if (list1 == null && list2 == null)
            {
                response = null;
            }
            else if (list1 == null)
            {
                response = list2.Distinct();
            }
            else if (list2 == null)
            {
                response = list1.Distinct();
            }
            else
            {
                response = list1.Union(list2);
            }

            return response;
        }

        private static bool IsMatch(int? field, string expression, RegexOptions options = RegexOptions.None)
        {
            if (field == null)
                return false;
            else
                return Regex.IsMatch(Convert.ToString(field.Value), expression, options);
        }

        private static bool IsMatch(string field, string expression, RegexOptions options = RegexOptions.None)
        {
            if (field == null)
                return false;
            else
                return Regex.IsMatch(field, expression, options);
        }

        // This is just context and tuple counts where they are integer values - easier that changing the parsing logic to just return the value
        private static int Count(int count)
        {
            return count;
        }


        private static int Count<T>(IEnumerable<T> values)
        {
            return values == null ? 0 : values.Where(f => f != null).Count();
        }


        private static int Count<T>(ICollection<T> values)
        {
            return values == null ? 0 : values.Where(f => f != null).Count();
        }


        private static bool exists(bool value)
        {
            return value;
        }

        private static string GetValueOrEmpty(bool? val)
        {
            return (val.HasValue) ? val.ToString().ToLower() : string.Empty;
        }

        private static string GetValueOrEmptyDate(DateTime? val)
        {
            return (val.HasValue) ? val.Value.ToString("yyyy-MM-dd") : string.Empty;
        }

        private static string GetValueOrEmptyDate(string val)
        {
            DateTime dateTimeOut;
            if (DateTime.TryParse(val, out dateTimeOut))
            {
                return dateTimeOut.ToString("yyyy-MM-dd");
            }
            return !string.IsNullOrWhiteSpace(val) ? val : string.Empty;
        }

        private static string GetValueOrEmptyDateTime(string val)
        {
            DateTime dateTimeOut;
            if (DateTime.TryParse(val, out dateTimeOut))
            {
                return dateTimeOut.ToString("yyyy-MM-ddTHH:mm:ssZ");
            }
            return !string.IsNullOrWhiteSpace(val) ? val : string.Empty;
        }

        private static string GetValueOrEmptyDateTime(DateTime? val)
        {
            return (val.HasValue) ? val.Value.ToString("yyyy-MM-ddTHH:mm:ssZ") : string.Empty;
        }

        private static string GetValueOrEmpty(string val)
        {
            return !string.IsNullOrWhiteSpace(val) ? val : string.Empty;
        }

        private static string GetValueOrEmpty(decimal? val)
        {
            return (val.HasValue) ? val.ToString() : string.Empty;
        }

        private static string GetValueOrEmpty(int? val)
        {
            return (val.HasValue) ? val.ToString() : string.Empty;
        }

        /// <summary>
        /// Get string value between [first] a and [last] b.
        /// </summary>
        public static string Between(string value, string a, string b)
        {
            int posA = value.IndexOf(a);
            int posB = value.LastIndexOf(b);
            if (posA == -1)
            {
                return "";
            }
            if (posB == -1)
            {
                return "";
            }
            int adjustedPosA = posA + a.Length;
            if (adjustedPosA >= posB)
            {
                return "";
            }
            return value.Substring(adjustedPosA, posB - adjustedPosA);
        }

        /// <summary>
        /// Get string value after [first] a.
        /// </summary>
        public static string Before(string value, string a)
        {
            int posA = value.IndexOf(a);
            if (posA == -1)
            {
                return "";
            }
            return value.Substring(0, posA);
        }

        /// <summary>
        /// Get string value after [last] a.
        /// </summary>
        public static string After(string value, string a)
        {
            int posA = value.LastIndexOf(a);
            if (posA == -1)
            {
                return "";
            }
            int adjustedPosA = posA + a.Length;
            if (adjustedPosA >= value.Length)
            {
                return "";
            }
            return value.Substring(adjustedPosA);
        }

        private static int Length(object field)
        {
            if (field == null)
                return 0;
            else
                return field.ToString().Trim().Length;
        }

        private static bool NotSameValues(IEnumerable<object> nodes)
        {
            if (nodes == null)
                return false;

            object[] nodesArray = nodes.Cast<object>().ToArray();
            return NotSameValues(nodesArray);
        }

        private static bool NotSameValues(params object[] nodes)
        {
            if (nodes == null)
                return false;

            return (from x in nodes select x).Distinct().Count() != 1;
        }

        private static bool HasDuplicateValues(IEnumerable<DateTime> nodes)
        {
            if (nodes == null)
                return false;

            object[] nodesArray = nodes.Cast<object>().ToArray();
            return HasDuplicateValues(nodesArray);
        }

        private static bool HasDuplicateValues(IEnumerable<DateTime?> nodes)
        {
            if (nodes == null)
                return false;

            object[] nodesArray = nodes.Cast<object>().ToArray();
            return HasDuplicateValues(nodesArray);
        }

        private static bool NotSameValues(IEnumerable<DateTime> nodes)
        {
            if (nodes == null)
                return false;

            object[] nodesArray = nodes.Cast<object>().ToArray();
            return NotSameValues(nodesArray);
        }

        private static bool NotSameValues(IEnumerable<DateTime?> nodes)
        {
            if (nodes == null)
                return false;

            object[] nodesArray = nodes.Cast<object>().ToArray();
            return NotSameValues(nodesArray);
        }

        private static bool HasDuplicateValues(IEnumerable<object> nodes)
        {
            if (nodes == null)
                return false;

            object[] nodesArray = nodes.Cast<object>().ToArray();
            return HasDuplicateValues(nodesArray);
        }

        private static bool HasDuplicateValues(params object[] nodes)
        {
            if (nodes == null)
                return false;

            nodes = nodes.Where(x => x != null).ToArray();
            return !((from x in nodes select x).Distinct().Count() == nodes.Count());

        }

        private int DuplicateValueIndex(IEnumerable<object> values)
        {
            int response = 0;
            var hashset = new HashSet<object>();
            foreach (var value in values)
            {
                if (!hashset.Add(value))
                {
                    return response;
                }
                response++;
            }
            return response;
        }

        private int DuplicateValueIndex<T>(IEnumerable<T?> values) where T : struct
        {
            int response = 0;
            var hashset = new HashSet<T?>();
            foreach (var value in values)
            {
                if (!hashset.Add(value))
                {
                    return response;
                }
                response++;
            }
            return response;
        }

        private static bool IsDate(object value)
        {
            DateTime dateValue;
            return (value != null && DateTime.TryParse(value.ToString(), out dateValue));
        }

        public DateTime AsDate(string dateAsString)
        {
            DateTime response = DateTime.MinValue;
            DateTime date;

            if (DateTime.TryParse(dateAsString, out date))
            {
                response = date;
            }

            return response;
        }

        public DateTime Date(DateTime? datetime)
        {
            return datetime.GetValueOrDefault().Date;
        }

        //The logic in After function expects "---" for day and "--" for month. 
        //Since hyphen was missing the date always returned null
        public DateTime? ConvertToDate(int day, int month, int year)
        {
            return ConvertToDate(day == 0 ? null : "---" + day.ToString(), month == 0 ? null : "--" + month.ToString(), year == 0 ? null : year.ToString());
        }

        public DateTime? ConvertToDate(string day, string month, string year)
        {
            DateTime? response;
            DateTime result;

            if (year == null || month == null || day == null)
            {
                return null;
            }
            string dateAsString = year + "-" + After(month, "--") + "-" + After(day, "---");
            if (DateTime.TryParse(dateAsString, out result))
            {
                response = result;
            }
            else
            {
                response = null;
            }
            return response;
        }


        public DateTime? ConvertToDate(string day, string month, int year)
        {
            DateTime? response;
            DateTime result;

            if (year == 0 || month == null || day == null)
            {
                return null;
            }
            string dateAsString = year.ToString() + "-" + After(month, "--") + "-" + After(day, "---");
            if (DateTime.TryParse(dateAsString, out result))
            {
                response = result;
            }
            else
            {
                response = null;
            }
            return response;
        }


        private static int Day(string dateAsString)
        {
            int response = 0;
            DateTime date;

            if (DateTime.TryParse(dateAsString, out date))
            {
                response = date.Day;
            }

            return response;
        }


        private static int Day(DateTime? date)
        {
            if (date == null)
                return 0;
            else
                return date.Value.Day;

        }


        private static string Month(string dateAsString)
        {
            string response = null;
            DateTime date;

            if (DateTime.TryParse(dateAsString, out date))
            {
                response = date.ToString("MMMM");
            }
            else
            {
                return "NotAMonth";
            }

            return response;
        }

        private static string Month(DateTime? date)
        {
            if (date == null)
                return "NotAMonth";
            else
                return date.Value.ToString("MMMM");
        }


        private static int MonthAsInt(string dateAsString)
        {
            int response = 0;
            DateTime date;

            if (DateTime.TryParse(dateAsString, out date))
            {
                response = date.Month;
            }

            return response;
        }

        private static int MonthAsInt(DateTime? date)
        {
            if (date == null)
            {
                return 0;
            }
            return date.Value.Month;
        }


        private static int Year(string dateAsString)
        {
            int response = 0;
            DateTime date;

            if (DateTime.TryParse(dateAsString, out date))
            {
                response = date.Year;
            }

            return response;
        }

        private static int Year(DateTime? date)
        {
            if (date == null)
                return 0;
            else
                return date.Value.Year;
        }


        private static int CurrentFinancialYear()
        {
            return DateToFinancialYear(DateTime.Now, 7);
        }

        private static int FinancialYear(string dateAsString)
        {
            return DateToFinancialYear(dateAsString, 7);
        }

        private static int FinancialYear(DateTime? date)
        {
            return DateToFinancialYear(date, 7);
        }

        private static int DateToFinancialYear(string dateAsString, int startingMonth)
        {
            int response = 0;
            DateTime date;
            if (DateTime.TryParse(dateAsString, out date))
            {
                response = DateToFinancialYear(date, startingMonth);
            }

            return response;
        }
        private static int DateToFinancialYear(DateTime? date, int startingMonth)
        {
            int response;
            if (date == null)
            {
                response = 0;
            }
            else
            {
                int year = date.Value.Year;
                int month = date.Value.Month;

                if (startingMonth > month)
                    response = year;
                else
                    response = year + 1;
            }
            return response;
        }


        private static int FBTYear(string dateAsString)
        {
            int response = 0;
            DateTime date;

            if (DateTime.TryParse(dateAsString, out date))
                response = FBTYear(date);
            return response;
        }

        private static int FBTYear(DateTime? date)
        {
            if (date == null)
            {
                return 0;
            }
            else
            {
                if (date.Value.Month > 3)
                    return date.Value.Year + 1;
                else
                    return date.Value.Year;
            }
        }

        private static DateTime? AddMonthsToDate(DateTime? dateTime, int months)
        {
            return dateTime == null ? null : (DateTime?)dateTime.Value.AddMonths(months);
        }

        private static bool IsNumeric(object value)
        {
            decimal numbervalue;
            return (value != null && decimal.TryParse(value.ToString(), out numbervalue));
        }

        private static bool NotMonetary(decimal? field, string sign, int digits, int decimals)
        {
            if (field == null)
            {
                return false;
            }
            else
            {
                string signExpression;
                string decimalExpression;
                int digitsToUse = digits - decimals;

                if (sign == "U")
                    signExpression = "^";
                else
                    signExpression = "^-?";

                if (decimals > 0)
                    decimalExpression = @"(\.\d{1," + decimals + "})?$";
                else
                    decimalExpression = @"$";

                return !(Regex.IsMatch(field.Value.ToString("0.#########################"), signExpression + @"\d{1," + digitsToUse + "}" + decimalExpression));
            }
        }

        private static bool NotNumeric(decimal? field, string sign, int digits, int decimals)
        {
            if (field == null)
            {
                return false;
            }
            else
            {
                string signExpression;
                string decimalExpression;
                int digitsToUse = digits - decimals;

                if (sign == "U")
                    signExpression = "^";
                else
                    signExpression = "^-?";

                if (decimals > 0)
                    decimalExpression = @"(\.\d{1," + decimals + "})?$";
                else
                    decimalExpression = @"$";

                return !(Regex.IsMatch(field.Value.ToString("0.#########################"), signExpression + @"\d{1," + digitsToUse + "}" + decimalExpression));
            }
        }

        private static bool NotNumeric(int? field, string sign, int digits, int decimals = 0)
        {
            if (field == null)
            {
                return false;
            }
            else
            {
                string signExpression;

                if (sign == "U")
                    signExpression = "^";
                else
                    signExpression = "^-?";

                return !(Regex.IsMatch(field.Value.ToString(), signExpression + @"\d{1," + digits + "}$"));
            }
        }

        private static bool NotNumeric(long? field, string sign, int digits, int decimals = 0)
        {
            if (field == null)
            {
                return false;
            }
            else
            {
                string signExpression;

                if (sign == "U")
                    signExpression = "^";
                else
                    signExpression = "^-?";

                return !(Regex.IsMatch(field.Value.ToString(), signExpression + @"\d{1," + digits + "}$"));
            }
        }


        private static bool OutsideRange(decimal field, decimal expression, int range)
        {
            bool response;

            response = (field < (expression - range)) || (field > (expression + range));

            return response;
        }


        private static bool FailsUSIAlgorithm(string usi, string abn)
        {
            bool response = true;
            if (usi == null || abn == null)
            {
                response = false;
            }
            else
            {
                usi = usi.Trim();
                abn = abn.Trim();
                if (usi.Length == 14)
                {
                    int numeric;
                    if (usi.Substring(0, 11) == abn && int.TryParse(usi.Substring(11, 3), out numeric))
                        response = false;

                }
                else if (usi.Length == 9)
                {

                    if (Regex.IsMatch(usi, @"^([a-zA-Z]{3}\d{4}[a-zA-Z]{2})$"))
                        response = false;
                }
            }
            return response;
        }

        private static bool FailsTANAlgorithm(string tan)
        {
            bool response;
            decimal decimalTan;

            if (tan == null)
                return false;

            tan = tan.Trim();

            if (!decimal.TryParse(tan, out decimalTan))
                return true;

            if (tan.Length != 8)
                return true;

            decimal tanSum =
                7 * int.Parse(tan.Substring(0, 1)) +
                9 * int.Parse(tan.Substring(1, 1)) +
                8 * int.Parse(tan.Substring(2, 1)) +
                4 * int.Parse(tan.Substring(3, 1)) +
                6 * int.Parse(tan.Substring(4, 1)) +
                3 * int.Parse(tan.Substring(5, 1)) +
                5 * int.Parse(tan.Substring(6, 1)) +
                1 * int.Parse(tan.Substring(7, 1));

            if ((tanSum % 11) == 0)
                response = false;
            else
                response = true;

            return response;
        }


        private static bool FailsABNAlgorithm(string abn)
        {
            bool response;
            decimal decimalAbn;

            if (abn == null)
                return false;

            abn = abn.Trim();

            if (!decimal.TryParse(abn, out decimalAbn))
                return true;

            if (abn.Length != 11)
                return true;

            decimal abnSum =
                10 * (int.Parse(abn.Substring(0, 1)) - 1) +
                1 * int.Parse(abn.Substring(1, 1)) +
                3 * int.Parse(abn.Substring(2, 1)) +
                5 * int.Parse(abn.Substring(3, 1)) +
                7 * int.Parse(abn.Substring(4, 1)) +
                9 * int.Parse(abn.Substring(5, 1)) +
                11 * int.Parse(abn.Substring(6, 1)) +
                13 * int.Parse(abn.Substring(7, 1)) +
                15 * int.Parse(abn.Substring(8, 1)) +
                17 * int.Parse(abn.Substring(9, 1)) +
                19 * int.Parse(abn.Substring(10, 1));

            if ((abnSum % 89) == 0)
                response = false;
            else
                response = true;

            return response;
        }
        private static bool FailsACNAlgorithm(string acn)
        {
            bool response;
            decimal decimalAbn;
            if (acn == null)
                return false;

            acn = acn.Trim();

            if (!decimal.TryParse(acn, out decimalAbn))
                return true;

            if (acn.Length != 9)
                return true;

            decimal abnSum =
                8 * int.Parse(acn.Substring(0, 1)) +
                7 * int.Parse(acn.Substring(1, 1)) +
                6 * int.Parse(acn.Substring(2, 1)) +
                5 * int.Parse(acn.Substring(3, 1)) +
                4 * int.Parse(acn.Substring(4, 1)) +
                3 * int.Parse(acn.Substring(5, 1)) +
                2 * int.Parse(acn.Substring(6, 1)) +
                1 * int.Parse(acn.Substring(7, 1));

            decimal checkDigit = int.Parse(acn.Substring(8, 1));
            decimal acnRemainder = abnSum % 10;

            if (((10 - acnRemainder) % 10) == checkDigit)
                response = false;
            else
                response = true;

            return response;
        }


        private static bool FailsTFNAlgorithm(string tfn)
        {
            bool response;
            decimal decimalTfn;

            if (tfn == null)
                return false;

            tfn = tfn.Trim();
            tfn = Regex.Replace(tfn, "^0+", "");

            if (!decimal.TryParse(tfn, out decimalTfn))
                return true;

            if (tfn.Length < 8)
                return true;


            decimal tfn1To7Sum =
                1 * int.Parse(tfn.Substring(0, 1)) +
                4 * int.Parse(tfn.Substring(1, 1)) +
                3 * int.Parse(tfn.Substring(2, 1)) +
                7 * int.Parse(tfn.Substring(3, 1)) +
                5 * int.Parse(tfn.Substring(4, 1)) +
                8 * int.Parse(tfn.Substring(5, 1)) +
                6 * int.Parse(tfn.Substring(6, 1));

            decimal tfn8 = 9 * int.Parse(tfn.Substring(7, 1));

            if (tfn.Length == 8)
            {
                decimal tFNLg8WSum9 = 10 * int.Parse(tfn.Substring(7, 1));
                decimal tFNLg8WSum = tfn1To7Sum + tFNLg8WSum9;

                if ((tFNLg8WSum % 11) == 0)
                    response = false;
                else
                    response = true;
            }
            else if (tfn.Length == 9)
            {
                decimal tfn9 = 10 * int.Parse(tfn.Substring(8, 1));
                decimal tFNLg9WSum = tfn1To7Sum + tfn8 + tfn9;

                if ((tFNLg9WSum % 11) == 0)
                    response = false;
                else
                    response = true;
            }
            else
            {
                response = true;
            }

            return response;
        }


        private static decimal ConditionalValue(bool expression, decimal? trueVal, decimal? falseVal)
        {
            return expression ? trueVal.GetValueOrDefault() : falseVal.GetValueOrDefault();
        }

        private static decimal AsNumeric(string value)
        {
            decimal numberValue;
            decimal.TryParse(value, out numberValue);
            return numberValue;
        }

        private static bool RegexMatch(int? field, string expression, string flags = "")
        {
            return IsMatch(field, expression, GetRegexOption(flags));
        }

        private static bool RegexMatch(string field, string expression, string flags = "")
        {
            return IsMatch(field, expression, GetRegexOption(flags));
        }

        private static RegexOptions GetRegexOption(string flags)
        {
            RegexOptions options = RegexOptions.None;

            char[] characters = flags.ToCharArray();

            foreach (char character in characters)
            {
                switch (character)
                {
                    case 'i':
                        options = options | RegexOptions.IgnoreCase;
                        break;
                    case 'm':
                        options = options | RegexOptions.Multiline;
                        break;
                    case 's':
                        options = options | RegexOptions.Singleline;
                        break;
                    case 'n':
                        options = options | RegexOptions.ExplicitCapture;
                        break;
                    case 'x':
                        options = options | RegexOptions.IgnorePatternWhitespace;
                        break;
                }
            }

            return options;
        }

        private static string Substring(object field, int start, int length)
        {
            return field.ToString().Substring(start, length);
        }

        /// <summary>
        /// Returns an occurrence index as [occurrenceIndex] of occurrenceIndex > 0, otherwise the empty string
        /// </summary>
        /// <param name="occurrenceIndex">Index of the occurrence.</param>
        /// <returns>Occurrence in XPath [#] format</returns>
        public string OccurrenceIndex(int occurrenceIndex)
        {
            return occurrenceIndex > 0 ? "[" + occurrenceIndex + "]" : "";
        }

        #endregion // Functions

        public FITR2021 ConsumedReport { get; private set; }

        private static Dictionary<string, ProcessMessageDocument> _processMessageDocuments = new Dictionary<string, ProcessMessageDocument>();

        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        public List<ProcessMessageDocument> ValidateReport(FITR2021 report)
        {

            List<ProcessMessageDocument> response = new List<ProcessMessageDocument>();
            ProcessMessageDocument processMessage;
            ProcessMessageParameter parameter;
            bool assertion;

            this.ConsumedReport = report;

            // ------------------------------------------------------------------------------
            // Validations are now done in code.
            // This version supports full validation of a report - for those rules that have successfully parsed and could be generated.
            // Generated Validations include:
            //   - Validation logic in C#.
            //   - Production of errors in EBMS, SBR and SWS format.
            // 
            // The generation at this stage does not support the following - and has to be completed manually (later versions of the generation will do this for you)
            //   - Business Rules that did not parse.  The ESR team will help support you where this happens as an effort is being made to rectify these
            //   - You will get TODO tasks for those that the parser could not cope with
            // ------------------------------------------------------------------------------


            #region VR.ATO.FITR.434007

            /*  VR.ATO.FITR.434007
            Mandatory field not supplied
    
            Legacy Rule Format:
            IF (FITR:RP:pyin.xx.xx:Report.TargetFinancial.Year = NULL)
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR1 = NULL
    
            Data Elements:
    
            ^FITR1 = RP:Report.TargetFinancial.Year
            */
            assertion = (report.FITR1 == null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.001001",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Mandatory field not supplied",
                    Location = "/xbrli:xbrl/tns:Report.TargetFinancial.Year",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434007" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR1", Value = GetValueOrEmpty(report.FITR1) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434007

            #region VR.ATO.FITR.434011

            /*  VR.ATO.FITR.434011
            Current postal address - Line 1 is required
    
            Legacy Rule Format:
            IF COUNT(((pyde.xx.xx:AddressDetails.Usage.Code = "POS") AND (pyde.xx.xx:AddressDetails.Currency.Code = "C") AND (pyde.xx.xx:AddressDetails.Line1.Text <> NULLORBLANK)) IN TUPLE(address2.xx.xx:AddressDetails)) <> 1
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR8 = BLANK
    
            Data Elements:
    
            ^FITR8 = RP:AddressDetails:AddressDetails.Line1.Text WHERE [AddressDetails.Usage.Code='POS'].[AddressDetails.Currency.Code='C']
            */
            assertion = (string.IsNullOrWhiteSpace(report.FITR8) == true);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.428023",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Current postal address must be supplied.",
                    LongDescription = @"Only one instance of current postal address must be supplied.",
                    Location = "/xbrli:xbrl/tns:AddressDetails" + OccurrenceIndex(report.AddressDetailsCollection_POS_C_RPOccurrenceIndex) + "/tns:AddressDetails.Line1.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434011" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR8", Value = report.FITR8 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434011

            #region VR.ATO.FITR.434016

            /*  VR.ATO.FITR.434016
            Trustee details - name must be present
    
            Legacy Rule Format:
            IF [FITR12] = NULLORBLANK AND [FITR14] = SET("009","010","022","053","085","118","168","198")
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR12 = BLANK AND InSet(^FITR14, '"009","010","022","053","085","118","168","198"')
    
            Data Elements:
    
            ^FITR12 = RP:Trustee:OrganisationNameDetails:OrganisationNameDetails.OrganisationalName.Text WHERE [OrganisationNameDetails.OrganisationalNameType.Code='MN'].[OrganisationNameDetails.Currency.Code='C']
    
            ^FITR14 = RP:Party.Type.Code
            */
            assertion = (string.IsNullOrWhiteSpace(report.FITR12) == true && IsMatch(report.FITR14, @"^(009|010|022|053|085|118|168|198)$"));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434016",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Trustee details - name must be present",
                    LongDescription = @"The Non-individual trustee's name must be present for all types of funds, except Non-regulated funds.",
                    Location = "/xbrli:xbrl/tns:Trustee/tns:OrganisationNameDetails" + OccurrenceIndex(report.Trustee_OrganisationNameDetailsCollection_MN_C_RPOccurrenceIndex) + "/tns:OrganisationNameDetails.OrganisationalName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434016" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR12", Value = report.FITR12 });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR14", Value = report.FITR14 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434016

            #region VR.ATO.FITR.434017

            /*  VR.ATO.FITR.434017
            Status of fund must be a valid 3 digit code 009, 010, 022, 053, 085, 118, 138, 168 or 197
    
            Legacy Rule Format:
            IF [FITR14] <> NULL AND [FITR14] <> SET("009","010","022","053","085","118","138","168","198") 
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR14 <> NULL AND NotInSet(^FITR14, '"009","010","022","053","085","118","138","168","198"')
    
            Data Elements:
    
            ^FITR14 = RP:Party.Type.Code
            */
            assertion = (report.FITR14 != null && !(IsMatch(report.FITR14, @"^(009|010|022|053|085|118|138|168|198)$", RegexOptions.IgnoreCase)));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.434017",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Status of fund must be a valid 3 digit code 009, 010, 022, 053, 085, 118, 138, 168 or 198",
                    LongDescription = @"Type of fund or trust must be one of the following codes: 009 - Approved Deposit Fund; 010 - APRA RegSupFnd Public Sector; 022 - APRA Regulated Super Fund Small APRA Fund; 053 - Corporate Fund; 085 - Eligible Rollover Fund; 118 - Industry fund; 138 - Non Regulated Superannuation Fund; 168 - Pooled Superannuation Trust; 198 - Retail Fund",
                    Location = "/xbrli:xbrl/tns:Party.Type.Code",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434017" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR14", Value = report.FITR14 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434017

            #region VR.ATO.FITR.434018

            /*  VR.ATO.FITR.434018
            Mandatory field not supplied
    
            Legacy Rule Format:
            IF [FITR14] = NULLORBLANK 
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR14 = BLANK
    
            Data Elements:
    
            ^FITR14 = RP:Party.Type.Code
            */
            assertion = (string.IsNullOrWhiteSpace(report.FITR14) == true);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.001001",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Mandatory field not supplied",
                    Location = "/xbrli:xbrl/tns:Party.Type.Code",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434018" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR14", Value = report.FITR14 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434018

            #region VR.ATO.FITR.434019

            /*  VR.ATO.FITR.434019
            Mandatory field not supplied
    
            Legacy Rule Format:
            IF [FITR15] = NULL
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR15 = NULL
    
            Data Elements:
    
            ^FITR15 = RP:SuperannuationFundDetails.AustralianSuperannuationFundStatus.Indicator
            */
            assertion = (report.FITR15 == null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.001001",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Mandatory field not supplied",
                    Location = "/xbrli:xbrl/tns:SuperannuationFundDetails.AustralianSuperannuationFundStatus.Indicator",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434019" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR15", Value = GetValueOrEmpty(report.FITR15) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434019

            #region VR.ATO.FITR.434020

            /*  VR.ATO.FITR.434020
            Mandatory field not supplied
    
            Legacy Rule Format:
            IF [FITR16] = NULLORBLANK 
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR16 = BLANK
    
            Data Elements:
    
            ^FITR16 = RP:SuperannuationFundDetails.FundBenefitStructure.Code
            */
            assertion = (string.IsNullOrWhiteSpace(report.FITR16) == true);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.001001",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Mandatory field not supplied",
                    Location = "/xbrli:xbrl/tns:SuperannuationFundDetails.FundBenefitStructure.Code",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434020" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR16", Value = report.FITR16 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434020

            #region VR.ATO.FITR.434022

            /*  VR.ATO.FITR.434022
            If the fund or trust wound up during the income year, the number of members must not be provided
    
            Legacy Rule Format:
            IF [FITR19] = TRUE AND [FITR208] <> NULL 
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR19 = TRUE AND ^FITR208 <> NULL
    
            Data Elements:
    
            ^FITR208 = RP(Instant):OrganisationControl.RegisteredMember.Count
    
            ^FITR19 = RP:OrganisationDetails.OperationCease.Indicator
            */
            assertion = (report.FITR19 == true && report.FITR208 != null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434297",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"If the fund or trust wound up during the income year, the number of members must not be provided",
                    Location = "/xbrli:xbrl/tns:OrganisationControl.RegisteredMember.Count",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434022" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR19", Value = GetValueOrEmpty(report.FITR19) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR208", Value = report.FITR208.GetValueOrDefault().ToString() });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434022

            #region VR.ATO.FITR.434023

            /*  VR.ATO.FITR.434023
            Mandatory field not supplied
    
            Legacy Rule Format:
            IF [FITR18] = NULL 
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR18 = NULL
    
            Data Elements:
    
            ^FITR18 = RP:OrganisationDetails.Establishment.Date
            */
            assertion = (report.FITR18 == null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.001001",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Mandatory field not supplied",
                    Location = "/xbrli:xbrl/tns:OrganisationDetails.Establishment.Date",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434023" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR18", Value = GetValueOrEmptyDate(report.FITR18) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434023

            #region VR.ATO.FITR.434024

            /*  VR.ATO.FITR.434024
            Date of establishment must not be after the financial year
    
            Legacy Rule Format:
            IF [FITR18] > ([FITR1]&"-06-30")  
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR18 > AsDate(Concat(^FITR1, '-06-30'))
    
            Data Elements:
    
            ^FITR18 = RP:OrganisationDetails.Establishment.Date
    
            ^FITR1 = RP:Report.TargetFinancial.Year
            */
            assertion = (report.FITR18.GetValueOrDefault() > AsDate(string.Concat(report.FITR1.GetValueOrDefault(), @"-06-30")));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434024",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Date of establishment must not be after the financial year",
                    Location = "/xbrli:xbrl/tns:OrganisationDetails.Establishment.Date",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434024" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR18", Value = GetValueOrEmptyDate(report.FITR18) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR1", Value = GetValueOrEmpty(report.FITR1) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434024

            #region VR.ATO.FITR.434025

            /*  VR.ATO.FITR.434025
            Mandatory field not supplied
    
            Legacy Rule Format:
            IF [FITR19] = NULL
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR19 = NULL
    
            Data Elements:
    
            ^FITR19 = RP:OrganisationDetails.OperationCease.Indicator
            */
            assertion = (report.FITR19 == null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.001001",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Mandatory field not supplied",
                    Location = "/xbrli:xbrl/tns:OrganisationDetails.OperationCease.Indicator",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434025" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR19", Value = GetValueOrEmpty(report.FITR19) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434025

            #region VR.ATO.FITR.434026

            /*  VR.ATO.FITR.434026
            Date wound up must be present if the fund wound up during the income year
    
            Legacy Rule Format:
            IF [FITR19] = TRUE AND [FITR20] = NULL  
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR19 = TRUE AND ^FITR20 = NULL
    
            Data Elements:
    
            ^FITR20 = RP:OrganisationDetails.OrganisationActivityStatusEnd.Date
    
            ^FITR19 = RP:OrganisationDetails.OperationCease.Indicator
            */
            assertion = (report.FITR19 == true && report.FITR20 == null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.434026",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Date wound up must be present if the fund wound up during the income year",
                    Location = "/xbrli:xbrl/tns:OrganisationDetails.OrganisationActivityStatusEnd.Date",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434026" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR19", Value = GetValueOrEmpty(report.FITR19) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR20", Value = GetValueOrEmptyDate(report.FITR20) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434026

            #region VR.ATO.FITR.434028

            /*  VR.ATO.FITR.434028
            Mandatory field not supplied
    
            Legacy Rule Format:
            IF [FITR21] = NULL
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR21 = NULL
    
            Data Elements:
    
            ^FITR21 = RP:CapitalGainsTax.Event.Indicator
            */
            assertion = (report.FITR21 == null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.001001",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Mandatory field not supplied",
                    Location = "/xbrli:xbrl/tns:CapitalGainsTax.Event.Indicator",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434028" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR21", Value = GetValueOrEmpty(report.FITR21) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434028

            #region VR.ATO.FITR.434035

            /*  VR.ATO.FITR.434035
            Answer to CGT event question must be 'yes' (true)
    
            Legacy Rule Format:
            IF [FITR23] > 0 AND [FITR21] = FALSE  
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR23 > 0 AND ^FITR21 = FALSE
    
            Data Elements:
    
            ^FITR21 = RP:CapitalGainsTax.Event.Indicator
    
            ^FITR23 = RP:Income.CapitalGainsNet.Amount
            */
            assertion = (report.FITR23.GetValueOrDefault() > 0 && report.FITR21 == false);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.434035",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Answer to CGT event question must be 'yes' (true)",
                    LongDescription = @"If a Net capital gain amount is entered, then the answer to the question 'Did you have a CGT event during the year' must be 'yes' (true)",
                    Location = "/xbrli:xbrl/tns:CapitalGainsTax.Event.Indicator",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434035" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR23", Value = GetValueOrEmpty(report.FITR23) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR21", Value = GetValueOrEmpty(report.FITR21) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434035

            #region VR.ATO.FITR.434037

            /*  VR.ATO.FITR.434037
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR25] <> NULL AND [FITR25] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR25, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR25 = RP:Income.Interest.Gross.Amount
            */
            assertion = NotMonetary(report.FITR25, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:Income.Interest.Gross.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434037" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR25", Value = GetValueOrEmpty(report.FITR25) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434037

            #region VR.ATO.FITR.434038

            /*  VR.ATO.FITR.434038
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR26] <> NULL AND [FITR26] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR26, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR26 = RP:Income.ForestryManagedInvestmentScheme.Amount
            */
            assertion = NotMonetary(report.FITR26, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:Income.ForestryManagedInvestmentScheme.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434038" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR26", Value = GetValueOrEmpty(report.FITR26) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434038

            #region VR.ATO.FITR.434040

            /*  VR.ATO.FITR.434040
            Monetary field - max 11 characters, signed, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR28] <> NULL AND [FITR28] <> MONETARY(S,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR28, 'S', 11, 0)
    
            Data Elements:
    
            ^FITR28 = RP.JFOREIGN:Income.Net.Amount
            */
            assertion = NotMonetary(report.FITR28, @"S", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:Income.Net.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434040" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR28", Value = GetValueOrEmpty(report.FITR28) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434040

            #region VR.ATO.FITR.434041

            /*  VR.ATO.FITR.434041
            Gross foreign income must not be less than Net foreign income.
    
            Legacy Rule Format:
            IF [FITR27] <> NULL AND [FITR27] < [FITR28]  
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR27 <> NULL AND ^FITR27 < ^FITR28
    
            Data Elements:
    
            ^FITR27 = RP.JFOREIGN:Income.Gross.Amount
    
            ^FITR28 = RP.JFOREIGN:Income.Net.Amount
            */
            assertion = (report.FITR27 != null && report.FITR27.GetValueOrDefault() < report.FITR28.GetValueOrDefault());
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.434041",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Gross foreign income must not be less than Net foreign income.",
                    Location = "/xbrli:xbrl/tns:Income.Gross.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434041" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR27", Value = GetValueOrEmpty(report.FITR27) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR28", Value = GetValueOrEmpty(report.FITR28) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434041

            #region VR.ATO.FITR.434042

            /*  VR.ATO.FITR.434042
            Gross foreign income must be present if Net foreign income is present
    
            Legacy Rule Format:
            IF [FITR28] > 0 AND [FITR27] = NULL  
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR28 > 0 AND ^FITR27 = NULL
    
            Data Elements:
    
            ^FITR27 = RP.JFOREIGN:Income.Gross.Amount
    
            ^FITR28 = RP.JFOREIGN:Income.Net.Amount
            */
            assertion = (report.FITR28.GetValueOrDefault() > 0 && report.FITR27 == null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.434042",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Gross foreign income must be present if Net foreign income is present",
                    Location = "/xbrli:xbrl/tns:Income.Gross.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434042" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR28", Value = GetValueOrEmpty(report.FITR28) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR27", Value = GetValueOrEmpty(report.FITR27) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434042

            #region VR.ATO.FITR.434043

            /*  VR.ATO.FITR.434043
            The Number of transfers from foreign funds must be present if an amount is present
    
            Legacy Rule Format:
            IF [FITR30] > 0 AND [FITR31] = NULL 
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR30 > 0 AND ^FITR31 = NULL
    
            Data Elements:
    
            ^FITR31 = RP:InternationalDealings.ForeignFundTransfers.Count
    
            ^FITR30 = RP:Equity.InternationalDealings.ForeignFundTransfers.Amount
            */
            assertion = (report.FITR30.GetValueOrDefault() > 0 && report.FITR31 == null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434043",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"The Number of transfers from foreign funds must be present if an amount is present",
                    Location = "/xbrli:xbrl/tns:InternationalDealings.ForeignFundTransfers.Count",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434043" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR30", Value = GetValueOrEmpty(report.FITR30) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR31", Value = report.FITR31.GetValueOrDefault().ToString() });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434043

            #region VR.ATO.FITR.434048

            /*  VR.ATO.FITR.434048
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR34] <> NULL AND [FITR34] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR34, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR34 = RP:Income.DividendsUnfranked.Amount
            */
            assertion = NotMonetary(report.FITR34, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:Income.DividendsUnfranked.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434048" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR34", Value = GetValueOrEmpty(report.FITR34) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434048

            #region VR.ATO.FITR.434049

            /*  VR.ATO.FITR.434049
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR35] <> NULL AND [FITR35] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR35, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR35 = RP:Income.DividendsFranked.Amount
            */
            assertion = NotMonetary(report.FITR35, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:Income.DividendsFranked.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434049" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR35", Value = GetValueOrEmpty(report.FITR35) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434049

            #region VR.ATO.FITR.434050

            /*  VR.ATO.FITR.434050
            Franking credit cannot be greater than 48% of dividends-franked amount.
    
            Legacy Rule Format:
            IF [FITR36] > 0 AND (([FITR35] * 0.48) < ([FITR36] - 1)) 
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR36 > 0 AND ((^FITR35 * 0.48) < (^FITR36 - 1))
    
            Data Elements:
    
            ^FITR36 = RP:IncomeTax.FrankingCredits.Amount
    
            ^FITR35 = RP:Income.DividendsFranked.Amount
            */
            assertion = (report.FITR36.GetValueOrDefault() > 0 && report.FITR35.GetValueOrDefault() * 0.48M < report.FITR36.GetValueOrDefault() - 1);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.430098",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Franking credit cannot be greater than 48% of dividends-franked amount.",
                    LongDescription = @"Franking credit must not exceed 48% of Dividends - Franked amount",
                    Location = "/xbrli:xbrl/tns:IncomeTax.FrankingCredits.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434050" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR36", Value = GetValueOrEmpty(report.FITR36) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR35", Value = GetValueOrEmpty(report.FITR35) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434050

            #region VR.ATO.FITR.434051

            /*  VR.ATO.FITR.434051
            Franking credit amount must be present
    
            Legacy Rule Format:
            IF [FITR35] > 0 AND [FITR36] = NULL 
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR35 > 0 AND ^FITR36 = NULL
    
            Data Elements:
    
            ^FITR36 = RP:IncomeTax.FrankingCredits.Amount
    
            ^FITR35 = RP:Income.DividendsFranked.Amount
            */
            assertion = (report.FITR35.GetValueOrDefault() > 0 && report.FITR36 == null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.430100",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Franking credit amount must be present",
                    LongDescription = @"If Dividends - Franked amount is present, then Franking credit must be present",
                    Location = "/xbrli:xbrl/tns:IncomeTax.FrankingCredits.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434051" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR35", Value = GetValueOrEmpty(report.FITR35) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR36", Value = GetValueOrEmpty(report.FITR36) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434051

            #region VR.ATO.FITR.434052

            /*  VR.ATO.FITR.434052
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR36] <> NULL AND [FITR36] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR36, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR36 = RP:IncomeTax.FrankingCredits.Amount
            */
            assertion = NotMonetary(report.FITR36, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:IncomeTax.FrankingCredits.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434052" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR36", Value = GetValueOrEmpty(report.FITR36) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434052

            #region VR.ATO.FITR.434053

            /*  VR.ATO.FITR.434053
            Transfer of liability to LIC or PST must not exceed the sum of Assessable contributions
    
            Legacy Rule Format:
            IF [FITR46] > ([FITR41] + [FITR42])  
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR46 > (^FITR41 + ^FITR42)
    
            Data Elements:
    
            ^FITR46 = RP:SuperannuationContribution.LiabilityTransferToLifeInsuranceCompanyOrPooledSuperannuationTrustAssessable.Amount
    
            ^FITR41 = RP:SuperannuationContribution.EmployerAssessable.Amount
    
            ^FITR42 = RP:SuperannuationContribution.PersonalAssessable.Amount
            */
            assertion = (report.FITR46.GetValueOrDefault() > report.FITR41.GetValueOrDefault() + report.FITR42.GetValueOrDefault());
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434053",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Transfer of liability to LIC or PST must not exceed the sum of Assessable contributions",
                    LongDescription = @"The 'Transfer of liability to life insurance company or PST' amount must not be greater than the sum of 'Assessable employer contributions' and 'Assessable personal contributions'",
                    Location = "/xbrli:xbrl/tns:SuperannuationContribution.LiabilityTransferToLifeInsuranceCompanyOrPooledSuperannuationTrustAssessable.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434053" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR46", Value = GetValueOrEmpty(report.FITR46) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR41", Value = GetValueOrEmpty(report.FITR41) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR42", Value = GetValueOrEmpty(report.FITR42) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434053

            #region VR.ATO.FITR.434054

            /*  VR.ATO.FITR.434054
            Transfer of liability to LIC or PST amount must be present
    
            Legacy Rule Format:
            IF [FITR84] = TRUE AND ([FITR46] = 0 OR [FITR46] = NULL) 
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR84 = TRUE AND (^FITR46 = 0 OR ^FITR46 = NULL)
    
            Data Elements:
    
            ^FITR46 = RP:SuperannuationContribution.LiabilityTransferToLifeInsuranceCompanyOrPooledSuperannuationTrustAssessable.Amount
    
            ^FITR84 = RP:IncomeTax.Superannuation.TransferredAssessable.Indicator
            */
            assertion = (report.FITR84 == true && (report.FITR46 == 0 || report.FITR46 == null));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434054",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Transfer of liability to LIC or PST amount must be present",
                    LongDescription = @"If the answer to the question 'Has the fund or trust, with the consent of the transferee, transferred assessable contributions...' is 'yes' (true), then a 'Transfer of liability to a LIC or PST' amount, other than zero, must be present",
                    Location = "/xbrli:xbrl/tns:SuperannuationContribution.LiabilityTransferToLifeInsuranceCompanyOrPooledSuperannuationTrustAssessable.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434054" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR84", Value = GetValueOrEmpty(report.FITR84) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR46", Value = GetValueOrEmpty(report.FITR46) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434054

            #region VR.ATO.FITR.434055

            /*  VR.ATO.FITR.434055
            Other income amount must be present
    
            Legacy Rule Format:
            IF [FITR48] <> NULL AND ([FITR47] = NULL OR [FITR47] = 0)   
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR48 <> NULL AND (^FITR47 = NULL OR ^FITR47 = 0)
    
            Data Elements:
    
            ^FITR47 = RP:Income.Other.Amount
    
            ^FITR48 = RP:Income.Other.Code
            */
            assertion = (report.FITR48 != null && (report.FITR47 == null || report.FITR47 == 0));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.434055",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Other income amount must be present",
                    LongDescription = @"The 'Other income' amount must be greater than zero if the 'Other income - code' is present",
                    Location = "/xbrli:xbrl/tns:Income.Other.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434055" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR48", Value = GetValueOrEmpty(report.FITR48) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR47", Value = GetValueOrEmpty(report.FITR47) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434055

            #region VR.ATO.FITR.434058

            /*  VR.ATO.FITR.434058
            Other income code must be present if Other income amount is present.
    
            Legacy Rule Format:
            IF [FITR47] > 0 AND [FITR48] = NULLORBLANK
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR47 > 0 AND ^FITR48 = BLANK
    
            Data Elements:
    
            ^FITR48 = RP:Income.Other.Code
    
            ^FITR47 = RP:Income.Other.Amount
            */
            assertion = (report.FITR47.GetValueOrDefault() > 0 && string.IsNullOrWhiteSpace(report.FITR48) == true);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434282",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Other income code must be present if Other income amount is present.",
                    LongDescription = @"Other income codes are:    'B' - Assessable balancing adjustment amount,    'C' - LIC capital gain amount,  'O' - Other income received not listed,    'R' - Rebate or refund of death or disability premiums,    'W' - Gross payments subject to foreign resident withholding, 'T' - TOFA gains or assessable TOFA balancing adjustment",
                    Location = "/xbrli:xbrl/tns:Income.Other.Code",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434058" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR47", Value = GetValueOrEmpty(report.FITR47) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR48", Value = GetValueOrEmpty(report.FITR48) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434058

            #region VR.ATO.FITR.434059

            /*  VR.ATO.FITR.434059
            Exempt current pension income must not be greater than Gross income.
    
            Legacy Rule Format:
            IF [FITR200] > ([FITR23] + [FITR24] + [FITR25] + [FITR26] + [FITR28] + [FITR29] + [FITR30] + [FITR32] + [FITR33] + [FITR34] + [FITR35] + [FITR36] + [FITR37] + [FITR38] + [FITR39] + [FITR40] + [FITR41] + [FITR42] + [FITR43] - [FITR44] - [FITR45] - [FITR46] + [FITR186] + [FITR47] + [FITR49] + [FITR50])  
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR200 > (^FITR23 + ^FITR24 + ^FITR25 + ^FITR26 + ^FITR28 + ^FITR29 + ^FITR30 + ^FITR32 + ^FITR33 + ^FITR34 + ^FITR35 + ^FITR36 + ^FITR37 + ^FITR38 + ^FITR39 + ^FITR40 + ^FITR41 + ^FITR42 + ^FITR43 - ^FITR44 - ^FITR45 - ^FITR46 + ^FITR186 + ^FITR47 + ^FITR49 + ^FITR50)
    
            Data Elements:
    
            ^FITR200 = RP:TaxConcession.PensionIncomeExemptDeduction.Amount
    
            ^FITR23 = RP:Income.CapitalGainsNet.Amount
    
            ^FITR24 = RP:Income.RentLeasingHiringGross.Amount
    
            ^FITR25 = RP:Income.Interest.Gross.Amount
    
            ^FITR26 = RP:Income.ForestryManagedInvestmentScheme.Amount
    
            ^FITR28 = RP.JFOREIGN:Income.Net.Amount
    
            ^FITR29 = RP:IncomeTax.FrankingCredits.ReceivedFromNewZealandCompanies.Amount
    
            ^FITR30 = RP:Equity.InternationalDealings.ForeignFundTransfers.Amount
    
            ^FITR32 = RP:Remuneration.ABNNotQuotedPaymentGross.Amount
    
            ^FITR33 = RP:Income.PartnershipDistributionGross.Amount
    
            ^FITR34 = RP:Income.DividendsUnfranked.Amount
    
            ^FITR35 = RP:Income.DividendsFranked.Amount
    
            ^FITR36 = RP:IncomeTax.FrankingCredits.Amount
    
            ^FITR37 = RP:Income.TrustDistributionUnfranked.Amount
    
            ^FITR38 = RP:Income.TrustDistributionFranked.Amount
    
            ^FITR39 = RP:IncomeTax.FrankingCredits.FrankingCreditShareReceivedIndirectly.Amount
    
            ^FITR40 = RP:Income.TrustDistributionOther.Amount
    
            ^FITR41 = RP:SuperannuationContribution.EmployerAssessable.Amount
    
            ^FITR42 = RP:SuperannuationContribution.PersonalAssessable.Amount
    
            ^FITR43 = RP:SuperannuationContribution.EmployerContributionsTaxFileNumberNotQuoted.Amount
    
            ^FITR44 = RP:SuperannuationContribution.ContributionsExcludedByTrustee.Amount
    
            ^FITR45 = RP:SuperannuationContribution.ReducedContributions.Amount
    
            ^FITR46 = RP:SuperannuationContribution.LiabilityTransferToLifeInsuranceCompanyOrPooledSuperannuationTrustAssessable.Amount
    
            ^FITR186 = RP:Income.ForeignExchangeGains.Amount
    
            ^FITR47 = RP:Income.Other.Amount
    
            ^FITR49 = RP:Income.AssessableIncomeDueToFundTaxStatusChange.Amount
    
            ^FITR50 = RP:Income.NonArmsLengthOtherNet.Amount
            */
            assertion = (report.FITR200.GetValueOrDefault() > report.FITR23.GetValueOrDefault() + report.FITR24.GetValueOrDefault() + report.FITR25.GetValueOrDefault() + report.FITR26.GetValueOrDefault() + report.FITR28.GetValueOrDefault() + report.FITR29.GetValueOrDefault() + report.FITR30.GetValueOrDefault() + report.FITR32.GetValueOrDefault() + report.FITR33.GetValueOrDefault() + report.FITR34.GetValueOrDefault() + report.FITR35.GetValueOrDefault() + report.FITR36.GetValueOrDefault() + report.FITR37.GetValueOrDefault() + report.FITR38.GetValueOrDefault() + report.FITR39.GetValueOrDefault() + report.FITR40.GetValueOrDefault() + report.FITR41.GetValueOrDefault() + report.FITR42.GetValueOrDefault() + report.FITR43.GetValueOrDefault() - report.FITR44.GetValueOrDefault() - report.FITR45.GetValueOrDefault() - report.FITR46.GetValueOrDefault() + report.FITR186.GetValueOrDefault() + report.FITR47.GetValueOrDefault() + report.FITR49.GetValueOrDefault() + report.FITR50.GetValueOrDefault());
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434278",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Exempt current pension income must not be greater than Gross income.",
                    Location = "/xbrli:xbrl/tns:TaxConcession.PensionIncomeExemptDeduction.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434059" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR200", Value = GetValueOrEmpty(report.FITR200) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR23", Value = GetValueOrEmpty(report.FITR23) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR24", Value = GetValueOrEmpty(report.FITR24) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR25", Value = GetValueOrEmpty(report.FITR25) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR26", Value = GetValueOrEmpty(report.FITR26) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR28", Value = GetValueOrEmpty(report.FITR28) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR29", Value = GetValueOrEmpty(report.FITR29) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR30", Value = GetValueOrEmpty(report.FITR30) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR32", Value = GetValueOrEmpty(report.FITR32) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR33", Value = GetValueOrEmpty(report.FITR33) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR34", Value = GetValueOrEmpty(report.FITR34) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR35", Value = GetValueOrEmpty(report.FITR35) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR36", Value = GetValueOrEmpty(report.FITR36) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR37", Value = GetValueOrEmpty(report.FITR37) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR38", Value = GetValueOrEmpty(report.FITR38) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR39", Value = GetValueOrEmpty(report.FITR39) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR40", Value = GetValueOrEmpty(report.FITR40) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR41", Value = GetValueOrEmpty(report.FITR41) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR42", Value = GetValueOrEmpty(report.FITR42) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR43", Value = GetValueOrEmpty(report.FITR43) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR44", Value = GetValueOrEmpty(report.FITR44) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR45", Value = GetValueOrEmpty(report.FITR45) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR46", Value = GetValueOrEmpty(report.FITR46) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR186", Value = GetValueOrEmpty(report.FITR186) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR47", Value = GetValueOrEmpty(report.FITR47) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR49", Value = GetValueOrEmpty(report.FITR49) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR50", Value = GetValueOrEmpty(report.FITR50) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434059

            #region VR.ATO.FITR.434060

            /*  VR.ATO.FITR.434060
            Exempt current pension income must not equal Gross income when Deductions are present
    
            Legacy Rule Format:
            IF [FITR200] = ([FITR23] + [FITR24] + [FITR25] + [FITR26] + [FITR28] + [FITR29] + [FITR30] + [FITR32] + [FITR33] + [FITR34] + [FITR35] + [FITR36] + [FITR37] + [FITR38] + [FITR39] + [FITR40] + [FITR41] + [FITR42] + [FITR43] - [FITR44] - [FITR45] - [FITR46] + [FITR186] + [FITR47] + [FITR49] + [FITR50]) AND ([FITR52] > 0 OR [FITR53] > 0 OR [FITR54] > 0 OR [FITR55] > 0 OR [FITR56] > 0 OR [FITR57] > 0 OR [FITR59] > 0 OR [FITR190] > 0 OR [FITR188] > 0 OR [FITR61] > 0  OR [FITR187] > 0 OR [FITR62] > 0 OR [FITR64] > 0)  
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR200 = (^FITR23 + ^FITR24 + ^FITR25 + ^FITR26 + ^FITR28 + ^FITR29 + ^FITR30 + ^FITR32 + ^FITR33 + ^FITR34 + ^FITR35 + ^FITR36 + ^FITR37 + ^FITR38 + ^FITR39 + ^FITR40 + ^FITR41 + ^FITR42 + ^FITR43 - ^FITR44 - ^FITR45 - ^FITR46 + ^FITR186 + ^FITR47 + ^FITR49 + ^FITR50) AND (^FITR52 > 0 OR ^FITR53 > 0 OR ^FITR54 > 0 OR ^FITR55 > 0 OR ^FITR56 > 0 OR ^FITR57 > 0 OR ^FITR59 > 0 OR ^FITR190 > 0 OR ^FITR188 > 0 OR ^FITR61 > 0  OR ^FITR187 > 0 OR ^FITR62 > 0 OR ^FITR64 > 0)
    
            Data Elements:
    
            ^FITR200 = RP:TaxConcession.PensionIncomeExemptDeduction.Amount
    
            ^FITR23 = RP:Income.CapitalGainsNet.Amount
    
            ^FITR24 = RP:Income.RentLeasingHiringGross.Amount
    
            ^FITR25 = RP:Income.Interest.Gross.Amount
    
            ^FITR26 = RP:Income.ForestryManagedInvestmentScheme.Amount
    
            ^FITR28 = RP.JFOREIGN:Income.Net.Amount
    
            ^FITR29 = RP:IncomeTax.FrankingCredits.ReceivedFromNewZealandCompanies.Amount
    
            ^FITR30 = RP:Equity.InternationalDealings.ForeignFundTransfers.Amount
    
            ^FITR32 = RP:Remuneration.ABNNotQuotedPaymentGross.Amount
    
            ^FITR33 = RP:Income.PartnershipDistributionGross.Amount
    
            ^FITR34 = RP:Income.DividendsUnfranked.Amount
    
            ^FITR35 = RP:Income.DividendsFranked.Amount
    
            ^FITR36 = RP:IncomeTax.FrankingCredits.Amount
    
            ^FITR37 = RP:Income.TrustDistributionUnfranked.Amount
    
            ^FITR38 = RP:Income.TrustDistributionFranked.Amount
    
            ^FITR39 = RP:IncomeTax.FrankingCredits.FrankingCreditShareReceivedIndirectly.Amount
    
            ^FITR40 = RP:Income.TrustDistributionOther.Amount
    
            ^FITR41 = RP:SuperannuationContribution.EmployerAssessable.Amount
    
            ^FITR42 = RP:SuperannuationContribution.PersonalAssessable.Amount
    
            ^FITR43 = RP:SuperannuationContribution.EmployerContributionsTaxFileNumberNotQuoted.Amount
    
            ^FITR44 = RP:SuperannuationContribution.ContributionsExcludedByTrustee.Amount
    
            ^FITR45 = RP:SuperannuationContribution.ReducedContributions.Amount
    
            ^FITR46 = RP:SuperannuationContribution.LiabilityTransferToLifeInsuranceCompanyOrPooledSuperannuationTrustAssessable.Amount
    
            ^FITR186 = RP:Income.ForeignExchangeGains.Amount
    
            ^FITR47 = RP:Income.Other.Amount
    
            ^FITR49 = RP:Income.AssessableIncomeDueToFundTaxStatusChange.Amount
    
            ^FITR50 = RP:Income.NonArmsLengthOtherNet.Amount
    
            ^FITR52 = RP.JAUS:Expense.Interest.Amount
    
            ^FITR53 = RP.JFOREIGN:Expense.Interest.Amount
    
            ^FITR54 = RP:Remuneration.WagesandSalaries.Payments.Amount
    
            ^FITR55 = RP:Expense.CapitalWorksDeduction.Amount
    
            ^FITR56 = RP:Expense.DepreciationAllowableDeduction.Amount
    
            ^FITR57 = RP:Expense.PremiumsDeathOrDisability.Amount
    
            ^FITR59 = RP:Expense.Investment.Amount
    
            ^FITR190 = RP:Expense.Operating.ManagementExpense.Amount
    
            ^FITR188 = RP:Expense.Operating.AdministrativeExpense.Amount
    
            ^FITR61 = RP:Expense.ForestryManagedInvestmentSchemeDeduction.Amount
    
            ^FITR187 = RP:Income.ForeignExchangeLosses.Amount
    
            ^FITR62 = RP:Expense.DeductibleOther.Amount
    
            ^FITR64 = RP:IncomeTax.Deduction.TaxLossesDeducted.Amount
            */
            assertion = (report.FITR200.GetValueOrDefault() == report.FITR23.GetValueOrDefault() + report.FITR24.GetValueOrDefault() + report.FITR25.GetValueOrDefault() + report.FITR26.GetValueOrDefault() + report.FITR28.GetValueOrDefault() + report.FITR29.GetValueOrDefault() + report.FITR30.GetValueOrDefault() + report.FITR32.GetValueOrDefault() + report.FITR33.GetValueOrDefault() + report.FITR34.GetValueOrDefault() + report.FITR35.GetValueOrDefault() + report.FITR36.GetValueOrDefault() + report.FITR37.GetValueOrDefault() + report.FITR38.GetValueOrDefault() + report.FITR39.GetValueOrDefault() + report.FITR40.GetValueOrDefault() + report.FITR41.GetValueOrDefault() + report.FITR42.GetValueOrDefault() + report.FITR43.GetValueOrDefault() - report.FITR44.GetValueOrDefault() - report.FITR45.GetValueOrDefault() - report.FITR46.GetValueOrDefault() + report.FITR186.GetValueOrDefault() + report.FITR47.GetValueOrDefault() + report.FITR49.GetValueOrDefault() + report.FITR50.GetValueOrDefault() && (report.FITR52.GetValueOrDefault() > 0 || report.FITR53.GetValueOrDefault() > 0 || report.FITR54.GetValueOrDefault() > 0 || report.FITR55.GetValueOrDefault() > 0 || report.FITR56.GetValueOrDefault() > 0 || report.FITR57.GetValueOrDefault() > 0 || report.FITR59.GetValueOrDefault() > 0 || report.FITR190.GetValueOrDefault() > 0 || report.FITR188.GetValueOrDefault() > 0 || report.FITR61.GetValueOrDefault() > 0 || report.FITR187.GetValueOrDefault() > 0 || report.FITR62.GetValueOrDefault() > 0 || report.FITR64.GetValueOrDefault() > 0));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434279",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Exempt current pension income must not equal Gross income when Deductions are present",
                    LongDescription = @"The 'Exempt current pension income' amount must not equal the Gross income when one or more Deduction amounts are greater than zero",
                    Location = "/xbrli:xbrl/tns:TaxConcession.PensionIncomeExemptDeduction.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434060" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR200", Value = GetValueOrEmpty(report.FITR200) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR23", Value = GetValueOrEmpty(report.FITR23) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR24", Value = GetValueOrEmpty(report.FITR24) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR25", Value = GetValueOrEmpty(report.FITR25) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR26", Value = GetValueOrEmpty(report.FITR26) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR28", Value = GetValueOrEmpty(report.FITR28) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR29", Value = GetValueOrEmpty(report.FITR29) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR30", Value = GetValueOrEmpty(report.FITR30) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR32", Value = GetValueOrEmpty(report.FITR32) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR33", Value = GetValueOrEmpty(report.FITR33) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR34", Value = GetValueOrEmpty(report.FITR34) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR35", Value = GetValueOrEmpty(report.FITR35) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR36", Value = GetValueOrEmpty(report.FITR36) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR37", Value = GetValueOrEmpty(report.FITR37) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR38", Value = GetValueOrEmpty(report.FITR38) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR39", Value = GetValueOrEmpty(report.FITR39) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR40", Value = GetValueOrEmpty(report.FITR40) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR41", Value = GetValueOrEmpty(report.FITR41) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR42", Value = GetValueOrEmpty(report.FITR42) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR43", Value = GetValueOrEmpty(report.FITR43) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR44", Value = GetValueOrEmpty(report.FITR44) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR45", Value = GetValueOrEmpty(report.FITR45) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR46", Value = GetValueOrEmpty(report.FITR46) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR186", Value = GetValueOrEmpty(report.FITR186) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR47", Value = GetValueOrEmpty(report.FITR47) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR49", Value = GetValueOrEmpty(report.FITR49) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR50", Value = GetValueOrEmpty(report.FITR50) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR52", Value = GetValueOrEmpty(report.FITR52) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR53", Value = GetValueOrEmpty(report.FITR53) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR54", Value = GetValueOrEmpty(report.FITR54) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR55", Value = GetValueOrEmpty(report.FITR55) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR56", Value = GetValueOrEmpty(report.FITR56) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR57", Value = GetValueOrEmpty(report.FITR57) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR59", Value = GetValueOrEmpty(report.FITR59) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR190", Value = GetValueOrEmpty(report.FITR190) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR188", Value = GetValueOrEmpty(report.FITR188) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR61", Value = GetValueOrEmpty(report.FITR61) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR187", Value = GetValueOrEmpty(report.FITR187) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR62", Value = GetValueOrEmpty(report.FITR62) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR64", Value = GetValueOrEmpty(report.FITR64) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434060

            #region VR.ATO.FITR.434061

            /*  VR.ATO.FITR.434061
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR53] <> NULL AND [FITR53] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR53, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR53 = RP.JFOREIGN:Expense.Interest.Amount
            */
            assertion = NotMonetary(report.FITR53, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:Expense.Interest.Amount[@contextRef='" + report.RPJFOREIGNId + "']",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434061" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR53", Value = GetValueOrEmpty(report.FITR53) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434061

            #region VR.ATO.FITR.434063

            /*  VR.ATO.FITR.434063
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR54] <> NULL AND [FITR54] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR54, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR54 = RP:Remuneration.WagesandSalaries.Payments.Amount
            */
            assertion = NotMonetary(report.FITR54, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:Remuneration.WagesandSalaries.Payments.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434063" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR54", Value = GetValueOrEmpty(report.FITR54) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434063

            #region VR.ATO.FITR.434064

            /*  VR.ATO.FITR.434064
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR55] <> NULL AND [FITR55] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR55, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR55 = RP:Expense.CapitalWorksDeduction.Amount
            */
            assertion = NotMonetary(report.FITR55, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:Expense.CapitalWorksDeduction.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434064" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR55", Value = GetValueOrEmpty(report.FITR55) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434064

            #region VR.ATO.FITR.434067

            /*  VR.ATO.FITR.434067
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR61] <> NULL AND [FITR61] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR61, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR61 = RP:Expense.ForestryManagedInvestmentSchemeDeduction.Amount
            */
            assertion = NotMonetary(report.FITR61, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:Expense.ForestryManagedInvestmentSchemeDeduction.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434067" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR61", Value = GetValueOrEmpty(report.FITR61) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434067

            #region VR.ATO.FITR.434068

            /*  VR.ATO.FITR.434068
            If 'Other deductions - code' is present, 'Other deductions' amount greater than zero must be present
    
            Legacy Rule Format:
            IF ([FITR62] = 0 OR [FITR62] = NULL) AND [FITR63] <> NULL
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^FITR62 = 0 OR ^FITR62 = NULL) AND ^FITR63 <> NULL
    
            Data Elements:
    
            ^FITR62 = RP:Expense.DeductibleOther.Amount
    
            ^FITR63 = RP:Expense.DeductibleOther.Code
            */
            assertion = ((report.FITR62 == 0 || report.FITR62 == null) && report.FITR63 != null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.434068",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"If 'Other deductions - code' is present, 'Other deductions' amount greater than zero must be present",
                    Location = "/xbrli:xbrl/tns:Expense.DeductibleOther.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434068" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR62", Value = GetValueOrEmpty(report.FITR62) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR63", Value = GetValueOrEmpty(report.FITR63) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434068

            #region VR.ATO.FITR.434069

            /*  VR.ATO.FITR.434069
            If an 'Other deductions' amount is present, then an 'Other deductions - Code' must be present
    
            Legacy Rule Format:
            IF [FITR62] > 0 AND [FITR63] = NULLORBLANK  
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR62 > 0 AND ^FITR63 = BLANK
    
            Data Elements:
    
            ^FITR62 = RP:Expense.DeductibleOther.Amount
    
            ^FITR63 = RP:Expense.DeductibleOther.Code
            */
            assertion = (report.FITR62.GetValueOrDefault() > 0 && string.IsNullOrWhiteSpace(report.FITR63) == true);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.434069",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"If an 'Other deductions' amount is present, an 'Other deductions' code must be present",
                    Location = "/xbrli:xbrl/tns:Expense.DeductibleOther.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434069" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR62", Value = GetValueOrEmpty(report.FITR62) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR63", Value = GetValueOrEmpty(report.FITR63) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434069

            #region VR.ATO.FITR.434070

            /*  VR.ATO.FITR.434070
            Mandatory field not supplied
    
            Legacy Rule Format:
            IF [FITR65] = NULL
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR65 = NULL
    
            Data Elements:
    
            ^FITR65 = RP:Income.Taxable.Amount
            */
            assertion = (report.FITR65 == null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.001001",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Mandatory field not supplied",
                    Location = "/xbrli:xbrl/tns:Income.Taxable.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434070" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR65", Value = GetValueOrEmpty(report.FITR65) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434070

            #region VR.ATO.FITR.434072

            /*  VR.ATO.FITR.434072
            Taxable income/loss amount must equal income minus deductions
    
            Legacy Rule Format:
            IF [FITR65] <> NULL AND ([FITR65] <> ([FITR23] + [FITR24] + [FITR25] + [FITR26] + [FITR28] + [FITR29] + [FITR30] + [FITR32] + [FITR33] + [FITR34] + [FITR35] + [FITR36] + [FITR37] + [FITR38] + [FITR39] + [FITR40] + ([FITR41] + [FITR42] + [FITR43] - [FITR44] - [FITR45] - [FITR46]) + [FITR186] + [FITR47] + [FITR49] + [FITR50]) - [FITR200] - ([FITR52] + [FITR53] + [FITR54] + [FITR55] + [FITR56] + [FITR57] + [FITR59] + [FITR190] + [FITR188] + [FITR61] + [FITR187] + [FITR62] + [FITR64]))
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR65 <> NULL AND (^FITR65 <> (^FITR23 + ^FITR24 + ^FITR25 + ^FITR26 + ^FITR28 + ^FITR29 + ^FITR30 + ^FITR32 + ^FITR33 + ^FITR34 + ^FITR35 + ^FITR36 + ^FITR37 + ^FITR38 + ^FITR39 + ^FITR40 + (^FITR41 + ^FITR42 + ^FITR43 - ^FITR44 - ^FITR45 - ^FITR46) + ^FITR186 + ^FITR47 + ^FITR49 + ^FITR50) - ^FITR200 - (^FITR52 + ^FITR53 + ^FITR54 + ^FITR55 + ^FITR56 + ^FITR57 + ^FITR59 + ^FITR190 + ^FITR188 + ^FITR61 + ^FITR187 + ^FITR62 + ^FITR64))
    
            Data Elements:
    
            ^FITR65 = RP:Income.Taxable.Amount
    
            ^FITR23 = RP:Income.CapitalGainsNet.Amount
    
            ^FITR24 = RP:Income.RentLeasingHiringGross.Amount
    
            ^FITR25 = RP:Income.Interest.Gross.Amount
    
            ^FITR26 = RP:Income.ForestryManagedInvestmentScheme.Amount
    
            ^FITR28 = RP.JFOREIGN:Income.Net.Amount
    
            ^FITR29 = RP:IncomeTax.FrankingCredits.ReceivedFromNewZealandCompanies.Amount
    
            ^FITR30 = RP:Equity.InternationalDealings.ForeignFundTransfers.Amount
    
            ^FITR32 = RP:Remuneration.ABNNotQuotedPaymentGross.Amount
    
            ^FITR33 = RP:Income.PartnershipDistributionGross.Amount
    
            ^FITR34 = RP:Income.DividendsUnfranked.Amount
    
            ^FITR35 = RP:Income.DividendsFranked.Amount
    
            ^FITR36 = RP:IncomeTax.FrankingCredits.Amount
    
            ^FITR37 = RP:Income.TrustDistributionUnfranked.Amount
    
            ^FITR38 = RP:Income.TrustDistributionFranked.Amount
    
            ^FITR39 = RP:IncomeTax.FrankingCredits.FrankingCreditShareReceivedIndirectly.Amount
    
            ^FITR40 = RP:Income.TrustDistributionOther.Amount
    
            ^FITR41 = RP:SuperannuationContribution.EmployerAssessable.Amount
    
            ^FITR42 = RP:SuperannuationContribution.PersonalAssessable.Amount
    
            ^FITR43 = RP:SuperannuationContribution.EmployerContributionsTaxFileNumberNotQuoted.Amount
    
            ^FITR44 = RP:SuperannuationContribution.ContributionsExcludedByTrustee.Amount
    
            ^FITR45 = RP:SuperannuationContribution.ReducedContributions.Amount
    
            ^FITR46 = RP:SuperannuationContribution.LiabilityTransferToLifeInsuranceCompanyOrPooledSuperannuationTrustAssessable.Amount
    
            ^FITR186 = RP:Income.ForeignExchangeGains.Amount
    
            ^FITR47 = RP:Income.Other.Amount
    
            ^FITR49 = RP:Income.AssessableIncomeDueToFundTaxStatusChange.Amount
    
            ^FITR50 = RP:Income.NonArmsLengthOtherNet.Amount
    
            ^FITR200 = RP:TaxConcession.PensionIncomeExemptDeduction.Amount
    
            ^FITR52 = RP.JAUS:Expense.Interest.Amount
    
            ^FITR53 = RP.JFOREIGN:Expense.Interest.Amount
    
            ^FITR54 = RP:Remuneration.WagesandSalaries.Payments.Amount
    
            ^FITR55 = RP:Expense.CapitalWorksDeduction.Amount
    
            ^FITR56 = RP:Expense.DepreciationAllowableDeduction.Amount
    
            ^FITR57 = RP:Expense.PremiumsDeathOrDisability.Amount
    
            ^FITR59 = RP:Expense.Investment.Amount
    
            ^FITR190 = RP:Expense.Operating.ManagementExpense.Amount
    
            ^FITR188 = RP:Expense.Operating.AdministrativeExpense.Amount
    
            ^FITR61 = RP:Expense.ForestryManagedInvestmentSchemeDeduction.Amount
    
            ^FITR187 = RP:Income.ForeignExchangeLosses.Amount
    
            ^FITR62 = RP:Expense.DeductibleOther.Amount
    
            ^FITR64 = RP:IncomeTax.Deduction.TaxLossesDeducted.Amount
            */
            assertion = (report.FITR65 != null && report.FITR65.GetValueOrDefault() != report.FITR23.GetValueOrDefault() + report.FITR24.GetValueOrDefault() + report.FITR25.GetValueOrDefault() + report.FITR26.GetValueOrDefault() + report.FITR28.GetValueOrDefault() + report.FITR29.GetValueOrDefault() + report.FITR30.GetValueOrDefault() + report.FITR32.GetValueOrDefault() + report.FITR33.GetValueOrDefault() + report.FITR34.GetValueOrDefault() + report.FITR35.GetValueOrDefault() + report.FITR36.GetValueOrDefault() + report.FITR37.GetValueOrDefault() + report.FITR38.GetValueOrDefault() + report.FITR39.GetValueOrDefault() + report.FITR40.GetValueOrDefault() + (report.FITR41.GetValueOrDefault() + report.FITR42.GetValueOrDefault() + report.FITR43.GetValueOrDefault() - report.FITR44.GetValueOrDefault() - report.FITR45.GetValueOrDefault() - report.FITR46.GetValueOrDefault()) + report.FITR186.GetValueOrDefault() + report.FITR47.GetValueOrDefault() + report.FITR49.GetValueOrDefault() + report.FITR50.GetValueOrDefault() - report.FITR200.GetValueOrDefault() - (report.FITR52.GetValueOrDefault() + report.FITR53.GetValueOrDefault() + report.FITR54.GetValueOrDefault() + report.FITR55.GetValueOrDefault() + report.FITR56.GetValueOrDefault() + report.FITR57.GetValueOrDefault() + report.FITR59.GetValueOrDefault() + report.FITR190.GetValueOrDefault() + report.FITR188.GetValueOrDefault() + report.FITR61.GetValueOrDefault() + report.FITR187.GetValueOrDefault() + report.FITR62.GetValueOrDefault() + report.FITR64.GetValueOrDefault()));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434072",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Taxable income/loss amount must equal income minus deductions",
                    LongDescription = @"The Taxable income or loss amount must be equal to the sum of all income amounts minus the sum of all deductions amounts",
                    Location = "/xbrli:xbrl/tns:Income.Taxable.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434072" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR65", Value = GetValueOrEmpty(report.FITR65) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR23", Value = GetValueOrEmpty(report.FITR23) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR24", Value = GetValueOrEmpty(report.FITR24) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR25", Value = GetValueOrEmpty(report.FITR25) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR26", Value = GetValueOrEmpty(report.FITR26) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR28", Value = GetValueOrEmpty(report.FITR28) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR29", Value = GetValueOrEmpty(report.FITR29) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR30", Value = GetValueOrEmpty(report.FITR30) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR32", Value = GetValueOrEmpty(report.FITR32) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR33", Value = GetValueOrEmpty(report.FITR33) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR34", Value = GetValueOrEmpty(report.FITR34) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR35", Value = GetValueOrEmpty(report.FITR35) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR36", Value = GetValueOrEmpty(report.FITR36) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR37", Value = GetValueOrEmpty(report.FITR37) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR38", Value = GetValueOrEmpty(report.FITR38) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR39", Value = GetValueOrEmpty(report.FITR39) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR40", Value = GetValueOrEmpty(report.FITR40) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR41", Value = GetValueOrEmpty(report.FITR41) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR42", Value = GetValueOrEmpty(report.FITR42) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR43", Value = GetValueOrEmpty(report.FITR43) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR44", Value = GetValueOrEmpty(report.FITR44) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR45", Value = GetValueOrEmpty(report.FITR45) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR46", Value = GetValueOrEmpty(report.FITR46) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR186", Value = GetValueOrEmpty(report.FITR186) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR47", Value = GetValueOrEmpty(report.FITR47) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR49", Value = GetValueOrEmpty(report.FITR49) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR50", Value = GetValueOrEmpty(report.FITR50) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR200", Value = GetValueOrEmpty(report.FITR200) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR52", Value = GetValueOrEmpty(report.FITR52) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR53", Value = GetValueOrEmpty(report.FITR53) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR54", Value = GetValueOrEmpty(report.FITR54) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR55", Value = GetValueOrEmpty(report.FITR55) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR56", Value = GetValueOrEmpty(report.FITR56) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR57", Value = GetValueOrEmpty(report.FITR57) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR59", Value = GetValueOrEmpty(report.FITR59) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR190", Value = GetValueOrEmpty(report.FITR190) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR188", Value = GetValueOrEmpty(report.FITR188) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR61", Value = GetValueOrEmpty(report.FITR61) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR187", Value = GetValueOrEmpty(report.FITR187) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR62", Value = GetValueOrEmpty(report.FITR62) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR64", Value = GetValueOrEmpty(report.FITR64) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434072

            #region VR.ATO.FITR.434073

            /*  VR.ATO.FITR.434073
            Tax losses carried forward amount must be present
    
            Legacy Rule Format:
            IF [FITR65] < 0 AND ([FITR200] = 0 OR [FITR200] = NULL) AND ([FITR76] = 0 OR [FITR76] = NULL)  
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR65 < 0 AND (^FITR200 = 0 OR ^FITR200 = NULL) AND (^FITR76 = 0 OR ^FITR76 = NULL)
    
            Data Elements:
    
            ^FITR76 = RP.Closing:Tax.Losses.CarriedForward.LaterIncomeYearsTotal.Amount
    
            ^FITR65 = RP:Income.Taxable.Amount
    
            ^FITR200 = RP:TaxConcession.PensionIncomeExemptDeduction.Amount
            */
            assertion = (report.FITR65.GetValueOrDefault() < 0 && (report.FITR200 == 0 || report.FITR200 == null) && (report.FITR76 == 0 || report.FITR76 == null));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.434073",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Tax losses carried forward amount must be present",
                    LongDescription = @"If the Taxable income or loss amount is less than zero, there must be an 'Exempt current pension income' amount or a 'Tax losses carried forward to later income years' amount",
                    Location = "/xbrli:xbrl/tns:Tax.Losses.CarriedForward.LaterIncomeYearsTotal.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434073" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR65", Value = GetValueOrEmpty(report.FITR65) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR200", Value = GetValueOrEmpty(report.FITR200) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR76", Value = GetValueOrEmpty(report.FITR76) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434073

            #region VR.ATO.FITR.434074

            /*  VR.ATO.FITR.434074
            Gross tax must not be less than 15% of Taxable income
    
            Legacy Rule Format:
            IF [FITR65] > 0 AND [FITR66] <> NULL AND ([FITR66] < ([FITR65] * 0.15) - 1)  
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR65 > 0 AND ^FITR66 <> NULL AND (^FITR66 < (^FITR65 * 0.15) - 1)
    
            Data Elements:
    
            ^FITR66 = RP:IncomeTax.TaxableGross.Amount
    
            ^FITR65 = RP:Income.Taxable.Amount
            */
            assertion = (report.FITR65.GetValueOrDefault() > 0 && report.FITR66 != null && report.FITR66.GetValueOrDefault() < report.FITR65.GetValueOrDefault() * 0.15M - 1);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.434074",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Gross tax must not be less than 15% of Taxable income",
                    LongDescription = @"If the Taxable income or loss amount is greater than zero, then the Gross tax amount must not be less than 15% of the Taxable income or loss amount",
                    Location = "/xbrli:xbrl/tns:IncomeTax.TaxableGross.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434074" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR65", Value = GetValueOrEmpty(report.FITR65) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR66", Value = GetValueOrEmpty(report.FITR66) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434074

            #region VR.ATO.FITR.434075

            /*  VR.ATO.FITR.434075
            Gross tax must not be greater than zero
    
            Legacy Rule Format:
            IF ([FITR43] = 0 AND [FITR65] <= 0) AND [FITR66] > 0  
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^FITR43 = 0 AND ^FITR65 <= 0) AND ^FITR66 > 0
    
            Data Elements:
    
            ^FITR66 = RP:IncomeTax.TaxableGross.Amount
    
            ^FITR43 = RP:SuperannuationContribution.EmployerContributionsTaxFileNumberNotQuoted.Amount
    
            ^FITR65 = RP:Income.Taxable.Amount
            */
            assertion = (report.FITR43 == 0 && report.FITR65.GetValueOrDefault() <= 0 && report.FITR66.GetValueOrDefault() > 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.434075",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Gross tax must not be greater than zero",
                    LongDescription = @"If a 'No-TFN quoted contributions' amount is not present, and 'Taxable income or loss' amount is less than or equal to zero, then the Gross Tax amount cannot be greater than zero",
                    Location = "/xbrli:xbrl/tns:IncomeTax.TaxableGross.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434075" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR43", Value = GetValueOrEmpty(report.FITR43) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR65", Value = GetValueOrEmpty(report.FITR65) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR66", Value = GetValueOrEmpty(report.FITR66) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434075

            #region VR.ATO.FITR.434078

            /*  VR.ATO.FITR.434078
            Monetary field - max 13 chars, no sign, may have a decimal point
    
            Legacy Rule Format:
            IF [FITR67] <> NULL AND [FITR67] <> MONETARY(U,13,2)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR67, 'U', 13, 2)
    
            Data Elements:
    
            ^FITR67 = RP:Income.InternationalDealings.TaxOffset.Amount
            */
            assertion = NotMonetary(report.FITR67, @"U", 13, 2);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:Income.InternationalDealings.TaxOffset.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434078" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR67", Value = GetValueOrEmpty(report.FITR67) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434078

            #region VR.ATO.FITR.434079

            /*  VR.ATO.FITR.434079
            Credit: Interest on early payments must not be less than 50 cents
    
            Legacy Rule Format:
            IF [FITR70] <> NULL AND [FITR70] < 0.50  
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR70 <> NULL AND ^FITR70 < 0.50
    
            Data Elements:
    
            ^FITR70 = RP:Income.Interest.EarlyPaymentCredit.Amount
            */
            assertion = (report.FITR70 != null && report.FITR70.GetValueOrDefault() < 0.50M);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.434079",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Credit: Interest on early payments must not be less than 50 cents",
                    Location = "/xbrli:xbrl/tns:Income.Interest.EarlyPaymentCredit.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434079" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR70", Value = GetValueOrEmpty(report.FITR70) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434079

            #region VR.ATO.FITR.434081

            /*  VR.ATO.FITR.434081
            Distribution or Other income amounts must be present
    
            Legacy Rule Format:
            IF [FITR71] > 0 AND ([FITR33] = NULL AND [FITR40] = NULL AND [FITR47] = NULL) 
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR71 > 0 AND (^FITR33 = NULL AND ^FITR40 = NULL AND ^FITR47 = NULL)
    
            Data Elements:
    
            ^FITR33 = RP:Income.PartnershipDistributionGross.Amount
    
            ^FITR71 = RP:IncomeTax.PayAsYouGoWithholding.CreditForAmountsWithheldFromForeignResidents.Amount
    
            ^FITR40 = RP:Income.TrustDistributionOther.Amount
    
            ^FITR47 = RP:Income.Other.Amount
            */
            assertion = (report.FITR71.GetValueOrDefault() > 0 && (report.FITR33 == null && report.FITR40 == null && report.FITR47 == null));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434081",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Distribution or Other income amounts must be present",
                    LongDescription = @"If 'Credit: foreign resident withholding' is greater than zero, then 'Gross distributions from partnerships', 'Trust distribution other amounts' and/or 'Other income' must be present",
                    Location = "/xbrli:xbrl/tns:Income.PartnershipDistributionGross.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434081" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR71", Value = GetValueOrEmpty(report.FITR71) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR33", Value = GetValueOrEmpty(report.FITR33) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR40", Value = GetValueOrEmpty(report.FITR40) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR47", Value = GetValueOrEmpty(report.FITR47) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434081

            #region VR.ATO.FITR.434082

            /*  VR.ATO.FITR.434082
            Credit: ABN/TFN not quoted amount must be present
    
            Legacy Rule Format:
            IF [FITR32] > 0 AND [FITR72] = NULL  
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR32 > 0 AND ^FITR72 = NULL
    
            Data Elements:
    
            ^FITR72 = RP:IncomeTax.PayAsYouGoWithholding.CreditForAmountsWithheldTFNNotQuotedAndABNNotQuoted.Amount
    
            ^FITR32 = RP:Remuneration.ABNNotQuotedPaymentGross.Amount
            */
            assertion = (report.FITR32.GetValueOrDefault() > 0 && report.FITR72 == null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.434082",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Credit: ABN/TFN not quoted amount must be present",
                    LongDescription = @"If 'Gross payments where ABN not quoted' is greater than zero, then 'Credit: TFN/ABN not quoted' amount must be present",
                    Location = "/xbrli:xbrl/tns:IncomeTax.PayAsYouGoWithholding.CreditForAmountsWithheldTFNNotQuotedAndABNNotQuoted.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434082" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR32", Value = GetValueOrEmpty(report.FITR32) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR72", Value = GetValueOrEmpty(report.FITR72) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434082

            #region VR.ATO.FITR.434083

            /*  VR.ATO.FITR.434083
            Mandatory field not supplied
    
            Legacy Rule Format:
            IF [FITR75] = NULL 
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR75 = NULL
    
            Data Elements:
    
            ^FITR75 = RP:IncomeTax.PayableOrRefundableTotal.Amount
            */
            assertion = (report.FITR75 == null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.001001",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Mandatory field not supplied",
                    Location = "/xbrli:xbrl/tns:IncomeTax.PayableOrRefundableTotal.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434083" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR75", Value = GetValueOrEmpty(report.FITR75) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434083

            #region VR.ATO.FITR.434091

            /*  VR.ATO.FITR.434091
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR78] <> NULL AND [FITR78] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR78, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR78 = RP.CFCLC:Income.InternationalDealings.Attributed.Amount
            */
            assertion = NotMonetary(report.FITR78, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:Income.InternationalDealings.Attributed.Amount[@contextRef='" + report.RPCFCLCId + "']",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434091" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR78", Value = GetValueOrEmpty(report.FITR78) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434091

            #region VR.ATO.FITR.434096

            /*  VR.ATO.FITR.434096
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR80] <> NULL AND [FITR80] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR80, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR80 = RP.CFCUC:Income.InternationalDealings.Attributed.Amount
            */
            assertion = NotMonetary(report.FITR80, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:Income.InternationalDealings.Attributed.Amount[@contextRef='" + report.RPCFCUCId + "']",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434096" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR80", Value = GetValueOrEmpty(report.FITR80) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434096

            #region VR.ATO.FITR.434097

            /*  VR.ATO.FITR.434097
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR59] <> NULL AND [FITR59] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR59, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR59 = RP:Expense.Investment.Amount
            */
            assertion = NotMonetary(report.FITR59, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:Expense.Investment.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434097" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR59", Value = GetValueOrEmpty(report.FITR59) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434097

            #region VR.ATO.FITR.434098

            /*  VR.ATO.FITR.434098
            If the fund has been wound-up during the year then 'Net assets available to pay benefits' must not be present.
    
            Legacy Rule Format:
            IF [FITR19] = TRUE AND [FITR83] <> NULL  
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR19 = TRUE AND ^FITR83 <> NULL
    
            Data Elements:
    
            ^FITR83 = RP:Equity.NetAssetsCalculated.Amount
    
            ^FITR19 = RP:OrganisationDetails.OperationCease.Indicator
            */
            assertion = (report.FITR19 == true && report.FITR83 != null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434098",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"If the fund has been wound-up during the year then 'Net assets available to pay benefits' must not be present.",
                    Location = "/xbrli:xbrl/tns:Equity.NetAssetsCalculated.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434098" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR19", Value = GetValueOrEmpty(report.FITR19) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR83", Value = GetValueOrEmpty(report.FITR83) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434098

            #region VR.ATO.FITR.434099

            /*  VR.ATO.FITR.434099
            Mandatory field not supplied
    
            Legacy Rule Format:
            IF [FITR84] = NULL 
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR84 = NULL
    
            Data Elements:
    
            ^FITR84 = RP:IncomeTax.Superannuation.TransferredAssessable.Indicator
            */
            assertion = (report.FITR84 == null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.001001",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Mandatory field not supplied",
                    Location = "/xbrli:xbrl/tns:IncomeTax.Superannuation.TransferredAssessable.Indicator",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434099" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR84", Value = GetValueOrEmpty(report.FITR84) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434099

            #region VR.ATO.FITR.434100

            /*  VR.ATO.FITR.434100
            Transfer of liabilities to LIC transferee details must be present
    
            Legacy Rule Format:
            IF [FITR84] = TRUE AND (COUNT(TUPLE(fitr.0008.lodge.req.xx.xx:Transferee)) = 0) 
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR84 = TRUE AND (Count(^Transferee) = 0)
    
            Data Elements:
    
            ^FITR84 = RP:IncomeTax.Superannuation.TransferredAssessable.Indicator
    
            ^Transferee = RP:Transferee
            */
            assertion = (report.FITR84 == true && Count(report.TransfereeCollectionCount) == 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434280",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Transfer of liabilities to LIC transferee details must be present",
                    LongDescription = @"If the answer to 'Has the fund or trust,... transferred assessable income contributions....' is 'yes' (true), then the name, ABN and amounts for at least one transferee must be present. Where there are more than two transferees, provide details of the two transferees that have the greatest market value",
                    Location = "/xbrli:xbrl/tns:IncomeTax.Superannuation.TransferredAssessable.Indicator",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434100" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR84", Value = GetValueOrEmpty(report.FITR84) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434100

            #region VR.ATO.FITR.434102

            /*  VR.ATO.FITR.434102
            If fund has not transferred contributions, Transferee details must not be present
    
            Legacy Rule Format:
            IF [FITR84] = FALSE AND (COUNT(TUPLE(fitr.0008.lodge.req.xx.xx:Transferee)) > 0) 
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR84 = FALSE AND (Count(^Transferee) > 0)
    
            Data Elements:
    
            ^Transferee = RP:Transferee
    
            ^FITR84 = RP:IncomeTax.Superannuation.TransferredAssessable.Indicator
            */
            assertion = (report.FITR84 == false && Count(report.TransfereeCollectionCount) > 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434102",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"If fund has not transferred contributions, Transferee details must not be present",
                    LongDescription = @"If the answer to 'Has the fund or trust,... transferred assessable income contributions ...' is 'no' (false), then the Name, ABN and Amounts for a trusferee must not be present",
                    Location = "/xbrli:xbrl/tns:Transferee" + OccurrenceIndex(report.TransfereeCollection == null ? 0 : (report.TransfereeCollection.Count() == 0 ? 0 : report.TransfereeCollection.Last().OccurrenceIndex)) + "/tns:OrganisationNameDetails/tns:OrganisationNameDetails.OrganisationalName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434102" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR84", Value = GetValueOrEmpty(report.FITR84) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434102

            #region VR.ATO.FITR.434103

            /*  VR.ATO.FITR.434103
            Answer to Transfer of liabilities question must be 'yes'
    
            Legacy Rule Format:
            IF [FITR46] > 0 AND [FITR84] = FALSE 
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR46 > 0 AND ^FITR84 = FALSE
    
            Data Elements:
    
            ^FITR84 = RP:IncomeTax.Superannuation.TransferredAssessable.Indicator
    
            ^FITR46 = RP:SuperannuationContribution.LiabilityTransferToLifeInsuranceCompanyOrPooledSuperannuationTrustAssessable.Amount
            */
            assertion = (report.FITR46.GetValueOrDefault() > 0 && report.FITR84 == false);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434103",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Answer to Transfer of liabilities question must be 'yes'",
                    LongDescription = @"If an amount is present at 'Transfer of liability to LIC or PST', then the answer to the question 'Has the fund or trust, ... transferred assessable income contributions ...' must be 'yes' (true).",
                    Location = "/xbrli:xbrl/tns:IncomeTax.Superannuation.TransferredAssessable.Indicator",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434103" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR46", Value = GetValueOrEmpty(report.FITR46) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR84", Value = GetValueOrEmpty(report.FITR84) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434103

            #region VR.ATO.FITR.434105

            /*  VR.ATO.FITR.434105
            Mandatory field not supplied
    
            Legacy Rule Format:
            IF [FITR90] = NULL 
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR90 = NULL
    
            Data Elements:
    
            ^FITR90 = RP:InternationalDealings.RelatedPartiesTransactions.Indicator
            */
            assertion = (report.FITR90 == null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.001001",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Mandatory field not supplied",
                    Location = "/xbrli:xbrl/tns:InternationalDealings.RelatedPartiesTransactions.Indicator",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434105" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR90", Value = GetValueOrEmpty(report.FITR90) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434105

            #region VR.ATO.FITR.434106

            /*  VR.ATO.FITR.434106
            Mandatory field not supplied
    
            Legacy Rule Format:
            IF [FITR91] = NULL
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR91 = NULL
    
            Data Elements:
    
            ^FITR91 = RP:InternationalDealings.RelatedPartiesTransactionsExcessAggregateValue.Indicator
            */
            assertion = (report.FITR91 == null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.001001",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Mandatory field not supplied",
                    Location = "/xbrli:xbrl/tns:InternationalDealings.RelatedPartiesTransactionsExcessAggregateValue.Indicator",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434106" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR91", Value = GetValueOrEmpty(report.FITR91) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434106

            #region VR.ATO.FITR.434110

            /*  VR.ATO.FITR.434110
            Answer to international related party dealings question must be 'yes'
    
            Legacy Rule Format:
            IF [FITR91] = TRUE AND [FITR90] = FALSE  
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR91 = TRUE AND ^FITR90 = FALSE
    
            Data Elements:
    
            ^FITR91 = RP:InternationalDealings.RelatedPartiesTransactionsExcessAggregateValue.Indicator
    
            ^FITR90 = RP:InternationalDealings.RelatedPartiesTransactions.Indicator
            */
            assertion = (report.FITR91 == true && report.FITR90 == false);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434249",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Answer to international related party dealings question must be true.",
                    LongDescription = @"If the answer to the question 'Was the aggregate amount of transactions...greater than $2 million', is 'yes' (true), then the answer to the question 'Did the fund have any transactions or dealing with international parties...', must be yes' (true)",
                    Location = "/xbrli:xbrl/tns:InternationalDealings.RelatedPartiesTransactionsExcessAggregateValue.Indicator",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434110" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR91", Value = GetValueOrEmpty(report.FITR91) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR90", Value = GetValueOrEmpty(report.FITR90) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434110

            #region VR.ATO.FITR.434114

            /*  VR.ATO.FITR.434114
            Mandatory field not supplied
    
            Legacy Rule Format:
            IF [FITR92] = NULL 
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR92 = NULL
    
            Data Elements:
    
            ^FITR92 = RP:InternationalDealings.DirectOrIndirectOverseasInterest.Indicator
            */
            assertion = (report.FITR92 == null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.001001",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Mandatory field not supplied",
                    Location = "/xbrli:xbrl/tns:InternationalDealings.DirectOrIndirectOverseasInterest.Indicator",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434114" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR92", Value = GetValueOrEmpty(report.FITR92) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434114

            #region VR.ATO.FITR.434119

            /*  VR.ATO.FITR.434119
            Answer to Overseas interests question must be 'yes' if Attributed foreign income is present
    
            Legacy Rule Format:
            IF ([FITR78] > 0 OR [FITR80] > 0) AND [FITR92] = FALSE
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^FITR78 > 0 OR ^FITR80 > 0) AND ^FITR92 = FALSE
    
            Data Elements:
    
            ^FITR92 = RP:InternationalDealings.DirectOrIndirectOverseasInterest.Indicator
    
            ^FITR78 = RP.CFCLC:Income.InternationalDealings.Attributed.Amount
    
            ^FITR80 = RP.CFCUC:Income.InternationalDealings.Attributed.Amount
            */
            assertion = ((report.FITR78.GetValueOrDefault() > 0 || report.FITR80.GetValueOrDefault() > 0) && report.FITR92 == false);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434119",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Answer to Overseas interests question must be 'yes' if Attributed foreign income is present",
                    LongDescription = @"If there is any Attributed foreign income amount present - 'Listed country', 'Section 404 country', 'Unlisted country', or 'FIF/FLP income' - then the answer to the question 'Did the fund have an overseas branch...', must be 'yes' (true)",
                    Location = "/xbrli:xbrl/tns:InternationalDealings.DirectOrIndirectOverseasInterest.Indicator",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434119" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR78", Value = GetValueOrEmpty(report.FITR78) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR80", Value = GetValueOrEmpty(report.FITR80) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR92", Value = GetValueOrEmpty(report.FITR92) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434119

            #region VR.ATO.FITR.434120

            /*  VR.ATO.FITR.434120
            Mandatory field not supplied
    
            Legacy Rule Format:
            IF [FITR93] = NULL 
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR93 = NULL
    
            Data Elements:
    
            ^FITR93 = RP:InternationalDealings.ForeignSourceIncomeExcessAggregateValue.Indicator
            */
            assertion = (report.FITR93 == null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.001001",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Mandatory field not supplied",
                    Location = "/xbrli:xbrl/tns:InternationalDealings.ForeignSourceIncomeExcessAggregateValue.Indicator",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434120" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR93", Value = GetValueOrEmpty(report.FITR93) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434120

            #region VR.ATO.FITR.434121

            /*  VR.ATO.FITR.434121
            Mandatory field not supplied
    
            Legacy Rule Format:
            IF [FITR94] = NULL
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR94 = NULL
    
            Data Elements:
    
            ^FITR94 = RP:InternationalDealings.TransactionsWithTaxHavenCountries.Indicator
            */
            assertion = (report.FITR94 == null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.001001",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Mandatory field not supplied",
                    Location = "/xbrli:xbrl/tns:InternationalDealings.TransactionsWithTaxHavenCountries.Indicator",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434121" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR94", Value = GetValueOrEmpty(report.FITR94) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434121

            #region VR.ATO.FITR.434122

            /*  VR.ATO.FITR.434122
            Mandatory field not supplied
    
            Legacy Rule Format:
            IF [FITR96] = NULL 
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR96 = NULL
    
            Data Elements:
    
            ^FITR96 = RP:IncomeTax.Deduction.ClaimedDeathOrDisabilityDeductionWhichRequiresCertificate.Indicator
            */
            assertion = (report.FITR96 == null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.001001",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Mandatory field not supplied",
                    Location = "/xbrli:xbrl/tns:IncomeTax.Deduction.ClaimedDeathOrDisabilityDeductionWhichRequiresCertificate.Indicator",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434122" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR96", Value = GetValueOrEmpty(report.FITR96) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434122

            #region VR.ATO.FITR.434123

            /*  VR.ATO.FITR.434123
            Answer to Death or diability -  'relevant certificate' question must be present
    
            Legacy Rule Format:
            IF [FITR96] = TRUE AND [FITR97] = NULL 
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR96 = TRUE AND ^FITR97 = NULL
    
            Data Elements:
    
            ^FITR97 = RP:IncomeTax.Deduction.ClaimedDeathOrDisabilityDeductionCertificateObtained.Indicator
    
            ^FITR96 = RP:IncomeTax.Deduction.ClaimedDeathOrDisabilityDeductionWhichRequiresCertificate.Indicator
            */
            assertion = (report.FITR96 == true && report.FITR97 == null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434123",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Answer to Death or diability -  'relevant certificate' question must be present",
                    LongDescription = @"If the answer to the question 'Is the fund or trust is claiming a deduction for premiums for death or disability....' is 'yes' (true), then an answer to the question 'If yes, has the fund or trust obtained the relevant certificate', must be present",
                    Location = "/xbrli:xbrl/tns:IncomeTax.Deduction.ClaimedDeathOrDisabilityDeductionCertificateObtained.Indicator",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434123" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR96", Value = GetValueOrEmpty(report.FITR96) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR97", Value = GetValueOrEmpty(report.FITR97) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434123

            #region VR.ATO.FITR.434124

            /*  VR.ATO.FITR.434124
            Answer to Death or diability -  'relevant certificate' question must NOT be present
    
            Legacy Rule Format:
            IF [FITR96] = FALSE AND [FITR97] <> NULL
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR96 = FALSE AND ^FITR97 <> NULL
    
            Data Elements:
    
            ^FITR97 = RP:IncomeTax.Deduction.ClaimedDeathOrDisabilityDeductionCertificateObtained.Indicator
    
            ^FITR96 = RP:IncomeTax.Deduction.ClaimedDeathOrDisabilityDeductionWhichRequiresCertificate.Indicator
            */
            assertion = (report.FITR96 == false && report.FITR97 != null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434124",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Answer to Death or diability -  'relevant certificate' question must NOT be present",
                    LongDescription = @"If the answer to the question 'Is the fund or trust is claiming a deduction for premiums for death or disability....' is 'no' (false), then an answer to the question 'If yes, has the fund or trust obtained the relevant certificate', must NOT be present",
                    Location = "/xbrli:xbrl/tns:IncomeTax.Deduction.ClaimedDeathOrDisabilityDeductionCertificateObtained.Indicator",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434124" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR96", Value = GetValueOrEmpty(report.FITR96) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR97", Value = GetValueOrEmpty(report.FITR97) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434124

            #region VR.ATO.FITR.434125

            /*  VR.ATO.FITR.434125
            Mandatory field not supplied
    
            Legacy Rule Format:
            IF [FITR98] = NULL 
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR98 = NULL
    
            Data Elements:
    
            ^FITR98 = RP:SuperannuationPayment.PaymentsToContributingEmployersAndAssociates.Indicator
            */
            assertion = (report.FITR98 == null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.001001",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Mandatory field not supplied",
                    Location = "/xbrli:xbrl/tns:SuperannuationPayment.PaymentsToContributingEmployersAndAssociates.Indicator",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434125" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR98", Value = GetValueOrEmpty(report.FITR98) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434125

            #region VR.ATO.FITR.434133

            /*  VR.ATO.FITR.434133
            Mandatory field not supplied
    
            Legacy Rule Format:
            IF [FITR103] = NULLORBLANK 
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR103 = BLANK
    
            Data Elements:
    
            ^FITR103 = RP:PersonNameDetails:PersonNameDetails.FamilyName.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
            */
            assertion = (string.IsNullOrWhiteSpace(report.FITR103) == true);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.001001",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Mandatory field not supplied",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_RPOccurrenceIndex) + "/tns:PersonNameDetails.FamilyName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434133" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR103", Value = report.FITR103 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434133

            #region VR.ATO.FITR.434136

            /*  VR.ATO.FITR.434136
            Mandatory field not supplied
    
            Legacy Rule Format:
            IF [FITR108] = NULL 
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR108 = NULL
    
            Data Elements:
    
            ^FITR108 = RP:ElectronicContactTelephone:ElectronicContact.Telephone.Minimal.Number WHERE [ElectronicContact.Telephone.Usage.Code='03']
            */
            assertion = (report.FITR108 == null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.001001",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Mandatory field not supplied",
                    Location = "/xbrli:xbrl/tns:ElectronicContactTelephone" + OccurrenceIndex(report.ElectronicContactTelephoneCollection_03_RPOccurrenceIndex) + "/tns:ElectronicContact.Telephone.Minimal.Number",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434136" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR108", Value = report.FITR108 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434136

            #region VR.ATO.FITR.434137

            /*  VR.ATO.FITR.434137
            Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters
    
            Legacy Rule Format:
            IF [FITR119] <> NULLORBLANK AND (LENGTH([FITR119]) > 12)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Length(^FITR119) > 12
    
            Data Elements:
    
            ^FITR119 = INT:Identifiers.TaxAgentClientReference.Text
            */
            assertion = (Length(report.FITR119) > 12);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.001011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field value supplied exceeds maximum field length",
                    Location = "/xbrli:xbrl/tns:Identifiers.TaxAgentClientReference.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434137" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR119", Value = GetValueOrEmpty(report.FITR119) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434137

            #region VR.ATO.FITR.434143

            /*  VR.ATO.FITR.434143
            Mandatory field not supplied
    
            Legacy Rule Format:
            IF [FITR66] = NULL
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR66 = NULL
    
            Data Elements:
    
            ^FITR66 = RP:IncomeTax.TaxableGross.Amount
            */
            assertion = (report.FITR66 == null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.001001",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Mandatory field not supplied",
                    Location = "/xbrli:xbrl/tns:IncomeTax.TaxableGross.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434143" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR66", Value = GetValueOrEmpty(report.FITR66) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434143

            #region VR.ATO.FITR.434145

            /*  VR.ATO.FITR.434145
            Date wound up must be within the year of assessment if the fund wound up during the income year
    
            Legacy Rule Format:
            IF [FITR20] <> NULL AND ([FITR20] < CONVERT_TO_DATE(1, 7, [FITR1]-1) OR [FITR20] > CONVERT_TO_DATE(30, 6, [FITR1]))
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR20 <> NULL AND (^FITR20 < ConvertToDate(1, 7, ^FITR1-1) OR ^FITR20 > ConvertToDate(30, 6, ^FITR1))
    
            Data Elements:
    
            ^FITR20 = RP:OrganisationDetails.OrganisationActivityStatusEnd.Date
    
            ^FITR1 = RP:Report.TargetFinancial.Year
            */
            assertion = (report.FITR20 != null && (report.FITR20.GetValueOrDefault() < ConvertToDate(1, 7, report.FITR1.GetValueOrDefault() - 1) || report.FITR20.GetValueOrDefault() > ConvertToDate(30, 6, report.FITR1.GetValueOrDefault())));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.434145",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Fund wound up date must be within year of assessment.",
                    LongDescription = @"The date the fund wound up must fall within the year of assessment.  For example, if the year of the return being lodged is 2010, then the date must not be before 1 July 2009 and must not be after 30 June 2010.",
                    Location = "/xbrli:xbrl/tns:OrganisationDetails.OrganisationActivityStatusEnd.Date",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434145" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR20", Value = GetValueOrEmptyDate(report.FITR20) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR1", Value = GetValueOrEmpty(report.FITR1) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434145

            #region VR.ATO.FITR.434146

            /*  VR.ATO.FITR.434146
            There must be one instance of the current Name of the superannuation fund or trust. This name must be submitted with a Currency Code 'C' and a Type of 'MN'
    
            Legacy Rule Format:
            IF COUNT(((RP:pyde.xx.xx:OrganisationNameDetails.OrganisationalNameType.Code = "MN") AND (RP:pyde.xx.xx:OrganisationNameDetails.OrganisationalNameCurrency.Code = "C") AND (RP:pyde.xx.xx:OrganisationNameDetails.OrganisationalName.Text  <> NULLORBLANK)) IN TUPLE xbrli\orgname2.xx.xx:OrganisationNameDetails) <> 1
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Not(Count(^OrganisationNameDetails) = 1 AND ^FITR6 <> BLANK)
    
            Data Elements:
    
            ^FITR6 = RP:OrganisationNameDetails:OrganisationNameDetails.OrganisationalName.Text WHERE [OrganisationNameDetails.OrganisationalNameType.Code='MN'].[OrganisationNameDetails.Currency.Code='C']
    
            ^OrganisationNameDetails = RP:OrganisationNameDetails
            */
            assertion = !((Count(report.OrganisationNameDetailsCollection_MN_C_RPCount) == 1 && string.IsNullOrWhiteSpace(report.FITR6) != true));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434146",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Name of superannuation fund or trust must be present",
                    LongDescription = @"There must be one instance of the current Name of the superannuation fund or trust. This name must be submitted with a Currency Code 'C' (current) and a Type of 'MN' (main name)",
                    Location = "/xbrli:xbrl/tns:OrganisationNameDetails" + OccurrenceIndex(report.OrganisationNameDetailsCollection_MN_C_RPLastOccurrenceIndex) + "/tns:OrganisationNameDetails.OrganisationalName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434146" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR6", Value = report.FITR6 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434146

            #region VR.ATO.FITR.434148

            /*  VR.ATO.FITR.434148
            The phone number of the intermediary must be present
    
            Legacy Rule Format:
            IF CONTEXT(INT) <> NULLORBLANK AND [FITR118] = NULL
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^Context <> BLANK AND ^FITR118 = NULL
    
            Data Elements:
    
            ^FITR118 = INT:ElectronicContactTelephone:ElectronicContact.Telephone.Minimal.Number WHERE [ElectronicContact.Telephone.Usage.Code='03']
    
            ^Context = INT
            */
            assertion = (report.INTExists != false && report.FITR118 == null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434148",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"The phone number of the intermediary must be present",
                    Location = "/xbrli:xbrl/tns:ElectronicContactTelephone" + OccurrenceIndex(report.ElectronicContactTelephoneCollection_03_INTOccurrenceIndex) + "/tns:ElectronicContact.Telephone.Minimal.Number",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434148" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR118", Value = report.FITR118 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434148

            #region VR.ATO.FITR.434163

            /*  VR.ATO.FITR.434163
            Incorrect number of XBRL contexts
    
            Legacy Rule Format:
            IF COUNT(CONTEXT(RP.CFCLC)) > 1 
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Count(^Context) > 1
    
            Data Elements:
    
            ^Context = RP:RP.CFCLC
            */
            assertion = (Count(report.RPCFCLCCount) > 1);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434163",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Invalid context. The number of Reporting party: Listed Country contexts must not exceed 1",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPCFCLCLastOccurrenceIndex),
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434163" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434163

            #region VR.ATO.FITR.434166

            /*  VR.ATO.FITR.434166
            Incorrect number of XBRL contexts
    
            Legacy Rule Format:
            IF COUNT(CONTEXT(RP.CFCUC)) > 1 
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Count(^Context) > 1
    
            Data Elements:
    
            ^Context = RP:RP.CFCUC
            */
            assertion = (Count(report.RPCFCUCCount) > 1);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434166",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Invalid context. The number of Reporting party: Unlisted Country contexts must not exceed 1",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPCFCUCLastOccurrenceIndex),
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434166" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434166

            #region VR.ATO.FITR.434168

            /*  VR.ATO.FITR.434168
            The number of transferees - as determined by the number of Transferee ABN's present - must not exceed 99
    
            Legacy Rule Format:
            IF COUNT([FITR86]) > 99
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Count(^FITR86) > 99
    
            Data Elements:
    
            ^FITR86 = RP:Transferee:Identifiers.AustralianBusinessNumber.Identifier
            */
            assertion = (Count(report.TransfereeCollection == null ? new List<string>() : report.TransfereeCollection.Select(f => f.FITR86)) > 99);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434168",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"The total number of transferees must not exceed 99",
                    Location = "/xbrli:xbrl/tns:Transferee/tns:Identifiers.AustralianBusinessNumber.Identifier",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434168" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434168

            #region VR.ATO.FITR.434169

            /*  VR.ATO.FITR.434169
            Monetary field - max 13 chars, no sign, may have a decimal point
    
            Legacy Rule Format:
            IF [FITR72] <> NULL AND [FITR72] <> MONETARY(U,13,2)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR72, 'U', 13, 2)
    
            Data Elements:
    
            ^FITR72 = RP:IncomeTax.PayAsYouGoWithholding.CreditForAmountsWithheldTFNNotQuotedAndABNNotQuoted.Amount
            */
            assertion = NotMonetary(report.FITR72, @"U", 13, 2);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:IncomeTax.PayAsYouGoWithholding.CreditForAmountsWithheldTFNNotQuotedAndABNNotQuoted.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434169" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR72", Value = GetValueOrEmpty(report.FITR72) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434169

            #region VR.ATO.FITR.434170

            /*  VR.ATO.FITR.434170
            Monetary field - max 13 chars, no sign, may have a decimal point
    
            Legacy Rule Format:
            IF [FITR73] <> NULL AND [FITR73] <> MONETARY(U,13,2)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR73, 'U', 13, 2)
    
            Data Elements:
    
            ^FITR73 = RP:IncomeTax.FrankingCredits.Refundable.Amount
            */
            assertion = NotMonetary(report.FITR73, @"U", 13, 2);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:IncomeTax.FrankingCredits.Refundable.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434170" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR73", Value = GetValueOrEmpty(report.FITR73) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434170

            #region VR.ATO.FITR.434171

            /*  VR.ATO.FITR.434171
            Monetary field - max 13 chars, no sign, may have a decimal point
    
            Legacy Rule Format:
            IF [FITR74] <> NULL AND [FITR74] <> MONETARY(U,13,2)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR74, 'U', 13, 2)
    
            Data Elements:
    
            ^FITR74 = RP:IncomeTax.LiabilityInstalmentsTotal.Amount
            */
            assertion = NotMonetary(report.FITR74, @"U", 13, 2);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:IncomeTax.LiabilityInstalmentsTotal.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434171" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR74", Value = GetValueOrEmpty(report.FITR74) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434171

            #region VR.ATO.FITR.434172

            /*  VR.ATO.FITR.434172
            Monetary field - max 13 chars, no sign, may have a decimal point
    
            Legacy Rule Format:
            IF [FITR75] <> NULL AND [FITR75] <> MONETARY(S,13,2)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR75, 'S', 13, 2)
    
            Data Elements:
    
            ^FITR75 = RP:IncomeTax.PayableOrRefundableTotal.Amount
            */
            assertion = NotMonetary(report.FITR75, @"S", 13, 2);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:IncomeTax.PayableOrRefundableTotal.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434172" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR75", Value = GetValueOrEmpty(report.FITR75) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434172

            #region VR.ATO.FITR.434173

            /*  VR.ATO.FITR.434173
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR76] <> NULL AND [FITR76] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR76, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR76 = RP.Closing:Tax.Losses.CarriedForward.LaterIncomeYearsTotal.Amount
            */
            assertion = NotMonetary(report.FITR76, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:Tax.Losses.CarriedForward.LaterIncomeYearsTotal.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434173" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR76", Value = GetValueOrEmpty(report.FITR76) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434173

            #region VR.ATO.FITR.434177

            /*  VR.ATO.FITR.434177
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR83] <> NULL AND [FITR83] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR83, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR83 = RP:Equity.NetAssetsCalculated.Amount
            */
            assertion = NotMonetary(report.FITR83, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:Equity.NetAssetsCalculated.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434177" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR83", Value = GetValueOrEmpty(report.FITR83) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434177

            #region VR.ATO.FITR.434180

            /*  VR.ATO.FITR.434180
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR89] <> NULL AND [FITR89] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR89, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR89 = RP:IncomeTax.Superannuation.MarketValueTotal.Amount
            */
            assertion = NotMonetary(report.FITR89, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:IncomeTax.Superannuation.MarketValueTotal.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434180" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR89", Value = GetValueOrEmpty(report.FITR89) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434180

            #region VR.ATO.FITR.434182

            /*  VR.ATO.FITR.434182
            Monetary field - max 13 chars, no sign, may have a decimal point
    
            Legacy Rule Format:
            IF [FITR131] <> NULL AND [FITR131] <> MONETARY(U,13,2)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR131, 'U', 13, 2)
    
            Data Elements:
    
            ^FITR131 = RP:IncomeTax.TaxOffsetTFNNotQuoted.Amount
            */
            assertion = NotMonetary(report.FITR131, @"U", 13, 2);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:IncomeTax.TaxOffsetTFNNotQuoted.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434182" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR131", Value = GetValueOrEmpty(report.FITR131) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434182

            #region VR.ATO.FITR.434183

            /*  VR.ATO.FITR.434183
            Monetary field - max 13 chars, no sign, may have a decimal point
    
            Legacy Rule Format:
            IF [FITR132] <> NULL AND [FITR132] <> MONETARY(U,13,2)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR132, 'U', 13, 2)
    
            Data Elements:
    
            ^FITR132 = RP:IncomeTax.TaxOffsetCalculatedInterestTFNNotQuoted.Amount
            */
            assertion = NotMonetary(report.FITR132, @"U", 13, 2);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:IncomeTax.TaxOffsetCalculatedInterestTFNNotQuoted.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434183" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR132", Value = GetValueOrEmpty(report.FITR132) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434183

            #region VR.ATO.FITR.434184

            /*  VR.ATO.FITR.434184
            Monetary field - max 13 chars, no sign, may have a decimal point
    
            Legacy Rule Format:
            IF [FITR133] <> NULL AND [FITR133] <> MONETARY(U,13,2)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR133, 'U', 13, 2)
    
            Data Elements:
    
            ^FITR133 = RP:IncomeTax.TaxOffsetEntitlementForNationalRentalAffordabilityScheme.Amount
            */
            assertion = NotMonetary(report.FITR133, @"U", 13, 2);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:IncomeTax.TaxOffsetEntitlementForNationalRentalAffordabilityScheme.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434184" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR133", Value = GetValueOrEmpty(report.FITR133) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434184

            #region VR.ATO.FITR.434185

            /*  VR.ATO.FITR.434185
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR134] <> NULL AND [FITR134] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR134, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR134 = RP.TOFA:Income.GainsTotal.Amount
            */
            assertion = NotMonetary(report.FITR134, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:Income.GainsTotal.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434185" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR134", Value = GetValueOrEmpty(report.FITR134) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434185

            #region VR.ATO.FITR.434186

            /*  VR.ATO.FITR.434186
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR135] <> NULL AND [FITR135] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR135, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR135 = RP.TOFA:Expense.LossesTotal.Amount
            */
            assertion = NotMonetary(report.FITR135, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:Expense.LossesTotal.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434186" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR135", Value = GetValueOrEmpty(report.FITR135) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434186

            #region VR.ATO.FITR.434188

            /*  VR.ATO.FITR.434188
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR23] <> NULL AND [FITR23] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR23, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR23 = RP:Income.CapitalGainsNet.Amount
            */
            assertion = NotMonetary(report.FITR23, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:Income.CapitalGainsNet.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434188" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR23", Value = GetValueOrEmpty(report.FITR23) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434188

            #region VR.ATO.FITR.434189

            /*  VR.ATO.FITR.434189
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR24] <> NULL AND [FITR24] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR24, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR24 = RP:Income.RentLeasingHiringGross.Amount
            */
            assertion = NotMonetary(report.FITR24, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:Income.RentLeasingHiringGross.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434189" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR24", Value = GetValueOrEmpty(report.FITR24) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434189

            #region VR.ATO.FITR.434190

            /*  VR.ATO.FITR.434190
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR27] <> NULL AND [FITR27] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR27, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR27 = RP.JFOREIGN:Income.Gross.Amount
            */
            assertion = NotMonetary(report.FITR27, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:Income.Gross.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434190" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR27", Value = GetValueOrEmpty(report.FITR27) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434190

            #region VR.ATO.FITR.434191

            /*  VR.ATO.FITR.434191
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR29] <> NULL AND [FITR29] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR29, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR29 = RP:IncomeTax.FrankingCredits.ReceivedFromNewZealandCompanies.Amount
            */
            assertion = NotMonetary(report.FITR29, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:IncomeTax.FrankingCredits.ReceivedFromNewZealandCompanies.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434191" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR29", Value = GetValueOrEmpty(report.FITR29) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434191

            #region VR.ATO.FITR.434192

            /*  VR.ATO.FITR.434192
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR30] <> NULL AND [FITR30] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR30, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR30 = RP:Equity.InternationalDealings.ForeignFundTransfers.Amount
            */
            assertion = NotMonetary(report.FITR30, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:Equity.InternationalDealings.ForeignFundTransfers.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434192" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR30", Value = GetValueOrEmpty(report.FITR30) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434192

            #region VR.ATO.FITR.434193

            /*  VR.ATO.FITR.434193
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR32] <> NULL AND [FITR32] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR32, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR32 = RP:Remuneration.ABNNotQuotedPaymentGross.Amount
            */
            assertion = NotMonetary(report.FITR32, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:Remuneration.ABNNotQuotedPaymentGross.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434193" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR32", Value = GetValueOrEmpty(report.FITR32) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434193

            #region VR.ATO.FITR.434194

            /*  VR.ATO.FITR.434194
            Monetary field - max 11 characters, signed, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR33] <> NULL AND [FITR33] <> MONETARY(S,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR33, 'S', 11, 0)
    
            Data Elements:
    
            ^FITR33 = RP:Income.PartnershipDistributionGross.Amount
            */
            assertion = NotMonetary(report.FITR33, @"S", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:Income.PartnershipDistributionGross.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434194" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR33", Value = GetValueOrEmpty(report.FITR33) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434194

            #region VR.ATO.FITR.434195

            /*  VR.ATO.FITR.434195
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR37] <> NULL AND [FITR37] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR37, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR37 = RP:Income.TrustDistributionUnfranked.Amount
            */
            assertion = NotMonetary(report.FITR37, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:Income.TrustDistributionUnfranked.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434195" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR37", Value = GetValueOrEmpty(report.FITR37) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434195

            #region VR.ATO.FITR.434196

            /*  VR.ATO.FITR.434196
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR38] <> NULL AND [FITR38] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR38, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR38 = RP:Income.TrustDistributionFranked.Amount
            */
            assertion = NotMonetary(report.FITR38, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:Income.TrustDistributionFranked.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434196" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR38", Value = GetValueOrEmpty(report.FITR38) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434196

            #region VR.ATO.FITR.434197

            /*  VR.ATO.FITR.434197
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR39] <> NULL AND [FITR39] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR39, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR39 = RP:IncomeTax.FrankingCredits.FrankingCreditShareReceivedIndirectly.Amount
            */
            assertion = NotMonetary(report.FITR39, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:IncomeTax.FrankingCredits.FrankingCreditShareReceivedIndirectly.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434197" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR39", Value = GetValueOrEmpty(report.FITR39) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434197

            #region VR.ATO.FITR.434198

            /*  VR.ATO.FITR.434198
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR40] <> NULL AND [FITR40] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR40, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR40 = RP:Income.TrustDistributionOther.Amount
            */
            assertion = NotMonetary(report.FITR40, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:Income.TrustDistributionOther.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434198" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR40", Value = GetValueOrEmpty(report.FITR40) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434198

            #region VR.ATO.FITR.434199

            /*  VR.ATO.FITR.434199
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR41] <> NULL AND [FITR41] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR41, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR41 = RP:SuperannuationContribution.EmployerAssessable.Amount
            */
            assertion = NotMonetary(report.FITR41, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:SuperannuationContribution.EmployerAssessable.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434199" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR41", Value = GetValueOrEmpty(report.FITR41) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434199

            #region VR.ATO.FITR.434200

            /*  VR.ATO.FITR.434200
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR42] <> NULL AND [FITR42] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR42, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR42 = RP:SuperannuationContribution.PersonalAssessable.Amount
            */
            assertion = NotMonetary(report.FITR42, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:SuperannuationContribution.PersonalAssessable.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434200" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR42", Value = GetValueOrEmpty(report.FITR42) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434200

            #region VR.ATO.FITR.434201

            /*  VR.ATO.FITR.434201
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR43] <> NULL AND [FITR43] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR43, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR43 = RP:SuperannuationContribution.EmployerContributionsTaxFileNumberNotQuoted.Amount
            */
            assertion = NotMonetary(report.FITR43, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:SuperannuationContribution.EmployerContributionsTaxFileNumberNotQuoted.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434201" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR43", Value = GetValueOrEmpty(report.FITR43) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434201

            #region VR.ATO.FITR.434202

            /*  VR.ATO.FITR.434202
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR44] <> NULL AND [FITR44] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR44, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR44 = RP:SuperannuationContribution.ContributionsExcludedByTrustee.Amount
            */
            assertion = NotMonetary(report.FITR44, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:SuperannuationContribution.ContributionsExcludedByTrustee.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434202" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR44", Value = GetValueOrEmpty(report.FITR44) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434202

            #region VR.ATO.FITR.434203

            /*  VR.ATO.FITR.434203
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR45] <> NULL AND [FITR45] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR45, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR45 = RP:SuperannuationContribution.ReducedContributions.Amount
            */
            assertion = NotMonetary(report.FITR45, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:SuperannuationContribution.ReducedContributions.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434203" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR45", Value = GetValueOrEmpty(report.FITR45) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434203

            #region VR.ATO.FITR.434204

            /*  VR.ATO.FITR.434204
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR46] <> NULL AND [FITR46] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR46, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR46 = RP:SuperannuationContribution.LiabilityTransferToLifeInsuranceCompanyOrPooledSuperannuationTrustAssessable.Amount
            */
            assertion = NotMonetary(report.FITR46, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:SuperannuationContribution.LiabilityTransferToLifeInsuranceCompanyOrPooledSuperannuationTrustAssessable.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434204" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR46", Value = GetValueOrEmpty(report.FITR46) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434204

            #region VR.ATO.FITR.434205

            /*  VR.ATO.FITR.434205
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR47] <> NULL AND [FITR47] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR47, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR47 = RP:Income.Other.Amount
            */
            assertion = NotMonetary(report.FITR47, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:Income.Other.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434205" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR47", Value = GetValueOrEmpty(report.FITR47) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434205

            #region VR.ATO.FITR.434206

            /*  VR.ATO.FITR.434206
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR49] <> NULL AND [FITR49] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR49, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR49 = RP:Income.AssessableIncomeDueToFundTaxStatusChange.Amount
            */
            assertion = NotMonetary(report.FITR49, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:Income.AssessableIncomeDueToFundTaxStatusChange.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434206" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR49", Value = GetValueOrEmpty(report.FITR49) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434206

            #region VR.ATO.FITR.434207

            /*  VR.ATO.FITR.434207
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR50] <> NULL AND [FITR50] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR50, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR50 = RP:Income.NonArmsLengthOtherNet.Amount
            */
            assertion = NotMonetary(report.FITR50, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:Income.NonArmsLengthOtherNet.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434207" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR50", Value = GetValueOrEmpty(report.FITR50) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434207

            #region VR.ATO.FITR.434209

            /*  VR.ATO.FITR.434209
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR52] <> NULL AND [FITR52] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR52, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR52 = RP.JAUS:Expense.Interest.Amount
            */
            assertion = NotMonetary(report.FITR52, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:Expense.Interest.Amount[@contextRef='" + report.RPJAUSId + "']",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434209" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR52", Value = GetValueOrEmpty(report.FITR52) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434209

            #region VR.ATO.FITR.434211

            /*  VR.ATO.FITR.434211
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR56] <> NULL AND [FITR56] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR56, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR56 = RP:Expense.DepreciationAllowableDeduction.Amount
            */
            assertion = NotMonetary(report.FITR56, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:Expense.DepreciationAllowableDeduction.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434211" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR56", Value = GetValueOrEmpty(report.FITR56) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434211

            #region VR.ATO.FITR.434212

            /*  VR.ATO.FITR.434212
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR57] <> NULL AND [FITR57] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR57, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR57 = RP:Expense.PremiumsDeathOrDisability.Amount
            */
            assertion = NotMonetary(report.FITR57, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:Expense.PremiumsDeathOrDisability.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434212" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR57", Value = GetValueOrEmpty(report.FITR57) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434212

            #region VR.ATO.FITR.434215

            /*  VR.ATO.FITR.434215
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR62] <> NULL AND [FITR62] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR62, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR62 = RP:Expense.DeductibleOther.Amount
            */
            assertion = NotMonetary(report.FITR62, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:Expense.DeductibleOther.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434215" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR62", Value = GetValueOrEmpty(report.FITR62) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434215

            #region VR.ATO.FITR.434216

            /*  VR.ATO.FITR.434216
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR64] <> NULL AND [FITR64] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR64, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR64 = RP:IncomeTax.Deduction.TaxLossesDeducted.Amount
            */
            assertion = NotMonetary(report.FITR64, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:IncomeTax.Deduction.TaxLossesDeducted.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434216" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR64", Value = GetValueOrEmpty(report.FITR64) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434216

            #region VR.ATO.FITR.434217

            /*  VR.ATO.FITR.434217
            Monetary field - max 11 characters, signed, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR65] <> NULL AND [FITR65] <> MONETARY(S,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR65, 'S', 11, 0)
    
            Data Elements:
    
            ^FITR65 = RP:Income.Taxable.Amount
            */
            assertion = NotMonetary(report.FITR65, @"S", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:Income.Taxable.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434217" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR65", Value = GetValueOrEmpty(report.FITR65) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434217

            #region VR.ATO.FITR.434218

            /*  VR.ATO.FITR.434218
            Monetary field - max 13 chars, no sign, may have a decimal point
    
            Legacy Rule Format:
            IF [FITR66] <> NULL AND [FITR66] <> MONETARY(U,13,2)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR66, 'U', 13, 2)
    
            Data Elements:
    
            ^FITR66 = RP:IncomeTax.TaxableGross.Amount
            */
            assertion = NotMonetary(report.FITR66, @"U", 13, 2);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:IncomeTax.TaxableGross.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434218" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR66", Value = GetValueOrEmpty(report.FITR66) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434218

            #region VR.ATO.FITR.434219

            /*  VR.ATO.FITR.434219
            Monetary field - max 13 chars, no sign, may have a decimal point
    
            Legacy Rule Format:
            IF [FITR68] <> NULL AND [FITR68] <> MONETARY(U,13,2)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR68, 'U', 13, 2)
    
            Data Elements:
    
            ^FITR68 = RP:IncomeTax.TaxOffsetsAndRebatesTotal.Amount
            */
            assertion = NotMonetary(report.FITR68, @"U", 13, 2);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:IncomeTax.TaxOffsetsAndRebatesTotal.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434219" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR68", Value = GetValueOrEmpty(report.FITR68) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434219

            #region VR.ATO.FITR.434220

            /*  VR.ATO.FITR.434220
            Monetary field - max 13 chars, no sign, may have a decimal point
    
            Legacy Rule Format:
            IF [FITR70] <> NULL AND [FITR70] <> MONETARY(U,13,2)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR70, 'U', 13, 2)
    
            Data Elements:
    
            ^FITR70 = RP:Income.Interest.EarlyPaymentCredit.Amount
            */
            assertion = NotMonetary(report.FITR70, @"U", 13, 2);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:Income.Interest.EarlyPaymentCredit.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434220" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR70", Value = GetValueOrEmpty(report.FITR70) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434220

            #region VR.ATO.FITR.434221

            /*  VR.ATO.FITR.434221
            Monetary field - max 13 chars, no sign, may have a decimal point
    
            Legacy Rule Format:
            IF [FITR71] <> NULL AND [FITR71] <> MONETARY(U,13,2)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR71, 'U', 13, 2)
    
            Data Elements:
    
            ^FITR71 = RP:IncomeTax.PayAsYouGoWithholding.CreditForAmountsWithheldFromForeignResidents.Amount
            */
            assertion = NotMonetary(report.FITR71, @"U", 13, 2);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:IncomeTax.PayAsYouGoWithholding.CreditForAmountsWithheldFromForeignResidents.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434221" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR71", Value = GetValueOrEmpty(report.FITR71) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434221

            #region VR.ATO.FITR.434222

            /*  VR.ATO.FITR.434222
            Overseas address indicator must be 'no' (false).
    
            Legacy Rule Format:
            IF [FITR143] = TRUE
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR143 = TRUE
    
            Data Elements:
    
            ^FITR143 = RP:AddressDetails:AddressDetails.OverseasAddress.Indicator WHERE [AddressDetails.Usage.Code='POS'].[AddressDetails.Currency.Code='C']
            */
            assertion = (report.FITR143 == true);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.436251",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Overseas address indicator must be 'no' (false).",
                    LongDescription = @"Current postal address must be an Australian address. The overseas indicator must therefore be 'no' (false)",
                    Location = "/xbrli:xbrl/tns:AddressDetails" + OccurrenceIndex(report.AddressDetailsCollection_POS_C_RPOccurrenceIndex) + "/tns:AddressDetails.OverseasAddress.Indicator",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434222" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR143", Value = GetValueOrEmpty(report.FITR143) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434222

            #region VR.ATO.FITR.434226

            /*  VR.ATO.FITR.434226
            There can only be a single non-individual Trustee reported.
    
            Legacy Rule Format:
            IF COUNT(fitr.0008.lodge.req.xx.xx:Trustee) > 1
              RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Count(^Trustee) > 1
    
            Data Elements:
    
            ^Trustee = RP:Trustee
            */
            assertion = (Count(report.TrusteeCollectionCount) > 1);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434226",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Only a single non-individual Trustee can be reported.",
                    Location = "/xbrli:xbrl/tns:Trustee" + OccurrenceIndex(report.TrusteeCollectionCount) + "/tns:Identifiers.AustralianBusinessNumber.Identifier",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434226" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434226

            #region VR.ATO.FITR.434239

            /*  VR.ATO.FITR.434239
            Monetary field - max 13 chars, no sign, may have a decimal point
    
            Legacy Rule Format:
            IF [FITR175] <> NULL AND [FITR175] <> MONETARY(U,13,2)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR175, 'U', 13, 2)
    
            Data Elements:
    
            ^FITR175 = RP:IncomeTax.PayAsYouGoWithholding.CreditForTaxWithheldFromCloselyHeldTrustShareReceivedIndirectly.Amount
            */
            assertion = NotMonetary(report.FITR175, @"U", 13, 2);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:IncomeTax.PayAsYouGoWithholding.CreditForTaxWithheldFromCloselyHeldTrustShareReceivedIndirectly.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434239" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR175", Value = GetValueOrEmpty(report.FITR175) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434239

            #region VR.ATO.FITR.434240

            /*  VR.ATO.FITR.434240
            Incorrect number of XBRL contexts
    
            Legacy Rule Format:
            IF COUNT(CONTEXT(RP.Closing)) > 1 
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Count(^Context) > 1
    
            Data Elements:
    
            ^Context = RP:RP.Closing
            */
            assertion = (Count(report.RPClosingCount) > 1);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434240",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Invalid context. The number of Reporting Party: Closing contexts must not exceed 1",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPClosingLastOccurrenceIndex),
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434240" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434240

            #region VR.ATO.FITR.434241

            /*  VR.ATO.FITR.434241
            Monetary field - max 13 chars, no sign, may have a decimal point
    
            Legacy Rule Format:
            IF [FITR69] <> NULL AND [FITR69] <> MONETARY(U,13,2)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR69, 'U', 13, 2)
    
            Data Elements:
    
            ^FITR69 = RP:InternationalDealings.AAMInterestSection102.Amount
            */
            assertion = NotMonetary(report.FITR69, @"U", 13, 2);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:InternationalDealings.AAMInterestSection102.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434241" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR69", Value = GetValueOrEmpty(report.FITR69) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434241

            #region VR.ATO.FITR.434242

            /*  VR.ATO.FITR.434242
            Monetary field - max 13 chars, no sign, may have a decimal point
    
            Legacy Rule Format:
            IF [FITR181] <> NULL AND [FITR181] <> MONETARY(U,13,2)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR181, 'U', 13, 2)
    
            Data Elements:
    
            ^FITR181 = RP:IncomeTax.TFNNotQuotedContributionsTax.Amount
            */
            assertion = NotMonetary(report.FITR181, @"U", 13, 2);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:IncomeTax.TFNNotQuotedContributionsTax.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434242" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR181", Value = GetValueOrEmpty(report.FITR181) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434242

            #region VR.ATO.FITR.434244

            /*  VR.ATO.FITR.434244
            Tax on no-TFN-quoted contributions must be No-TFN-quoted contributions amount multiplied by 0.02 for non-Australian superannuation funds
    
            Legacy Rule Format:
            IF [FITR15] = FALSE AND [FITR181]<>([FITR43] * 0.02)+/-1
             RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR15 = FALSE AND OutsideRange(^FITR181, ^FITR43 * 0.02, 1)
    
            Data Elements:
    
            ^FITR181 = RP:IncomeTax.TFNNotQuotedContributionsTax.Amount
    
            ^FITR15 = RP:SuperannuationFundDetails.AustralianSuperannuationFundStatus.Indicator
    
            ^FITR43 = RP:SuperannuationContribution.EmployerContributionsTaxFileNumberNotQuoted.Amount
            */
            assertion = (report.FITR15 == false && OutsideRange(report.FITR181.GetValueOrDefault(), report.FITR43.GetValueOrDefault() * 0.02M, 1));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434307",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"The amount at Tax on no-TFN-quoted contributions is incorrect",
                    Location = "/xbrli:xbrl/tns:IncomeTax.TFNNotQuotedContributionsTax.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434244" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR15", Value = GetValueOrEmpty(report.FITR15) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR181", Value = GetValueOrEmpty(report.FITR181) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR43", Value = GetValueOrEmpty(report.FITR43) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434244

            #region VR.ATO.FITR.434246

            /*  VR.ATO.FITR.434246
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR200] <> NULL AND [FITR200] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR200, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR200 = RP:TaxConcession.PensionIncomeExemptDeduction.Amount
            */
            assertion = NotMonetary(report.FITR200, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:TaxConcession.PensionIncomeExemptDeduction.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434246" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR200", Value = GetValueOrEmpty(report.FITR200) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434246

            #region VR.ATO.FITR.434247

            /*  VR.ATO.FITR.434247
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR190] <> NULL AND [FITR190] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR190, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR190 = RP:Expense.Operating.ManagementExpense.Amount
            */
            assertion = NotMonetary(report.FITR190, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:Expense.Operating.ManagementExpense.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434247" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR190", Value = GetValueOrEmpty(report.FITR190) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434247

            #region VR.ATO.FITR.434248

            /*  VR.ATO.FITR.434248
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR188] <> NULL AND [FITR188] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR188, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR188 = RP:Expense.Operating.AdministrativeExpense.Amount
            */
            assertion = NotMonetary(report.FITR188, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:Expense.Operating.AdministrativeExpense.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434248" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR188", Value = GetValueOrEmpty(report.FITR188) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434248

            #region VR.ATO.FITR.434252

            /*  VR.ATO.FITR.434252
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR186] <> NULL AND [FITR186] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR186, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR186 = RP:Income.ForeignExchangeGains.Amount
            */
            assertion = NotMonetary(report.FITR186, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:Income.ForeignExchangeGains.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434252" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR186", Value = GetValueOrEmpty(report.FITR186) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434252

            #region VR.ATO.FITR.434253

            /*  VR.ATO.FITR.434253
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR187] <> NULL AND [FITR187] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR187, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR187 = RP:Income.ForeignExchangeLosses.Amount
            */
            assertion = NotMonetary(report.FITR187, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:Income.ForeignExchangeLosses.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434253" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR187", Value = GetValueOrEmpty(report.FITR187) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434253

            #region VR.ATO.FITR.434254

            /*  VR.ATO.FITR.434254
            Monetary field - max 13 chars, no sign, may have a decimal point
    
            Legacy Rule Format:
            IF [FITR182] <> NULL AND [FITR182] <> MONETARY(U,13,2)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR182, 'U', 13, 2)
    
            Data Elements:
    
            ^FITR182 = RP:IncomeTax.TotalTaxOnTaxableIncome.Amount
            */
            assertion = NotMonetary(report.FITR182, @"U", 13, 2);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:IncomeTax.TotalTaxOnTaxableIncome.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434254" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR182", Value = GetValueOrEmpty(report.FITR182) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434254

            #region VR.ATO.FITR.434255

            /*  VR.ATO.FITR.434255
            Monetary field - max 13 chars, no sign, may have a decimal point
    
            Legacy Rule Format:
            IF [FITR189] <> NULL AND [FITR189] <> MONETARY(U,13,2)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR189, 'U', 13, 2)
    
            Data Elements:
    
            ^FITR189 = RP:IncomeTax.TaxOffsetNonRefundableNonCarryForward.Amount
            */
            assertion = NotMonetary(report.FITR189, @"U", 13, 2);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:IncomeTax.TaxOffsetNonRefundableNonCarryForward.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434255" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR189", Value = GetValueOrEmpty(report.FITR189) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434255

            #region VR.ATO.FITR.434256

            /*  VR.ATO.FITR.434256
            Monetary field - max 13 chars, no sign, may have a decimal point
    
            Legacy Rule Format:
            IF [FITR183] <> NULL AND [FITR183] <> MONETARY(U,13,2)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR183, 'U', 13, 2)
    
            Data Elements:
    
            ^FITR183 = RP:IncomeTax.SubtotalTaxPayable.Amount
            */
            assertion = NotMonetary(report.FITR183, @"U", 13, 2);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:IncomeTax.SubtotalTaxPayable.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434256" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR183", Value = GetValueOrEmpty(report.FITR183) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434256

            #region VR.ATO.FITR.434257

            /*  VR.ATO.FITR.434257
            Monetary field - max 13 chars, no sign, may have a decimal point
    
            Legacy Rule Format:
            IF [FITR184] <> NULL AND [FITR184] <> MONETARY(U,13,2)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR184, 'U', 13, 2)
    
            Data Elements:
    
            ^FITR184 = RP:IncomeTax.TotalRefundableTaxOffsets.Amount
            */
            assertion = NotMonetary(report.FITR184, @"U", 13, 2);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:IncomeTax.TotalRefundableTaxOffsets.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434257" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR184", Value = GetValueOrEmpty(report.FITR184) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434257

            #region VR.ATO.FITR.434258

            /*  VR.ATO.FITR.434258
            Monetary field - max 13 chars, no sign, may have a decimal point
    
            Legacy Rule Format:
            IF [FITR199] <> NULL AND [FITR199] <> MONETARY(U,13,2)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR199, 'U', 13, 2)
    
            Data Elements:
    
            ^FITR199 = RP:IncomeTax.Payable.Amount
            */
            assertion = NotMonetary(report.FITR199, @"U", 13, 2);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:IncomeTax.Payable.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434258" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR199", Value = GetValueOrEmpty(report.FITR199) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434258

            #region VR.ATO.FITR.434259

            /*  VR.ATO.FITR.434259
            Monetary field - max 13 chars, no sign, may have a decimal point
    
            Legacy Rule Format:
            IF [FITR201] <> NULL AND [FITR201] <> MONETARY(U,13,2)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR201, 'U', 13, 2)
    
            Data Elements:
    
            ^FITR201 = RP:IncomeTax.Deduction.CreditsEligibleTotal.Amount
            */
            assertion = NotMonetary(report.FITR201, @"U", 13, 2);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:IncomeTax.Deduction.CreditsEligibleTotal.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434259" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR201", Value = GetValueOrEmpty(report.FITR201) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434259

            #region VR.ATO.FITR.434260

            /*  VR.ATO.FITR.434260
            Monetary field - max 13 chars, no sign, may have a decimal point
    
            Legacy Rule Format:
            IF [FITR185] <> NULL AND [FITR185] <> MONETARY(U,13,2)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR185, 'U', 13, 2)
    
            Data Elements:
    
            ^FITR185 = RP:IncomeTax.TotalRemainderOfRefundableTaxOffsets.Amount
            */
            assertion = NotMonetary(report.FITR185, @"U", 13, 2);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:IncomeTax.TotalRemainderOfRefundableTaxOffsets.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434260" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR185", Value = GetValueOrEmpty(report.FITR185) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434260

            #region VR.ATO.FITR.434261

            /*  VR.ATO.FITR.434261
            Gross Tax must be equal to sum of Tax on taxable income and Tax on no-TFN-quoted contributions
    
            Legacy Rule Format:
            IF [FITR66] <> ([FITR182] + [FITR181])
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR66 <> (^FITR182 + ^FITR181)
    
            Data Elements:
    
            ^FITR66 = RP:IncomeTax.TaxableGross.Amount
    
            ^FITR182 = RP:IncomeTax.TotalTaxOnTaxableIncome.Amount
    
            ^FITR181 = RP:IncomeTax.TFNNotQuotedContributionsTax.Amount
            */
            assertion = (report.FITR66.GetValueOrDefault() != report.FITR182.GetValueOrDefault() + report.FITR181.GetValueOrDefault());
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434300",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Gross Tax amount incorrect.",
                    LongDescription = @"Gross Tax must be equal to sum of Tax on taxable income and Tax on no-TFN-quoted contributions",
                    Location = "/xbrli:xbrl/tns:IncomeTax.TaxableGross.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434261" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR66", Value = GetValueOrEmpty(report.FITR66) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR182", Value = GetValueOrEmpty(report.FITR182) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR181", Value = GetValueOrEmpty(report.FITR181) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434261

            #region VR.ATO.FITR.434262

            /*  VR.ATO.FITR.434262
            Non-refundable non-carry forward tax offsets must be equal to sum of foreign income tax offset and rebates and tax offsets.
    
            Legacy Rule Format:
            IF ([FITR67] <> NULL OR [FITR68] <> NULL) AND ([FITR189] <> ([FITR67] + [FITR68]))
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^FITR67 <> NULL OR ^FITR68 <> NULL) AND (^FITR189 <> (^FITR67 + ^FITR68))
    
            Data Elements:
    
            ^FITR189 = RP:IncomeTax.TaxOffsetNonRefundableNonCarryForward.Amount
    
            ^FITR67 = RP:Income.InternationalDealings.TaxOffset.Amount
    
            ^FITR68 = RP:IncomeTax.TaxOffsetsAndRebatesTotal.Amount
            */
            assertion = ((report.FITR67 != null || report.FITR68 != null) && report.FITR189.GetValueOrDefault() != report.FITR67.GetValueOrDefault() + report.FITR68.GetValueOrDefault());
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434262",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Non-refundable non-carry forward tax offsets amount incorrect.",
                    LongDescription = @"Non-refundable non-carry forward tax offsets amount must be equal to sum of Foreign income tax offset and Rebates and tax offsets.",
                    Location = "/xbrli:xbrl/tns:IncomeTax.TaxOffsetNonRefundableNonCarryForward.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434262" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR67", Value = GetValueOrEmpty(report.FITR67) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR68", Value = GetValueOrEmpty(report.FITR68) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR189", Value = GetValueOrEmpty(report.FITR189) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434262

            #region VR.ATO.FITR.434263

            /*  VR.ATO.FITR.434263
            Refundable tax offsets amount must be equal to sum of Complying fund's franking credits tax offset, No-TFN tax offset, National rental affordability scheme tax offset and Exploration credit tax offset.
    
            Legacy Rule Format:
            IF [FITR184] <> ([FITR73] + [FITR131] + [FITR133] + [FITR206])
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR184 <> (^FITR73 + ^FITR131 + ^FITR133 + ^FITR206)
    
            Data Elements:
    
            ^FITR184 = RP:IncomeTax.TotalRefundableTaxOffsets.Amount
    
            ^FITR73 = RP:IncomeTax.FrankingCredits.Refundable.Amount
    
            ^FITR131 = RP:IncomeTax.TaxOffsetTFNNotQuoted.Amount
    
            ^FITR133 = RP:IncomeTax.TaxOffsetEntitlementForNationalRentalAffordabilityScheme.Amount
    
            ^FITR206 = RP:IncomeTax.TaxOffsetExplorationCredits.Amount
            */
            assertion = (report.FITR184.GetValueOrDefault() != report.FITR73.GetValueOrDefault() + report.FITR131.GetValueOrDefault() + report.FITR133.GetValueOrDefault() + report.FITR206.GetValueOrDefault());
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434263",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Refundable tax offsets amount incorrect.",
                    LongDescription = @"Refundable tax offsets amount must be equal to sum of Complying fund's franking credits tax offset, No-TFN tax offset, National rental affordability scheme tax offset and Exploration credit tax offset.",
                    Location = "/xbrli:xbrl/tns:IncomeTax.TotalRefundableTaxOffsets.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434263" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR184", Value = GetValueOrEmpty(report.FITR184) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR73", Value = GetValueOrEmpty(report.FITR73) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR131", Value = GetValueOrEmpty(report.FITR131) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR133", Value = GetValueOrEmpty(report.FITR133) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR206", Value = GetValueOrEmpty(report.FITR206) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434263

            #region VR.ATO.FITR.434264

            /*  VR.ATO.FITR.434264
            Subtotal 2 amount must be equal to Gross Tax minus Non-refundable non-carry forward tax offset amount if Non-refundable non-carry forward tax offset amount is less than Gross Tax
    
            Legacy Rule Format:
            IF ([FITR189] < [FITR66]) AND ([FITR183] <> ([FITR66] - [FITR189]))
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^FITR189 < ^FITR66) AND (^FITR183 <> (^FITR66 - ^FITR189))
    
            Data Elements:
    
            ^FITR183 = RP:IncomeTax.SubtotalTaxPayable.Amount
    
            ^FITR189 = RP:IncomeTax.TaxOffsetNonRefundableNonCarryForward.Amount
    
            ^FITR66 = RP:IncomeTax.TaxableGross.Amount
            */
            assertion = (report.FITR189.GetValueOrDefault() < report.FITR66.GetValueOrDefault() && report.FITR183.GetValueOrDefault() != report.FITR66.GetValueOrDefault() - report.FITR189.GetValueOrDefault());
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434344",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Subtotal 2 amount incorrect.",
                    LongDescription = @"If the Non-refundable non-carry forward tax offset is less than Gross Tax, then Subtotal 2 must be equal to Gross Tax less Non-refundable non-carry forward tax offset amount.",
                    Location = "/xbrli:xbrl/tns:IncomeTax.SubtotalTaxPayable.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434264" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR189", Value = GetValueOrEmpty(report.FITR189) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR66", Value = GetValueOrEmpty(report.FITR66) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR183", Value = GetValueOrEmpty(report.FITR183) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434264

            #region VR.ATO.FITR.434265

            /*  VR.ATO.FITR.434265
            Subtotal 2 amount must be zero if Non-refundable non-carry forward tax offset amount is greater than or equal to Gross Tax
    
            Legacy Rule Format:
            IF ([FITR189] >= [FITR66]) AND ([FITR183] <> 0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^FITR189 >= ^FITR66) AND (^FITR183 <> 0)
    
            Data Elements:
    
            ^FITR183 = RP:IncomeTax.SubtotalTaxPayable.Amount
    
            ^FITR189 = RP:IncomeTax.TaxOffsetNonRefundableNonCarryForward.Amount
    
            ^FITR66 = RP:IncomeTax.TaxableGross.Amount
            */
            assertion = (report.FITR189.GetValueOrDefault() >= report.FITR66.GetValueOrDefault() && report.FITR183 != 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434345",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Subtotal 2 amount must equal zero.",
                    LongDescription = @"If Non-refundable non-carry forward tax offset amount is greater than or equal to Gross Tax, then Subtotal 2 amount must equal zero.",
                    Location = "/xbrli:xbrl/tns:IncomeTax.SubtotalTaxPayable.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434265" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR189", Value = GetValueOrEmpty(report.FITR189) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR66", Value = GetValueOrEmpty(report.FITR66) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR183", Value = GetValueOrEmpty(report.FITR183) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434265

            #region VR.ATO.FITR.434266

            /*  VR.ATO.FITR.434266
            Tax payable amount must be zero if Subtotal 3 amount is zero.
    
            Legacy Rule Format:
            IF [FITR216] = 0 AND [FITR199] <> 0
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR216 = 0 AND ^FITR199 <> 0
    
            Data Elements:
    
            ^FITR199 = RP:IncomeTax.Payable.Amount
    
            ^FITR216 = RP:IncomeTax.TaxOffsetCarriedForwardSubTotal.Amount
            */
            assertion = (report.FITR216 == 0 && report.FITR199 != 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434339",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Tax payable amount must equal zero.",
                    LongDescription = @"If Subtotal 3 amount equals zero then Tax payable amount must equal zero.",
                    Location = "/xbrli:xbrl/tns:IncomeTax.Payable.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434266" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR216", Value = GetValueOrEmpty(report.FITR216) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR199", Value = GetValueOrEmpty(report.FITR199) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434266

            #region VR.ATO.FITR.434267

            /*  VR.ATO.FITR.434267
            Tax payable amount must be equal to Subtotal 3 amount minus Refundable tax offsets if Refundable tax offsets amount is less than Subtotal 3
    
            Legacy Rule Format:
            IF ([FITR184] < [FITR216]) AND ([FITR199] <> ([FITR216] - [FITR184]))
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^FITR184 < ^FITR216) AND (^FITR199 <> (^FITR216 - ^FITR184))
    
            Data Elements:
    
            ^FITR199 = RP:IncomeTax.Payable.Amount
    
            ^FITR184 = RP:IncomeTax.TotalRefundableTaxOffsets.Amount
    
            ^FITR216 = RP:IncomeTax.TaxOffsetCarriedForwardSubTotal.Amount
            */
            assertion = (report.FITR184.GetValueOrDefault() < report.FITR216.GetValueOrDefault() && report.FITR199.GetValueOrDefault() != report.FITR216.GetValueOrDefault() - report.FITR184.GetValueOrDefault());
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434340",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Tax payable amount incorrect.",
                    LongDescription = @"If Refundable tax offsets amount is less than Subtotal 3 then Tax payable amount must be equal to Subtotal 3 amount less Refundable tax offsets.",
                    Location = "/xbrli:xbrl/tns:IncomeTax.Payable.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434267" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR184", Value = GetValueOrEmpty(report.FITR184) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR216", Value = GetValueOrEmpty(report.FITR216) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR199", Value = GetValueOrEmpty(report.FITR199) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434267

            #region VR.ATO.FITR.434268

            /*  VR.ATO.FITR.434268
            Tax payable amount must be zero if Refundable tax offsets amount is greater than or equal to Subtotal 3
    
            Legacy Rule Format:
            IF ([FITR184] >= [FITR216]) AND ([FITR199] <> 0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^FITR184 >= ^FITR216) AND (^FITR199 <> 0)
    
            Data Elements:
    
            ^FITR199 = RP:IncomeTax.Payable.Amount
    
            ^FITR184 = RP:IncomeTax.TotalRefundableTaxOffsets.Amount
    
            ^FITR216 = RP:IncomeTax.TaxOffsetCarriedForwardSubTotal.Amount
            */
            assertion = (report.FITR184.GetValueOrDefault() >= report.FITR216.GetValueOrDefault() && report.FITR199 != 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434341",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Tax payable amount must equal zero.",
                    LongDescription = @"If Refundable tax offsets amount is greater than or equal to Subtotal 3 then Tax payable amount must equal zero.",
                    Location = "/xbrli:xbrl/tns:IncomeTax.Payable.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434268" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR184", Value = GetValueOrEmpty(report.FITR184) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR216", Value = GetValueOrEmpty(report.FITR216) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR199", Value = GetValueOrEmpty(report.FITR199) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434268

            #region VR.ATO.FITR.434269

            /*  VR.ATO.FITR.434269
            If Refundable tax offsets amount is greater than or equal to Subtotal 3, then any excess of refundable tax offsets over Subtotal 3 amount must be transferred to Tax offset refunds.
    
            Legacy Rule Format:
            IF ([FITR184] >= [FITR216]) AND ([FITR185] <> ([FITR184] - [FITR216]))
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^FITR184 >= ^FITR216) AND (^FITR185 <> (^FITR184 - ^FITR216))
    
            Data Elements:
    
            ^FITR185 = RP:IncomeTax.TotalRemainderOfRefundableTaxOffsets.Amount
    
            ^FITR184 = RP:IncomeTax.TotalRefundableTaxOffsets.Amount
    
            ^FITR216 = RP:IncomeTax.TaxOffsetCarriedForwardSubTotal.Amount
            */
            assertion = (report.FITR184.GetValueOrDefault() >= report.FITR216.GetValueOrDefault() && report.FITR185.GetValueOrDefault() != report.FITR184.GetValueOrDefault() - report.FITR216.GetValueOrDefault());
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434342",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Tax offset refunds (Remainder of refundable tax offsets) amount incorrect.",
                    LongDescription = @"If Refundable tax offsets amount is greater than or equal to Subtotal 3, then any excess of refundable tax offsets over Subtotal 3 amount must be transferred to Tax offset refunds.",
                    Location = "/xbrli:xbrl/tns:IncomeTax.TotalRemainderOfRefundableTaxOffsets.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434269" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR184", Value = GetValueOrEmpty(report.FITR184) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR216", Value = GetValueOrEmpty(report.FITR216) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR185", Value = GetValueOrEmpty(report.FITR185) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434269

            #region VR.ATO.FITR.434270

            /*  VR.ATO.FITR.434270
            Eligible credits must equal the sum of Credit for interest on early payments, Credit for tax withheld - foreign resident withholding (excluding capital gains), Credit for tax withheld - where ABN or TFN not quoted (non-individual), Credit for TFN amounts withheld from payments from closely held trusts, Credit for interest on no-TFN tax offset and Credit for amounts withheld from foreign resident capital gains withholding.
    
            Legacy Rule Format:
            IF [FITR201] <> ([FITR70] + [FITR71] + [FITR72] + [FITR175] + [FITR132] + [FITR209])
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR201 <> (^FITR70 + ^FITR71 + ^FITR72 + ^FITR175 + ^FITR132 + ^FITR209)
    
            Data Elements:
    
            ^FITR201 = RP:IncomeTax.Deduction.CreditsEligibleTotal.Amount
    
            ^FITR70 = RP:Income.Interest.EarlyPaymentCredit.Amount
    
            ^FITR71 = RP:IncomeTax.PayAsYouGoWithholding.CreditForAmountsWithheldFromForeignResidents.Amount
    
            ^FITR72 = RP:IncomeTax.PayAsYouGoWithholding.CreditForAmountsWithheldTFNNotQuotedAndABNNotQuoted.Amount
    
            ^FITR175 = RP:IncomeTax.PayAsYouGoWithholding.CreditForTaxWithheldFromCloselyHeldTrustShareReceivedIndirectly.Amount
    
            ^FITR132 = RP:IncomeTax.TaxOffsetCalculatedInterestTFNNotQuoted.Amount
    
            ^FITR209 = RP:IncomeTax.PayAsYouGoWithholding.CreditForCapitalGainsWithheldFromForeignResidents.Amount
            */
            assertion = (report.FITR201.GetValueOrDefault() != report.FITR70.GetValueOrDefault() + report.FITR71.GetValueOrDefault() + report.FITR72.GetValueOrDefault() + report.FITR175.GetValueOrDefault() + report.FITR132.GetValueOrDefault() + report.FITR209.GetValueOrDefault());
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434317",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Eligible credits amount incorrect.",
                    LongDescription = @"Eligible credits must equal the sum of Credit for interest on early payments, Credit for tax withheld - foreign resident withholding (excluding capital gains), Credit for tax withheld - where ABN or TFN not quoted (non-individual), Credit for TFN amounts withheld from payments from closely held trusts, Credit for interest on no-TFN tax offset and Credit for amounts withheld from foreign resident capital gains withholding.",
                    Location = "/xbrli:xbrl/tns:IncomeTax.Deduction.CreditsEligibleTotal.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434270" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR201", Value = GetValueOrEmpty(report.FITR201) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR70", Value = GetValueOrEmpty(report.FITR70) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR71", Value = GetValueOrEmpty(report.FITR71) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR72", Value = GetValueOrEmpty(report.FITR72) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR175", Value = GetValueOrEmpty(report.FITR175) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR132", Value = GetValueOrEmpty(report.FITR132) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR209", Value = GetValueOrEmpty(report.FITR209) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434270

            #region VR.ATO.FITR.434271

            /*  VR.ATO.FITR.434271
            Amount due or refundable must equal Tax Payable plus Section 102AAM interest charge less Eligible credits, remainder of refundable tax offsets and PAYG instalments raised.
    
            Legacy Rule Format:
            IF [FITR75] <> NULL AND [FITR75] <> ([FITR199] + [FITR69] - ([FITR201] + [FITR185] + [FITR74]))
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR75 <> NULL AND ^FITR75 <> (^FITR199 + ^FITR69 - (^FITR201 + ^FITR185 + ^FITR74))
    
            Data Elements:
    
            ^FITR75 = RP:IncomeTax.PayableOrRefundableTotal.Amount
    
            ^FITR199 = RP:IncomeTax.Payable.Amount
    
            ^FITR69 = RP:InternationalDealings.AAMInterestSection102.Amount
    
            ^FITR201 = RP:IncomeTax.Deduction.CreditsEligibleTotal.Amount
    
            ^FITR185 = RP:IncomeTax.TotalRemainderOfRefundableTaxOffsets.Amount
    
            ^FITR74 = RP:IncomeTax.LiabilityInstalmentsTotal.Amount
            */
            assertion = (report.FITR75 != null && report.FITR75.GetValueOrDefault() != report.FITR199.GetValueOrDefault() + report.FITR69.GetValueOrDefault() - (report.FITR201.GetValueOrDefault() + report.FITR185.GetValueOrDefault() + report.FITR74.GetValueOrDefault()));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434271",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Amount due or refundable incorrect.",
                    LongDescription = @"Amount due or refundable must equal Tax Payable plus Section 102AAM interest charge less Eligible credits, Remainder of refundable tax offsets and PAYG instalments raised.",
                    Location = "/xbrli:xbrl/tns:IncomeTax.PayableOrRefundableTotal.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434271" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR75", Value = GetValueOrEmpty(report.FITR75) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR199", Value = GetValueOrEmpty(report.FITR199) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR69", Value = GetValueOrEmpty(report.FITR69) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR201", Value = GetValueOrEmpty(report.FITR201) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR185", Value = GetValueOrEmpty(report.FITR185) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR74", Value = GetValueOrEmpty(report.FITR74) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434271

            #region VR.ATO.FITR.434275

            /*  VR.ATO.FITR.434275
            Number of transfers received from foreign funds for the current income year must be between 0 and 999.
    
            Legacy Rule Format:
            IF [FITR31] <> NULL AND [FITR31] <> SET (0-999)
              RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR31 <> NULL AND (^FITR31 < 0 OR ^FITR31 > 999)
    
            Data Elements:
    
            ^FITR31 = RP:InternationalDealings.ForeignFundTransfers.Count
            */
            assertion = (report.FITR31 != null && (report.FITR31.GetValueOrDefault() < 0 || report.FITR31.GetValueOrDefault() > 999));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434275",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Transfers from foreign funds number must be between 0 and 999",
                    Location = "/xbrli:xbrl/tns:InternationalDealings.ForeignFundTransfers.Count",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434275" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR31", Value = report.FITR31.GetValueOrDefault().ToString() });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434275

            #region VR.ATO.FITR.434276

            /*  VR.ATO.FITR.434276
            Mandatory field not supplied
    
            Legacy Rule Format:
            IF [FITR199] = NULL 
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR199 = NULL
    
            Data Elements:
    
            ^FITR199 = RP:IncomeTax.Payable.Amount
            */
            assertion = (report.FITR199 == null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.001001",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Mandatory field not supplied",
                    Location = "/xbrli:xbrl/tns:IncomeTax.Payable.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434276" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR199", Value = GetValueOrEmpty(report.FITR199) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434276

            #region VR.ATO.FITR.434277

            /*  VR.ATO.FITR.434277
            Schedule A must not exceed 3000 characters
    
            Legacy Rule Format:
            IF [FITR202] <> NULLORBLANK AND LENGTH([FITR202]) > 3000
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Length(^FITR202) > 3000
    
            Data Elements:
    
            ^FITR202 = RP:RegulatoryDisclosures.GeneralInformationAboutFinancialStatements.Text
            */
            assertion = (Length(report.FITR202) > 3000);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434277",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"The Other attachments schedule must not exceed 3000 characters",
                    Location = "/xbrli:xbrl/tns:RegulatoryDisclosures.GeneralInformationAboutFinancialStatements.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434277" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR202", Value = GetValueOrEmpty(report.FITR202) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434277

            #region VR.ATO.FITR.434286

            /*  VR.ATO.FITR.434286
            Exemption or rollover code must be unique
    
            Legacy Rule Format:
            IF [FITR204] <> NULL AND [FITR204] = ANY OTHER OCCURRENCE OF [FITR204]
              RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            HasDuplicateValues(^FITR204)
    
            Data Elements:
    
            ^FITR204 = RP:CapitalGainsTax.ExemptionOrRolloverApplied.Code
            */
            assertion = HasDuplicateValues(report.FITR204Collection);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434286",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Exemption or rollover code must be unique",
                    Location = "/xbrli:xbrl/tns:CapitalGainsTax.ExemptionOrRolloverApplied.Code" + OccurrenceIndex(DuplicateValueIndex(report.FITR204Collection) + 1) + "",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434286" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434286

            #region VR.ATO.FITR.434290

            /*  VR.ATO.FITR.434290
            Other deductions code must be A - Deductible balancing adjustment,
            B - Contribution that is a fringe benefit,
            C - Exclusion of personal contributions,
            E - Environment protection expenditure,
            I - Listed investment company capital gain amount,
            N - Deduction relating to foreign non-assessable non-exempt income,
            O - Other expenses not listed elsewhere,
            R - Return of contributions by non-complying funds,
            S - Superannuation (Financial Assistance Funding) Levy Act 1993, or
            T - TOFA losses
    
            Legacy Rule Format:
            IF [FITR63] <> NULL AND [FITR63] <> SET("A - Deductible balancing adjustment",
            "B - Contribution that is a fringe benefit",
            "C - Exclusion of personal contributions",
            "E - Environment protection expenditure",
            "I - Listed investment company capital gain amount",
            "N - Deduction relating to foreign non-assessable non-exempt income",
            "O - Other expenses not listed elsewhere",
            "R - Return of contributions by non-complying funds",
            "S - Superannuation (Financial Assistance Funding) Levy Act 1993",
            "T - TOFA losses")
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR63 <> NULL AND NotInSet(^FITR63, '"A - Deductible balancing adjustment","B - Contribution that is a fringe benefit","C - Exclusion of personal contributions","E - Environment protection expenditure","I - Listed investment company capital gain amount","N - Deduction relating to foreign non-assessable non-exempt income","O - Other expenses not listed elsewhere","R - Return of contributions by non-complying funds","S - Superannuation (Financial Assistance Funding) Levy Act 1993","T - TOFA losses"')
    
            Data Elements:
    
            ^FITR63 = RP:Expense.DeductibleOther.Code
            */
            assertion = (report.FITR63 != null && !(IsMatch(report.FITR63, @"^(A - Deductible balancing adjustment|B - Contribution that is a fringe benefit|C - Exclusion of personal contributions|E - Environment protection expenditure|I - Listed investment company capital gain amount|N - Deduction relating to foreign non-assessable non-exempt income|O - Other expenses not listed elsewhere|R - Return of contributions by non-complying funds|S - Superannuation \(Financial Assistance Funding\) Levy Act 1993|T - TOFA losses)$", RegexOptions.IgnoreCase)));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434312",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Other deductions code must be A, B, C, E, I, N, O, R, S or T",
                    LongDescription = @"Other deductions code must be A - Deductible balancing adjustment,
B - Contribution that is a fringe benefit,
C - Exclusion of personal contributions,
E - Environment protection expenditure,
I - Listed investment company capital gain amount,
N - Deduction relating to foreign non-assessable non-exempt income,
O - Other expenses not listed elsewhere,
R - Return of contributions by non-complying funds,
S - Superannuation (Financial Assistance Funding) Levy Act 1993, or
T - TOFA losses",
                    Location = "/xbrli:xbrl/tns:Expense.DeductibleOther.Code",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434290" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR63", Value = GetValueOrEmpty(report.FITR63) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434290

            #region VR.ATO.FITR.434291

            /*  VR.ATO.FITR.434291
            Other income code must be B, C, O, R, W or T
    
            Legacy Rule Format:
            IF [FITR48] <> NULL AND [FITR48] <> SET("B", "C", "O", "R", "T", "W" )
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR48 <> NULL AND NotInSet(^FITR48, '"B", "C", "O", "R", "T", "W"')
    
            Data Elements:
    
            ^FITR48 = RP:Income.Other.Code
            */
            assertion = (report.FITR48 != null && !(IsMatch(report.FITR48, @"^(B|C|O|R|T|W)$", RegexOptions.IgnoreCase)));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434291",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Other income code must be B, C, O, R, W or T",
                    LongDescription = @"Other income codes are:    'B' - Assessable balancing adjustment amount,    'C' - LIC capital gain amount,  'O' - Other income received not listed,    'R' - Rebate or refund of death or disability premiums,    'W' - Gross payments subject to foreign resident withholding, 'T' - TOFA gains or assessable TOFA balancing adjustment",
                    Location = "/xbrli:xbrl/tns:Income.Other.Code",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434291" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR48", Value = GetValueOrEmpty(report.FITR48) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434291

            #region VR.ATO.FITR.434292

            /*  VR.ATO.FITR.434292
            If Refundable tax offsets amount is less than Subtotal 3, then Tax offset refunds amount must equal zero.
    
            Legacy Rule Format:
            IF [FITR184] < [FITR216] AND [FITR185] > 0
             RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR184 < ^FITR216 AND ^FITR185 > 0
    
            Data Elements:
    
            ^FITR185 = RP:IncomeTax.TotalRemainderOfRefundableTaxOffsets.Amount
    
            ^FITR184 = RP:IncomeTax.TotalRefundableTaxOffsets.Amount
    
            ^FITR216 = RP:IncomeTax.TaxOffsetCarriedForwardSubTotal.Amount
            */
            assertion = (report.FITR184.GetValueOrDefault() < report.FITR216.GetValueOrDefault() && report.FITR185.GetValueOrDefault() > 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434343",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Tax offset refunds (Remainder of refundable tax offsets) must be zero.",
                    LongDescription = @"If Refundable tax offsets amount is less than Subtotal 3, then Tax offset refunds amount must equal zero.",
                    Location = "/xbrli:xbrl/tns:IncomeTax.TotalRemainderOfRefundableTaxOffsets.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434292" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR184", Value = GetValueOrEmpty(report.FITR184) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR216", Value = GetValueOrEmpty(report.FITR216) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR185", Value = GetValueOrEmpty(report.FITR185) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434292

            #region VR.ATO.FITR.434293

            /*  VR.ATO.FITR.434293
            Mandatory field not supplied
    
            Legacy Rule Format:
            IF  [FITR185] = NULL
              RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR185 = NULL
    
            Data Elements:
    
            ^FITR185 = RP:IncomeTax.TotalRemainderOfRefundableTaxOffsets.Amount
            */
            assertion = (report.FITR185 == null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.001001",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Mandatory field not supplied",
                    Location = "/xbrli:xbrl/tns:IncomeTax.TotalRemainderOfRefundableTaxOffsets.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434293" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR185", Value = GetValueOrEmpty(report.FITR185) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434293

            #region VR.ATO.FITR.434301

            /*  VR.ATO.FITR.434301
            Monetary field - max 13 chars, no sign, may have a decimal point
    
            Legacy Rule Format:
            IF [FITR206] <> NULL AND [FITR206] <> MONETARY(U,13,2)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR206, 'U', 13, 2)
    
            Data Elements:
    
            ^FITR206 = RP:IncomeTax.TaxOffsetExplorationCredits.Amount
            */
            assertion = NotMonetary(report.FITR206, @"U", 13, 2);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:IncomeTax.TaxOffsetExplorationCredits.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434301" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR206", Value = GetValueOrEmpty(report.FITR206) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434301

            #region VR.ATO.FITR.434302

            /*  VR.ATO.FITR.434302
            Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
            Legacy Rule Format:
            IF [FITR207] <> NULL AND [FITR207] <> MONETARY(U,11,0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR207, 'U', 11, 0)
    
            Data Elements:
    
            ^FITR207 = RP.Closing(Instant):Capital.Losses.CarriedForward.Net.Amount
            */
            assertion = NotMonetary(report.FITR207, @"U", 11, 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:Capital.Losses.CarriedForward.Net.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434302" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR207", Value = GetValueOrEmpty(report.FITR207) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434302

            #region VR.ATO.FITR.434309

            /*  VR.ATO.FITR.434309
            The total number RP(Instant) supplied cannot exceed 1
    
            Legacy Rule Format:
            IF COUNT(CONTEXT(RP(Instant))) > 1
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Count(^Context) > 1
    
            Data Elements:
    
            ^Context = RP(Instant)
            */
            assertion = (Count(report.RPInstantCount) > 1);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434309",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"The number of RP(Instant) contexts must not exceed 1",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPInstantLastOccurrenceIndex),
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434309" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434309

            #region VR.ATO.FITR.434310

            /*  VR.ATO.FITR.434310
            A TOFA gain must be present
    
            Legacy Rule Format:
            IF ([FITR48] ="T" AND ([FITR134] = NULL OR
            [FITR134] = 0))
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^FITR48 = 'T' AND (^FITR134 = NULL OR ^FITR134 = 0))
    
            Data Elements:
    
            ^FITR134 = RP.TOFA:Income.GainsTotal.Amount
    
            ^FITR48 = RP:Income.Other.Code
            */
            assertion = (report.FITR48 == @"T" && (report.FITR134 == null || report.FITR134 == 0));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434310",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"A TOFA gain must be present",
                    Location = "/xbrli:xbrl/tns:Income.GainsTotal.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434310" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR48", Value = GetValueOrEmpty(report.FITR48) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR134", Value = GetValueOrEmpty(report.FITR134) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434310

            #region VR.ATO.FITR.434311

            /*  VR.ATO.FITR.434311
            A TOFA loss must be present
    
            Legacy Rule Format:
            IF ([FITR63] ="T - TOFA losses" AND ([FITR135] = NULL OR
            [FITR135] = 0))    
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^FITR63 = 'T - TOFA losses' AND (^FITR135 = NULL OR ^FITR135 = 0))
    
            Data Elements:
    
            ^FITR135 = RP.TOFA:Expense.LossesTotal.Amount
    
            ^FITR63 = RP:Expense.DeductibleOther.Code
            */
            assertion = (report.FITR63 == @"T - TOFA losses" && (report.FITR135 == null || report.FITR135 == 0));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434311",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"A TOFA loss must be present",
                    Location = "/xbrli:xbrl/tns:Expense.LossesTotal.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434311" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR63", Value = GetValueOrEmpty(report.FITR63) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR135", Value = GetValueOrEmpty(report.FITR135) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434311

            #region VR.ATO.FITR.434312

            /*  VR.ATO.FITR.434312
            Monetary field - max 13 chars, no sign, may have a decimal point
    
            Legacy Rule Format:
            IF [FITR209] <> NULL AND [FITR209] <> MONETARY(U,13,2)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR209, 'U', 13, 2)
    
            Data Elements:
    
            ^FITR209 = RP:IncomeTax.PayAsYouGoWithholding.CreditForCapitalGainsWithheldFromForeignResidents.Amount
            */
            assertion = NotMonetary(report.FITR209, @"U", 13, 2);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:IncomeTax.PayAsYouGoWithholding.CreditForCapitalGainsWithheldFromForeignResidents.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434312" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR209", Value = GetValueOrEmpty(report.FITR209) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434312

            #region VR.ATO.FITR.434314

            /*  VR.ATO.FITR.434314
            Tax on no-TFN-quoted contributions must be No-TFN- quoted contributions amount multiplied by 0.32 or 0.02 for Australian superannuation funds
    
            Legacy Rule Format:
            IF [FITR15] = TRUE AND  [FITR43] > 0 AND 
            ([FITR181] <> ([FITR43]*0.32)+/-1 AND [FITR181] <> ([FITR43]*0.02)+/-1)
             RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR15 = TRUE AND  ^FITR43 > 0 AND (OutsideRange(^FITR181, ^FITR43 * 0.32, 1) AND OutsideRange(^FITR181, ^FITR43 * 0.02, 1))
    
            Data Elements:
    
            ^FITR181 = RP:IncomeTax.TFNNotQuotedContributionsTax.Amount
    
            ^FITR15 = RP:SuperannuationFundDetails.AustralianSuperannuationFundStatus.Indicator
    
            ^FITR43 = RP:SuperannuationContribution.EmployerContributionsTaxFileNumberNotQuoted.Amount
            */
            assertion = (report.FITR15 == true && report.FITR43.GetValueOrDefault() > 0 && (OutsideRange(report.FITR181.GetValueOrDefault(), report.FITR43.GetValueOrDefault() * 0.32M, 1) && OutsideRange(report.FITR181.GetValueOrDefault(), report.FITR43.GetValueOrDefault() * 0.02M, 1)));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434314",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"The amount at Tax on no-TFN-quoted contributions is incorrect",
                    LongDescription = @"A 'yes' response has been given to the Australian superannuation fund question however the No-TFN quoted contributions amount is greater than zero and the Tax on no-TFN contributions amount is not equal to No-TFN quoted contributions amount multiplied by 0.02 or 0.32",
                    Location = "/xbrli:xbrl/tns:IncomeTax.TFNNotQuotedContributionsTax.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434314" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR15", Value = GetValueOrEmpty(report.FITR15) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR43", Value = GetValueOrEmpty(report.FITR43) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR181", Value = GetValueOrEmpty(report.FITR181) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434314

            #region VR.ATO.FITR.434315

            /*  VR.ATO.FITR.434315
            Gross tax must not be less than 45% of Taxable income
    
            Legacy Rule Format:
            IF (([FITR15] = FALSE OR [FITR14] = "138") AND [FITR65] > 0 AND [FITR66] < ([FITR65] * 0.45))
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ((^FITR15 = FALSE OR ^FITR14 = '138') AND ^FITR65 > 0 AND ^FITR66 < (^FITR65 * 0.45))
    
            Data Elements:
    
            ^FITR65 = RP:Income.Taxable.Amount
    
            ^FITR15 = RP:SuperannuationFundDetails.AustralianSuperannuationFundStatus.Indicator
    
            ^FITR14 = RP:Party.Type.Code
    
            ^FITR66 = RP:IncomeTax.TaxableGross.Amount
            */
            assertion = ((report.FITR15 == false || report.FITR14 == @"138") && report.FITR65.GetValueOrDefault() > 0 && report.FITR66.GetValueOrDefault() < report.FITR65.GetValueOrDefault() * 0.45M);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434315",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Gross tax must not be less than 45% of Taxable income",
                    LongDescription = @"Gross tax amount must not be less than 45% of the Taxable income or loss amount if the Status of the fund is 'Non-regulated' (code 138) or if the fund is not an Australian superannuation fund",
                    Location = "/xbrli:xbrl/tns:Income.Taxable.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434315" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR15", Value = GetValueOrEmpty(report.FITR15) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR14", Value = report.FITR14 });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR65", Value = GetValueOrEmpty(report.FITR65) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR66", Value = GetValueOrEmpty(report.FITR66) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434315

            #region VR.ATO.FITR.434316

            /*  VR.ATO.FITR.434316
            Year of return must be either the current or next (future) lodgment year associated with this service. If a future year lodgment is required, then it must be lodged no later than the 15th of June of the current year, unless it is an amendment. After the 15th of June, a return for that lodgment year must be submitted from the 1st of July, using the new service associated to that specific lodgment year.
    
            Legacy Rule Format:
            IF ([FITR1] <> SET(2021,2022))
            OR ([FITR1] = 2022 AND DATE(Today) > 2022-06-15 
            AND [FITR171] <> TRUE)
             RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (NotInSet(^FITR1, '"2021","2022"')) OR (^FITR1 = 2022 AND Today() > AsDate('2022-06-15') AND ^FITR171 <> TRUE)
    
            Data Elements:
    
            ^FITR1 = RP:Report.TargetFinancial.Year
    
            ^FITR171 = RP:Report.Amendment.Indicator
            */
            assertion = (!(IsMatch(report.FITR1.GetValueOrDefault().ToString(), @"^(2021|2022)$", RegexOptions.IgnoreCase)) || report.FITR1.GetValueOrDefault() == 2022 && DateTime.Now.Date > AsDate(@"2022-06-15") && report.FITR171 != true);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.500127",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Year of return is invalid",
                    LongDescription = @"Year of return must be either the current or next (future) lodgment year associated with this service. If a future year lodgment is required, then it must be lodged no later than the 15th of June of the current year, unless it is an amendment. After the 15th of June, a return for that lodgment year must be submitted from the 1st of July, using the new service associated to that specific lodgment year.",
                    Location = "/xbrli:xbrl/tns:Report.TargetFinancial.Year",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434316" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR1", Value = GetValueOrEmpty(report.FITR1) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR171", Value = GetValueOrEmpty(report.FITR171) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434316

            #region VR.ATO.FITR.434329

            /*  VR.ATO.FITR.434329
            Tax on no-TFN-quoted contributions must be supplied.
    
            Legacy Rule Format:
            IF [FITR181] = NULL
             RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR181 = NULL
    
            Data Elements:
    
            ^FITR181 = RP:IncomeTax.TFNNotQuotedContributionsTax.Amount
            */
            assertion = (report.FITR181 == null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434329",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Tax on No-TFN-quoted contributions must be supplied. A value must be included in the field, even if the value is zero.",
                    Location = "/xbrli:xbrl/tns:IncomeTax.TFNNotQuotedContributionsTax.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434329" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR181", Value = GetValueOrEmpty(report.FITR181) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434329

            #region VR.ATO.FITR.434330

            /*  VR.ATO.FITR.434330
            Gross Tax must be greater than zero if 'No-TFN-quoted contributions' amount is greater than zero.
    
            Legacy Rule Format:
            IF [FITR43] > 0 AND [FITR66] = 0
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR43 > 0 AND ^FITR66 = 0
    
            Data Elements:
    
            ^FITR66 = RP:IncomeTax.TaxableGross.Amount
    
            ^FITR43 = RP:SuperannuationContribution.EmployerContributionsTaxFileNumberNotQuoted.Amount
            */
            assertion = (report.FITR43.GetValueOrDefault() > 0 && report.FITR66 == 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434330",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Gross Tax must be greater than zero if 'No-TFN-quoted contributions' amount is greater than zero.",
                    Location = "/xbrli:xbrl/tns:IncomeTax.TaxableGross.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434330" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR43", Value = GetValueOrEmpty(report.FITR43) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR66", Value = GetValueOrEmpty(report.FITR66) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434330

            #region VR.ATO.FITR.434331

            /*  VR.ATO.FITR.434331
            Monetary field - max 13 chars, no sign, may have a decimal point
    
            Legacy Rule Format:
            IF [FITR213] <> NULL AND [FITR213] <> MONETARY(U,13,2)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR213, 'U', 13, 2)
    
            Data Elements:
    
            ^FITR213 = RP.LimitedPartners:TaxOffsetClaim.NonRefundableOther.Amount
            */
            assertion = NotMonetary(report.FITR213, @"U", 13, 2);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:TaxOffsetClaim.NonRefundableOther.Amount[@contextRef='" + report.RPLimitedPartnersId + "']",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434331" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR213", Value = GetValueOrEmpty(report.FITR213) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434331

            #region VR.ATO.FITR.434332

            /*  VR.ATO.FITR.434332
            Monetary field - max 13 chars, no sign, may have a decimal point
    
            Legacy Rule Format:
            IF [FITR214] <> NULL AND [FITR214] <> MONETARY(U,13,2)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR214, 'U', 13, 2)
    
            Data Elements:
    
            ^FITR214 = RP.Investor:TaxOffsetClaim.NonRefundableOther.Amount
            */
            assertion = NotMonetary(report.FITR214, @"U", 13, 2);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:TaxOffsetClaim.NonRefundableOther.Amount[@contextRef='" + report.RPInvestorId + "']",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434332" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR214", Value = GetValueOrEmpty(report.FITR214) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434332

            #region VR.ATO.FITR.434333

            /*  VR.ATO.FITR.434333
            The sum of Early stage investor tax offset and Early stage investor tax offset carried forward from previous year cannot exceed $200,000.
    
            Legacy Rule Format:
            IF ([FITR214] <> NULL OR [FITR218] <> NULL) AND ([FITR214] + [FITR218] > 200000)
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^FITR214 <> NULL OR ^FITR218 <> NULL) AND (^FITR214 + ^FITR218 > 200000)
    
            Data Elements:
    
            ^FITR214 = RP.Investor:TaxOffsetClaim.NonRefundableOther.Amount
    
            ^FITR218 = RP.Investor.YPrevious:TaxOffsetClaim.NonRefundableOther.Amount
            */
            assertion = ((report.FITR214 != null || report.FITR218 != null) && report.FITR214.GetValueOrDefault() + report.FITR218.GetValueOrDefault() > 200000);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434333",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"The Early stage investor tax offset is capped at $200,000 per investor and their affiliates combined each income year.",
                    LongDescription = @"The Early stage investor tax offset is capped at $200,000 per investor and their affiliates combined each income year. This $200,000 limit includes the sum of both the current year tax offset and the amount of early stage investor tax offset carried forward from the previous year.",
                    Location = "/xbrli:xbrl/tns:TaxOffsetClaim.NonRefundableOther.Amount[@contextRef='" + report.RPInvestorId + "']",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434333" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR214", Value = GetValueOrEmpty(report.FITR214) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR218", Value = GetValueOrEmpty(report.FITR218) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434333

            #region VR.ATO.FITR.434334

            /*  VR.ATO.FITR.434334
            Monetary field - max 13 chars, no sign, may have a decimal point
    
            Legacy Rule Format:
            IF [FITR215] <> NULL AND [FITR215] <> MONETARY(U,13,2)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR215, 'U', 13, 2)
    
            Data Elements:
    
            ^FITR215 = RP:IncomeTax.TaxOffsetNonRefundableCarryForward.Amount
            */
            assertion = NotMonetary(report.FITR215, @"U", 13, 2);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:IncomeTax.TaxOffsetNonRefundableCarryForward.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434334" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR215", Value = GetValueOrEmpty(report.FITR215) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434334

            #region VR.ATO.FITR.434335

            /*  VR.ATO.FITR.434335
            The amount for Non-refundable carry forward tax offsets must be equal to sum of Early stage Venture Capital Limited Partnership (ESVCLP) tax offset, Early stage venture capital limited partnership tax offset carried forward from previous year, Early stage investor tax offset, and Early stage investor tax offset carried forward from previous year.
    
            Legacy Rule Format:
            IF ([FITR213] <> NULL OR [FITR214] <> NULL OR [FITR217] <> NULL OR [FITR218] <> NULL) AND ([FITR215] <> ([FITR213] + [FITR214] + [FITR217] + [FITR218]))
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^FITR213 <> NULL OR ^FITR214 <> NULL OR ^FITR217 <> NULL OR ^FITR218 <> NULL) AND (^FITR215 <> (^FITR213 + ^FITR214 + ^FITR217 + ^FITR218))
    
            Data Elements:
    
            ^FITR215 = RP:IncomeTax.TaxOffsetNonRefundableCarryForward.Amount
    
            ^FITR213 = RP.LimitedPartners:TaxOffsetClaim.NonRefundableOther.Amount
    
            ^FITR214 = RP.Investor:TaxOffsetClaim.NonRefundableOther.Amount
    
            ^FITR217 = RP.LimitedPartners.YPrevious:TaxOffsetClaim.NonRefundableOther.Amount
    
            ^FITR218 = RP.Investor.YPrevious:TaxOffsetClaim.NonRefundableOther.Amount
            */
            assertion = ((report.FITR213 != null || report.FITR214 != null || report.FITR217 != null || report.FITR218 != null) && report.FITR215.GetValueOrDefault() != report.FITR213.GetValueOrDefault() + report.FITR214.GetValueOrDefault() + report.FITR217.GetValueOrDefault() + report.FITR218.GetValueOrDefault());
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434335",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Non-refundable carry forward tax offsets amount incorrect.",
                    LongDescription = @"Non-refundable carry forward tax offsets amount must be equal to sum of Early stage Venture Capital Limited Partnership (ESVCLP) tax offset, Early stage venture capital limited partnership tax offset carried forward from previous year, Early stage investor tax offset, and Early stage investor tax offset carried forward from previous year.",
                    Location = "/xbrli:xbrl/tns:IncomeTax.TaxOffsetNonRefundableCarryForward.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434335" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR213", Value = GetValueOrEmpty(report.FITR213) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR214", Value = GetValueOrEmpty(report.FITR214) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR217", Value = GetValueOrEmpty(report.FITR217) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR218", Value = GetValueOrEmpty(report.FITR218) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR215", Value = GetValueOrEmpty(report.FITR215) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434335

            #region VR.ATO.FITR.434336

            /*  VR.ATO.FITR.434336
            Monetary field - max 13 chars, no sign, may have a decimal point
    
            Legacy Rule Format:
            IF [FITR216] <> NULL AND [FITR216] <> MONETARY(U,13,2)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR216, 'U', 13, 2)
    
            Data Elements:
    
            ^FITR216 = RP:IncomeTax.TaxOffsetCarriedForwardSubTotal.Amount
            */
            assertion = NotMonetary(report.FITR216, @"U", 13, 2);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:IncomeTax.TaxOffsetCarriedForwardSubTotal.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434336" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR216", Value = GetValueOrEmpty(report.FITR216) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434336

            #region VR.ATO.FITR.434337

            /*  VR.ATO.FITR.434337
            If Non-refundable carry forward tax offset amount is less than Subtotal 2 amount, then Subtotal 3 amount must be equal to Subtotal 2 amount minus Non-refundable carry forward tax offset amount
    
            Legacy Rule Format:
            IF ([FITR215] < [FITR183]) AND ([FITR216] <> ([FITR183] - [FITR215]))
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^FITR215 < ^FITR183) AND (^FITR216 <> (^FITR183 - ^FITR215))
    
            Data Elements:
    
            ^FITR216 = RP:IncomeTax.TaxOffsetCarriedForwardSubTotal.Amount
    
            ^FITR215 = RP:IncomeTax.TaxOffsetNonRefundableCarryForward.Amount
    
            ^FITR183 = RP:IncomeTax.SubtotalTaxPayable.Amount
            */
            assertion = (report.FITR215.GetValueOrDefault() < report.FITR183.GetValueOrDefault() && report.FITR216.GetValueOrDefault() != report.FITR183.GetValueOrDefault() - report.FITR215.GetValueOrDefault());
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434337",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Subtotal 3 amount incorrect.",
                    LongDescription = @"If Non-refundable carry forward tax offset amount is less than Subtotal 2 amount, then Subtotal 3 amount must be equal to Subtotal 2 amount minus Non-refundable carry forward tax offset amount.",
                    Location = "/xbrli:xbrl/tns:IncomeTax.TaxOffsetCarriedForwardSubTotal.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434337" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR215", Value = GetValueOrEmpty(report.FITR215) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR183", Value = GetValueOrEmpty(report.FITR183) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR216", Value = GetValueOrEmpty(report.FITR216) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434337

            #region VR.ATO.FITR.434338

            /*  VR.ATO.FITR.434338
            If Non-refundable carry forward tax offset amount is greater than or equal to Subtotal 2 amount, then Subtotal 3 amount must be zero
    
            Legacy Rule Format:
            IF ([FITR215] >= [FITR183]) AND ([FITR216] <> 0)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^FITR215 >= ^FITR183) AND (^FITR216 <> 0)
    
            Data Elements:
    
            ^FITR216 = RP:IncomeTax.TaxOffsetCarriedForwardSubTotal.Amount
    
            ^FITR215 = RP:IncomeTax.TaxOffsetNonRefundableCarryForward.Amount
    
            ^FITR183 = RP:IncomeTax.SubtotalTaxPayable.Amount
            */
            assertion = (report.FITR215.GetValueOrDefault() >= report.FITR183.GetValueOrDefault() && report.FITR216 != 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434338",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Subtotal 3 amount must equal zero.",
                    LongDescription = @"If Non-refundable carry forward tax offset amount is greater than or equal to Subtotal 2 amount, then Subtotal 3 amount must be zero.",
                    Location = "/xbrli:xbrl/tns:IncomeTax.TaxOffsetCarriedForwardSubTotal.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434338" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR215", Value = GetValueOrEmpty(report.FITR215) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR183", Value = GetValueOrEmpty(report.FITR183) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR216", Value = GetValueOrEmpty(report.FITR216) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434338

            #region VR.ATO.FITR.434346

            /*  VR.ATO.FITR.434346
            No-TFN-quoted contributions must be supplied.
    
            Legacy Rule Format:
            IF [FITR43] = NULL
             RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR43 = NULL
    
            Data Elements:
    
            ^FITR43 = RP:SuperannuationContribution.EmployerContributionsTaxFileNumberNotQuoted.Amount
            */
            assertion = (report.FITR43 == null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434346",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"No-TFN-quoted contributions must be supplied.",
                    LongDescription = @"No-TFN-quoted contributions must be supplied. A value must be included in the field, even if the value is zero.",
                    Location = "/xbrli:xbrl/tns:SuperannuationContribution.EmployerContributionsTaxFileNumberNotQuoted.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434346" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR43", Value = GetValueOrEmpty(report.FITR43) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434346

            #region VR.ATO.FITR.434347

            /*  VR.ATO.FITR.434347
            Monetary field - max 13 chars, no sign, may have a decimal point
    
            Legacy Rule Format:
            IF [FITR218] <> NULL AND [FITR218] <> MONETARY(U,13,2)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR218, 'U', 13, 2)
    
            Data Elements:
    
            ^FITR218 = RP.Investor.YPrevious:TaxOffsetClaim.NonRefundableOther.Amount
            */
            assertion = NotMonetary(report.FITR218, @"U", 13, 2);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:TaxOffsetClaim.NonRefundableOther.Amount[@contextRef='" + report.RPInvestorYPreviousId + "']",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434347" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR218", Value = GetValueOrEmpty(report.FITR218) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434347

            #region VR.ATO.FITR.434348

            /*  VR.ATO.FITR.434348
            Monetary field - max 13 chars, no sign, may have a decimal point
    
            Legacy Rule Format:
            IF [FITR217] <> NULL AND [FITR217] <> MONETARY(U,13,2)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            NotMonetary(^FITR217, 'U', 13, 2)
    
            Data Elements:
    
            ^FITR217 = RP.LimitedPartners.YPrevious:TaxOffsetClaim.NonRefundableOther.Amount
            */
            assertion = NotMonetary(report.FITR217, @"U", 13, 2);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.400011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                    LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                    Location = "/xbrli:xbrl/tns:TaxOffsetClaim.NonRefundableOther.Amount[@contextRef='" + report.RPLimitedPartnersYPreviousId + "']",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434348" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR217", Value = GetValueOrEmpty(report.FITR217) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434348

            #region VR.ATO.FITR.434350

            /*  VR.ATO.FITR.434350
            If the answer to the Country by Country Reporting Entity indicator is 'yes' (true), then the Significant Global Entity indicator must be 'yes' (true)
    
            Legacy Rule Format:
            IF [FITR212] <> TRUE AND [FITR219] = TRUE
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR212 <> TRUE AND ^FITR219 = TRUE
    
            Data Elements:
    
            ^FITR212 = RP:OrganisationDetails.SignificantGlobalEntityStatus.Indicator
    
            ^FITR219 = RP:InternationalDealings.CountryByCountryReportingEntityStatus.Indicator
            */
            assertion = (report.FITR212 != true && report.FITR219 == true);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.434350",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Answer to Significant Global Entity indicator must be 'yes' (true)",
                    LongDescription = @"If the answer to the Country by Country Reporting Entity indicator is 'yes' (true), then the Significant Global Entity indicator must be 'yes' (true)",
                    Location = "/xbrli:xbrl/tns:OrganisationDetails.SignificantGlobalEntityStatus.Indicator",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434350" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR212", Value = GetValueOrEmpty(report.FITR212) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR219", Value = GetValueOrEmpty(report.FITR219) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.434350

            #region VR.ATO.FITR.W00001

            /*  VR.ATO.FITR.W00001
            The answer to the Significant Global Entity indicator is 'yes' (true) and Country by Country Reporting Entity indicator has not been provided. Check and correct if necessary.
    
            Legacy Rule Format:
            IF [FITR219] = NULL AND [FITR212] = TRUE
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR219 = NULL AND ^FITR212 = TRUE
    
            Data Elements:
    
            ^FITR219 = RP:InternationalDealings.CountryByCountryReportingEntityStatus.Indicator
    
            ^FITR212 = RP:OrganisationDetails.SignificantGlobalEntityStatus.Indicator
            */
            assertion = (report.FITR219 == null && report.FITR212 == true);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.FITR.W00001",
                    Severity = ProcessMessageSeverity.Warning,
                    Description = @"Check Country by Country Reporting Entity indicator",
                    LongDescription = @"The answer to the Significant Global Entity indicator is 'yes' (true) and Country by Country Reporting Entity indicator has not been provided. Check and correct if necessary.",
                    Location = "/xbrli:xbrl/tns:InternationalDealings.CountryByCountryReportingEntityStatus.Indicator",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.W00001" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR219", Value = GetValueOrEmpty(report.FITR219) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR212", Value = GetValueOrEmpty(report.FITR212) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.FITR.W00001

            #region VR.ATO.GEN.000201

            /*  VR.ATO.GEN.000201
            Period start date must be prior to Period end date
    
            Legacy Rule Format:
            IF period.startDate >= period.endDate WHERE CONTEXT(ALL)
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^StartDate >= ^EndDate
    
            Data Elements:
    
            ^StartDate = INT
    
            ^EndDate = INT
            */
            assertion = ((report.INTExists == true) && (report.INTStartDate >= report.INTEndDate));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.200009",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Period start date must be prior to Period end date",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.INTOccurrenceIndex) + "/xbrli:period/xbrli:startDate",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.000201" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.000201

            #region VR.ATO.GEN.000201

            /*  VR.ATO.GEN.000201
            Period start date must be prior to Period end date
    
            Legacy Rule Format:
            IF period.startDate >= period.endDate WHERE CONTEXT(ALL)
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^StartDate >= ^EndDate
    
            Data Elements:
    
            ^StartDate = RP
    
            ^EndDate = RP
            */
            assertion = ((report.RPExists == true) && (report.RPStartDate >= report.RPEndDate));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.200009",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Period start date must be prior to Period end date",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPOccurrenceIndex) + "/xbrli:period/xbrli:startDate",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.000201" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.000201

            #region VR.ATO.GEN.000201

            /*  VR.ATO.GEN.000201
            Period start date must be prior to Period end date
    
            Legacy Rule Format:
            IF period.startDate >= period.endDate WHERE CONTEXT(ALL)
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^StartDate >= ^EndDate
    
            Data Elements:
    
            ^StartDate = RP:RP.CFCLC
    
            ^EndDate = RP:RP.CFCLC
            */
            assertion = ((report.RPCFCLCExists == true) && (report.RPCFCLCStartDate >= report.RPCFCLCEndDate));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.200009",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Period start date must be prior to Period end date",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPCFCLCOccurrenceIndex) + "/xbrli:period/xbrli:startDate",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.000201" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.000201

            #region VR.ATO.GEN.000201

            /*  VR.ATO.GEN.000201
            Period start date must be prior to Period end date
    
            Legacy Rule Format:
            IF period.startDate >= period.endDate WHERE CONTEXT(ALL)
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^StartDate >= ^EndDate
    
            Data Elements:
    
            ^StartDate = RP:RP.CFCUC
    
            ^EndDate = RP:RP.CFCUC
            */
            assertion = ((report.RPCFCUCExists == true) && (report.RPCFCUCStartDate >= report.RPCFCUCEndDate));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.200009",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Period start date must be prior to Period end date",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPCFCUCOccurrenceIndex) + "/xbrli:period/xbrli:startDate",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.000201" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.000201

            #region VR.ATO.GEN.000201

            /*  VR.ATO.GEN.000201
            Period start date must be prior to Period end date
    
            Legacy Rule Format:
            IF period.startDate >= period.endDate WHERE CONTEXT(ALL)
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^StartDate >= ^EndDate
    
            Data Elements:
    
            ^StartDate = RP:RP.Closing
    
            ^EndDate = RP:RP.Closing
            */
            assertion = ((report.RPClosingExists == true) && (report.RPClosingStartDate >= report.RPClosingEndDate));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.200009",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Period start date must be prior to Period end date",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPClosingOccurrenceIndex) + "/xbrli:period/xbrli:startDate",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.000201" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.000201

            #region VR.ATO.GEN.000201

            /*  VR.ATO.GEN.000201
            Period start date must be prior to Period end date
    
            Legacy Rule Format:
            IF period.startDate >= period.endDate WHERE CONTEXT(ALL)
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^StartDate >= ^EndDate
    
            Data Elements:
    
            ^StartDate = RP:RP.Investor
    
            ^EndDate = RP:RP.Investor
            */
            assertion = ((report.RPInvestorExists == true) && (report.RPInvestorStartDate >= report.RPInvestorEndDate));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.200009",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Period start date must be prior to Period end date",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPInvestorOccurrenceIndex) + "/xbrli:period/xbrli:startDate",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.000201" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.000201

            #region VR.ATO.GEN.000201

            /*  VR.ATO.GEN.000201
            Period start date must be prior to Period end date
    
            Legacy Rule Format:
            IF period.startDate >= period.endDate WHERE CONTEXT(ALL)
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^StartDate >= ^EndDate
    
            Data Elements:
    
            ^StartDate = RP.Investor:RP.Investor.YPrevious
    
            ^EndDate = RP.Investor:RP.Investor.YPrevious
            */
            assertion = ((report.RPInvestorYPreviousExists == true) && (report.RPInvestorYPreviousStartDate >= report.RPInvestorYPreviousEndDate));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.200009",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Period start date must be prior to Period end date",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPInvestorYPreviousOccurrenceIndex) + "/xbrli:period/xbrli:startDate",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.000201" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.000201

            #region VR.ATO.GEN.000201

            /*  VR.ATO.GEN.000201
            Period start date must be prior to Period end date
    
            Legacy Rule Format:
            IF period.startDate >= period.endDate WHERE CONTEXT(ALL)
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^StartDate >= ^EndDate
    
            Data Elements:
    
            ^StartDate = RP:RP.JAUS
    
            ^EndDate = RP:RP.JAUS
            */
            assertion = ((report.RPJAUSExists == true) && (report.RPJAUSStartDate >= report.RPJAUSEndDate));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.200009",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Period start date must be prior to Period end date",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPJAUSOccurrenceIndex) + "/xbrli:period/xbrli:startDate",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.000201" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.000201

            #region VR.ATO.GEN.000201

            /*  VR.ATO.GEN.000201
            Period start date must be prior to Period end date
    
            Legacy Rule Format:
            IF period.startDate >= period.endDate WHERE CONTEXT(ALL)
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^StartDate >= ^EndDate
    
            Data Elements:
    
            ^StartDate = RP:RP.JFOREIGN
    
            ^EndDate = RP:RP.JFOREIGN
            */
            assertion = ((report.RPJFOREIGNExists == true) && (report.RPJFOREIGNStartDate >= report.RPJFOREIGNEndDate));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.200009",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Period start date must be prior to Period end date",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPJFOREIGNOccurrenceIndex) + "/xbrli:period/xbrli:startDate",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.000201" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.000201

            #region VR.ATO.GEN.000201

            /*  VR.ATO.GEN.000201
            Period start date must be prior to Period end date
    
            Legacy Rule Format:
            IF period.startDate >= period.endDate WHERE CONTEXT(ALL)
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^StartDate >= ^EndDate
    
            Data Elements:
    
            ^StartDate = RP:RP.LimitedPartners
    
            ^EndDate = RP:RP.LimitedPartners
            */
            assertion = ((report.RPLimitedPartnersExists == true) && (report.RPLimitedPartnersStartDate >= report.RPLimitedPartnersEndDate));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.200009",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Period start date must be prior to Period end date",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPLimitedPartnersOccurrenceIndex) + "/xbrli:period/xbrli:startDate",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.000201" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.000201

            #region VR.ATO.GEN.000201

            /*  VR.ATO.GEN.000201
            Period start date must be prior to Period end date
    
            Legacy Rule Format:
            IF period.startDate >= period.endDate WHERE CONTEXT(ALL)
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^StartDate >= ^EndDate
    
            Data Elements:
    
            ^StartDate = RP.LimitedPartners:RP.LimitedPartners.YPrevious
    
            ^EndDate = RP.LimitedPartners:RP.LimitedPartners.YPrevious
            */
            assertion = ((report.RPLimitedPartnersYPreviousExists == true) && (report.RPLimitedPartnersYPreviousStartDate >= report.RPLimitedPartnersYPreviousEndDate));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.200009",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Period start date must be prior to Period end date",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPLimitedPartnersYPreviousOccurrenceIndex) + "/xbrli:period/xbrli:startDate",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.000201" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.000201

            #region VR.ATO.GEN.000201

            /*  VR.ATO.GEN.000201
            Period start date must be prior to Period end date
    
            Legacy Rule Format:
            IF period.startDate >= period.endDate WHERE CONTEXT(ALL)
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^StartDate >= ^EndDate
    
            Data Elements:
    
            ^StartDate = RP:RP.TOFA
    
            ^EndDate = RP:RP.TOFA
            */
            assertion = ((report.RPTOFAExists == true) && (report.RPTOFAStartDate >= report.RPTOFAEndDate));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.200009",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Period start date must be prior to Period end date",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPTOFAOccurrenceIndex) + "/xbrli:period/xbrli:startDate",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.000201" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.000201

            #region VR.ATO.GEN.000209

            /*  VR.ATO.GEN.000209
            Invalid context. The number of Reporting party contexts must equal 1
    
            Legacy Rule Format:
            IF COUNT(RP) <> 1
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Count(^Context) <> 1
    
            Data Elements:
    
            ^Context = RP
            */
            assertion = (Count(report.RPCount) != 1);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.430296",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Invalid context. The number of Reporting party contexts must equal 1",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPLastOccurrenceIndex),
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.000209" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.000209

            #region VR.ATO.GEN.000227

            /*  VR.ATO.GEN.000227
            Invalid context. The number of Intermediary contexts must not exceed 1
    
            Legacy Rule Format:
            IF COUNT(INT) > 1
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Count(^Context) > 1
    
            Data Elements:
    
            ^Context = INT
            */
            assertion = (Count(report.INTCount) > 1);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.436280",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Invalid context. The number of Intermediary contexts must not exceed 1",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.INTLastOccurrenceIndex),
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.000227" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.000227

            #region VR.ATO.GEN.000406

            /*  VR.ATO.GEN.000406
            Address line 1 contains incorrect 'care of' reference.  The only acceptable 'care of' reference is C/-
    
            Legacy Rule Format:
            IF (pyde.xx.xx:AddressDetails.Line1.Text <> NULLORBLANK) AND (pyde.xx.xx:AddressDetails.Line1.Text STARTSWITH SET("C/O ","C/ ","Care Of ","CO ")) OR (pyde.xx.xx:AddressDetails.Line1.Text CONTAINS SET(" C/O "," C/ "," Care Of "))
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^FITR8 <> BLANK) AND (StartsWithSet(^FITR8,'"C/O ","C/ ","Care Of ","CO "')) OR (ContainsSet(^FITR8, '" C/O "," C/ "," Care Of "'))
    
            Data Elements:
    
            ^FITR8 = RP:AddressDetails:AddressDetails.Line1.Text WHERE [AddressDetails.Usage.Code='POS'].[AddressDetails.Currency.Code='C']
            */
            assertion = (string.IsNullOrWhiteSpace(report.FITR8) != true && IsMatch(report.FITR8, @"^(C/O |C/ |Care Of |CO )", RegexOptions.IgnoreCase) || IsMatch(report.FITR8, @" C/O | C/ | Care Of ", RegexOptions.IgnoreCase));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000406",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Address line 1 contains incorrect 'care of' reference.  The only acceptable 'care of' reference is C/-",
                    Location = "/xbrli:xbrl/tns:AddressDetails" + OccurrenceIndex(report.AddressDetailsCollection_POS_C_RPOccurrenceIndex) + "/tns:AddressDetails.Line1.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.000406" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR8", Value = report.FITR8 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.000406

            #region VR.ATO.GEN.000408

            /*  VR.ATO.GEN.000408
            Care of' is not valid for address line 2
    
            Legacy Rule Format:
            IF (pyde.xx.xx:AddressDetails.Line2.Text <> NULLORBLANK) AND (pyde.xx.xx:AddressDetails.Line2.Text STARTSWITH SET("C/- ","C/O ","C/ ","Care Of ","CO ") OR pyde.xx.xx:AddressDetails.Line2.Text CONTAINS SET(" C/- "," C/O "," C/ "," Care Of "))
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^FITR146 <> BLANK) AND (StartsWithSet(^FITR146,'"C/- ","C/O ","C/ ","Care Of ","CO "') OR ContainsSet(^FITR146, '" C/- "," C/O "," C/ "," Care Of "'))
    
            Data Elements:
    
            ^FITR146 = RP:AddressDetails:AddressDetails.Line2.Text WHERE [AddressDetails.Usage.Code='POS'].[AddressDetails.Currency.Code='C']
            */
            assertion = (string.IsNullOrWhiteSpace(report.FITR146) != true && (IsMatch(report.FITR146, @"^(C/- |C/O |C/ |Care Of |CO )", RegexOptions.IgnoreCase) || IsMatch(report.FITR146, @" C/- | C/O | C/ | Care Of ", RegexOptions.IgnoreCase)));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000408",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Care of' is not valid for address line 2",
                    Location = "/xbrli:xbrl/tns:AddressDetails" + OccurrenceIndex(report.AddressDetailsCollection_POS_C_RPOccurrenceIndex) + "/tns:AddressDetails.Line2.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.000408" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR146", Value = GetValueOrEmpty(report.FITR146) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.000408

            #region VR.ATO.GEN.000458

            /*  VR.ATO.GEN.000458
            Suffix is invalid
    
            Legacy Rule Format:
            IF (pyde.xx.xx:PersonNameDetails.NameSuffix.Text <> NULLORBLANK) AND (pyde.xx.xx:PersonNameDetails.NameSuffix.Text <> SET(DOMAIN(SUFFIX CODES)))
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^FITR113 <> BLANK) AND (NotInDomain(^FITR113, 'SUFFIX CODES'))
    
            Data Elements:
    
            ^FITR113 = INT:PersonNameDetails:PersonNameDetails.NameSuffix.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
            */
            assertion = (string.IsNullOrWhiteSpace(report.FITR113) != true && !(IsMatch(report.FITR113, @"^(ESQ|II|III|IV|JNR|JP|MHA|MHR|MLA|MLC|MP|QC|SNR)$", RegexOptions.IgnoreCase)));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000458",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Suffix is invalid",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_INTOccurrenceIndex) + "/tns:PersonNameDetails.NameSuffix.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.000458" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR113", Value = GetValueOrEmpty(report.FITR113) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.000458

            #region VR.ATO.GEN.000458

            /*  VR.ATO.GEN.000458
            Suffix is invalid
    
            Legacy Rule Format:
            IF (pyde.xx.xx:PersonNameDetails.NameSuffix.Text <> NULLORBLANK) AND (pyde.xx.xx:PersonNameDetails.NameSuffix.Text <> SET(DOMAIN(SUFFIX CODES)))
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^FITR104 <> BLANK) AND (NotInDomain(^FITR104, 'SUFFIX CODES'))
    
            Data Elements:
    
            ^FITR104 = RP:PersonNameDetails:PersonNameDetails.NameSuffix.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
            */
            assertion = (string.IsNullOrWhiteSpace(report.FITR104) != true && !(IsMatch(report.FITR104, @"^(ESQ|II|III|IV|JNR|JP|MHA|MHR|MLA|MLC|MP|QC|SNR)$", RegexOptions.IgnoreCase)));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000458",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Suffix is invalid",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_RPOccurrenceIndex) + "/tns:PersonNameDetails.NameSuffix.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.000458" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR104", Value = GetValueOrEmpty(report.FITR104) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.000458

            #region VR.ATO.GEN.000459

            /*  VR.ATO.GEN.000459
            Title is invalid
    
            Legacy Rule Format:
            IF (pyde.xx.xx:PersonNameDetails.Title.Text <> NULLORBLANK) AND (pyde.xx.xx:PersonNameDetails.Title.Text <> SET(DOMAIN(TITLE CODES)))
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^FITR111 <> BLANK) AND (NotInDomain(^FITR111, 'TITLE CODES'))
    
            Data Elements:
    
            ^FITR111 = INT:PersonNameDetails:PersonNameDetails.Title.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
            */
            assertion = (string.IsNullOrWhiteSpace(report.FITR111) != true && !(IsMatch(report.FITR111, @"^(2LT|AB|ABBOT|AC|ACM|ADM|AIR CDRE|ALDERMAN|AM|ARCHBISHOP|ARCHDEACON|ASSOC PROF|AVM|BARON|BARONESS|BISHOP|BR|BRIG|CANON|CAPT|CARDINAL|CDRE|CDT|CHAP|CMDR|CMM|COL|CONST|COUNT|COUNTESS|CPL|CPO|DAME|DEACON|DEACONESS|DEAN|DEPUTY SUPT|DR|DUCHESS|DUKE|EARL|EF|FLGOFF|FLT LT|FR|FSGT|GEN|GNR|GP CAPT|HON JUDGE|HON JUST|HON|HRH|INSP|JUDGE|JUST|LAC|LACW|LADY|LBDR|LCPL|LORD|LS|LT CMDR|LT COL|LT GEN|LT|MADAM|MAJ GEN|MAJ|MASTER|MATRON|MAYOR|MAYORESS|MIDN|MISS|MON|MOST REV|MR|MRS|MS|PASTOR|PATRIARCH|PLT OFF|PO|PRIOR|PROF|PTE|RABBI|RADM|RECTOR|REV|RF|RT HON|RT REV MON|RT REV|RTREV BISHOP|SBLT|SEN|SGT|SIR|SMN|SNR CONST|SQN LDR|SR|SSGT|SUPR|SWAMI|TF|VADM|VERY REV|VICAR|VISCOUNT|WG CDR|WO|WO1|WO2)$", RegexOptions.IgnoreCase)));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000459",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Title is invalid",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_INTOccurrenceIndex) + "/tns:PersonNameDetails.Title.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.000459" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR111", Value = GetValueOrEmpty(report.FITR111) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.000459

            #region VR.ATO.GEN.000459

            /*  VR.ATO.GEN.000459
            Title is invalid
    
            Legacy Rule Format:
            IF (pyde.xx.xx:PersonNameDetails.Title.Text <> NULLORBLANK) AND (pyde.xx.xx:PersonNameDetails.Title.Text <> SET(DOMAIN(TITLE CODES)))
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^FITR102 <> BLANK) AND (NotInDomain(^FITR102, 'TITLE CODES'))
    
            Data Elements:
    
            ^FITR102 = RP:PersonNameDetails:PersonNameDetails.Title.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
            */
            assertion = (string.IsNullOrWhiteSpace(report.FITR102) != true && !(IsMatch(report.FITR102, @"^(2LT|AB|ABBOT|AC|ACM|ADM|AIR CDRE|ALDERMAN|AM|ARCHBISHOP|ARCHDEACON|ASSOC PROF|AVM|BARON|BARONESS|BISHOP|BR|BRIG|CANON|CAPT|CARDINAL|CDRE|CDT|CHAP|CMDR|CMM|COL|CONST|COUNT|COUNTESS|CPL|CPO|DAME|DEACON|DEACONESS|DEAN|DEPUTY SUPT|DR|DUCHESS|DUKE|EARL|EF|FLGOFF|FLT LT|FR|FSGT|GEN|GNR|GP CAPT|HON JUDGE|HON JUST|HON|HRH|INSP|JUDGE|JUST|LAC|LACW|LADY|LBDR|LCPL|LORD|LS|LT CMDR|LT COL|LT GEN|LT|MADAM|MAJ GEN|MAJ|MASTER|MATRON|MAYOR|MAYORESS|MIDN|MISS|MON|MOST REV|MR|MRS|MS|PASTOR|PATRIARCH|PLT OFF|PO|PRIOR|PROF|PTE|RABBI|RADM|RECTOR|REV|RF|RT HON|RT REV MON|RT REV|RTREV BISHOP|SBLT|SEN|SGT|SIR|SMN|SNR CONST|SQN LDR|SR|SSGT|SUPR|SWAMI|TF|VADM|VERY REV|VICAR|VISCOUNT|WG CDR|WO|WO1|WO2)$", RegexOptions.IgnoreCase)));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000459",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Title is invalid",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_RPOccurrenceIndex) + "/tns:PersonNameDetails.Title.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.000459" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR102", Value = GetValueOrEmpty(report.FITR102) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.000459

            #region VR.ATO.GEN.001023

            /*  VR.ATO.GEN.001023
            TFN in entity identifier of context does not match the TFN provided for the Reporting Party
    
            Legacy Rule Format:
            IF (entity.identifier.TFN WHERE CONTEXT(ALL)) <> RP:entity.identifier.TFN
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^Context <> NULL AND ^OtherTFN <> ^TFN
    
            Data Elements:
    
            ^Context = RP(Instant)
    
            ^OtherTFN = RP(Instant)
    
            ^TFN = RP
            */
            assertion = (report.RPInstantExists != false && report.RPInstantIdentifierTFN != report.RPIdentifierTFN);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.001023",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Unexpected TFN mismatch with Reporting Party",
                    LongDescription = @"TFN in entity identifier of context does not match the TFN provided for the Reporting Party",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPInstantOccurrenceIndex) + "/xbrli:entity/xbrli:identifier",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.001023" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.001023

            #region VR.ATO.GEN.001023

            /*  VR.ATO.GEN.001023
            TFN in entity identifier of context does not match the TFN provided for the Reporting Party
    
            Legacy Rule Format:
            IF (entity.identifier.TFN WHERE CONTEXT(ALL)) <> RP:entity.identifier.TFN
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^Context <> NULL AND ^OtherTFN <> ^TFN
    
            Data Elements:
    
            ^Context = RP:RP.CFCLC
    
            ^OtherTFN = RP:RP.CFCLC
    
            ^TFN = RP
            */
            assertion = (report.RPCFCLCExists != false && report.RPCFCLCIdentifierTFN != report.RPIdentifierTFN);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.001023",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Unexpected TFN mismatch with Reporting Party",
                    LongDescription = @"TFN in entity identifier of context does not match the TFN provided for the Reporting Party",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPCFCLCOccurrenceIndex) + "/xbrli:entity/xbrli:identifier",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.001023" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.001023

            #region VR.ATO.GEN.001023

            /*  VR.ATO.GEN.001023
            TFN in entity identifier of context does not match the TFN provided for the Reporting Party
    
            Legacy Rule Format:
            IF (entity.identifier.TFN WHERE CONTEXT(ALL)) <> RP:entity.identifier.TFN
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^Context <> NULL AND ^OtherTFN <> ^TFN
    
            Data Elements:
    
            ^Context = RP:RP.CFCUC
    
            ^OtherTFN = RP:RP.CFCUC
    
            ^TFN = RP
            */
            assertion = (report.RPCFCUCExists != false && report.RPCFCUCIdentifierTFN != report.RPIdentifierTFN);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.001023",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Unexpected TFN mismatch with Reporting Party",
                    LongDescription = @"TFN in entity identifier of context does not match the TFN provided for the Reporting Party",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPCFCUCOccurrenceIndex) + "/xbrli:entity/xbrli:identifier",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.001023" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.001023

            #region VR.ATO.GEN.001023

            /*  VR.ATO.GEN.001023
            TFN in entity identifier of context does not match the TFN provided for the Reporting Party
    
            Legacy Rule Format:
            IF (entity.identifier.TFN WHERE CONTEXT(ALL)) <> RP:entity.identifier.TFN
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^Context <> NULL AND ^OtherTFN <> ^TFN
    
            Data Elements:
    
            ^Context = RP:RP.Closing
    
            ^OtherTFN = RP:RP.Closing
    
            ^TFN = RP
            */
            assertion = (report.RPClosingExists != false && report.RPClosingIdentifierTFN != report.RPIdentifierTFN);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.001023",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Unexpected TFN mismatch with Reporting Party",
                    LongDescription = @"TFN in entity identifier of context does not match the TFN provided for the Reporting Party",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPClosingOccurrenceIndex) + "/xbrli:entity/xbrli:identifier",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.001023" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.001023

            #region VR.ATO.GEN.001023

            /*  VR.ATO.GEN.001023
            TFN in entity identifier of context does not match the TFN provided for the Reporting Party
    
            Legacy Rule Format:
            IF (entity.identifier.TFN WHERE CONTEXT(ALL)) <> RP:entity.identifier.TFN
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^Context <> NULL AND ^OtherTFN <> ^TFN
    
            Data Elements:
    
            ^Context = RP.Closing(Instant)
    
            ^OtherTFN = RP.Closing(Instant)
    
            ^TFN = RP
            */
            assertion = (report.RPClosingInstantExists != false && report.RPClosingInstantIdentifierTFN != report.RPIdentifierTFN);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.001023",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Unexpected TFN mismatch with Reporting Party",
                    LongDescription = @"TFN in entity identifier of context does not match the TFN provided for the Reporting Party",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPClosingInstantOccurrenceIndex) + "/xbrli:entity/xbrli:identifier",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.001023" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.001023

            #region VR.ATO.GEN.001023

            /*  VR.ATO.GEN.001023
            TFN in entity identifier of context does not match the TFN provided for the Reporting Party
    
            Legacy Rule Format:
            IF (entity.identifier.TFN WHERE CONTEXT(ALL)) <> RP:entity.identifier.TFN
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^Context <> NULL AND ^OtherTFN <> ^TFN
    
            Data Elements:
    
            ^Context = RP:RP.Investor
    
            ^OtherTFN = RP:RP.Investor
    
            ^TFN = RP
            */
            assertion = (report.RPInvestorExists != false && report.RPInvestorIdentifierTFN != report.RPIdentifierTFN);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.001023",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Unexpected TFN mismatch with Reporting Party",
                    LongDescription = @"TFN in entity identifier of context does not match the TFN provided for the Reporting Party",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPInvestorOccurrenceIndex) + "/xbrli:entity/xbrli:identifier",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.001023" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.001023

            #region VR.ATO.GEN.001023

            /*  VR.ATO.GEN.001023
            TFN in entity identifier of context does not match the TFN provided for the Reporting Party
    
            Legacy Rule Format:
            IF (entity.identifier.TFN WHERE CONTEXT(ALL)) <> RP:entity.identifier.TFN
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^Context <> NULL AND ^OtherTFN <> ^TFN
    
            Data Elements:
    
            ^Context = RP.Investor:RP.Investor.YPrevious
    
            ^OtherTFN = RP.Investor:RP.Investor.YPrevious
    
            ^TFN = RP
            */
            assertion = (report.RPInvestorYPreviousExists != false && report.RPInvestorYPreviousIdentifierTFN != report.RPIdentifierTFN);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.001023",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Unexpected TFN mismatch with Reporting Party",
                    LongDescription = @"TFN in entity identifier of context does not match the TFN provided for the Reporting Party",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPInvestorYPreviousOccurrenceIndex) + "/xbrli:entity/xbrli:identifier",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.001023" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.001023

            #region VR.ATO.GEN.001023

            /*  VR.ATO.GEN.001023
            TFN in entity identifier of context does not match the TFN provided for the Reporting Party
    
            Legacy Rule Format:
            IF (entity.identifier.TFN WHERE CONTEXT(ALL)) <> RP:entity.identifier.TFN
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^Context <> NULL AND ^OtherTFN <> ^TFN
    
            Data Elements:
    
            ^Context = RP:RP.JAUS
    
            ^OtherTFN = RP:RP.JAUS
    
            ^TFN = RP
            */
            assertion = (report.RPJAUSExists != false && report.RPJAUSIdentifierTFN != report.RPIdentifierTFN);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.001023",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Unexpected TFN mismatch with Reporting Party",
                    LongDescription = @"TFN in entity identifier of context does not match the TFN provided for the Reporting Party",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPJAUSOccurrenceIndex) + "/xbrli:entity/xbrli:identifier",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.001023" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.001023

            #region VR.ATO.GEN.001023

            /*  VR.ATO.GEN.001023
            TFN in entity identifier of context does not match the TFN provided for the Reporting Party
    
            Legacy Rule Format:
            IF (entity.identifier.TFN WHERE CONTEXT(ALL)) <> RP:entity.identifier.TFN
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^Context <> NULL AND ^OtherTFN <> ^TFN
    
            Data Elements:
    
            ^Context = RP:RP.JFOREIGN
    
            ^OtherTFN = RP:RP.JFOREIGN
    
            ^TFN = RP
            */
            assertion = (report.RPJFOREIGNExists != false && report.RPJFOREIGNIdentifierTFN != report.RPIdentifierTFN);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.001023",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Unexpected TFN mismatch with Reporting Party",
                    LongDescription = @"TFN in entity identifier of context does not match the TFN provided for the Reporting Party",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPJFOREIGNOccurrenceIndex) + "/xbrli:entity/xbrli:identifier",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.001023" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.001023

            #region VR.ATO.GEN.001023

            /*  VR.ATO.GEN.001023
            TFN in entity identifier of context does not match the TFN provided for the Reporting Party
    
            Legacy Rule Format:
            IF (entity.identifier.TFN WHERE CONTEXT(ALL)) <> RP:entity.identifier.TFN
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^Context <> NULL AND ^OtherTFN <> ^TFN
    
            Data Elements:
    
            ^Context = RP:RP.LimitedPartners
    
            ^OtherTFN = RP:RP.LimitedPartners
    
            ^TFN = RP
            */
            assertion = (report.RPLimitedPartnersExists != false && report.RPLimitedPartnersIdentifierTFN != report.RPIdentifierTFN);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.001023",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Unexpected TFN mismatch with Reporting Party",
                    LongDescription = @"TFN in entity identifier of context does not match the TFN provided for the Reporting Party",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPLimitedPartnersOccurrenceIndex) + "/xbrli:entity/xbrli:identifier",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.001023" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.001023

            #region VR.ATO.GEN.001023

            /*  VR.ATO.GEN.001023
            TFN in entity identifier of context does not match the TFN provided for the Reporting Party
    
            Legacy Rule Format:
            IF (entity.identifier.TFN WHERE CONTEXT(ALL)) <> RP:entity.identifier.TFN
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^Context <> NULL AND ^OtherTFN <> ^TFN
    
            Data Elements:
    
            ^Context = RP.LimitedPartners:RP.LimitedPartners.YPrevious
    
            ^OtherTFN = RP.LimitedPartners:RP.LimitedPartners.YPrevious
    
            ^TFN = RP
            */
            assertion = (report.RPLimitedPartnersYPreviousExists != false && report.RPLimitedPartnersYPreviousIdentifierTFN != report.RPIdentifierTFN);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.001023",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Unexpected TFN mismatch with Reporting Party",
                    LongDescription = @"TFN in entity identifier of context does not match the TFN provided for the Reporting Party",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPLimitedPartnersYPreviousOccurrenceIndex) + "/xbrli:entity/xbrli:identifier",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.001023" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.001023

            #region VR.ATO.GEN.001023

            /*  VR.ATO.GEN.001023
            TFN in entity identifier of context does not match the TFN provided for the Reporting Party
    
            Legacy Rule Format:
            IF (entity.identifier.TFN WHERE CONTEXT(ALL)) <> RP:entity.identifier.TFN
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^Context <> NULL AND ^OtherTFN <> ^TFN
    
            Data Elements:
    
            ^Context = RP:RP.TOFA
    
            ^OtherTFN = RP:RP.TOFA
    
            ^TFN = RP
            */
            assertion = (report.RPTOFAExists != false && report.RPTOFAIdentifierTFN != report.RPIdentifierTFN);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.001023",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Unexpected TFN mismatch with Reporting Party",
                    LongDescription = @"TFN in entity identifier of context does not match the TFN provided for the Reporting Party",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPTOFAOccurrenceIndex) + "/xbrli:entity/xbrli:identifier",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.001023" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.001023

            #region VR.ATO.GEN.010028

            /*  VR.ATO.GEN.010028
            If the Amendment Indicator is 'yes' (true), then the Amendment Sequence Number, Amendment Type and Amendment Reason must be present and not blank
    
            Legacy Rule Format:
            IF pyin.xx.xx:Report.Amendment.Indicator = TRUE AND (pyin.xx.xx:Report.AmendmentSequence.Number = NULL OR pyin.xx.xx:Report.AmendmentType.Code = NULLORBLANK OR pyin.xx.xx:Report.AmendmentReason.Text = NULLORBLANK)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR171 = TRUE AND (^FITR174 = NULL OR ^FITR172 = BLANK OR ^FITR173 = BLANK)
    
            Data Elements:
    
            ^FITR171 = RP:Report.Amendment.Indicator
    
            ^FITR174 = RP:Report.AmendmentSequence.Number
    
            ^FITR172 = RP:Report.AmendmentType.Code
    
            ^FITR173 = RP:Report.AmendmentReason.Text
            */
            assertion = (report.FITR171 == true && (report.FITR174 == null || string.IsNullOrWhiteSpace(report.FITR172) == true || string.IsNullOrWhiteSpace(report.FITR173) == true));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.010028",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"If the return is an amendment, sequence number, type and reason must be present",
                    LongDescription = @"If the Amendment Indicator is 'yes' (true), then the Amendment Sequence Number, Amendment Type and Amendment Reason must be present and not blank.",
                    Location = "/xbrli:xbrl/tns:Report.Amendment.Indicator",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.010028" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR171", Value = GetValueOrEmpty(report.FITR171) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR174", Value = report.FITR174.GetValueOrDefault().ToString() });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR172", Value = GetValueOrEmpty(report.FITR172) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR173", Value = GetValueOrEmpty(report.FITR173) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.010028

            #region VR.ATO.GEN.010029

            /*  VR.ATO.GEN.010029
            If an Amendment Sequence Number, Amendment Type or Amendment Reason is present, then the Amendment Indicator must be 'yes' (true)
    
            Legacy Rule Format:
            IF pyin.xx.xx:Report.Amendment.Indicator <> TRUE AND (pyin.xx.xx:Report.AmendmentSequence.Number <> NULL OR pyin.xx.xx:Report.AmendmentType.Code <> NULLORBLANK OR pyin.xx.xx:Report.AmendmentReason.Text <> NULLORBLANK)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR171 <> TRUE AND (^FITR174 <> NULL OR ^FITR172 <> BLANK OR ^FITR173 <> BLANK)
    
            Data Elements:
    
            ^FITR171 = RP:Report.Amendment.Indicator
    
            ^FITR174 = RP:Report.AmendmentSequence.Number
    
            ^FITR172 = RP:Report.AmendmentType.Code
    
            ^FITR173 = RP:Report.AmendmentReason.Text
            */
            assertion = (report.FITR171 != true && (report.FITR174 != null || string.IsNullOrWhiteSpace(report.FITR172) != true || string.IsNullOrWhiteSpace(report.FITR173) != true));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.010029",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"If the return is an amendment, Amendment Indicator must be 'yes' (true)",
                    LongDescription = @"If an Amendment Sequence Number, Amendment Type or Amendment Reason is present, then the Amendment Indicator must be 'yes' (true)",
                    Location = "/xbrli:xbrl/tns:Report.Amendment.Indicator",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.010029" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR171", Value = GetValueOrEmpty(report.FITR171) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR174", Value = report.FITR174.GetValueOrDefault().ToString() });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR172", Value = GetValueOrEmpty(report.FITR172) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR173", Value = GetValueOrEmpty(report.FITR173) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.010029

            #region VR.ATO.GEN.010030

            /*  VR.ATO.GEN.010030
            Amendment sequence number must be a number between 1 and 9
    
            Legacy Rule Format:
            IF pyin.xx.xx:Report.AmendmentSequence.Number <> NULL AND pyin.xx.xx:Report.AmendmentSequence.Number <> SET(1-9)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR174 <> NULL AND (^FITR174 < 1 OR ^FITR174 > 9)
    
            Data Elements:
    
            ^FITR174 = RP:Report.AmendmentSequence.Number
            */
            assertion = (report.FITR174 != null && (report.FITR174.GetValueOrDefault() < 1 || report.FITR174.GetValueOrDefault() > 9));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.010030",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Amendment sequence number must be a number between 1 and 9",
                    Location = "/xbrli:xbrl/tns:Report.AmendmentSequence.Number",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.010030" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR174", Value = report.FITR174.GetValueOrDefault().ToString() });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.010030

            #region VR.ATO.GEN.408010

            /*  VR.ATO.GEN.408010
            Name field cannot contain 'space hyphen space'.
    
            Legacy Rule Format:
            WHERE IN TUPLE(personunstructuredname1.xx.xx:PersonUnstructuredName)
            IF (pyde.xx.xx:PersonUnstructuredName.FullName.Text CONTAINS " - ")
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Contains(^FITR169, ' - ')
    
            Data Elements:
    
            ^FITR169 = INT:Declaration:PersonUnstructuredName:PersonUnstructuredName.FullName.Text WHERE [PersonUnstructuredName.Usage.Code='DeclarationSignatory']
            */
            assertion = (report.FITR169 == null ? false : report.FITR169.ToUpperInvariant().Contains(@" - "));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.408010",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Name field cannot contain 'space hyphen space'.",
                    LongDescription = @"Name field cannot contain 'space hyphen space'.",
                    Location = "/xbrli:xbrl/tns:Declaration" + OccurrenceIndex(report.DeclarationCollection_TrueAndCorrect_INTOccurrenceIndex) + "/tns:PersonUnstructuredName" + OccurrenceIndex(report.Declaration_PersonUnstructuredNameCollection_DeclarationSignatory_INTOccurrenceIndex) + "/tns:PersonUnstructuredName.FullName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.408010" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR169", Value = report.FITR169 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.408010

            #region VR.ATO.GEN.408010

            /*  VR.ATO.GEN.408010
            Name field cannot contain 'space hyphen space'.
    
            Legacy Rule Format:
            WHERE IN TUPLE(personunstructuredname1.xx.xx:PersonUnstructuredName)
            IF (pyde.xx.xx:PersonUnstructuredName.FullName.Text CONTAINS " - ")
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Contains(^FITR167, ' - ')
    
            Data Elements:
    
            ^FITR167 = RP:Declaration:PersonUnstructuredName:PersonUnstructuredName.FullName.Text WHERE [PersonUnstructuredName.Usage.Code='DeclarationSignatory']
            */
            assertion = (report.FITR167 == null ? false : report.FITR167.ToUpperInvariant().Contains(@" - "));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.408010",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Name field cannot contain 'space hyphen space'.",
                    LongDescription = @"Name field cannot contain 'space hyphen space'.",
                    Location = "/xbrli:xbrl/tns:Declaration" + OccurrenceIndex(report.DeclarationCollection_TrueAndCorrect_RPOccurrenceIndex) + "/tns:PersonUnstructuredName" + OccurrenceIndex(report.Declaration_PersonUnstructuredNameCollection_DeclarationSignatory_RPOccurrenceIndex) + "/tns:PersonUnstructuredName.FullName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.408010" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR167", Value = report.FITR167 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.408010

            #region VR.ATO.GEN.410002

            /*  VR.ATO.GEN.410002
            Address Line 2 must be present if Address Line 1 contains 'C/-'
    
            Legacy Rule Format:
            WHERE IN TUPLE(addressdetails2.xx.xx:AddressDetails)
            IF ((pyde.xx.xx:AddressDetails.Line1.Text CONTAINS "C/-") AND (pyde.xx.xx:AddressDetails.Line2.Text = NULLORBLANK))
              RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ((Contains(^FITR8, 'C/-')) AND (^FITR146 = BLANK))
    
            Data Elements:
    
            ^FITR8 = RP:AddressDetails:AddressDetails.Line1.Text WHERE [AddressDetails.Usage.Code='POS'].[AddressDetails.Currency.Code='C']
    
            ^FITR146 = RP:AddressDetails:AddressDetails.Line2.Text WHERE [AddressDetails.Usage.Code='POS'].[AddressDetails.Currency.Code='C']
            */
            assertion = ((report.FITR8 == null ? false : report.FITR8.ToUpperInvariant().Contains(@"C/-")) && string.IsNullOrWhiteSpace(report.FITR146) == true);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.410002",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Address Line 2 must be present if Address Line 1 contains 'C/-'",
                    Location = "/xbrli:xbrl/tns:AddressDetails" + OccurrenceIndex(report.AddressDetailsCollection_POS_C_RPOccurrenceIndex) + "/tns:AddressDetails.Line1.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.410002" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR8", Value = report.FITR8 });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR146", Value = GetValueOrEmpty(report.FITR146) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.410002

            #region VR.ATO.GEN.410008

            /*  VR.ATO.GEN.410008
            Address Line 3 must be left blank
    
            Legacy Rule Format:
            IF (pyde.xx.xx:AddressDetails.Line3.Text <> NULLORBLANK)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR147 <> BLANK
    
            Data Elements:
    
            ^FITR147 = RP:AddressDetails:AddressDetails.Line3.Text WHERE [AddressDetails.Usage.Code='POS'].[AddressDetails.Currency.Code='C']
            */
            assertion = (string.IsNullOrWhiteSpace(report.FITR147) != true);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.410008",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Address Line 3 must be left blank",
                    LongDescription = @"Lines 3 and 4 of the address are not used on this form",
                    Location = "/xbrli:xbrl/tns:AddressDetails" + OccurrenceIndex(report.AddressDetailsCollection_POS_C_RPOccurrenceIndex) + "/tns:AddressDetails.Line3.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.410008" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR147", Value = GetValueOrEmpty(report.FITR147) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.410008

            #region VR.ATO.GEN.410013

            /*  VR.ATO.GEN.410013
            Address Line 4 must be left blank
    
            Legacy Rule Format:
            IF (pyde.xx.xx:AddressDetails.Line4.Text <> NULLORBLANK)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR148 <> BLANK
    
            Data Elements:
    
            ^FITR148 = RP:AddressDetails:AddressDetails.Line4.Text WHERE [AddressDetails.Usage.Code='POS'].[AddressDetails.Currency.Code='C']
            */
            assertion = (string.IsNullOrWhiteSpace(report.FITR148) != true);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.410013",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Address Line 4 must be left blank",
                    LongDescription = @"Lines 3 and 4 of the address are not used on this form",
                    Location = "/xbrli:xbrl/tns:AddressDetails" + OccurrenceIndex(report.AddressDetailsCollection_POS_C_RPOccurrenceIndex) + "/tns:AddressDetails.Line4.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.410013" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR148", Value = GetValueOrEmpty(report.FITR148) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.410013

            #region VR.ATO.GEN.410038

            /*  VR.ATO.GEN.410038
            Non-individual name field cannot contain 'Pship', 'P'ship or 'P/Ship'
    
            Legacy Rule Format:
            IF (pyde.xx.xx:OrganisationNameDetails.OrganisationalName.Text STARTSWITH "T/A ") AND (pyde.xx.xx:OrganisationNameDetails.OrganisationalName.Text ENDSWITH SET(" Pship"," P'ship"," P/ship"))
              RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (StartsWith(^FITR116, 'T/A ')) AND (EndsWithSet(^FITR116, '" Pship"," P'ship"," P/ship"'))
    
            Data Elements:
    
            ^FITR116 = INT:OrganisationNameDetails:OrganisationNameDetails.OrganisationalName.Text WHERE [OrganisationNameDetails.OrganisationalNameType.Code='MN'].[OrganisationNameDetails.Currency.Code='C']
            */
            assertion = (IsMatch(report.FITR116, @"^(T/A )", RegexOptions.IgnoreCase) && IsMatch(report.FITR116, @"( Pship| P'ship| P/ship)$", RegexOptions.IgnoreCase));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.410038",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Non-individual name field cannot start with 'T/A' and end with 'Pship', 'P'ship or 'P/Ship'",
                    Location = "/xbrli:xbrl/tns:OrganisationNameDetails" + OccurrenceIndex(report.OrganisationNameDetailsCollection_MN_C_INTOccurrenceIndex) + "/tns:OrganisationNameDetails.OrganisationalName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.410038" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR116", Value = report.FITR116 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.410038

            #region VR.ATO.GEN.410038

            /*  VR.ATO.GEN.410038
            Non-individual name field cannot contain 'Pship', 'P'ship or 'P/Ship'
    
            Legacy Rule Format:
            IF (pyde.xx.xx:OrganisationNameDetails.OrganisationalName.Text STARTSWITH "T/A ") AND (pyde.xx.xx:OrganisationNameDetails.OrganisationalName.Text ENDSWITH SET(" Pship"," P'ship"," P/ship"))
              RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (StartsWith(^FITR12, 'T/A ')) AND (EndsWithSet(^FITR12, '" Pship"," P'ship"," P/ship"'))
    
            Data Elements:
    
            ^FITR12 = RP:Trustee:OrganisationNameDetails:OrganisationNameDetails.OrganisationalName.Text WHERE [OrganisationNameDetails.OrganisationalNameType.Code='MN'].[OrganisationNameDetails.Currency.Code='C']
            */
            assertion = (IsMatch(report.FITR12, @"^(T/A )", RegexOptions.IgnoreCase) && IsMatch(report.FITR12, @"( Pship| P'ship| P/ship)$", RegexOptions.IgnoreCase));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.410038",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Non-individual name field cannot start with 'T/A' and end with 'Pship', 'P'ship or 'P/Ship'",
                    Location = "/xbrli:xbrl/tns:Trustee/tns:OrganisationNameDetails" + OccurrenceIndex(report.Trustee_OrganisationNameDetailsCollection_MN_C_RPOccurrenceIndex) + "/tns:OrganisationNameDetails.OrganisationalName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.410038" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR12", Value = report.FITR12 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.410038

            #region VR.ATO.GEN.410038

            /*  VR.ATO.GEN.410038
            Non-individual name field cannot contain 'Pship', 'P'ship or 'P/Ship'
    
            Legacy Rule Format:
            IF (pyde.xx.xx:OrganisationNameDetails.OrganisationalName.Text STARTSWITH "T/A ") AND (pyde.xx.xx:OrganisationNameDetails.OrganisationalName.Text ENDSWITH SET(" Pship"," P'ship"," P/ship"))
              RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (StartsWith(^FITR6, 'T/A ')) AND (EndsWithSet(^FITR6, '" Pship"," P'ship"," P/ship"'))
    
            Data Elements:
    
            ^FITR6 = RP:OrganisationNameDetails:OrganisationNameDetails.OrganisationalName.Text WHERE [OrganisationNameDetails.OrganisationalNameType.Code='MN'].[OrganisationNameDetails.Currency.Code='C']
            */
            assertion = (IsMatch(report.FITR6, @"^(T/A )", RegexOptions.IgnoreCase) && IsMatch(report.FITR6, @"( Pship| P'ship| P/ship)$", RegexOptions.IgnoreCase));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.410038",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Non-individual name field cannot start with 'T/A' and end with 'Pship', 'P'ship or 'P/Ship'",
                    Location = "/xbrli:xbrl/tns:OrganisationNameDetails" + OccurrenceIndex(report.OrganisationNameDetailsCollection_MN_C_RPOccurrenceIndex) + "/tns:OrganisationNameDetails.OrganisationalName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.410038" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR6", Value = report.FITR6 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.410038

            #region VR.ATO.GEN.410039

            /*  VR.ATO.GEN.410039
            Organisation name must contain at least one alpha or numeric character
    
            Legacy Rule Format:
            IF (pyde.xx.xx:OrganisationNameDetails.OrganisationalName.Text  <> NULLORBLANK) AND (pyde.xx.xx:OrganisationNameDetails.OrganisationalName.Text  DOES NOT CONTAIN SET("a-z", "A-Z", "0-9"))
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^FITR116  <> BLANK) AND (NotContainsSet(^FITR116, '"a-z", "A-Z", "0-9"'))
    
            Data Elements:
    
            ^FITR116 = INT:OrganisationNameDetails:OrganisationNameDetails.OrganisationalName.Text WHERE [OrganisationNameDetails.OrganisationalNameType.Code='MN'].[OrganisationNameDetails.Currency.Code='C']
            */
            assertion = (string.IsNullOrWhiteSpace(report.FITR116) != true && !(IsMatch(report.FITR116, @"[a-z]|[A-Z]|[0-9]", RegexOptions.IgnoreCase)));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.410039",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Organisation name must contain at least one alpha or numeric character",
                    Location = "/xbrli:xbrl/tns:OrganisationNameDetails" + OccurrenceIndex(report.OrganisationNameDetailsCollection_MN_C_INTOccurrenceIndex) + "/tns:OrganisationNameDetails.OrganisationalName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.410039" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR116", Value = report.FITR116 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.410039

            #region VR.ATO.GEN.410039

            /*  VR.ATO.GEN.410039
            Organisation name must contain at least one alpha or numeric character
    
            Legacy Rule Format:
            IF (pyde.xx.xx:OrganisationNameDetails.OrganisationalName.Text  <> NULLORBLANK) AND (pyde.xx.xx:OrganisationNameDetails.OrganisationalName.Text  DOES NOT CONTAIN SET("a-z", "A-Z", "0-9"))
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^FITR12  <> BLANK) AND (NotContainsSet(^FITR12, '"a-z", "A-Z", "0-9"'))
    
            Data Elements:
    
            ^FITR12 = RP:Trustee:OrganisationNameDetails:OrganisationNameDetails.OrganisationalName.Text WHERE [OrganisationNameDetails.OrganisationalNameType.Code='MN'].[OrganisationNameDetails.Currency.Code='C']
            */
            assertion = (string.IsNullOrWhiteSpace(report.FITR12) != true && !(IsMatch(report.FITR12, @"[a-z]|[A-Z]|[0-9]", RegexOptions.IgnoreCase)));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.410039",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Organisation name must contain at least one alpha or numeric character",
                    Location = "/xbrli:xbrl/tns:Trustee/tns:OrganisationNameDetails" + OccurrenceIndex(report.Trustee_OrganisationNameDetailsCollection_MN_C_RPOccurrenceIndex) + "/tns:OrganisationNameDetails.OrganisationalName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.410039" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR12", Value = report.FITR12 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.410039

            #region VR.ATO.GEN.410039

            /*  VR.ATO.GEN.410039
            Organisation name must contain at least one alpha or numeric character
    
            Legacy Rule Format:
            IF (pyde.xx.xx:OrganisationNameDetails.OrganisationalName.Text  <> NULLORBLANK) AND (pyde.xx.xx:OrganisationNameDetails.OrganisationalName.Text  DOES NOT CONTAIN SET("a-z", "A-Z", "0-9"))
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^FITR6  <> BLANK) AND (NotContainsSet(^FITR6, '"a-z", "A-Z", "0-9"'))
    
            Data Elements:
    
            ^FITR6 = RP:OrganisationNameDetails:OrganisationNameDetails.OrganisationalName.Text WHERE [OrganisationNameDetails.OrganisationalNameType.Code='MN'].[OrganisationNameDetails.Currency.Code='C']
            */
            assertion = (string.IsNullOrWhiteSpace(report.FITR6) != true && !(IsMatch(report.FITR6, @"[a-z]|[A-Z]|[0-9]", RegexOptions.IgnoreCase)));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.410039",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Organisation name must contain at least one alpha or numeric character",
                    Location = "/xbrli:xbrl/tns:OrganisationNameDetails" + OccurrenceIndex(report.OrganisationNameDetailsCollection_MN_C_RPOccurrenceIndex) + "/tns:OrganisationNameDetails.OrganisationalName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.410039" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR6", Value = report.FITR6 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.410039

            #region VR.ATO.GEN.410131

            /*  VR.ATO.GEN.410131
            Other Given Name must contain at least one alphabetical character
    
            Legacy Rule Format:
            IF (pyde.xx.xx:PersonNameDetails.OtherGivenName.Text <> NULLORBLANK) AND (pyde.xx.xx:PersonNameDetails.OtherGivenName.Text DOES NOT CONTAIN SET("A-Z","a-z"))
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^FITR115 <> BLANK) AND (NotContainsSet(^FITR115, '"A-Z","a-z"'))
    
            Data Elements:
    
            ^FITR115 = INT:PersonNameDetails:PersonNameDetails.OtherGivenName.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
            */
            assertion = (string.IsNullOrWhiteSpace(report.FITR115) != true && !(IsMatch(report.FITR115, @"[A-Z]|[a-z]", RegexOptions.IgnoreCase)));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.410131",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Other Given Name must contain at least one alphabetical character",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_INTOccurrenceIndex) + "/tns:PersonNameDetails.OtherGivenName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.410131" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR115", Value = GetValueOrEmpty(report.FITR115) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.410131

            #region VR.ATO.GEN.410131

            /*  VR.ATO.GEN.410131
            Other Given Name must contain at least one alphabetical character
    
            Legacy Rule Format:
            IF (pyde.xx.xx:PersonNameDetails.OtherGivenName.Text <> NULLORBLANK) AND (pyde.xx.xx:PersonNameDetails.OtherGivenName.Text DOES NOT CONTAIN SET("A-Z","a-z"))
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^FITR106 <> BLANK) AND (NotContainsSet(^FITR106, '"A-Z","a-z"'))
    
            Data Elements:
    
            ^FITR106 = RP:PersonNameDetails:PersonNameDetails.OtherGivenName.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
            */
            assertion = (string.IsNullOrWhiteSpace(report.FITR106) != true && !(IsMatch(report.FITR106, @"[A-Z]|[a-z]", RegexOptions.IgnoreCase)));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.410131",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Other Given Name must contain at least one alphabetical character",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_RPOccurrenceIndex) + "/tns:PersonNameDetails.OtherGivenName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.410131" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR106", Value = GetValueOrEmpty(report.FITR106) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.410131

            #region VR.ATO.GEN.410194

            /*  VR.ATO.GEN.410194
            Address Line 1 must not exceed 38 characters
    
            Legacy Rule Format:
            IF LENGTH(pyde.xx.xx:AddressDetails.Line1.Text) > 38
                 RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Length(^FITR8) > 38
    
            Data Elements:
    
            ^FITR8 = RP:AddressDetails:AddressDetails.Line1.Text WHERE [AddressDetails.Usage.Code='POS'].[AddressDetails.Currency.Code='C']
            */
            assertion = (Length(report.FITR8) > 38);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.410194",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Address Line 1 must not exceed 38 characters",
                    LongDescription = @"Although SBR sets a limit of 50 characters for the Line 1 of an address, the maximum allowable for this form is 38",
                    Location = "/xbrli:xbrl/tns:AddressDetails" + OccurrenceIndex(report.AddressDetailsCollection_POS_C_RPOccurrenceIndex) + "/tns:AddressDetails.Line1.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.410194" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR8", Value = report.FITR8 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.410194

            #region VR.ATO.GEN.410195

            /*  VR.ATO.GEN.410195
            Address Line 2 must not exceed 38 characters
    
            Legacy Rule Format:
            IF LENGTH(pyde.xx.xx:AddressDetails.Line2.Text) > 38
                 RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Length(^FITR146) > 38
    
            Data Elements:
    
            ^FITR146 = RP:AddressDetails:AddressDetails.Line2.Text WHERE [AddressDetails.Usage.Code='POS'].[AddressDetails.Currency.Code='C']
            */
            assertion = (Length(report.FITR146) > 38);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.410195",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Address Line 2 must not exceed 38 characters",
                    LongDescription = @"Although SBR sets a limit of 50 characters for the Line 2 of an address, the maximum allowable for this form is 38",
                    Location = "/xbrli:xbrl/tns:AddressDetails" + OccurrenceIndex(report.AddressDetailsCollection_POS_C_RPOccurrenceIndex) + "/tns:AddressDetails.Line2.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.410195" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR146", Value = GetValueOrEmpty(report.FITR146) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.410195

            #region VR.ATO.GEN.410200

            /*  VR.ATO.GEN.410200
            Family name cannot contain "Exec for", "Rep for" or "Trustee for"
    
            Legacy Rule Format:
            IF (pyde.xx.xx:PersonNameDetails.FamilyName.Text = FOUND("Exec for","Rep for","Trustee for"))
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (FoundSet(^FITR112, '"Exec for","Rep for","Trustee for"'))
    
            Data Elements:
    
            ^FITR112 = INT:PersonNameDetails:PersonNameDetails.FamilyName.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
            */
            assertion = IsMatch(report.FITR112, @"(^(Exec for|Rep for|Trustee for)$)|(^(.*)( Exec for | Rep for | Trustee for )(.*)$)|(^(Exec for |Rep for |Trustee for )(.*)$)|(^(.*)( Exec for| Rep for| Trustee for)$)", RegexOptions.IgnoreCase);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000424",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Family name cannot contain ""Exec for"", ""Rep for"" or ""Trustee for""",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_INTOccurrenceIndex) + "/tns:PersonNameDetails.FamilyName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.410200" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR112", Value = report.FITR112 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.410200

            #region VR.ATO.GEN.410200

            /*  VR.ATO.GEN.410200
            Family name cannot contain "Exec for", "Rep for" or "Trustee for"
    
            Legacy Rule Format:
            IF (pyde.xx.xx:PersonNameDetails.FamilyName.Text = FOUND("Exec for","Rep for","Trustee for"))
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (FoundSet(^FITR103, '"Exec for","Rep for","Trustee for"'))
    
            Data Elements:
    
            ^FITR103 = RP:PersonNameDetails:PersonNameDetails.FamilyName.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
            */
            assertion = IsMatch(report.FITR103, @"(^(Exec for|Rep for|Trustee for)$)|(^(.*)( Exec for | Rep for | Trustee for )(.*)$)|(^(Exec for |Rep for |Trustee for )(.*)$)|(^(.*)( Exec for| Rep for| Trustee for)$)", RegexOptions.IgnoreCase);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000424",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Family name cannot contain ""Exec for"", ""Rep for"" or ""Trustee for""",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_RPOccurrenceIndex) + "/tns:PersonNameDetails.FamilyName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.410200" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR103", Value = report.FITR103 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.410200

            #region VR.ATO.GEN.410201

            /*  VR.ATO.GEN.410201
            Family name cannot contain "MR", "MRS", "MISS" or "MS"
    
            Legacy Rule Format:
            IF (pyde.xx.xx:PersonNameDetails.FamilyName.Text = FOUND("MR","MRS","MISS","MS"))
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (FoundSet(^FITR112, '"MR","MRS","MISS","MS"'))
    
            Data Elements:
    
            ^FITR112 = INT:PersonNameDetails:PersonNameDetails.FamilyName.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
            */
            assertion = IsMatch(report.FITR112, @"(^(MR|MRS|MISS|MS)$)|(^(.*)( MR | MRS | MISS | MS )(.*)$)|(^(MR |MRS |MISS |MS )(.*)$)|(^(.*)( MR| MRS| MISS| MS)$)", RegexOptions.IgnoreCase);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000426",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Family name cannot contain ""MR"", ""MRS"", ""MISS"" or ""MS""",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_INTOccurrenceIndex) + "/tns:PersonNameDetails.FamilyName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.410201" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR112", Value = report.FITR112 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.410201

            #region VR.ATO.GEN.410201

            /*  VR.ATO.GEN.410201
            Family name cannot contain "MR", "MRS", "MISS" or "MS"
    
            Legacy Rule Format:
            IF (pyde.xx.xx:PersonNameDetails.FamilyName.Text = FOUND("MR","MRS","MISS","MS"))
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (FoundSet(^FITR103, '"MR","MRS","MISS","MS"'))
    
            Data Elements:
    
            ^FITR103 = RP:PersonNameDetails:PersonNameDetails.FamilyName.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
            */
            assertion = IsMatch(report.FITR103, @"(^(MR|MRS|MISS|MS)$)|(^(.*)( MR | MRS | MISS | MS )(.*)$)|(^(MR |MRS |MISS |MS )(.*)$)|(^(.*)( MR| MRS| MISS| MS)$)", RegexOptions.IgnoreCase);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000426",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Family name cannot contain ""MR"", ""MRS"", ""MISS"" or ""MS""",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_RPOccurrenceIndex) + "/tns:PersonNameDetails.FamilyName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.410201" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR103", Value = report.FITR103 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.410201

            #region VR.ATO.GEN.410203

            /*  VR.ATO.GEN.410203
            First name cannot contain "Exec for", "Rep for" or "Trustee for"
    
            Legacy Rule Format:
            IF (pyde.xx.xx:PersonNameDetails.GivenName.Text = FOUND("Exec for","Rep for","Trustee for"))
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (FoundSet(^FITR114, '"Exec for","Rep for","Trustee for"'))
    
            Data Elements:
    
            ^FITR114 = INT:PersonNameDetails:PersonNameDetails.GivenName.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
            */
            assertion = IsMatch(report.FITR114, @"(^(Exec for|Rep for|Trustee for)$)|(^(.*)( Exec for | Rep for | Trustee for )(.*)$)|(^(Exec for |Rep for |Trustee for )(.*)$)|(^(.*)( Exec for| Rep for| Trustee for)$)", RegexOptions.IgnoreCase);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000437",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"First name cannot contain ""Exec for"", ""Rep for"" or ""Trustee for""",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_INTOccurrenceIndex) + "/tns:PersonNameDetails.GivenName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.410203" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR114", Value = report.FITR114 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.410203

            #region VR.ATO.GEN.410203

            /*  VR.ATO.GEN.410203
            First name cannot contain "Exec for", "Rep for" or "Trustee for"
    
            Legacy Rule Format:
            IF (pyde.xx.xx:PersonNameDetails.GivenName.Text = FOUND("Exec for","Rep for","Trustee for"))
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (FoundSet(^FITR105, '"Exec for","Rep for","Trustee for"'))
    
            Data Elements:
    
            ^FITR105 = RP:PersonNameDetails:PersonNameDetails.GivenName.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
            */
            assertion = IsMatch(report.FITR105, @"(^(Exec for|Rep for|Trustee for)$)|(^(.*)( Exec for | Rep for | Trustee for )(.*)$)|(^(Exec for |Rep for |Trustee for )(.*)$)|(^(.*)( Exec for| Rep for| Trustee for)$)", RegexOptions.IgnoreCase);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000437",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"First name cannot contain ""Exec for"", ""Rep for"" or ""Trustee for""",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_RPOccurrenceIndex) + "/tns:PersonNameDetails.GivenName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.410203" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR105", Value = report.FITR105 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.410203

            #region VR.ATO.GEN.410204

            /*  VR.ATO.GEN.410204
            First name cannot contain "MR", "MRS", "MISS" or "MS"
    
            Legacy Rule Format:
            IF (pyde.xx.xx:PersonNameDetails.GivenName.Text = FOUND("MR","MRS","MISS","MS"))
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (FoundSet(^FITR114, '"MR","MRS","MISS","MS"'))
    
            Data Elements:
    
            ^FITR114 = INT:PersonNameDetails:PersonNameDetails.GivenName.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
            */
            assertion = IsMatch(report.FITR114, @"(^(MR|MRS|MISS|MS)$)|(^(.*)( MR | MRS | MISS | MS )(.*)$)|(^(MR |MRS |MISS |MS )(.*)$)|(^(.*)( MR| MRS| MISS| MS)$)", RegexOptions.IgnoreCase);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000438",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"First name cannot contain ""MR"", ""MRS"", ""MISS"" or ""MS""",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_INTOccurrenceIndex) + "/tns:PersonNameDetails.GivenName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.410204" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR114", Value = report.FITR114 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.410204

            #region VR.ATO.GEN.410204

            /*  VR.ATO.GEN.410204
            First name cannot contain "MR", "MRS", "MISS" or "MS"
    
            Legacy Rule Format:
            IF (pyde.xx.xx:PersonNameDetails.GivenName.Text = FOUND("MR","MRS","MISS","MS"))
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (FoundSet(^FITR105, '"MR","MRS","MISS","MS"'))
    
            Data Elements:
    
            ^FITR105 = RP:PersonNameDetails:PersonNameDetails.GivenName.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
            */
            assertion = IsMatch(report.FITR105, @"(^(MR|MRS|MISS|MS)$)|(^(.*)( MR | MRS | MISS | MS )(.*)$)|(^(MR |MRS |MISS |MS )(.*)$)|(^(.*)( MR| MRS| MISS| MS)$)", RegexOptions.IgnoreCase);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000438",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"First name cannot contain ""MR"", ""MRS"", ""MISS"" or ""MS""",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_RPOccurrenceIndex) + "/tns:PersonNameDetails.GivenName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.410204" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR105", Value = report.FITR105 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.410204

            #region VR.ATO.GEN.410205

            /*  VR.ATO.GEN.410205
            Address Line 1 must not contain 'AS ABOVE'
    
            Legacy Rule Format:
            IF (pyde.xx.xx:AddressDetails.Line1.Text = FOUND("AS ABOVE"))
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (FoundSet(^FITR8, '"AS ABOVE"'))
    
            Data Elements:
    
            ^FITR8 = RP:AddressDetails:AddressDetails.Line1.Text WHERE [AddressDetails.Usage.Code='POS'].[AddressDetails.Currency.Code='C']
            */
            assertion = IsMatch(report.FITR8, @"(^(AS ABOVE)$)|(^(.*)( AS ABOVE )(.*)$)|(^(AS ABOVE )(.*)$)|(^(.*)( AS ABOVE)$)", RegexOptions.IgnoreCase);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.410001",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Address Line 1 must not contain 'AS ABOVE'",
                    Location = "/xbrli:xbrl/tns:AddressDetails" + OccurrenceIndex(report.AddressDetailsCollection_POS_C_RPOccurrenceIndex) + "/tns:AddressDetails.Line1.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.410205" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR8", Value = report.FITR8 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.410205

            #region VR.ATO.GEN.410208

            /*  VR.ATO.GEN.410208
            Other given name cannot contain "Exec for", "Rep for" or "Trustee for"
    
            Legacy Rule Format:
            IF (pyde.xx.xx:PersonNameDetails.OtherGivenName.Text = FOUND("Exec for","Rep for","Trustee for"))
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (FoundSet(^FITR115, '"Exec for","Rep for","Trustee for"'))
    
            Data Elements:
    
            ^FITR115 = INT:PersonNameDetails:PersonNameDetails.OtherGivenName.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
            */
            assertion = IsMatch(report.FITR115, @"(^(Exec for|Rep for|Trustee for)$)|(^(.*)( Exec for | Rep for | Trustee for )(.*)$)|(^(Exec for |Rep for |Trustee for )(.*)$)|(^(.*)( Exec for| Rep for| Trustee for)$)", RegexOptions.IgnoreCase);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000449",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Other given name cannot contain ""Exec for"", ""Rep for"" or ""Trustee for""",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_INTOccurrenceIndex) + "/tns:PersonNameDetails.OtherGivenName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.410208" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR115", Value = GetValueOrEmpty(report.FITR115) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.410208

            #region VR.ATO.GEN.410208

            /*  VR.ATO.GEN.410208
            Other given name cannot contain "Exec for", "Rep for" or "Trustee for"
    
            Legacy Rule Format:
            IF (pyde.xx.xx:PersonNameDetails.OtherGivenName.Text = FOUND("Exec for","Rep for","Trustee for"))
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (FoundSet(^FITR106, '"Exec for","Rep for","Trustee for"'))
    
            Data Elements:
    
            ^FITR106 = RP:PersonNameDetails:PersonNameDetails.OtherGivenName.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
            */
            assertion = IsMatch(report.FITR106, @"(^(Exec for|Rep for|Trustee for)$)|(^(.*)( Exec for | Rep for | Trustee for )(.*)$)|(^(Exec for |Rep for |Trustee for )(.*)$)|(^(.*)( Exec for| Rep for| Trustee for)$)", RegexOptions.IgnoreCase);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000449",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Other given name cannot contain ""Exec for"", ""Rep for"" or ""Trustee for""",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_RPOccurrenceIndex) + "/tns:PersonNameDetails.OtherGivenName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.410208" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR106", Value = GetValueOrEmpty(report.FITR106) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.410208

            #region VR.ATO.GEN.410209

            /*  VR.ATO.GEN.410209
            Other given name cannot contain "MR", "MRS", "MISS" or "MS"
    
            Legacy Rule Format:
            IF (pyde.xx.xx:PersonNameDetails.OtherGivenName.Text = FOUND("MR","MRS","MISS","MS"))
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (FoundSet(^FITR115, '"MR","MRS","MISS","MS"'))
    
            Data Elements:
    
            ^FITR115 = INT:PersonNameDetails:PersonNameDetails.OtherGivenName.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
            */
            assertion = IsMatch(report.FITR115, @"(^(MR|MRS|MISS|MS)$)|(^(.*)( MR | MRS | MISS | MS )(.*)$)|(^(MR |MRS |MISS |MS )(.*)$)|(^(.*)( MR| MRS| MISS| MS)$)", RegexOptions.IgnoreCase);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000450",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Other given name cannot contain ""MR"", ""MRS"", ""MISS"" or ""MS""",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_INTOccurrenceIndex) + "/tns:PersonNameDetails.OtherGivenName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.410209" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR115", Value = GetValueOrEmpty(report.FITR115) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.410209

            #region VR.ATO.GEN.410209

            /*  VR.ATO.GEN.410209
            Other given name cannot contain "MR", "MRS", "MISS" or "MS"
    
            Legacy Rule Format:
            IF (pyde.xx.xx:PersonNameDetails.OtherGivenName.Text = FOUND("MR","MRS","MISS","MS"))
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (FoundSet(^FITR106, '"MR","MRS","MISS","MS"'))
    
            Data Elements:
    
            ^FITR106 = RP:PersonNameDetails:PersonNameDetails.OtherGivenName.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
            */
            assertion = IsMatch(report.FITR106, @"(^(MR|MRS|MISS|MS)$)|(^(.*)( MR | MRS | MISS | MS )(.*)$)|(^(MR |MRS |MISS |MS )(.*)$)|(^(.*)( MR| MRS| MISS| MS)$)", RegexOptions.IgnoreCase);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000450",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Other given name cannot contain ""MR"", ""MRS"", ""MISS"" or ""MS""",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_RPOccurrenceIndex) + "/tns:PersonNameDetails.OtherGivenName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.410209" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR106", Value = GetValueOrEmpty(report.FITR106) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.410209

            #region VR.ATO.GEN.410213

            /*  VR.ATO.GEN.410213
            Country Name must be left blank
    
            Legacy Rule Format:
            IF (pyde.xx.xx:AddressDetails.CountryName.Text <> NULLORBLANK)
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR149 <> BLANK
    
            Data Elements:
    
            ^FITR149 = RP:AddressDetails:AddressDetails.CountryName.Text WHERE [AddressDetails.Usage.Code='POS'].[AddressDetails.Currency.Code='C']
            */
            assertion = (string.IsNullOrWhiteSpace(report.FITR149) != true);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.410213",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Country Name must be left blank",
                    LongDescription = @"For SBR submissions to the ATO, the Country Name must be left blank. To supply the country for an overseas address please supply a Country Code",
                    Location = "/xbrli:xbrl/tns:AddressDetails" + OccurrenceIndex(report.AddressDetailsCollection_POS_C_RPOccurrenceIndex) + "/tns:AddressDetails.CountryName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.410213" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR149", Value = GetValueOrEmpty(report.FITR149) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.410213

            #region VR.ATO.GEN.428016

            /*  VR.ATO.GEN.428016
            Tax File Number has failed the algorithm check.
    
            Legacy Rule Format:
            IF (RP:entity.identifier.TFN <> NULLORBLANK) AND (TFNALGORITHM (RP:entity.identifier.TFN) = FALSE)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^TFN <> BLANK) AND (FailsTFNAlgorithm(^TFN))
    
            Data Elements:
    
            ^TFN = RP
            */
            assertion = (string.IsNullOrWhiteSpace(report.RPIdentifierTFN) != true && FailsTFNAlgorithm(report.RPIdentifierTFN));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.428016",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Tax File Number has failed the algorithm check",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPOccurrenceIndex) + "/xbrli:entity/xbrli:identifier",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.428016" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.428016

            #region VR.ATO.GEN.428230

            /*  VR.ATO.GEN.428230
            Suburb/Town/City cannot contain state code
    
            Legacy Rule Format:
            IF (pyde.xx.xx:AddressDetails.LocalityName.Text = FOUND("QLD","NSW","VIC","SA","WA","NT","ACT","TAS"))
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (FoundSet(^FITR9, '"QLD","NSW","VIC","SA","WA","NT","ACT","TAS"'))
    
            Data Elements:
    
            ^FITR9 = RP:AddressDetails:AddressDetails.LocalityName.Text WHERE [AddressDetails.Usage.Code='POS'].[AddressDetails.Currency.Code='C']
            */
            assertion = IsMatch(report.FITR9, @"(^(QLD|NSW|VIC|SA|WA|NT|ACT|TAS)$)|(^(.*)( QLD | NSW | VIC | SA | WA | NT | ACT | TAS )(.*)$)|(^(QLD |NSW |VIC |SA |WA |NT |ACT |TAS )(.*)$)|(^(.*)( QLD| NSW| VIC| SA| WA| NT| ACT| TAS)$)", RegexOptions.IgnoreCase);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000411",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Suburb/Town/City cannot contain state code",
                    Location = "/xbrli:xbrl/tns:AddressDetails" + OccurrenceIndex(report.AddressDetailsCollection_POS_C_RPOccurrenceIndex) + "/tns:AddressDetails.LocalityName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.428230" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR9", Value = report.FITR9 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.428230

            #region VR.ATO.GEN.428247

            /*  VR.ATO.GEN.428247
            The Tax Agent number must pass the Tax Agent Number algorithm check
    
            Legacy Rule Format:
            IF ((pyid.xx.xx:Identifiers.TaxAgentNumber.Identifier) <> NULLORBLANK) AND (TANALGORITHM(pyid.xx.xx:Identifiers.TaxAgentNumber.Identifier) = FALSE)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^FITR2 <> BLANK) AND (FailsTANAlgorithm(^FITR2))
    
            Data Elements:
    
            ^FITR2 = INT:Identifiers.TaxAgentNumber.Identifier
            */
            assertion = (string.IsNullOrWhiteSpace(report.FITR2) != true && FailsTANAlgorithm(report.FITR2));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.410009",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"The Tax Agent number has failed the Tax Agent number algorithm check",
                    Location = "/xbrli:xbrl/tns:Identifiers.TaxAgentNumber.Identifier",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.428247" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR2", Value = GetValueOrEmpty(report.FITR2) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.428247

            #region VR.ATO.GEN.428254

            /*  VR.ATO.GEN.428254
            Address line 1 cannot contain "UNKNOWN"
    
            Legacy Rule Format:
            IF (pyde.xx.xx:AddressDetails.Line1.Text CONTAINS "UNKNOWN")
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Contains(^FITR8, 'UNKNOWN')
    
            Data Elements:
    
            ^FITR8 = RP:AddressDetails:AddressDetails.Line1.Text WHERE [AddressDetails.Usage.Code='POS'].[AddressDetails.Currency.Code='C']
            */
            assertion = (report.FITR8 == null ? false : report.FITR8.ToUpperInvariant().Contains(@"UNKNOWN"));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000405",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Address line 1 cannot contain ""UNKNOWN""",
                    Location = "/xbrli:xbrl/tns:AddressDetails" + OccurrenceIndex(report.AddressDetailsCollection_POS_C_RPOccurrenceIndex) + "/tns:AddressDetails.Line1.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.428254" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR8", Value = report.FITR8 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.428254

            #region VR.ATO.GEN.428255

            /*  VR.ATO.GEN.428255
            Address line 2 cannot contain "UNKNOWN"
    
            Legacy Rule Format:
            IF (pyde.xx.xx:AddressDetails.Line2.Text CONTAINS "UNKNOWN")
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Contains(^FITR146, 'UNKNOWN')
    
            Data Elements:
    
            ^FITR146 = RP:AddressDetails:AddressDetails.Line2.Text WHERE [AddressDetails.Usage.Code='POS'].[AddressDetails.Currency.Code='C']
            */
            assertion = (report.FITR146 == null ? false : report.FITR146.ToUpperInvariant().Contains(@"UNKNOWN"));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000407",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Address line 2 cannot contain ""UNKNOWN""",
                    Location = "/xbrli:xbrl/tns:AddressDetails" + OccurrenceIndex(report.AddressDetailsCollection_POS_C_RPOccurrenceIndex) + "/tns:AddressDetails.Line2.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.428255" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR146", Value = GetValueOrEmpty(report.FITR146) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.428255

            #region VR.ATO.GEN.428256

            /*  VR.ATO.GEN.428256
            Suburb/Town/City cannot contain "UNKNOWN"
    
            Legacy Rule Format:
            IF (pyde.xx.xx:AddressDetails.LocalityName.Text CONTAINS "UNKNOWN")
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Contains(^FITR9, 'UNKNOWN')
    
            Data Elements:
    
            ^FITR9 = RP:AddressDetails:AddressDetails.LocalityName.Text WHERE [AddressDetails.Usage.Code='POS'].[AddressDetails.Currency.Code='C']
            */
            assertion = (report.FITR9 == null ? false : report.FITR9.ToUpperInvariant().Contains(@"UNKNOWN"));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000409",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Suburb/Town/City cannot contain ""UNKNOWN""",
                    Location = "/xbrli:xbrl/tns:AddressDetails" + OccurrenceIndex(report.AddressDetailsCollection_POS_C_RPOccurrenceIndex) + "/tns:AddressDetails.LocalityName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.428256" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR9", Value = report.FITR9 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.428256

            #region VR.ATO.GEN.428258

            /*  VR.ATO.GEN.428258
            Non Individual Name cannot contain space hyphen space combinations
    
            Legacy Rule Format:
            IF (pyde.xx.xx:OrganisationNameDetails.OrganisationalName.Text CONTAINS  " - " )
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Contains(^FITR116, ' - ')
    
            Data Elements:
    
            ^FITR116 = INT:OrganisationNameDetails:OrganisationNameDetails.OrganisationalName.Text WHERE [OrganisationNameDetails.OrganisationalNameType.Code='MN'].[OrganisationNameDetails.Currency.Code='C']
            */
            assertion = (report.FITR116 == null ? false : report.FITR116.ToUpperInvariant().Contains(@" - "));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000413",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Non Individual Name cannot contain space hyphen space combinations",
                    Location = "/xbrli:xbrl/tns:OrganisationNameDetails" + OccurrenceIndex(report.OrganisationNameDetailsCollection_MN_C_INTOccurrenceIndex) + "/tns:OrganisationNameDetails.OrganisationalName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.428258" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR116", Value = report.FITR116 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.428258

            #region VR.ATO.GEN.428258

            /*  VR.ATO.GEN.428258
            Non Individual Name cannot contain space hyphen space combinations
    
            Legacy Rule Format:
            IF (pyde.xx.xx:OrganisationNameDetails.OrganisationalName.Text CONTAINS  " - " )
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Contains(^FITR12, ' - ')
    
            Data Elements:
    
            ^FITR12 = RP:Trustee:OrganisationNameDetails:OrganisationNameDetails.OrganisationalName.Text WHERE [OrganisationNameDetails.OrganisationalNameType.Code='MN'].[OrganisationNameDetails.Currency.Code='C']
            */
            assertion = (report.FITR12 == null ? false : report.FITR12.ToUpperInvariant().Contains(@" - "));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000413",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Non Individual Name cannot contain space hyphen space combinations",
                    Location = "/xbrli:xbrl/tns:Trustee/tns:OrganisationNameDetails" + OccurrenceIndex(report.Trustee_OrganisationNameDetailsCollection_MN_C_RPOccurrenceIndex) + "/tns:OrganisationNameDetails.OrganisationalName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.428258" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR12", Value = report.FITR12 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.428258

            #region VR.ATO.GEN.428258

            /*  VR.ATO.GEN.428258
            Non Individual Name cannot contain space hyphen space combinations
    
            Legacy Rule Format:
            IF (pyde.xx.xx:OrganisationNameDetails.OrganisationalName.Text CONTAINS  " - " )
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Contains(^FITR6, ' - ')
    
            Data Elements:
    
            ^FITR6 = RP:OrganisationNameDetails:OrganisationNameDetails.OrganisationalName.Text WHERE [OrganisationNameDetails.OrganisationalNameType.Code='MN'].[OrganisationNameDetails.Currency.Code='C']
            */
            assertion = (report.FITR6 == null ? false : report.FITR6.ToUpperInvariant().Contains(@" - "));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000413",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Non Individual Name cannot contain space hyphen space combinations",
                    Location = "/xbrli:xbrl/tns:OrganisationNameDetails" + OccurrenceIndex(report.OrganisationNameDetailsCollection_MN_C_RPOccurrenceIndex) + "/tns:OrganisationNameDetails.OrganisationalName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.428258" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR6", Value = report.FITR6 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.428258

            #region VR.ATO.GEN.428259

            /*  VR.ATO.GEN.428259
            Non Individual Name cannot contain "P/L"
    
            Legacy Rule Format:
            IF (pyde.xx.xx:OrganisationNameDetails.OrganisationalName.Text CONTAINS "P/L")
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Contains(^FITR116, 'P/L')
    
            Data Elements:
    
            ^FITR116 = INT:OrganisationNameDetails:OrganisationNameDetails.OrganisationalName.Text WHERE [OrganisationNameDetails.OrganisationalNameType.Code='MN'].[OrganisationNameDetails.Currency.Code='C']
            */
            assertion = (report.FITR116 == null ? false : report.FITR116.ToUpperInvariant().Contains(@"P/L"));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000414",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Non Individual Name cannot contain ""P/L""",
                    Location = "/xbrli:xbrl/tns:OrganisationNameDetails" + OccurrenceIndex(report.OrganisationNameDetailsCollection_MN_C_INTOccurrenceIndex) + "/tns:OrganisationNameDetails.OrganisationalName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.428259" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR116", Value = report.FITR116 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.428259

            #region VR.ATO.GEN.428259

            /*  VR.ATO.GEN.428259
            Non Individual Name cannot contain "P/L"
    
            Legacy Rule Format:
            IF (pyde.xx.xx:OrganisationNameDetails.OrganisationalName.Text CONTAINS "P/L")
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Contains(^FITR12, 'P/L')
    
            Data Elements:
    
            ^FITR12 = RP:Trustee:OrganisationNameDetails:OrganisationNameDetails.OrganisationalName.Text WHERE [OrganisationNameDetails.OrganisationalNameType.Code='MN'].[OrganisationNameDetails.Currency.Code='C']
            */
            assertion = (report.FITR12 == null ? false : report.FITR12.ToUpperInvariant().Contains(@"P/L"));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000414",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Non Individual Name cannot contain ""P/L""",
                    Location = "/xbrli:xbrl/tns:Trustee/tns:OrganisationNameDetails" + OccurrenceIndex(report.Trustee_OrganisationNameDetailsCollection_MN_C_RPOccurrenceIndex) + "/tns:OrganisationNameDetails.OrganisationalName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.428259" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR12", Value = report.FITR12 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.428259

            #region VR.ATO.GEN.428259

            /*  VR.ATO.GEN.428259
            Non Individual Name cannot contain "P/L"
    
            Legacy Rule Format:
            IF (pyde.xx.xx:OrganisationNameDetails.OrganisationalName.Text CONTAINS "P/L")
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Contains(^FITR6, 'P/L')
    
            Data Elements:
    
            ^FITR6 = RP:OrganisationNameDetails:OrganisationNameDetails.OrganisationalName.Text WHERE [OrganisationNameDetails.OrganisationalNameType.Code='MN'].[OrganisationNameDetails.Currency.Code='C']
            */
            assertion = (report.FITR6 == null ? false : report.FITR6.ToUpperInvariant().Contains(@"P/L"));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000414",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Non Individual Name cannot contain ""P/L""",
                    Location = "/xbrli:xbrl/tns:OrganisationNameDetails" + OccurrenceIndex(report.OrganisationNameDetailsCollection_MN_C_RPOccurrenceIndex) + "/tns:OrganisationNameDetails.OrganisationalName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.428259" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR6", Value = report.FITR6 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.428259

            #region VR.ATO.GEN.428260

            /*  VR.ATO.GEN.428260
            Non Individual Name cannot contain repeated hyphen, apostrophe, or space characters
    
            Legacy Rule Format:
            IF (pyde.xx.xx:OrganisationNameDetails.OrganisationalName.Text CONTAINS SET("--","''","  ") )
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (ContainsSet(^FITR116, '"--","''","  "'))
    
            Data Elements:
    
            ^FITR116 = INT:OrganisationNameDetails:OrganisationNameDetails.OrganisationalName.Text WHERE [OrganisationNameDetails.OrganisationalNameType.Code='MN'].[OrganisationNameDetails.Currency.Code='C']
            */
            assertion = IsMatch(report.FITR116, @"--|''|  ", RegexOptions.IgnoreCase);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000416",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Non Individual Name cannot contain repeated hyphen, apostrophe, or space characters",
                    Location = "/xbrli:xbrl/tns:OrganisationNameDetails" + OccurrenceIndex(report.OrganisationNameDetailsCollection_MN_C_INTOccurrenceIndex) + "/tns:OrganisationNameDetails.OrganisationalName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.428260" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR116", Value = report.FITR116 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.428260

            #region VR.ATO.GEN.428260

            /*  VR.ATO.GEN.428260
            Non Individual Name cannot contain repeated hyphen, apostrophe, or space characters
    
            Legacy Rule Format:
            IF (pyde.xx.xx:OrganisationNameDetails.OrganisationalName.Text CONTAINS SET("--","''","  ") )
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (ContainsSet(^FITR12, '"--","''","  "'))
    
            Data Elements:
    
            ^FITR12 = RP:Trustee:OrganisationNameDetails:OrganisationNameDetails.OrganisationalName.Text WHERE [OrganisationNameDetails.OrganisationalNameType.Code='MN'].[OrganisationNameDetails.Currency.Code='C']
            */
            assertion = IsMatch(report.FITR12, @"--|''|  ", RegexOptions.IgnoreCase);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000416",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Non Individual Name cannot contain repeated hyphen, apostrophe, or space characters",
                    Location = "/xbrli:xbrl/tns:Trustee/tns:OrganisationNameDetails" + OccurrenceIndex(report.Trustee_OrganisationNameDetailsCollection_MN_C_RPOccurrenceIndex) + "/tns:OrganisationNameDetails.OrganisationalName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.428260" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR12", Value = report.FITR12 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.428260

            #region VR.ATO.GEN.428260

            /*  VR.ATO.GEN.428260
            Non Individual Name cannot contain repeated hyphen, apostrophe, or space characters
    
            Legacy Rule Format:
            IF (pyde.xx.xx:OrganisationNameDetails.OrganisationalName.Text CONTAINS SET("--","''","  ") )
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (ContainsSet(^FITR6, '"--","''","  "'))
    
            Data Elements:
    
            ^FITR6 = RP:OrganisationNameDetails:OrganisationNameDetails.OrganisationalName.Text WHERE [OrganisationNameDetails.OrganisationalNameType.Code='MN'].[OrganisationNameDetails.Currency.Code='C']
            */
            assertion = IsMatch(report.FITR6, @"--|''|  ", RegexOptions.IgnoreCase);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000416",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Non Individual Name cannot contain repeated hyphen, apostrophe, or space characters",
                    Location = "/xbrli:xbrl/tns:OrganisationNameDetails" + OccurrenceIndex(report.OrganisationNameDetailsCollection_MN_C_RPOccurrenceIndex) + "/tns:OrganisationNameDetails.OrganisationalName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.428260" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR6", Value = report.FITR6 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.428260

            #region VR.ATO.GEN.428261

            /*  VR.ATO.GEN.428261
            Non Individual Name cannot end with "T/A", "T/A P'ship", "T/A Pship", "T/A P/Ship" or "T/A Partnership"
    
            Legacy Rule Format:
            IF (pyde.xx.xx:OrganisationNameDetails.OrganisationalName.Text ENDSWITH SET(" T/A"," T/A P'ship"," T/A Pship"," T/A P/Ship"," T/A Partnership"))
              RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (EndsWithSet(^FITR116, '" T/A"," T/A P'ship"," T/A Pship"," T/A P/Ship"," T/A Partnership"'))
    
            Data Elements:
    
            ^FITR116 = INT:OrganisationNameDetails:OrganisationNameDetails.OrganisationalName.Text WHERE [OrganisationNameDetails.OrganisationalNameType.Code='MN'].[OrganisationNameDetails.Currency.Code='C']
            */
            assertion = IsMatch(report.FITR116, @"( T/A| T/A P'ship| T/A Pship| T/A P/Ship| T/A Partnership)$", RegexOptions.IgnoreCase);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000417",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Non Individual Name cannot end with ""T/A"", ""T/A P'ship"", ""T/A Pship"", ""T/A P/Ship"" or ""T/A Partnership""",
                    Location = "/xbrli:xbrl/tns:OrganisationNameDetails" + OccurrenceIndex(report.OrganisationNameDetailsCollection_MN_C_INTOccurrenceIndex) + "/tns:OrganisationNameDetails.OrganisationalName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.428261" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR116", Value = report.FITR116 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.428261

            #region VR.ATO.GEN.428261

            /*  VR.ATO.GEN.428261
            Non Individual Name cannot end with "T/A", "T/A P'ship", "T/A Pship", "T/A P/Ship" or "T/A Partnership"
    
            Legacy Rule Format:
            IF (pyde.xx.xx:OrganisationNameDetails.OrganisationalName.Text ENDSWITH SET(" T/A"," T/A P'ship"," T/A Pship"," T/A P/Ship"," T/A Partnership"))
              RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (EndsWithSet(^FITR12, '" T/A"," T/A P'ship"," T/A Pship"," T/A P/Ship"," T/A Partnership"'))
    
            Data Elements:
    
            ^FITR12 = RP:Trustee:OrganisationNameDetails:OrganisationNameDetails.OrganisationalName.Text WHERE [OrganisationNameDetails.OrganisationalNameType.Code='MN'].[OrganisationNameDetails.Currency.Code='C']
            */
            assertion = IsMatch(report.FITR12, @"( T/A| T/A P'ship| T/A Pship| T/A P/Ship| T/A Partnership)$", RegexOptions.IgnoreCase);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000417",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Non Individual Name cannot end with ""T/A"", ""T/A P'ship"", ""T/A Pship"", ""T/A P/Ship"" or ""T/A Partnership""",
                    Location = "/xbrli:xbrl/tns:Trustee/tns:OrganisationNameDetails" + OccurrenceIndex(report.Trustee_OrganisationNameDetailsCollection_MN_C_RPOccurrenceIndex) + "/tns:OrganisationNameDetails.OrganisationalName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.428261" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR12", Value = report.FITR12 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.428261

            #region VR.ATO.GEN.428261

            /*  VR.ATO.GEN.428261
            Non Individual Name cannot end with "T/A", "T/A P'ship", "T/A Pship", "T/A P/Ship" or "T/A Partnership"
    
            Legacy Rule Format:
            IF (pyde.xx.xx:OrganisationNameDetails.OrganisationalName.Text ENDSWITH SET(" T/A"," T/A P'ship"," T/A Pship"," T/A P/Ship"," T/A Partnership"))
              RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (EndsWithSet(^FITR6, '" T/A"," T/A P'ship"," T/A Pship"," T/A P/Ship"," T/A Partnership"'))
    
            Data Elements:
    
            ^FITR6 = RP:OrganisationNameDetails:OrganisationNameDetails.OrganisationalName.Text WHERE [OrganisationNameDetails.OrganisationalNameType.Code='MN'].[OrganisationNameDetails.Currency.Code='C']
            */
            assertion = IsMatch(report.FITR6, @"( T/A| T/A P'ship| T/A Pship| T/A P/Ship| T/A Partnership)$", RegexOptions.IgnoreCase);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000417",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Non Individual Name cannot end with ""T/A"", ""T/A P'ship"", ""T/A Pship"", ""T/A P/Ship"" or ""T/A Partnership""",
                    Location = "/xbrli:xbrl/tns:OrganisationNameDetails" + OccurrenceIndex(report.OrganisationNameDetailsCollection_MN_C_RPOccurrenceIndex) + "/tns:OrganisationNameDetails.OrganisationalName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.428261" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR6", Value = report.FITR6 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.428261

            #region VR.ATO.GEN.428262

            /*  VR.ATO.GEN.428262
            Family Name cannot contain "space hyphen space"
    
            Legacy Rule Format:
            IF (pyde.xx.xx:PersonNameDetails.FamilyName.Text CONTAINS " - ")
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Contains(^FITR112, ' - ')
    
            Data Elements:
    
            ^FITR112 = INT:PersonNameDetails:PersonNameDetails.FamilyName.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
            */
            assertion = (report.FITR112 == null ? false : report.FITR112.ToUpperInvariant().Contains(@" - "));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000423",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Family Name cannot contain ""space hyphen space""",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_INTOccurrenceIndex) + "/tns:PersonNameDetails.FamilyName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.428262" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR112", Value = report.FITR112 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.428262

            #region VR.ATO.GEN.428262

            /*  VR.ATO.GEN.428262
            Family Name cannot contain "space hyphen space"
    
            Legacy Rule Format:
            IF (pyde.xx.xx:PersonNameDetails.FamilyName.Text CONTAINS " - ")
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Contains(^FITR103, ' - ')
    
            Data Elements:
    
            ^FITR103 = RP:PersonNameDetails:PersonNameDetails.FamilyName.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
            */
            assertion = (report.FITR103 == null ? false : report.FITR103.ToUpperInvariant().Contains(@" - "));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000423",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Family Name cannot contain ""space hyphen space""",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_RPOccurrenceIndex) + "/tns:PersonNameDetails.FamilyName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.428262" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR103", Value = report.FITR103 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.428262

            #region VR.ATO.GEN.428263

            /*  VR.ATO.GEN.428263
            Family name cannot contain repeated hyphen, apostrophe, or space characters
    
            Legacy Rule Format:
            IF (pyde.xx.xx:PersonNameDetails.FamilyName.Text CONTAINS SET("--","''","  ") )
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (ContainsSet(^FITR112, '"--","''","  "'))
    
            Data Elements:
    
            ^FITR112 = INT:PersonNameDetails:PersonNameDetails.FamilyName.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
            */
            assertion = IsMatch(report.FITR112, @"--|''|  ", RegexOptions.IgnoreCase);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000427",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Family name cannot contain repeated hyphen, apostrophe, or space characters",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_INTOccurrenceIndex) + "/tns:PersonNameDetails.FamilyName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.428263" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR112", Value = report.FITR112 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.428263

            #region VR.ATO.GEN.428263

            /*  VR.ATO.GEN.428263
            Family name cannot contain repeated hyphen, apostrophe, or space characters
    
            Legacy Rule Format:
            IF (pyde.xx.xx:PersonNameDetails.FamilyName.Text CONTAINS SET("--","''","  ") )
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (ContainsSet(^FITR103, '"--","''","  "'))
    
            Data Elements:
    
            ^FITR103 = RP:PersonNameDetails:PersonNameDetails.FamilyName.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
            */
            assertion = IsMatch(report.FITR103, @"--|''|  ", RegexOptions.IgnoreCase);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000427",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Family name cannot contain repeated hyphen, apostrophe, or space characters",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_RPOccurrenceIndex) + "/tns:PersonNameDetails.FamilyName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.428263" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR103", Value = report.FITR103 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.428263

            #region VR.ATO.GEN.428264

            /*  VR.ATO.GEN.428264
            First name cannot contain space hyphen space
    
            Legacy Rule Format:
            IF (pyde.xx.xx:PersonNameDetails.GivenName.Text CONTAINS  " - ")
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Contains(^FITR114, ' - ')
    
            Data Elements:
    
            ^FITR114 = INT:PersonNameDetails:PersonNameDetails.GivenName.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
            */
            assertion = (report.FITR114 == null ? false : report.FITR114.ToUpperInvariant().Contains(@" - "));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000434",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"First name cannot contain space hyphen space",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_INTOccurrenceIndex) + "/tns:PersonNameDetails.GivenName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.428264" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR114", Value = report.FITR114 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.428264

            #region VR.ATO.GEN.428264

            /*  VR.ATO.GEN.428264
            First name cannot contain space hyphen space
    
            Legacy Rule Format:
            IF (pyde.xx.xx:PersonNameDetails.GivenName.Text CONTAINS  " - ")
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Contains(^FITR105, ' - ')
    
            Data Elements:
    
            ^FITR105 = RP:PersonNameDetails:PersonNameDetails.GivenName.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
            */
            assertion = (report.FITR105 == null ? false : report.FITR105.ToUpperInvariant().Contains(@" - "));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000434",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"First name cannot contain space hyphen space",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_RPOccurrenceIndex) + "/tns:PersonNameDetails.GivenName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.428264" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR105", Value = report.FITR105 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.428264

            #region VR.ATO.GEN.428265

            /*  VR.ATO.GEN.428265
            First name cannot contain repeated hyphen, apostrophe, or space characters
    
            Legacy Rule Format:
            IF (pyde.xx.xx:PersonNameDetails.GivenName.Text CONTAINS SET("--","''","  ") )
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (ContainsSet(^FITR114, '"--","''","  "'))
    
            Data Elements:
    
            ^FITR114 = INT:PersonNameDetails:PersonNameDetails.GivenName.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
            */
            assertion = IsMatch(report.FITR114, @"--|''|  ", RegexOptions.IgnoreCase);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000439",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"First name cannot contain repeated hyphen, apostrophe, or space characters",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_INTOccurrenceIndex) + "/tns:PersonNameDetails.GivenName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.428265" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR114", Value = report.FITR114 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.428265

            #region VR.ATO.GEN.428265

            /*  VR.ATO.GEN.428265
            First name cannot contain repeated hyphen, apostrophe, or space characters
    
            Legacy Rule Format:
            IF (pyde.xx.xx:PersonNameDetails.GivenName.Text CONTAINS SET("--","''","  ") )
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (ContainsSet(^FITR105, '"--","''","  "'))
    
            Data Elements:
    
            ^FITR105 = RP:PersonNameDetails:PersonNameDetails.GivenName.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
            */
            assertion = IsMatch(report.FITR105, @"--|''|  ", RegexOptions.IgnoreCase);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000439",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"First name cannot contain repeated hyphen, apostrophe, or space characters",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_RPOccurrenceIndex) + "/tns:PersonNameDetails.GivenName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.428265" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR105", Value = report.FITR105 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.428265

            #region VR.ATO.GEN.428266

            /*  VR.ATO.GEN.428266
            Other given name cannot contain space hyphen space
    
            Legacy Rule Format:
            IF (pyde.xx.xx:PersonNameDetails.OtherGivenName.Text CONTAINS  " - ")
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Contains(^FITR115, ' - ')
    
            Data Elements:
    
            ^FITR115 = INT:PersonNameDetails:PersonNameDetails.OtherGivenName.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
            */
            assertion = (report.FITR115 == null ? false : report.FITR115.ToUpperInvariant().Contains(@" - "));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000446",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Other given name cannot contain space hyphen space",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_INTOccurrenceIndex) + "/tns:PersonNameDetails.OtherGivenName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.428266" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR115", Value = GetValueOrEmpty(report.FITR115) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.428266

            #region VR.ATO.GEN.428266

            /*  VR.ATO.GEN.428266
            Other given name cannot contain space hyphen space
    
            Legacy Rule Format:
            IF (pyde.xx.xx:PersonNameDetails.OtherGivenName.Text CONTAINS  " - ")
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Contains(^FITR106, ' - ')
    
            Data Elements:
    
            ^FITR106 = RP:PersonNameDetails:PersonNameDetails.OtherGivenName.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
            */
            assertion = (report.FITR106 == null ? false : report.FITR106.ToUpperInvariant().Contains(@" - "));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000446",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Other given name cannot contain space hyphen space",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_RPOccurrenceIndex) + "/tns:PersonNameDetails.OtherGivenName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.428266" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR106", Value = GetValueOrEmpty(report.FITR106) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.428266

            #region VR.ATO.GEN.428267

            /*  VR.ATO.GEN.428267
            Other given name cannot contain repeated hyphen, apostrophe, or space characters
    
            Legacy Rule Format:
            IF (pyde.xx.xx:PersonNameDetails.OtherGivenName.Text CONTAINS SET("--","''","  ") )
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (ContainsSet(^FITR115, '"--","''","  "'))
    
            Data Elements:
    
            ^FITR115 = INT:PersonNameDetails:PersonNameDetails.OtherGivenName.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
            */
            assertion = IsMatch(report.FITR115, @"--|''|  ", RegexOptions.IgnoreCase);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000451",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Other given name cannot contain repeated hyphen, apostrophe, or space characters",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_INTOccurrenceIndex) + "/tns:PersonNameDetails.OtherGivenName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.428267" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR115", Value = GetValueOrEmpty(report.FITR115) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.428267

            #region VR.ATO.GEN.428267

            /*  VR.ATO.GEN.428267
            Other given name cannot contain repeated hyphen, apostrophe, or space characters
    
            Legacy Rule Format:
            IF (pyde.xx.xx:PersonNameDetails.OtherGivenName.Text CONTAINS SET("--","''","  ") )
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (ContainsSet(^FITR106, '"--","''","  "'))
    
            Data Elements:
    
            ^FITR106 = RP:PersonNameDetails:PersonNameDetails.OtherGivenName.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
            */
            assertion = IsMatch(report.FITR106, @"--|''|  ", RegexOptions.IgnoreCase);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000451",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Other given name cannot contain repeated hyphen, apostrophe, or space characters",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_RPOccurrenceIndex) + "/tns:PersonNameDetails.OtherGivenName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.428267" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR106", Value = GetValueOrEmpty(report.FITR106) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.428267

            #region VR.ATO.GEN.430206

            /*  VR.ATO.GEN.430206
            Hours taken to prepare and complete form must be 1 to 999
    
            Legacy Rule Format:
            IF (pyin.xx.xx:Report.CompletionHours.Number <> NULL) AND (pyin.xx.xx:Report.CompletionHours.Number <> SET (1-999))
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^FITR110 <> NULL) AND ((^FITR110 < 1 OR ^FITR110 > 999))
    
            Data Elements:
    
            ^FITR110 = RP:Report.CompletionHours.Number
            */
            assertion = (report.FITR110 != null && (report.FITR110.GetValueOrDefault() < 1 || report.FITR110.GetValueOrDefault() > 999));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.430206",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Hours taken to prepare and complete form must be 1 to 999",
                    Location = "/xbrli:xbrl/tns:Report.CompletionHours.Number",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.430206" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR110", Value = report.FITR110.GetValueOrDefault().ToString() });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.430206

            #region VR.ATO.GEN.430222

            /*  VR.ATO.GEN.430222
            Email address must not start or end with an @ symbol
    
            Legacy Rule Format:
            IF (pyde.xx.xx:ElectronicContact.ElectronicMail.Address.Text STARTSWITH "@") OR (pyde.xx.xx:ElectronicContact.ElectronicMail.Address.Text ENDSWITH "@")
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (StartsWith(^FITR109, '@')) OR (EndsWith(^FITR109, '@'))
    
            Data Elements:
    
            ^FITR109 = RP:ElectronicContactElectronicMail:ElectronicContact.ElectronicMail.Address.Text WHERE [ElectronicContact.ElectronicMail.Usage.Code='03']
            */
            assertion = (IsMatch(report.FITR109, @"^(@)", RegexOptions.IgnoreCase) || IsMatch(report.FITR109, @"(@)$", RegexOptions.IgnoreCase));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.430222",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Email address must not start or end with an @ symbol",
                    Location = "/xbrli:xbrl/tns:ElectronicContactElectronicMail" + OccurrenceIndex(report.ElectronicContactElectronicMailCollection_03_RPOccurrenceIndex) + "/tns:ElectronicContact.ElectronicMail.Address.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.430222" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR109", Value = report.FITR109 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.430222

            #region VR.ATO.GEN.430223

            /*  VR.ATO.GEN.430223
            Email address must not contain more than one @ symbol
    
            Legacy Rule Format:
            IF pyde.xx.xx:ElectronicContact.ElectronicMail.Address.Text CONTAINS MORE THAN ONE "@"
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ContainsMultiple(^FITR109, '@')
    
            Data Elements:
    
            ^FITR109 = RP:ElectronicContactElectronicMail:ElectronicContact.ElectronicMail.Address.Text WHERE [ElectronicContact.ElectronicMail.Usage.Code='03']
            */
            assertion = IsMatch(report.FITR109, @"@(.*)@", RegexOptions.IgnoreCase);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.430223",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Email address must not contain more than one @ symbol",
                    Location = "/xbrli:xbrl/tns:ElectronicContactElectronicMail" + OccurrenceIndex(report.ElectronicContactElectronicMailCollection_03_RPOccurrenceIndex) + "/tns:ElectronicContact.ElectronicMail.Address.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.430223" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR109", Value = report.FITR109 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.430223

            #region VR.ATO.GEN.430224

            /*  VR.ATO.GEN.430224
            Email address must not start or end with a dot (.)
    
            Legacy Rule Format:
            IF (pyde.xx.xx:ElectronicContact.ElectronicMail.Address.Text STARTSWITH ".") OR (pyde.xx.xx:ElectronicContact.ElectronicMail.Address.Text ENDSWITH ".") 
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (StartsWith(^FITR109, '.')) OR (EndsWith(^FITR109, '.'))
    
            Data Elements:
    
            ^FITR109 = RP:ElectronicContactElectronicMail:ElectronicContact.ElectronicMail.Address.Text WHERE [ElectronicContact.ElectronicMail.Usage.Code='03']
            */
            assertion = (IsMatch(report.FITR109, @"^(\.)", RegexOptions.IgnoreCase) || IsMatch(report.FITR109, @"(\.)$", RegexOptions.IgnoreCase));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.430224",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Email address must not start or end with a dot (.)",
                    Location = "/xbrli:xbrl/tns:ElectronicContactElectronicMail" + OccurrenceIndex(report.ElectronicContactElectronicMailCollection_03_RPOccurrenceIndex) + "/tns:ElectronicContact.ElectronicMail.Address.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.430224" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR109", Value = report.FITR109 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.430224

            #region VR.ATO.GEN.430226

            /*  VR.ATO.GEN.430226
            Email address must not contain more than one consecutive dots (..)
    
            Legacy Rule Format:
            IF pyde.xx.xx:ElectronicContact.ElectronicMail.Address.Text CONTAINS ".."
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Contains(^FITR109, '..')
    
            Data Elements:
    
            ^FITR109 = RP:ElectronicContactElectronicMail:ElectronicContact.ElectronicMail.Address.Text WHERE [ElectronicContact.ElectronicMail.Usage.Code='03']
            */
            assertion = (report.FITR109 == null ? false : report.FITR109.ToUpperInvariant().Contains(@".."));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.430226",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Email address must not contain more than one consecutive dots (..)",
                    Location = "/xbrli:xbrl/tns:ElectronicContactElectronicMail" + OccurrenceIndex(report.ElectronicContactElectronicMailCollection_03_RPOccurrenceIndex) + "/tns:ElectronicContact.ElectronicMail.Address.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.430226" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR109", Value = report.FITR109 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.430226

            #region VR.ATO.GEN.430227

            /*  VR.ATO.GEN.430227
            Email address must not contain a dot followed by a @ symbol (.@)
    
            Legacy Rule Format:
            IF pyde.xx.xx:ElectronicContact.ElectronicMail.Address.Text CONTAINS ".@"
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Contains(^FITR109, '.@')
    
            Data Elements:
    
            ^FITR109 = RP:ElectronicContactElectronicMail:ElectronicContact.ElectronicMail.Address.Text WHERE [ElectronicContact.ElectronicMail.Usage.Code='03']
            */
            assertion = (report.FITR109 == null ? false : report.FITR109.ToUpperInvariant().Contains(@".@"));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.430227",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Email address must not contain a dot followed by a @ symbol (.@)",
                    Location = "/xbrli:xbrl/tns:ElectronicContactElectronicMail" + OccurrenceIndex(report.ElectronicContactElectronicMailCollection_03_RPOccurrenceIndex) + "/tns:ElectronicContact.ElectronicMail.Address.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.430227" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR109", Value = report.FITR109 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.430227

            #region VR.ATO.GEN.430228

            /*  VR.ATO.GEN.430228
            Email address must not contain a @ symbol followed by a dot (@.)
    
            Legacy Rule Format:
            IF pyde.xx.xx:ElectronicContact.ElectronicMail.Address.Text CONTAINS "@."
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Contains(^FITR109, '@.')
    
            Data Elements:
    
            ^FITR109 = RP:ElectronicContactElectronicMail:ElectronicContact.ElectronicMail.Address.Text WHERE [ElectronicContact.ElectronicMail.Usage.Code='03']
            */
            assertion = (report.FITR109 == null ? false : report.FITR109.ToUpperInvariant().Contains(@"@."));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.430228",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Email address must not contain a @ symbol followed by a dot (@.)",
                    Location = "/xbrli:xbrl/tns:ElectronicContactElectronicMail" + OccurrenceIndex(report.ElectronicContactElectronicMailCollection_03_RPOccurrenceIndex) + "/tns:ElectronicContact.ElectronicMail.Address.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.430228" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR109", Value = report.FITR109 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.430228

            #region VR.ATO.GEN.430229

            /*  VR.ATO.GEN.430229
            Email address must not contain an invalid character
    
            Legacy Rule Format:
            IF pyde.xx.xx:ElectronicContact.ElectronicMail.Address.Text CONTAINS SET("(", ")", "<", ">", ",", ";", ":", "\",  """, "]", "[")
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ContainsSet(^FITR109, '"(", ")", "<", ">", ",", ";", ":", "\",  """, "]", "["')
    
            Data Elements:
    
            ^FITR109 = RP:ElectronicContactElectronicMail:ElectronicContact.ElectronicMail.Address.Text WHERE [ElectronicContact.ElectronicMail.Usage.Code='03']
            */
            assertion = IsMatch(report.FITR109, @"\(|\)|<|>|,|;|:|\\|""|\]|\[", RegexOptions.IgnoreCase);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.430229",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Email address must not contain an invalid character",
                    LongDescription = @"An email address must not contain any of the following invalid characters: ( ) < > , ; : \ "" [ ]",
                    Location = "/xbrli:xbrl/tns:ElectronicContactElectronicMail" + OccurrenceIndex(report.ElectronicContactElectronicMailCollection_03_RPOccurrenceIndex) + "/tns:ElectronicContact.ElectronicMail.Address.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.430229" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR109", Value = report.FITR109 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.430229

            #region VR.ATO.GEN.430230

            /*  VR.ATO.GEN.430230
            Email address must contain an @ symbol
    
            Legacy Rule Format:
            IF (pyde.xx.xx:ElectronicContact.ElectronicMail.Address.Text <> NULLORBLANK) AND (pyde.xx.xx:ElectronicContact.ElectronicMail.Address.Text DOES NOT CONTAIN "@")
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^ElectronicContactElectronicMail <> NULL) AND ((^FITR109 <> BLANK) AND (NotContains(^FITR109, '@')))
    
            Data Elements:
    
            ^FITR109 = RP:ElectronicContactElectronicMail:ElectronicContact.ElectronicMail.Address.Text WHERE [ElectronicContact.ElectronicMail.Usage.Code='03']
    
            ^ElectronicContactElectronicMail = RP:ElectronicContactElectronicMail
            */
            assertion = (report.ElectronicContactElectronicMailCollection_03_RPExists != false && (string.IsNullOrWhiteSpace(report.FITR109) != true && !(report.FITR109 == null ? true : report.FITR109.ToUpperInvariant().Contains(@"@"))));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.430230",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Email address must contain an @ symbol",
                    Location = "/xbrli:xbrl/tns:ElectronicContactElectronicMail" + OccurrenceIndex(report.ElectronicContactElectronicMailCollection_03_RPOccurrenceIndex) + "/tns:ElectronicContact.ElectronicMail.Address.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.430230" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR109", Value = report.FITR109 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.430230

            #region VR.ATO.GEN.430231

            /*  VR.ATO.GEN.430231
            An email address must not start with a space or include a space within it
    
            Legacy Rule Format:
            IF pyde.xx.xx:ElectronicContact.ElectronicMail.Address.Text = FOUND(" ")
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Contains(^FITR109, ' ')
    
            Data Elements:
    
            ^FITR109 = RP:ElectronicContactElectronicMail:ElectronicContact.ElectronicMail.Address.Text WHERE [ElectronicContact.ElectronicMail.Usage.Code='03']
            */
            assertion = (report.FITR109 == null ? false : report.FITR109.ToUpperInvariant().Contains(@" "));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.430231",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"An email address must not start with a space or include a space within it",
                    Location = "/xbrli:xbrl/tns:ElectronicContactElectronicMail" + OccurrenceIndex(report.ElectronicContactElectronicMailCollection_03_RPOccurrenceIndex) + "/tns:ElectronicContact.ElectronicMail.Address.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.430231" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR109", Value = report.FITR109 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.430231

            #region VR.ATO.GEN.430232

            /*  VR.ATO.GEN.430232
            Email address must contain a dot (.)
    
            Legacy Rule Format:
            IF (pyde.xx.xx:ElectronicContact.ElectronicMail.Address.Text <> NULLORBLANK) AND (pyde.xx.xx:ElectronicContact.ElectronicMail.Address.Text DOES NOT CONTAIN ".")
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^ElectronicContactElectronicMail <> NULL) AND ((^FITR109 <> BLANK) AND (NotContains(^FITR109, '.')))
    
            Data Elements:
    
            ^FITR109 = RP:ElectronicContactElectronicMail:ElectronicContact.ElectronicMail.Address.Text WHERE [ElectronicContact.ElectronicMail.Usage.Code='03']
    
            ^ElectronicContactElectronicMail = RP:ElectronicContactElectronicMail
            */
            assertion = (report.ElectronicContactElectronicMailCollection_03_RPExists != false && (string.IsNullOrWhiteSpace(report.FITR109) != true && !(report.FITR109 == null ? true : report.FITR109.ToUpperInvariant().Contains(@"."))));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.430232",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Email address must contain a dot (.)",
                    Location = "/xbrli:xbrl/tns:ElectronicContactElectronicMail" + OccurrenceIndex(report.ElectronicContactElectronicMailCollection_03_RPOccurrenceIndex) + "/tns:ElectronicContact.ElectronicMail.Address.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.430232" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR109", Value = report.FITR109 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.430232

            #region VR.ATO.GEN.430246

            /*  VR.ATO.GEN.430246
            Suburb/Town/City cannot contain numeric characters
    
            Legacy Rule Format:
            WHERE IN TUPLE(addressdetails2.xx.xx:AddressDetails)
            IF (pyde.xx.xx:AddressDetails.OverseasAddress.Indicator = FALSE) AND (pyde.xx.xx:AddressDetails.LocalityName.Text CONTAINS NUMERICS SET(0-9))
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^FITR143 = FALSE) AND (ContainsSet(^FITR9, '0-9'))
    
            Data Elements:
    
            ^FITR9 = RP:AddressDetails:AddressDetails.LocalityName.Text WHERE [AddressDetails.Usage.Code='POS'].[AddressDetails.Currency.Code='C']
    
            ^FITR143 = RP:AddressDetails:AddressDetails.OverseasAddress.Indicator WHERE [AddressDetails.Usage.Code='POS'].[AddressDetails.Currency.Code='C']
            */
            assertion = (report.FITR143 == false && IsMatch(report.FITR9, @"[0-9]", RegexOptions.IgnoreCase));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000410",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Suburb/Town/City cannot contain numeric characters",
                    Location = "/xbrli:xbrl/tns:AddressDetails" + OccurrenceIndex(report.AddressDetailsCollection_POS_C_RPOccurrenceIndex) + "/tns:AddressDetails.LocalityName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.430246" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR143", Value = GetValueOrEmpty(report.FITR143) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR9", Value = report.FITR9 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.430246

            #region VR.ATO.GEN.430252

            /*  VR.ATO.GEN.430252
            Unstructured name must not be more than 200 characters
    
            Legacy Rule Format:
            IF LENGTH(pyde.xx.xx:PersonUnstructuredName.FullName.Text) > 200
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Length(^FITR169) > 200
    
            Data Elements:
    
            ^FITR169 = INT:Declaration:PersonUnstructuredName:PersonUnstructuredName.FullName.Text WHERE [PersonUnstructuredName.Usage.Code='DeclarationSignatory']
            */
            assertion = (Length(report.FITR169) > 200);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.430252",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Unstructured name must not be more than 200 characters",
                    Location = "/xbrli:xbrl/tns:Declaration" + OccurrenceIndex(report.DeclarationCollection_TrueAndCorrect_INTOccurrenceIndex) + "/tns:PersonUnstructuredName" + OccurrenceIndex(report.Declaration_PersonUnstructuredNameCollection_DeclarationSignatory_INTOccurrenceIndex) + "/tns:PersonUnstructuredName.FullName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.430252" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR169", Value = report.FITR169 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.430252

            #region VR.ATO.GEN.430252

            /*  VR.ATO.GEN.430252
            Unstructured name must not be more than 200 characters
    
            Legacy Rule Format:
            IF LENGTH(pyde.xx.xx:PersonUnstructuredName.FullName.Text) > 200
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Length(^FITR167) > 200
    
            Data Elements:
    
            ^FITR167 = RP:Declaration:PersonUnstructuredName:PersonUnstructuredName.FullName.Text WHERE [PersonUnstructuredName.Usage.Code='DeclarationSignatory']
            */
            assertion = (Length(report.FITR167) > 200);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.430252",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Unstructured name must not be more than 200 characters",
                    Location = "/xbrli:xbrl/tns:Declaration" + OccurrenceIndex(report.DeclarationCollection_TrueAndCorrect_RPOccurrenceIndex) + "/tns:PersonUnstructuredName" + OccurrenceIndex(report.Declaration_PersonUnstructuredNameCollection_DeclarationSignatory_RPOccurrenceIndex) + "/tns:PersonUnstructuredName.FullName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.430252" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR167", Value = report.FITR167 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.430252

            #region VR.ATO.GEN.430255

            /*  VR.ATO.GEN.430255
            Signature date must not be later than today
    
            Legacy Rule Format:
            IF pyin.xx.xx:Declaration.Signature.Date > DATE(TODAY)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR125 > Today()
    
            Data Elements:
    
            ^FITR125 = INT:Declaration:Declaration.Signature.Date WHERE [Declaration.StatementType.Code='TrueAndCorrect']
            */
            assertion = (report.FITR125.GetValueOrDefault() > DateTime.Now.Date);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.430255",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Signature date must not be later than today",
                    Location = "/xbrli:xbrl/tns:Declaration" + OccurrenceIndex(report.DeclarationCollection_TrueAndCorrect_INTOccurrenceIndex) + "/tns:Declaration.Signature.Date",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.430255" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR125", Value = GetValueOrEmptyDate(report.FITR125) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.430255

            #region VR.ATO.GEN.430255

            /*  VR.ATO.GEN.430255
            Signature date must not be later than today
    
            Legacy Rule Format:
            IF pyin.xx.xx:Declaration.Signature.Date > DATE(TODAY)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR123 > Today()
    
            Data Elements:
    
            ^FITR123 = RP:Declaration:Declaration.Signature.Date WHERE [Declaration.StatementType.Code='TrueAndCorrect']
            */
            assertion = (report.FITR123.GetValueOrDefault() > DateTime.Now.Date);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.430255",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Signature date must not be later than today",
                    Location = "/xbrli:xbrl/tns:Declaration" + OccurrenceIndex(report.DeclarationCollection_TrueAndCorrect_RPOccurrenceIndex) + "/tns:Declaration.Signature.Date",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.430255" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR123", Value = GetValueOrEmptyDate(report.FITR123) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.430255

            #region VR.ATO.GEN.430266

            /*  VR.ATO.GEN.430266
            Invalid context. The number of Reporting party: Australian contexts must not exceed 1
    
            Legacy Rule Format:
            IF COUNT(CONTEXT(RP.JAus)) > 1 
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Count(^Context) > 1
    
            Data Elements:
    
            ^Context = RP:RP.JAUS
            */
            assertion = (Count(report.RPJAUSCount) > 1);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.430266",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Invalid context. The number of Reporting party: Australian contexts must not exceed 1",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPJAUSLastOccurrenceIndex),
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.430266" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.430266

            #region VR.ATO.GEN.430267

            /*  VR.ATO.GEN.430267
            Invalid context. The number of Reporting party: Foreign contexts must not exceed 1
    
            Legacy Rule Format:
            IF COUNT(CONTEXT(RP.JForeign)) > 1 
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Count(^Context) > 1
    
            Data Elements:
    
            ^Context = RP:RP.JFOREIGN
            */
            assertion = (Count(report.RPJFOREIGNCount) > 1);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.430267",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Invalid context. The number of Reporting party: Foreign contexts must not exceed 1",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPJFOREIGNLastOccurrenceIndex),
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.430267" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.430267

            #region VR.ATO.GEN.430280

            /*  VR.ATO.GEN.430280
            Invalid context. The number of Reporting Party Instant: Closing contexts must not exceed 1
    
            Legacy Rule Format:
            IF COUNT(CONTEXT(RP.Closing(Instant))) > 1 
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Count(^Context) > 1
    
            Data Elements:
    
            ^Context = RP.Closing(Instant)
            */
            assertion = (Count(report.RPClosingInstantCount) > 1);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.430280",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Invalid context. The number of Reporting Party Instant: Closing contexts must not exceed 1",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPClosingInstantLastOccurrenceIndex),
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.430280" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.430280

            #region VR.ATO.GEN.430302

            /*  VR.ATO.GEN.430302
            Invalid context. The number of Reporting party: TOFA contexts must not exceed 1
    
            Legacy Rule Format:
            IF COUNT(CONTEXT(RP.TOFA)) > 1 
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Count(^Context) > 1
    
            Data Elements:
    
            ^Context = RP:RP.TOFA
            */
            assertion = (Count(report.RPTOFACount) > 1);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.430302",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Invalid context. The number of Reporting party: TOFA contexts must not exceed 1",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPTOFALastOccurrenceIndex),
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.430302" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.430302

            #region VR.ATO.GEN.430318

            /*  VR.ATO.GEN.430318
            ABN in the context definition must pass the ABN algorithm check
    
            Legacy Rule Format:
            IF (entity.identifier.ABN <> NULLORBLANK) AND (ABNALGORITHM(entity.identifier.ABN) = FALSE)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^ABN <> BLANK) AND (FailsABNAlgorithm(^ABN))
    
            Data Elements:
    
            ^ABN = INT
            */
            assertion = (string.IsNullOrWhiteSpace(report.INTIdentifierABN) != true && FailsABNAlgorithm(report.INTIdentifierABN));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.430318",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"ABN in context must be a valid ABN",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.INTOccurrenceIndex) + "/xbrli:entity/xbrli:identifier",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.430318" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.430318

            #region VR.ATO.GEN.432037

            /*  VR.ATO.GEN.432037
            Bank account name must not contain invalid characters _ ! @ $ % [ ] ? = : ; " ,
    
            Legacy Rule Format:
            IF pyid.xx.xx:FinancialInstitutionAccount.FinancialInstitutionAccountName.Text CONTAINS SET("_", "!", "@", "$", "%", "]", "[", "?", "=", ":", ";", """, ",")
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ContainsSet(^FITR178, '"_", "!", "@", "$", "%", "]", "[", "?", "=", ":", ";", """, ","')
    
            Data Elements:
    
            ^FITR178 = RP:FinancialInstitutionAccount:FinancialInstitutionAccount.FinancialInstitutionAccountName.Text
            */
            assertion = IsMatch(report.FITR178, @"_|!|@|\$|%|\]|\[|\?|=|:|;|""|,", RegexOptions.IgnoreCase);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.432037",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Bank account name must not contain invalid characters _ ! @ $ % [ ] ? = : ; "" ,",
                    Location = "/xbrli:xbrl/tns:FinancialInstitutionAccount/tns:FinancialInstitutionAccount.FinancialInstitutionAccountName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.432037" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR178", Value = report.FITR178 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.432037

            #region VR.ATO.GEN.432040

            /*  VR.ATO.GEN.432040
            Only one EFT bank account may be present
    
            Legacy Rule Format:
            IF COUNT (TUPLE(financialinstitutionaccount1.xx.xx:FinancialInstitutionAccount)) > 1
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Count(^FinancialInstitutionAccount) > 1
    
            Data Elements:
    
            ^FinancialInstitutionAccount = RP:FinancialInstitutionAccount
            */
            assertion = (Count(report.FinancialInstitutionAccountCollectionCount) > 1);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.432040",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Only one EFT bank account may be present",
                    Location = "/xbrli:xbrl/tns:FinancialInstitutionAccount" + OccurrenceIndex(report.FinancialInstitutionAccountCollectionCount) + "/tns:FinancialInstitutionAccount.FinancialInstitutionAccountName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.432040" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.432040

            #region VR.ATO.GEN.432041

            /*  VR.ATO.GEN.432041
            Financial institution account name cannot be greater than 32 characters
    
            Legacy Rule Format:
            IF LENGTH(pyid.xx.xx:FinancialInstitutionAccount.FinancialInstitutionAccountName.Text) > 32
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Length(^FITR178) > 32
    
            Data Elements:
    
            ^FITR178 = RP:FinancialInstitutionAccount:FinancialInstitutionAccount.FinancialInstitutionAccountName.Text
            */
            assertion = (Length(report.FITR178) > 32);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.432041",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Bank account name must not be longer than 32 characters",
                    LongDescription = @"32 characters is the Banking Standard and set by the Banking Industry",
                    Location = "/xbrli:xbrl/tns:FinancialInstitutionAccount/tns:FinancialInstitutionAccount.FinancialInstitutionAccountName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.432041" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR178", Value = report.FITR178 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.432041

            #region VR.ATO.GEN.432267

            /*  VR.ATO.GEN.432267
            Financial institution name and branch name must not be present
    
            Legacy Rule Format:
            IF (pyid.xx.xx:FinancialInstitutionAccount.FinancialInstitutionBranchName.Text) <> NULLORBLANK OR (pyid.xx.xx:FinancialInstitutionAccount.FinancialInstitutionName.Text <> NULLORBLANK)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR179 <> BLANK OR (^FITR180 <> BLANK)
    
            Data Elements:
    
            ^FITR180 = RP:FinancialInstitutionAccount:FinancialInstitutionAccount.FinancialInstitutionName.Text
    
            ^FITR179 = RP:FinancialInstitutionAccount:FinancialInstitutionAccount.FinancialInstitutionBranchName.Text
            */
            assertion = (string.IsNullOrWhiteSpace(report.FITR179) != true || string.IsNullOrWhiteSpace(report.FITR180) != true);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.432267",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Financial institution name and branch name must not be present",
                    Location = "/xbrli:xbrl/tns:FinancialInstitutionAccount/tns:FinancialInstitutionAccount.FinancialInstitutionName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.432267" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR179", Value = GetValueOrEmpty(report.FITR179) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR180", Value = GetValueOrEmpty(report.FITR180) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.432267

            #region VR.ATO.GEN.432268

            /*  VR.ATO.GEN.432268
            Based on the return calculation details a credit or refund is due. In order for the ATO to pay this credit or refund, bank account details must be provided.
    
            Legacy Rule Format:
            IF (RP:rvctc1.xx.xx:IncomeTax.PayableOrRefundableTotal.Amount < 0) AND
            COUNT TUPLE (fininstacc1.xx.xx:FinancialInstitutionAccount) <> 1
             RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^FITR75 < 0) AND Count(^FinancialInstitutionAccount) <> 1
    
            Data Elements:
    
            ^FITR75 = RP:IncomeTax.PayableOrRefundableTotal.Amount
    
            ^FinancialInstitutionAccount = RP:FinancialInstitutionAccount
            */
            assertion = (report.FITR75.GetValueOrDefault() < 0 && Count(report.FinancialInstitutionAccountCollectionCount) != 1);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.432268",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Financial institution details must be entered when a credit (refund) is due.",
                    LongDescription = @"Based on the return calculation details a credit or refund is due. In order for the ATO to pay this credit or refund, Australian financial institution account details must be provided. If account details are not available please phone the Tax Office on: For self lodgers 13 28 61 options 2 2 and for tax agents 13 72 86 options 1 3 1 2.",
                    Location = "/xbrli:xbrl/tns:IncomeTax.PayableOrRefundableTotal.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.432268" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR75", Value = GetValueOrEmpty(report.FITR75) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.432268

            #region VR.ATO.GEN.432417

            /*  VR.ATO.GEN.432417
            ABN for reporting party must be provided when there is no intermediary
    
            Legacy Rule Format:
            IF CONTEXT(INT) = NULL AND RP:pyid.xx.xx:Identifiers.AustralianBusinessNumber.Identifier = NULLORBLANK
                    RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^Context = NULL AND ^FITR7 = BLANK
    
            Data Elements:
    
            ^FITR7 = RP:Identifiers.AustralianBusinessNumber.Identifier
    
            ^Context = INT
            */
            assertion = (report.INTExists == false && string.IsNullOrWhiteSpace(report.FITR7) == true);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.432417",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"ABN for reporting party must be provided when there is no intermediary",
                    Location = "/xbrli:xbrl/tns:Identifiers.AustralianBusinessNumber.Identifier",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.432417" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR7", Value = GetValueOrEmpty(report.FITR7) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.432417

            #region VR.ATO.GEN.434223

            /*  VR.ATO.GEN.434223
            ABN must pass the ABN algorithm
    
            Legacy Rule Format:
            IF (pyid.xx.xx:Identifiers.AustralianBusinessNumber.Identifier <> NULLORBLANK) and (ABNALGORITHM(pyid.xx.xx:Identifiers.AustralianBusinessNumber.Identifier) = FALSE)
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^FITR13 <> BLANK) and (FailsABNAlgorithm(^FITR13))
    
            Data Elements:
    
            ^FITR13 = RP:Trustee:Identifiers.AustralianBusinessNumber.Identifier
            */
            assertion = (string.IsNullOrWhiteSpace(report.FITR13) != true && FailsABNAlgorithm(report.FITR13));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000477",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"ABN is invalid",
                    Location = "/xbrli:xbrl/tns:Trustee/tns:Identifiers.AustralianBusinessNumber.Identifier",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.434223" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR13", Value = GetValueOrEmpty(report.FITR13) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.434223

            #region VR.ATO.GEN.434223

            /*  VR.ATO.GEN.434223
            ABN must pass the ABN algorithm
    
            Legacy Rule Format:
            IF (pyid.xx.xx:Identifiers.AustralianBusinessNumber.Identifier <> NULLORBLANK) and (ABNALGORITHM(pyid.xx.xx:Identifiers.AustralianBusinessNumber.Identifier) = FALSE)
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^FITR7 <> BLANK) and (FailsABNAlgorithm(^FITR7))
    
            Data Elements:
    
            ^FITR7 = RP:Identifiers.AustralianBusinessNumber.Identifier
            */
            assertion = (string.IsNullOrWhiteSpace(report.FITR7) != true && FailsABNAlgorithm(report.FITR7));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.000477",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"ABN is invalid",
                    Location = "/xbrli:xbrl/tns:Identifiers.AustralianBusinessNumber.Identifier",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.434223" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR7", Value = GetValueOrEmpty(report.FITR7) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.434223

            #region VR.ATO.GEN.438019

            /*  VR.ATO.GEN.438019
            If an exemption or rollover has been applied then the exemption or rollover code(s) must be supplied.
    
            Legacy Rule Format:
            IF (RP:rvctc3.xx.xx:CapitalGainsTax.ExemptionOrRolloverApplied.Indicator = TRUE) AND (COUNT(RP:rvctc3.xx.xx:CapitalGainsTax.ExemptionOrRolloverApplied.Code <> NULL) = 0)
              RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^FITR203 = TRUE) AND (Count(^FITR204) = 0)
    
            Data Elements:
    
            ^FITR203 = RP:CapitalGainsTax.ExemptionOrRolloverApplied.Indicator
    
            ^FITR204 = RP:CapitalGainsTax.ExemptionOrRolloverApplied.Code
            */
            assertion = (report.FITR203 == true && Count(report.FITR204Collection) == 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438019",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Exemption or rollover code not supplied",
                    LongDescription = @"If an exemption or rollover has been applied then the exemption or rollover code(s) must be supplied",
                    Location = "/xbrli:xbrl/tns:CapitalGainsTax.ExemptionOrRolloverApplied.Indicator",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438019" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR203", Value = GetValueOrEmpty(report.FITR203) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438019

            #region VR.ATO.GEN.438021

            /*  VR.ATO.GEN.438021
            If an exemption or rollover code(s) is provided then the exemption or rollover applied indicator must be true
    
            Legacy Rule Format:
            IF (COUNT(RP:rvctc3.xx.xx:CapitalGainsTax.ExemptionOrRolloverApplied.Code <> NULL) > 0) AND (RP:rvctc3.xx.xx:CapitalGainsTax.ExemptionOrRolloverApplied.Indicator <> TRUE)
              RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (Count(^FITR204) > 0) AND (^FITR203 <> TRUE)
    
            Data Elements:
    
            ^FITR203 = RP:CapitalGainsTax.ExemptionOrRolloverApplied.Indicator
    
            ^FITR204 = RP:CapitalGainsTax.ExemptionOrRolloverApplied.Code
            */
            assertion = (Count(report.FITR204Collection) > 0 && report.FITR203 != true);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438021",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Exemption or rollover applied indicator must be true",
                    LongDescription = @"If an exemption or rollover code(s) is provided then the exemption or rollover applied indicator must be true",
                    Location = "/xbrli:xbrl/tns:CapitalGainsTax.ExemptionOrRolloverApplied.Indicator",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438021" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR203", Value = GetValueOrEmpty(report.FITR203) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438021

            #region VR.ATO.GEN.438029

            /*  VR.ATO.GEN.438029
            Entity identifier scheme for ReportingParty must be "http://www.ato.gov.au/tfn"
    
            Legacy Rule Format:
            IF (RprtPyType.xx.xx:ReportPartyTypeDimension = "RprtPyType.xx.xx:ReportingParty") AND (entity.identifier.scheme <> "http://www.ato.gov.au/tfn")
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^Context <> NULL) AND (^IdentifierScheme <> 'http://www.ato.gov.au/tfn')
    
            Data Elements:
    
            ^IdentifierScheme = RP
    
            ^Context = RP
            */
            assertion = (report.RPExists != false && report.RPIdentifierScheme != @"http://www.ato.gov.au/tfn");
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438029",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Entity identifier scheme for ReportingParty must be ""http://www.ato.gov.au/tfn""",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPOccurrenceIndex) + "/xbrli:entity/xbrli:identifier/@scheme",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438029" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438029

            #region VR.ATO.GEN.438029

            /*  VR.ATO.GEN.438029
            Entity identifier scheme for ReportingParty must be "http://www.ato.gov.au/tfn"
    
            Legacy Rule Format:
            IF (RprtPyType.xx.xx:ReportPartyTypeDimension = "RprtPyType.xx.xx:ReportingParty") AND (entity.identifier.scheme <> "http://www.ato.gov.au/tfn")
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^Context <> NULL) AND (^IdentifierScheme <> 'http://www.ato.gov.au/tfn')
    
            Data Elements:
    
            ^IdentifierScheme = RP(Instant)
    
            ^Context = RP(Instant)
            */
            assertion = (report.RPInstantExists != false && report.RPInstantIdentifierScheme != @"http://www.ato.gov.au/tfn");
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438029",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Entity identifier scheme for ReportingParty must be ""http://www.ato.gov.au/tfn""",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPInstantOccurrenceIndex) + "/xbrli:entity/xbrli:identifier/@scheme",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438029" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438029

            #region VR.ATO.GEN.438029

            /*  VR.ATO.GEN.438029
            Entity identifier scheme for ReportingParty must be "http://www.ato.gov.au/tfn"
    
            Legacy Rule Format:
            IF (RprtPyType.xx.xx:ReportPartyTypeDimension = "RprtPyType.xx.xx:ReportingParty") AND (entity.identifier.scheme <> "http://www.ato.gov.au/tfn")
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^Context <> NULL) AND (^IdentifierScheme <> 'http://www.ato.gov.au/tfn')
    
            Data Elements:
    
            ^IdentifierScheme = RP:RP.CFCLC
    
            ^Context = RP:RP.CFCLC
            */
            assertion = (report.RPCFCLCExists != false && report.RPCFCLCIdentifierScheme != @"http://www.ato.gov.au/tfn");
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438029",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Entity identifier scheme for ReportingParty must be ""http://www.ato.gov.au/tfn""",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPCFCLCOccurrenceIndex) + "/xbrli:entity/xbrli:identifier/@scheme",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438029" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438029

            #region VR.ATO.GEN.438029

            /*  VR.ATO.GEN.438029
            Entity identifier scheme for ReportingParty must be "http://www.ato.gov.au/tfn"
    
            Legacy Rule Format:
            IF (RprtPyType.xx.xx:ReportPartyTypeDimension = "RprtPyType.xx.xx:ReportingParty") AND (entity.identifier.scheme <> "http://www.ato.gov.au/tfn")
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^Context <> NULL) AND (^IdentifierScheme <> 'http://www.ato.gov.au/tfn')
    
            Data Elements:
    
            ^IdentifierScheme = RP:RP.CFCUC
    
            ^Context = RP:RP.CFCUC
            */
            assertion = (report.RPCFCUCExists != false && report.RPCFCUCIdentifierScheme != @"http://www.ato.gov.au/tfn");
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438029",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Entity identifier scheme for ReportingParty must be ""http://www.ato.gov.au/tfn""",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPCFCUCOccurrenceIndex) + "/xbrli:entity/xbrli:identifier/@scheme",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438029" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438029

            #region VR.ATO.GEN.438029

            /*  VR.ATO.GEN.438029
            Entity identifier scheme for ReportingParty must be "http://www.ato.gov.au/tfn"
    
            Legacy Rule Format:
            IF (RprtPyType.xx.xx:ReportPartyTypeDimension = "RprtPyType.xx.xx:ReportingParty") AND (entity.identifier.scheme <> "http://www.ato.gov.au/tfn")
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^Context <> NULL) AND (^IdentifierScheme <> 'http://www.ato.gov.au/tfn')
    
            Data Elements:
    
            ^IdentifierScheme = RP:RP.Closing
    
            ^Context = RP:RP.Closing
            */
            assertion = (report.RPClosingExists != false && report.RPClosingIdentifierScheme != @"http://www.ato.gov.au/tfn");
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438029",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Entity identifier scheme for ReportingParty must be ""http://www.ato.gov.au/tfn""",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPClosingOccurrenceIndex) + "/xbrli:entity/xbrli:identifier/@scheme",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438029" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438029

            #region VR.ATO.GEN.438029

            /*  VR.ATO.GEN.438029
            Entity identifier scheme for ReportingParty must be "http://www.ato.gov.au/tfn"
    
            Legacy Rule Format:
            IF (RprtPyType.xx.xx:ReportPartyTypeDimension = "RprtPyType.xx.xx:ReportingParty") AND (entity.identifier.scheme <> "http://www.ato.gov.au/tfn")
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^Context <> NULL) AND (^IdentifierScheme <> 'http://www.ato.gov.au/tfn')
    
            Data Elements:
    
            ^IdentifierScheme = RP.Closing(Instant)
    
            ^Context = RP.Closing(Instant)
            */
            assertion = (report.RPClosingInstantExists != false && report.RPClosingInstantIdentifierScheme != @"http://www.ato.gov.au/tfn");
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438029",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Entity identifier scheme for ReportingParty must be ""http://www.ato.gov.au/tfn""",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPClosingInstantOccurrenceIndex) + "/xbrli:entity/xbrli:identifier/@scheme",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438029" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438029

            #region VR.ATO.GEN.438029

            /*  VR.ATO.GEN.438029
            Entity identifier scheme for ReportingParty must be "http://www.ato.gov.au/tfn"
    
            Legacy Rule Format:
            IF (RprtPyType.xx.xx:ReportPartyTypeDimension = "RprtPyType.xx.xx:ReportingParty") AND (entity.identifier.scheme <> "http://www.ato.gov.au/tfn")
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^Context <> NULL) AND (^IdentifierScheme <> 'http://www.ato.gov.au/tfn')
    
            Data Elements:
    
            ^IdentifierScheme = RP:RP.Investor
    
            ^Context = RP:RP.Investor
            */
            assertion = (report.RPInvestorExists != false && report.RPInvestorIdentifierScheme != @"http://www.ato.gov.au/tfn");
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438029",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Entity identifier scheme for ReportingParty must be ""http://www.ato.gov.au/tfn""",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPInvestorOccurrenceIndex) + "/xbrli:entity/xbrli:identifier/@scheme",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438029" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438029

            #region VR.ATO.GEN.438029

            /*  VR.ATO.GEN.438029
            Entity identifier scheme for ReportingParty must be "http://www.ato.gov.au/tfn"
    
            Legacy Rule Format:
            IF (RprtPyType.xx.xx:ReportPartyTypeDimension = "RprtPyType.xx.xx:ReportingParty") AND (entity.identifier.scheme <> "http://www.ato.gov.au/tfn")
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^Context <> NULL) AND (^IdentifierScheme <> 'http://www.ato.gov.au/tfn')
    
            Data Elements:
    
            ^IdentifierScheme = RP.Investor:RP.Investor.YPrevious
    
            ^Context = RP.Investor:RP.Investor.YPrevious
            */
            assertion = (report.RPInvestorYPreviousExists != false && report.RPInvestorYPreviousIdentifierScheme != @"http://www.ato.gov.au/tfn");
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438029",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Entity identifier scheme for ReportingParty must be ""http://www.ato.gov.au/tfn""",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPInvestorYPreviousOccurrenceIndex) + "/xbrli:entity/xbrli:identifier/@scheme",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438029" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438029

            #region VR.ATO.GEN.438029

            /*  VR.ATO.GEN.438029
            Entity identifier scheme for ReportingParty must be "http://www.ato.gov.au/tfn"
    
            Legacy Rule Format:
            IF (RprtPyType.xx.xx:ReportPartyTypeDimension = "RprtPyType.xx.xx:ReportingParty") AND (entity.identifier.scheme <> "http://www.ato.gov.au/tfn")
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^Context <> NULL) AND (^IdentifierScheme <> 'http://www.ato.gov.au/tfn')
    
            Data Elements:
    
            ^IdentifierScheme = RP:RP.JAUS
    
            ^Context = RP:RP.JAUS
            */
            assertion = (report.RPJAUSExists != false && report.RPJAUSIdentifierScheme != @"http://www.ato.gov.au/tfn");
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438029",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Entity identifier scheme for ReportingParty must be ""http://www.ato.gov.au/tfn""",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPJAUSOccurrenceIndex) + "/xbrli:entity/xbrli:identifier/@scheme",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438029" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438029

            #region VR.ATO.GEN.438029

            /*  VR.ATO.GEN.438029
            Entity identifier scheme for ReportingParty must be "http://www.ato.gov.au/tfn"
    
            Legacy Rule Format:
            IF (RprtPyType.xx.xx:ReportPartyTypeDimension = "RprtPyType.xx.xx:ReportingParty") AND (entity.identifier.scheme <> "http://www.ato.gov.au/tfn")
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^Context <> NULL) AND (^IdentifierScheme <> 'http://www.ato.gov.au/tfn')
    
            Data Elements:
    
            ^IdentifierScheme = RP:RP.JFOREIGN
    
            ^Context = RP:RP.JFOREIGN
            */
            assertion = (report.RPJFOREIGNExists != false && report.RPJFOREIGNIdentifierScheme != @"http://www.ato.gov.au/tfn");
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438029",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Entity identifier scheme for ReportingParty must be ""http://www.ato.gov.au/tfn""",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPJFOREIGNOccurrenceIndex) + "/xbrli:entity/xbrli:identifier/@scheme",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438029" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438029

            #region VR.ATO.GEN.438029

            /*  VR.ATO.GEN.438029
            Entity identifier scheme for ReportingParty must be "http://www.ato.gov.au/tfn"
    
            Legacy Rule Format:
            IF (RprtPyType.xx.xx:ReportPartyTypeDimension = "RprtPyType.xx.xx:ReportingParty") AND (entity.identifier.scheme <> "http://www.ato.gov.au/tfn")
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^Context <> NULL) AND (^IdentifierScheme <> 'http://www.ato.gov.au/tfn')
    
            Data Elements:
    
            ^IdentifierScheme = RP:RP.LimitedPartners
    
            ^Context = RP:RP.LimitedPartners
            */
            assertion = (report.RPLimitedPartnersExists != false && report.RPLimitedPartnersIdentifierScheme != @"http://www.ato.gov.au/tfn");
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438029",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Entity identifier scheme for ReportingParty must be ""http://www.ato.gov.au/tfn""",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPLimitedPartnersOccurrenceIndex) + "/xbrli:entity/xbrli:identifier/@scheme",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438029" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438029

            #region VR.ATO.GEN.438029

            /*  VR.ATO.GEN.438029
            Entity identifier scheme for ReportingParty must be "http://www.ato.gov.au/tfn"
    
            Legacy Rule Format:
            IF (RprtPyType.xx.xx:ReportPartyTypeDimension = "RprtPyType.xx.xx:ReportingParty") AND (entity.identifier.scheme <> "http://www.ato.gov.au/tfn")
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^Context <> NULL) AND (^IdentifierScheme <> 'http://www.ato.gov.au/tfn')
    
            Data Elements:
    
            ^IdentifierScheme = RP.LimitedPartners:RP.LimitedPartners.YPrevious
    
            ^Context = RP.LimitedPartners:RP.LimitedPartners.YPrevious
            */
            assertion = (report.RPLimitedPartnersYPreviousExists != false && report.RPLimitedPartnersYPreviousIdentifierScheme != @"http://www.ato.gov.au/tfn");
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438029",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Entity identifier scheme for ReportingParty must be ""http://www.ato.gov.au/tfn""",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPLimitedPartnersYPreviousOccurrenceIndex) + "/xbrli:entity/xbrli:identifier/@scheme",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438029" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438029

            #region VR.ATO.GEN.438029

            /*  VR.ATO.GEN.438029
            Entity identifier scheme for ReportingParty must be "http://www.ato.gov.au/tfn"
    
            Legacy Rule Format:
            IF (RprtPyType.xx.xx:ReportPartyTypeDimension = "RprtPyType.xx.xx:ReportingParty") AND (entity.identifier.scheme <> "http://www.ato.gov.au/tfn")
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^Context <> NULL) AND (^IdentifierScheme <> 'http://www.ato.gov.au/tfn')
    
            Data Elements:
    
            ^IdentifierScheme = RP:RP.TOFA
    
            ^Context = RP:RP.TOFA
            */
            assertion = (report.RPTOFAExists != false && report.RPTOFAIdentifierScheme != @"http://www.ato.gov.au/tfn");
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438029",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Entity identifier scheme for ReportingParty must be ""http://www.ato.gov.au/tfn""",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPTOFAOccurrenceIndex) + "/xbrli:entity/xbrli:identifier/@scheme",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438029" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438029

            #region VR.ATO.GEN.438030

            /*  VR.ATO.GEN.438030
            Scheme for Intermediary must be set to http://www.ato.gov.au/abn
    
            Legacy Rule Format:
            IF (RprtPyType.xx.xx:ReportPartyTypeDimension = “RprtPyType.xx.xx:Intermediary”) AND (entity.identifier.scheme <> “http://www.ato.gov.au/abn”)
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^Context <> NULL AND ^IdentifierScheme <> 'http://www.ato.gov.au/abn'
    
            Data Elements:
    
            ^IdentifierScheme = INT
    
            ^Context = INT
            */
            assertion = (report.INTExists != false && report.INTIdentifierScheme != @"http://www.ato.gov.au/abn");
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438030",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Entity identifier scheme for Intermediary must be ""http://www.ato.gov.au/abn""",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.INTOccurrenceIndex) + "/xbrli:entity/xbrli:identifier/@scheme",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438030" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438030

            #region VR.ATO.GEN.438034

            /*  VR.ATO.GEN.438034
            Both taxable income and tax losses carried forward to later income years are present.
    
            Legacy Rule Format:
            IF bafpr1.xx.xx:Income.Taxable.Amount > 0 AND
            rvctc3.xx.xx:Tax.Losses.CarriedForward.LaterIncomeYearsTotal.Amount > 0
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR65 > 0 AND ^FITR76 > 0
    
            Data Elements:
    
            ^FITR76 = RP.Closing:Tax.Losses.CarriedForward.LaterIncomeYearsTotal.Amount
    
            ^FITR65 = RP:Income.Taxable.Amount
            */
            assertion = (report.FITR65.GetValueOrDefault() > 0 && report.FITR76.GetValueOrDefault() > 0);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438034",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Both taxable income and tax losses carried forward to later income years are present.",
                    LongDescription = @"Tax losses carried forward to later income years cannot be present with a positive taxable income.",
                    Location = "/xbrli:xbrl/tns:Tax.Losses.CarriedForward.LaterIncomeYearsTotal.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438034" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR65", Value = GetValueOrEmpty(report.FITR65) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR76", Value = GetValueOrEmpty(report.FITR76) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438034

            #region VR.ATO.GEN.438035

            /*  VR.ATO.GEN.438035
            Unstructured name must not  contain invalid characters ~ ^ < > + , * 
            Invalid characters ~^<>+  have been removed from the techincal business rule as they will be picked by the XBRL validator.
    
            Legacy Rule Format:
            IF (pyde.xx.xx:PersonUnstructuredName.FullName.Text <> NULLORBLANK AND pyde.xx.xx:PersonUnstructuredName.FullName.Text CONTAINS SET(",","*"))
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ContainsSet(^FITR169, '",","*"')
    
            Data Elements:
    
            ^FITR169 = INT:Declaration:PersonUnstructuredName:PersonUnstructuredName.FullName.Text WHERE [PersonUnstructuredName.Usage.Code='DeclarationSignatory']
            */
            assertion = IsMatch(report.FITR169, @",|\*", RegexOptions.IgnoreCase);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438035",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Unstructured name must not contain invalid characters ~ ^ < > + , *",
                    Location = "/xbrli:xbrl/tns:Declaration" + OccurrenceIndex(report.DeclarationCollection_TrueAndCorrect_INTOccurrenceIndex) + "/tns:PersonUnstructuredName" + OccurrenceIndex(report.Declaration_PersonUnstructuredNameCollection_DeclarationSignatory_INTOccurrenceIndex) + "/tns:PersonUnstructuredName.FullName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438035" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR169", Value = report.FITR169 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438035

            #region VR.ATO.GEN.438035

            /*  VR.ATO.GEN.438035
            Unstructured name must not  contain invalid characters ~ ^ < > + , * 
            Invalid characters ~^<>+  have been removed from the techincal business rule as they will be picked by the XBRL validator.
    
            Legacy Rule Format:
            IF (pyde.xx.xx:PersonUnstructuredName.FullName.Text <> NULLORBLANK AND pyde.xx.xx:PersonUnstructuredName.FullName.Text CONTAINS SET(",","*"))
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ContainsSet(^FITR167, '",","*"')
    
            Data Elements:
    
            ^FITR167 = RP:Declaration:PersonUnstructuredName:PersonUnstructuredName.FullName.Text WHERE [PersonUnstructuredName.Usage.Code='DeclarationSignatory']
            */
            assertion = IsMatch(report.FITR167, @",|\*", RegexOptions.IgnoreCase);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438035",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Unstructured name must not contain invalid characters ~ ^ < > + , *",
                    Location = "/xbrli:xbrl/tns:Declaration" + OccurrenceIndex(report.DeclarationCollection_TrueAndCorrect_RPOccurrenceIndex) + "/tns:PersonUnstructuredName" + OccurrenceIndex(report.Declaration_PersonUnstructuredNameCollection_DeclarationSignatory_RPOccurrenceIndex) + "/tns:PersonUnstructuredName.FullName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438035" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR167", Value = report.FITR167 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438035

            #region VR.ATO.GEN.438036

            /*  VR.ATO.GEN.438036
            Tax on taxable income must be supplied
    
            Legacy Rule Format:
            IF (RP:rvctc1.xx.xx:IncomeTax.TotalTaxOnTaxableIncome.Amount) = NULL
             RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR182 = NULL
    
            Data Elements:
    
            ^FITR182 = RP:IncomeTax.TotalTaxOnTaxableIncome.Amount
            */
            assertion = (report.FITR182 == null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438036",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Tax on taxable income must be supplied. A value must be included in the field, even if the value is zero.",
                    Location = "/xbrli:xbrl/tns:IncomeTax.TotalTaxOnTaxableIncome.Amount",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438036" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR182", Value = GetValueOrEmpty(report.FITR182) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438036

            #region VR.ATO.GEN.438052

            /*  VR.ATO.GEN.438052
            Date of closing balance cannot be later than the end date of Reporting period
    
            Legacy Rule Format:
            IF RP.Closing(Instant):period.instantDate <> RP:period.endDate
              RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^InstantDate1 <> ^EndDate
    
            Data Elements:
    
            ^EndDate = RP
    
            ^InstantDate1 = RP.Closing(Instant)
            */
            assertion = ((report.RPClosingInstantExists == true) && (report.RPClosingInstantInstantDate != report.RPEndDate));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438052",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Date of closing balance must match with the end date of Reporting period",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPOccurrenceIndex) + "/xbrli:period/xbrli:endDate",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438052" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438052

            #region VR.ATO.GEN.438053

            /*  VR.ATO.GEN.438053
            Period start date within context must match start date within reporting party context
    
            Legacy Rule Format:
            IF period.startDate  <> RP:period.startDate
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^Context <> NULL) AND (^StartDate  <> ^StartDate1)
    
            Data Elements:
    
            ^StartDate = INT
    
            ^StartDate1 = RP
    
            ^Context = INT
            */
            assertion = (report.INTExists != false && report.INTStartDate != report.RPStartDate);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438053",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Period start date must match Reporting Party period start date",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.INTOccurrenceIndex) + "/xbrli:period/xbrli:startDate",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438053" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438053

            #region VR.ATO.GEN.438053

            /*  VR.ATO.GEN.438053
            Period start date within context must match start date within reporting party context
    
            Legacy Rule Format:
            IF period.startDate  <> RP:period.startDate
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^Context <> NULL) AND (^StartDate  <> ^StartDate1)
    
            Data Elements:
    
            ^StartDate = RP:RP.CFCLC
    
            ^StartDate1 = RP
    
            ^Context = RP:RP.CFCLC
            */
            assertion = (report.RPCFCLCExists != false && report.RPCFCLCStartDate != report.RPStartDate);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438053",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Period start date must match Reporting Party period start date",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPCFCLCOccurrenceIndex) + "/xbrli:period/xbrli:startDate",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438053" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438053

            #region VR.ATO.GEN.438053

            /*  VR.ATO.GEN.438053
            Period start date within context must match start date within reporting party context
    
            Legacy Rule Format:
            IF period.startDate  <> RP:period.startDate
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^Context <> NULL) AND (^StartDate  <> ^StartDate1)
    
            Data Elements:
    
            ^StartDate = RP:RP.CFCUC
    
            ^StartDate1 = RP
    
            ^Context = RP:RP.CFCUC
            */
            assertion = (report.RPCFCUCExists != false && report.RPCFCUCStartDate != report.RPStartDate);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438053",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Period start date must match Reporting Party period start date",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPCFCUCOccurrenceIndex) + "/xbrli:period/xbrli:startDate",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438053" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438053

            #region VR.ATO.GEN.438053

            /*  VR.ATO.GEN.438053
            Period start date within context must match start date within reporting party context
    
            Legacy Rule Format:
            IF period.startDate  <> RP:period.startDate
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^Context <> NULL) AND (^StartDate  <> ^StartDate1)
    
            Data Elements:
    
            ^StartDate = RP:RP.Closing
    
            ^StartDate1 = RP
    
            ^Context = RP:RP.Closing
            */
            assertion = (report.RPClosingExists != false && report.RPClosingStartDate != report.RPStartDate);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438053",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Period start date must match Reporting Party period start date",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPClosingOccurrenceIndex) + "/xbrli:period/xbrli:startDate",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438053" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438053

            #region VR.ATO.GEN.438053

            /*  VR.ATO.GEN.438053
            Period start date within context must match start date within reporting party context
    
            Legacy Rule Format:
            IF period.startDate  <> RP:period.startDate
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^Context <> NULL) AND (^StartDate  <> ^StartDate1)
    
            Data Elements:
    
            ^StartDate = RP:RP.Investor
    
            ^StartDate1 = RP
    
            ^Context = RP:RP.Investor
            */
            assertion = (report.RPInvestorExists != false && report.RPInvestorStartDate != report.RPStartDate);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438053",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Period start date must match Reporting Party period start date",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPInvestorOccurrenceIndex) + "/xbrli:period/xbrli:startDate",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438053" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438053

            #region VR.ATO.GEN.438053

            /*  VR.ATO.GEN.438053
            Period start date within context must match start date within reporting party context
    
            Legacy Rule Format:
            IF period.startDate  <> RP:period.startDate
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^Context <> NULL) AND (^StartDate  <> ^StartDate1)
    
            Data Elements:
    
            ^StartDate = RP.Investor:RP.Investor.YPrevious
    
            ^StartDate1 = RP
    
            ^Context = RP.Investor:RP.Investor.YPrevious
            */
            assertion = (report.RPInvestorYPreviousExists != false && report.RPInvestorYPreviousStartDate != report.RPStartDate);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438053",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Period start date must match Reporting Party period start date",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPInvestorYPreviousOccurrenceIndex) + "/xbrli:period/xbrli:startDate",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438053" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438053

            #region VR.ATO.GEN.438053

            /*  VR.ATO.GEN.438053
            Period start date within context must match start date within reporting party context
    
            Legacy Rule Format:
            IF period.startDate  <> RP:period.startDate
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^Context <> NULL) AND (^StartDate  <> ^StartDate1)
    
            Data Elements:
    
            ^StartDate = RP:RP.JAUS
    
            ^StartDate1 = RP
    
            ^Context = RP:RP.JAUS
            */
            assertion = (report.RPJAUSExists != false && report.RPJAUSStartDate != report.RPStartDate);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438053",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Period start date must match Reporting Party period start date",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPJAUSOccurrenceIndex) + "/xbrli:period/xbrli:startDate",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438053" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438053

            #region VR.ATO.GEN.438053

            /*  VR.ATO.GEN.438053
            Period start date within context must match start date within reporting party context
    
            Legacy Rule Format:
            IF period.startDate  <> RP:period.startDate
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^Context <> NULL) AND (^StartDate  <> ^StartDate1)
    
            Data Elements:
    
            ^StartDate = RP:RP.JFOREIGN
    
            ^StartDate1 = RP
    
            ^Context = RP:RP.JFOREIGN
            */
            assertion = (report.RPJFOREIGNExists != false && report.RPJFOREIGNStartDate != report.RPStartDate);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438053",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Period start date must match Reporting Party period start date",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPJFOREIGNOccurrenceIndex) + "/xbrli:period/xbrli:startDate",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438053" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438053

            #region VR.ATO.GEN.438053

            /*  VR.ATO.GEN.438053
            Period start date within context must match start date within reporting party context
    
            Legacy Rule Format:
            IF period.startDate  <> RP:period.startDate
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^Context <> NULL) AND (^StartDate  <> ^StartDate1)
    
            Data Elements:
    
            ^StartDate = RP:RP.LimitedPartners
    
            ^StartDate1 = RP
    
            ^Context = RP:RP.LimitedPartners
            */
            assertion = (report.RPLimitedPartnersExists != false && report.RPLimitedPartnersStartDate != report.RPStartDate);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438053",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Period start date must match Reporting Party period start date",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPLimitedPartnersOccurrenceIndex) + "/xbrli:period/xbrli:startDate",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438053" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438053

            #region VR.ATO.GEN.438053

            /*  VR.ATO.GEN.438053
            Period start date within context must match start date within reporting party context
    
            Legacy Rule Format:
            IF period.startDate  <> RP:period.startDate
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^Context <> NULL) AND (^StartDate  <> ^StartDate1)
    
            Data Elements:
    
            ^StartDate = RP.LimitedPartners:RP.LimitedPartners.YPrevious
    
            ^StartDate1 = RP
    
            ^Context = RP.LimitedPartners:RP.LimitedPartners.YPrevious
            */
            assertion = (report.RPLimitedPartnersYPreviousExists != false && report.RPLimitedPartnersYPreviousStartDate != report.RPStartDate);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438053",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Period start date must match Reporting Party period start date",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPLimitedPartnersYPreviousOccurrenceIndex) + "/xbrli:period/xbrli:startDate",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438053" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438053

            #region VR.ATO.GEN.438053

            /*  VR.ATO.GEN.438053
            Period start date within context must match start date within reporting party context
    
            Legacy Rule Format:
            IF period.startDate  <> RP:period.startDate
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^Context <> NULL) AND (^StartDate  <> ^StartDate1)
    
            Data Elements:
    
            ^StartDate = RP:RP.TOFA
    
            ^StartDate1 = RP
    
            ^Context = RP:RP.TOFA
            */
            assertion = (report.RPTOFAExists != false && report.RPTOFAStartDate != report.RPStartDate);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438053",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Period start date must match Reporting Party period start date",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPTOFAOccurrenceIndex) + "/xbrli:period/xbrli:startDate",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438053" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438053

            #region VR.ATO.GEN.438054

            /*  VR.ATO.GEN.438054
            Period end date within context must match end date within reporting party context
    
            Legacy Rule Format:
            IF period.endDate <> RP:period.endDate
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^Context <> NULL) AND (^EndDate1 <> ^EndDate)
    
            Data Elements:
    
            ^Context = INT
    
            ^EndDate1 = INT
    
            ^EndDate = RP
            */
            assertion = (report.INTExists != false && report.INTEndDate != report.RPEndDate);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438054",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Period end date must match Reporting Party period end date",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.INTOccurrenceIndex) + "/xbrli:period/xbrli:endDate",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438054" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438054

            #region VR.ATO.GEN.438054

            /*  VR.ATO.GEN.438054
            Period end date within context must match end date within reporting party context
    
            Legacy Rule Format:
            IF period.endDate <> RP:period.endDate
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^Context <> NULL) AND (^EndDate1 <> ^EndDate)
    
            Data Elements:
    
            ^Context = RP:RP.CFCLC
    
            ^EndDate1 = RP:RP.CFCLC
    
            ^EndDate = RP
            */
            assertion = (report.RPCFCLCExists != false && report.RPCFCLCEndDate != report.RPEndDate);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438054",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Period end date must match Reporting Party period end date",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPCFCLCOccurrenceIndex) + "/xbrli:period/xbrli:endDate",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438054" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438054

            #region VR.ATO.GEN.438054

            /*  VR.ATO.GEN.438054
            Period end date within context must match end date within reporting party context
    
            Legacy Rule Format:
            IF period.endDate <> RP:period.endDate
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^Context <> NULL) AND (^EndDate1 <> ^EndDate)
    
            Data Elements:
    
            ^Context = RP:RP.CFCUC
    
            ^EndDate1 = RP:RP.CFCUC
    
            ^EndDate = RP
            */
            assertion = (report.RPCFCUCExists != false && report.RPCFCUCEndDate != report.RPEndDate);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438054",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Period end date must match Reporting Party period end date",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPCFCUCOccurrenceIndex) + "/xbrli:period/xbrli:endDate",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438054" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438054

            #region VR.ATO.GEN.438054

            /*  VR.ATO.GEN.438054
            Period end date within context must match end date within reporting party context
    
            Legacy Rule Format:
            IF period.endDate <> RP:period.endDate
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^Context <> NULL) AND (^EndDate1 <> ^EndDate)
    
            Data Elements:
    
            ^Context = RP:RP.Closing
    
            ^EndDate1 = RP:RP.Closing
    
            ^EndDate = RP
            */
            assertion = (report.RPClosingExists != false && report.RPClosingEndDate != report.RPEndDate);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438054",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Period end date must match Reporting Party period end date",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPClosingOccurrenceIndex) + "/xbrli:period/xbrli:endDate",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438054" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438054

            #region VR.ATO.GEN.438054

            /*  VR.ATO.GEN.438054
            Period end date within context must match end date within reporting party context
    
            Legacy Rule Format:
            IF period.endDate <> RP:period.endDate
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^Context <> NULL) AND (^EndDate1 <> ^EndDate)
    
            Data Elements:
    
            ^Context = RP:RP.Investor
    
            ^EndDate1 = RP:RP.Investor
    
            ^EndDate = RP
            */
            assertion = (report.RPInvestorExists != false && report.RPInvestorEndDate != report.RPEndDate);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438054",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Period end date must match Reporting Party period end date",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPInvestorOccurrenceIndex) + "/xbrli:period/xbrli:endDate",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438054" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438054

            #region VR.ATO.GEN.438054

            /*  VR.ATO.GEN.438054
            Period end date within context must match end date within reporting party context
    
            Legacy Rule Format:
            IF period.endDate <> RP:period.endDate
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^Context <> NULL) AND (^EndDate1 <> ^EndDate)
    
            Data Elements:
    
            ^Context = RP.Investor:RP.Investor.YPrevious
    
            ^EndDate1 = RP.Investor:RP.Investor.YPrevious
    
            ^EndDate = RP
            */
            assertion = (report.RPInvestorYPreviousExists != false && report.RPInvestorYPreviousEndDate != report.RPEndDate);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438054",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Period end date must match Reporting Party period end date",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPInvestorYPreviousOccurrenceIndex) + "/xbrli:period/xbrli:endDate",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438054" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438054

            #region VR.ATO.GEN.438054

            /*  VR.ATO.GEN.438054
            Period end date within context must match end date within reporting party context
    
            Legacy Rule Format:
            IF period.endDate <> RP:period.endDate
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^Context <> NULL) AND (^EndDate1 <> ^EndDate)
    
            Data Elements:
    
            ^Context = RP:RP.JAUS
    
            ^EndDate1 = RP:RP.JAUS
    
            ^EndDate = RP
            */
            assertion = (report.RPJAUSExists != false && report.RPJAUSEndDate != report.RPEndDate);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438054",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Period end date must match Reporting Party period end date",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPJAUSOccurrenceIndex) + "/xbrli:period/xbrli:endDate",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438054" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438054

            #region VR.ATO.GEN.438054

            /*  VR.ATO.GEN.438054
            Period end date within context must match end date within reporting party context
    
            Legacy Rule Format:
            IF period.endDate <> RP:period.endDate
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^Context <> NULL) AND (^EndDate1 <> ^EndDate)
    
            Data Elements:
    
            ^Context = RP:RP.JFOREIGN
    
            ^EndDate1 = RP:RP.JFOREIGN
    
            ^EndDate = RP
            */
            assertion = (report.RPJFOREIGNExists != false && report.RPJFOREIGNEndDate != report.RPEndDate);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438054",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Period end date must match Reporting Party period end date",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPJFOREIGNOccurrenceIndex) + "/xbrli:period/xbrli:endDate",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438054" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438054

            #region VR.ATO.GEN.438054

            /*  VR.ATO.GEN.438054
            Period end date within context must match end date within reporting party context
    
            Legacy Rule Format:
            IF period.endDate <> RP:period.endDate
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^Context <> NULL) AND (^EndDate1 <> ^EndDate)
    
            Data Elements:
    
            ^Context = RP:RP.LimitedPartners
    
            ^EndDate1 = RP:RP.LimitedPartners
    
            ^EndDate = RP
            */
            assertion = (report.RPLimitedPartnersExists != false && report.RPLimitedPartnersEndDate != report.RPEndDate);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438054",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Period end date must match Reporting Party period end date",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPLimitedPartnersOccurrenceIndex) + "/xbrli:period/xbrli:endDate",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438054" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438054

            #region VR.ATO.GEN.438054

            /*  VR.ATO.GEN.438054
            Period end date within context must match end date within reporting party context
    
            Legacy Rule Format:
            IF period.endDate <> RP:period.endDate
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^Context <> NULL) AND (^EndDate1 <> ^EndDate)
    
            Data Elements:
    
            ^Context = RP.LimitedPartners:RP.LimitedPartners.YPrevious
    
            ^EndDate1 = RP.LimitedPartners:RP.LimitedPartners.YPrevious
    
            ^EndDate = RP
            */
            assertion = (report.RPLimitedPartnersYPreviousExists != false && report.RPLimitedPartnersYPreviousEndDate != report.RPEndDate);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438054",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Period end date must match Reporting Party period end date",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPLimitedPartnersYPreviousOccurrenceIndex) + "/xbrli:period/xbrli:endDate",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438054" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438054

            #region VR.ATO.GEN.438054

            /*  VR.ATO.GEN.438054
            Period end date within context must match end date within reporting party context
    
            Legacy Rule Format:
            IF period.endDate <> RP:period.endDate
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^Context <> NULL) AND (^EndDate1 <> ^EndDate)
    
            Data Elements:
    
            ^Context = RP:RP.TOFA
    
            ^EndDate1 = RP:RP.TOFA
    
            ^EndDate = RP
            */
            assertion = (report.RPTOFAExists != false && report.RPTOFAEndDate != report.RPEndDate);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438054",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Period end date must match Reporting Party period end date",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPTOFAOccurrenceIndex) + "/xbrli:period/xbrli:endDate",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438054" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438054

            #region VR.ATO.GEN.438058

            /*  VR.ATO.GEN.438058
            Invalid context. The number of Reporting party: Investor contexts must not exceed 1.
    
            Legacy Rule Format:
            IF COUNT(CONTEXT(RP.Investor)) > 1 
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Count(^Context) > 1
    
            Data Elements:
    
            ^Context = RP:RP.Investor
            */
            assertion = (Count(report.RPInvestorCount) > 1);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438058",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Invalid context. The number of Reporting party: Investor contexts must not exceed 1.",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPInvestorLastOccurrenceIndex),
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438058" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438058

            #region VR.ATO.GEN.438059

            /*  VR.ATO.GEN.438059
            Invalid context. The number of Reporting party: Limited Partners contexts must not exceed 1.
    
            Legacy Rule Format:
            IF COUNT(CONTEXT(RP.LimitedPartners)) > 1 
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Count(^Context) > 1
    
            Data Elements:
    
            ^Context = RP:RP.LimitedPartners
            */
            assertion = (Count(report.RPLimitedPartnersCount) > 1);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438059",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Invalid context. The number of Reporting party: Limited Partners contexts must not exceed 1.",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPLimitedPartnersLastOccurrenceIndex),
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438059" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438059

            #region VR.ATO.GEN.438063

            /*  VR.ATO.GEN.438063
            If an Intermediary is present, Intermediary declaration statement accepted indicator must be supplied and must be “True”.
    
            Legacy Rule Format:
            IF INT<>NULL AND INT:pyin.xx.xx:Declaration.StatementAccepted.Indicator <> TRUE
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^INT <> NULL AND ^FITR165 <> TRUE
    
            Data Elements:
    
            ^FITR165 = INT:Declaration:Declaration.StatementAccepted.Indicator WHERE [Declaration.StatementType.Code='TrueAndCorrect']
    
            ^INT = INT
            */
            assertion = (report.INTExists != false && report.FITR165 != true);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438063",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Declaration statement accepted indicator is required.",
                    LongDescription = @"If an Intermediary is present, Intermediary declaration statement accepted indicator must be supplied and must be ""True"".",
                    Location = "/xbrli:xbrl/tns:Declaration" + OccurrenceIndex(report.DeclarationCollection_TrueAndCorrect_INTOccurrenceIndex) + "/tns:Declaration.StatementAccepted.Indicator",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438063" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR165", Value = GetValueOrEmpty(report.FITR165) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438063

            #region VR.ATO.GEN.438064

            /*  VR.ATO.GEN.438064
            If an Intermediary is not present, Reporting Party declaration statement accepted indicator must be supplied and must be “True”.
    
            Legacy Rule Format:
            IF INT=NULL AND RP:pyin.xx.xx:Declaration.StatementAccepted.Indicator <> TRUE
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^INT = NULL AND ^FITR162 <> TRUE
    
            Data Elements:
    
            ^FITR162 = RP:Declaration:Declaration.StatementAccepted.Indicator WHERE [Declaration.StatementType.Code='TrueAndCorrect']
    
            ^INT = INT
            */
            assertion = (report.INTExists == false && report.FITR162 != true);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438064",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Declaration statement accepted indicator is required.",
                    LongDescription = @"If an Intermediary is not present, Reporting Party declaration statement accepted indicator must be supplied and must be ""True"".",
                    Location = "/xbrli:xbrl/tns:Declaration" + OccurrenceIndex(report.DeclarationCollection_TrueAndCorrect_RPOccurrenceIndex) + "/tns:Declaration.StatementAccepted.Indicator",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438064" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR162", Value = GetValueOrEmpty(report.FITR162) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438064

            #region VR.ATO.GEN.438065

            /*  VR.ATO.GEN.438065
            If an Intermediary is present, Intermediary declaration signatory identifier must be supplied.
    
            Legacy Rule Format:
            IF INT<>NULL AND INT:pyin.xx.xx:Declaration.SignatoryIdentifier.Text = NULLORBLANK
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^INT <> NULL AND ^FITR163 = BLANK
    
            Data Elements:
    
            ^FITR163 = INT:Declaration:Declaration.SignatoryIdentifier.Text WHERE [Declaration.StatementType.Code='TrueAndCorrect']
    
            ^INT = INT
            */
            assertion = (report.INTExists != false && string.IsNullOrWhiteSpace(report.FITR163) == true);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438065",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Declaration signatory identifier is required.",
                    LongDescription = @"If an Intermediary is present, Intermediary declaration signatory identifier must be supplied.",
                    Location = "/xbrli:xbrl/tns:Declaration" + OccurrenceIndex(report.DeclarationCollection_TrueAndCorrect_INTOccurrenceIndex) + "/tns:Declaration.SignatoryIdentifier.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438065" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR163", Value = report.FITR163 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438065

            #region VR.ATO.GEN.438066

            /*  VR.ATO.GEN.438066
            If an Intermediary is not present, Reporting Party declaration signatory identifier must be supplied.
    
            Legacy Rule Format:
            IF INT=NULL AND RP:pyin.xx.xx:Declaration.SignatoryIdentifier.Text = NULLORBLANK
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^INT = NULL AND ^FITR160 = BLANK
    
            Data Elements:
    
            ^FITR160 = RP:Declaration:Declaration.SignatoryIdentifier.Text WHERE [Declaration.StatementType.Code='TrueAndCorrect']
    
            ^INT = INT
            */
            assertion = (report.INTExists == false && string.IsNullOrWhiteSpace(report.FITR160) == true);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438066",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Declaration signatory identifier is required.",
                    LongDescription = @"If an Intermediary is not present, Reporting Party declaration signatory identifier must be supplied.",
                    Location = "/xbrli:xbrl/tns:Declaration" + OccurrenceIndex(report.DeclarationCollection_TrueAndCorrect_RPOccurrenceIndex) + "/tns:Declaration.SignatoryIdentifier.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438066" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR160", Value = report.FITR160 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438066

            #region VR.ATO.GEN.438067

            /*  VR.ATO.GEN.438067
            If an Intermediary is present, Intermediary declaration signature date must be supplied.
    
            Legacy Rule Format:
            IF INT<>NULL AND INT:pyin.xx.xx:Declaration.Signature.Date = NULL
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^INT <> NULL AND ^FITR125 = NULL
    
            Data Elements:
    
            ^FITR125 = INT:Declaration:Declaration.Signature.Date WHERE [Declaration.StatementType.Code='TrueAndCorrect']
    
            ^INT = INT
            */
            assertion = (report.INTExists != false && report.FITR125 == null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438067",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Declaration signature date is required.",
                    LongDescription = @"If an Intermediary is present, Intermediary declaration signature date must be supplied.",
                    Location = "/xbrli:xbrl/tns:Declaration" + OccurrenceIndex(report.DeclarationCollection_TrueAndCorrect_INTOccurrenceIndex) + "/tns:Declaration.Signature.Date",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438067" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR125", Value = GetValueOrEmptyDate(report.FITR125) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438067

            #region VR.ATO.GEN.438068

            /*  VR.ATO.GEN.438068
            If an Intermediary is not present, Reporting Party declaration signature date must be supplied.
    
            Legacy Rule Format:
            IF INT=NULL AND RP:pyin.xx.xx:Declaration.Signature.Date = NULL
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^INT = NULL AND ^FITR123 = NULL
    
            Data Elements:
    
            ^FITR123 = RP:Declaration:Declaration.Signature.Date WHERE [Declaration.StatementType.Code='TrueAndCorrect']
    
            ^INT = INT
            */
            assertion = (report.INTExists == false && report.FITR123 == null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438068",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Declaration signature date is required.",
                    LongDescription = @"If an Intermediary is not present, Reporting Party declaration signature date must be supplied.",
                    Location = "/xbrli:xbrl/tns:Declaration" + OccurrenceIndex(report.DeclarationCollection_TrueAndCorrect_RPOccurrenceIndex) + "/tns:Declaration.Signature.Date",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438068" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR123", Value = GetValueOrEmptyDate(report.FITR123) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438068

            #region VR.ATO.GEN.438081

            /*  VR.ATO.GEN.438081
            Invalid context. The number of Reporting party:Investor:YPrevious contexts must not exceed 1.
    
            Legacy Rule Format:
            IF COUNT(CONTEXT(RP.Investor.YPrevious)) > 1 
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Count(^Context) > 1
    
            Data Elements:
    
            ^Context = RP.Investor:RP.Investor.YPrevious
            */
            assertion = (Count(report.RPInvestorYPreviousCount) > 1);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438081",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Invalid context. The number of Reporting party:Investor:YPrevious contexts must not exceed 1.",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPInvestorYPreviousLastOccurrenceIndex),
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438081" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438081

            #region VR.ATO.GEN.438082

            /*  VR.ATO.GEN.438082
            Invalid context. The number of Reporting party:Limited Partners:YPrevious contexts must not exceed 1.
    
            Legacy Rule Format:
            IF COUNT(CONTEXT(RP.LimitedPartners.YPrevious)) > 1 
            RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            Count(^Context) > 1
    
            Data Elements:
    
            ^Context = RP.LimitedPartners:RP.LimitedPartners.YPrevious
            */
            assertion = (Count(report.RPLimitedPartnersYPreviousCount) > 1);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.438082",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Invalid context. The number of Reporting party:Limited Partners:YPrevious contexts must not exceed 1.",
                    Location = "/xbrli:xbrl/xbrli:context" + OccurrenceIndex(report.RPLimitedPartnersYPreviousLastOccurrenceIndex),
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.438082" } },
                };
                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.438082

            #region VR.ATO.GEN.500130

            /*  VR.ATO.GEN.500130
            Person unstructured name must be provided.
    
            Legacy Rule Format:
            WHERE IN TUPLE(personunstructuredname1.xx.xx:PersonUnstructuredName)
            IF pyde.xx.xx:PersonUnstructuredName.FullName.Text = BLANK
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^PersonUnstructuredName <> NULL) AND (^FITR169 = BLANK)
    
            Data Elements:
    
            ^FITR169 = INT:Declaration:PersonUnstructuredName:PersonUnstructuredName.FullName.Text WHERE [PersonUnstructuredName.Usage.Code='DeclarationSignatory']
    
            ^PersonUnstructuredName = INT:Declaration:PersonUnstructuredName
            */
            assertion = (report.Declaration_PersonUnstructuredNameCollection_DeclarationSignatory_INTExists != false && string.IsNullOrWhiteSpace(report.FITR169) == true);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.500130",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Person unstructured name must be provided.",
                    Location = "/xbrli:xbrl/tns:Declaration" + OccurrenceIndex(report.DeclarationCollection_TrueAndCorrect_INTOccurrenceIndex) + "/tns:PersonUnstructuredName" + OccurrenceIndex(report.Declaration_PersonUnstructuredNameCollection_DeclarationSignatory_INTOccurrenceIndex) + "/tns:PersonUnstructuredName.FullName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.500130" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR169", Value = report.FITR169 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.500130

            #region VR.ATO.GEN.500130

            /*  VR.ATO.GEN.500130
            Person unstructured name must be provided.
    
            Legacy Rule Format:
            WHERE IN TUPLE(personunstructuredname1.xx.xx:PersonUnstructuredName)
            IF pyde.xx.xx:PersonUnstructuredName.FullName.Text = BLANK
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^PersonUnstructuredName <> NULL) AND (^FITR167 = BLANK)
    
            Data Elements:
    
            ^FITR167 = RP:Declaration:PersonUnstructuredName:PersonUnstructuredName.FullName.Text WHERE [PersonUnstructuredName.Usage.Code='DeclarationSignatory']
    
            ^PersonUnstructuredName = RP:Declaration:PersonUnstructuredName
            */
            assertion = (report.Declaration_PersonUnstructuredNameCollection_DeclarationSignatory_RPExists != false && string.IsNullOrWhiteSpace(report.FITR167) == true);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.500130",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Person unstructured name must be provided.",
                    Location = "/xbrli:xbrl/tns:Declaration" + OccurrenceIndex(report.DeclarationCollection_TrueAndCorrect_RPOccurrenceIndex) + "/tns:PersonUnstructuredName" + OccurrenceIndex(report.Declaration_PersonUnstructuredNameCollection_DeclarationSignatory_RPOccurrenceIndex) + "/tns:PersonUnstructuredName.FullName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.500130" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR167", Value = report.FITR167 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.500130

            #region VR.ATO.GEN.500132

            /*  VR.ATO.GEN.500132
            Organisation name must be present
    
            Legacy Rule Format:
            WHERE IN TUPLE(organisationname2.xx.xx:OrganisationNameDetails)
            IF pyde.xx.xx:OrganisationNameDetails.OrganisationalName.Text = BLANK
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^OrganisationNameDetails <> NULL) AND (^FITR116 = BLANK)
    
            Data Elements:
    
            ^FITR116 = INT:OrganisationNameDetails:OrganisationNameDetails.OrganisationalName.Text WHERE [OrganisationNameDetails.OrganisationalNameType.Code='MN'].[OrganisationNameDetails.Currency.Code='C']
    
            ^OrganisationNameDetails = INT:OrganisationNameDetails
            */
            assertion = (report.OrganisationNameDetailsCollection_MN_C_INTExists != false && string.IsNullOrWhiteSpace(report.FITR116) == true);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.500132",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Organisation name must be present.",
                    Location = "/xbrli:xbrl/tns:OrganisationNameDetails" + OccurrenceIndex(report.OrganisationNameDetailsCollection_MN_C_INTOccurrenceIndex) + "/tns:OrganisationNameDetails.OrganisationalName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.500132" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR116", Value = report.FITR116 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.500132

            #region VR.ATO.GEN.500132

            /*  VR.ATO.GEN.500132
            Organisation name must be present
    
            Legacy Rule Format:
            WHERE IN TUPLE(organisationname2.xx.xx:OrganisationNameDetails)
            IF pyde.xx.xx:OrganisationNameDetails.OrganisationalName.Text = BLANK
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^OrganisationNameDetails <> NULL) AND (^FITR12 = BLANK)
    
            Data Elements:
    
            ^FITR12 = RP:Trustee:OrganisationNameDetails:OrganisationNameDetails.OrganisationalName.Text WHERE [OrganisationNameDetails.OrganisationalNameType.Code='MN'].[OrganisationNameDetails.Currency.Code='C']
    
            ^OrganisationNameDetails = RP:Trustee:OrganisationNameDetails
            */
            assertion = (report.Trustee_OrganisationNameDetailsCollection_MN_C_RPExists != false && string.IsNullOrWhiteSpace(report.FITR12) == true);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.500132",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Organisation name must be present.",
                    Location = "/xbrli:xbrl/tns:Trustee/tns:OrganisationNameDetails" + OccurrenceIndex(report.Trustee_OrganisationNameDetailsCollection_MN_C_RPOccurrenceIndex) + "/tns:OrganisationNameDetails.OrganisationalName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.500132" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR12", Value = report.FITR12 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.500132

            #region VR.ATO.GEN.500132

            /*  VR.ATO.GEN.500132
            Organisation name must be present
    
            Legacy Rule Format:
            WHERE IN TUPLE(organisationname2.xx.xx:OrganisationNameDetails)
            IF pyde.xx.xx:OrganisationNameDetails.OrganisationalName.Text = BLANK
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^OrganisationNameDetails <> NULL) AND (^FITR6 = BLANK)
    
            Data Elements:
    
            ^FITR6 = RP:OrganisationNameDetails:OrganisationNameDetails.OrganisationalName.Text WHERE [OrganisationNameDetails.OrganisationalNameType.Code='MN'].[OrganisationNameDetails.Currency.Code='C']
    
            ^OrganisationNameDetails = RP:OrganisationNameDetails
            */
            assertion = (report.OrganisationNameDetailsCollection_MN_C_RPExists != false && string.IsNullOrWhiteSpace(report.FITR6) == true);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.500132",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Organisation name must be present.",
                    Location = "/xbrli:xbrl/tns:OrganisationNameDetails" + OccurrenceIndex(report.OrganisationNameDetailsCollection_MN_C_RPOccurrenceIndex) + "/tns:OrganisationNameDetails.OrganisationalName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.500132" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR6", Value = report.FITR6 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.500132

            #region VR.ATO.GEN.500133

            /*  VR.ATO.GEN.500133
            Address Line 1 must be present
    
            Legacy Rule Format:
            WHERE IN TUPLE(addressdetails2.xx.xx:AddressDetails)
            IF pyde.xx.xx:AddressDetails.Line1.Text = BLANK
              RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^AddressDetails <> NULL) AND (^FITR8 = BLANK)
    
            Data Elements:
    
            ^FITR8 = RP:AddressDetails:AddressDetails.Line1.Text WHERE [AddressDetails.Usage.Code='POS'].[AddressDetails.Currency.Code='C']
    
            ^AddressDetails = RP:AddressDetails
            */
            assertion = (report.AddressDetailsCollection_POS_C_RPExists != false && string.IsNullOrWhiteSpace(report.FITR8) == true);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.500133",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Address Line 1 must be present.",
                    Location = "/xbrli:xbrl/tns:AddressDetails" + OccurrenceIndex(report.AddressDetailsCollection_POS_C_RPOccurrenceIndex) + "/tns:AddressDetails.Line1.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.500133" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR8", Value = report.FITR8 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.500133

            #region VR.ATO.GEN.500134

            /*  VR.ATO.GEN.500134
            Suburb/Town/City must be present in an address
    
            Legacy Rule Format:
            WHERE IN TUPLE(addressdetails2.xx.xx:AddressDetails)
            IF pyde.xx.xx:AddressDetails.LocalityName.Text = BLANK
              RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^AddressDetails <> NULL) AND (^FITR9 = BLANK)
    
            Data Elements:
    
            ^FITR9 = RP:AddressDetails:AddressDetails.LocalityName.Text WHERE [AddressDetails.Usage.Code='POS'].[AddressDetails.Currency.Code='C']
    
            ^AddressDetails = RP:AddressDetails
            */
            assertion = (report.AddressDetailsCollection_POS_C_RPExists != false && string.IsNullOrWhiteSpace(report.FITR9) == true);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.500134",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Suburb/Town/City must be present in an address.",
                    Location = "/xbrli:xbrl/tns:AddressDetails" + OccurrenceIndex(report.AddressDetailsCollection_POS_C_RPOccurrenceIndex) + "/tns:AddressDetails.LocalityName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.500134" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR9", Value = report.FITR9 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.500134

            #region VR.ATO.GEN.500135

            /*  VR.ATO.GEN.500135
            Family Name must be present
    
            Legacy Rule Format:
            WHERE IN TUPLE(personstructuredname3.xx.xx:PersonNameDetails)
            IF pyde.xx.xx:PersonNameDetails.FamilyName.Text = BLANK
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^PersonNameDetails <> NULL) AND (^FITR112 = BLANK)
    
            Data Elements:
    
            ^FITR112 = INT:PersonNameDetails:PersonNameDetails.FamilyName.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
    
            ^PersonNameDetails = INT:PersonNameDetails
            */
            assertion = (report.PersonNameDetailsCollection_LGL_Contact_C_INTExists != false && string.IsNullOrWhiteSpace(report.FITR112) == true);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.500135",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Family Name must be present.",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_INTOccurrenceIndex) + "/tns:PersonNameDetails.FamilyName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.500135" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR112", Value = report.FITR112 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.500135

            #region VR.ATO.GEN.500135

            /*  VR.ATO.GEN.500135
            Family Name must be present
    
            Legacy Rule Format:
            WHERE IN TUPLE(personstructuredname3.xx.xx:PersonNameDetails)
            IF pyde.xx.xx:PersonNameDetails.FamilyName.Text = BLANK
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^PersonNameDetails <> NULL) AND (^FITR103 = BLANK)
    
            Data Elements:
    
            ^FITR103 = RP:PersonNameDetails:PersonNameDetails.FamilyName.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
    
            ^PersonNameDetails = RP:PersonNameDetails
            */
            assertion = (report.PersonNameDetailsCollection_LGL_Contact_C_RPExists != false && string.IsNullOrWhiteSpace(report.FITR103) == true);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.500135",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Family Name must be present.",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_RPOccurrenceIndex) + "/tns:PersonNameDetails.FamilyName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.500135" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR103", Value = report.FITR103 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.500135

            #region VR.ATO.GEN.500136

            /*  VR.ATO.GEN.500136
            BSB Number must be greater than 012000.
    
            Legacy Rule Format:
            WHERE IN TUPLE(financialinstitutionaccount1.xx.xx:FinancialInstitutionAccount) 
            IF pyid.xx.xx:FinancialInstitutionAccount.BankStateBranch.Number <> SET(012001- 999999)
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^FinancialInstitutionAccount <> NULL) AND (^FITR176 <> BLANK AND (IsNumeric(^FITR176) = 'FALSE' OR AsNumeric(^FITR176) < 12001 OR AsNumeric(^FITR176) > 999999 OR Length(^FITR176) <> 6))
    
            Data Elements:
    
            ^FITR176 = RP:FinancialInstitutionAccount:FinancialInstitutionAccount.BankStateBranch.Number
    
            ^FinancialInstitutionAccount = RP:FinancialInstitutionAccount
            */
            assertion = (report.FinancialInstitutionAccountCollectionExists != false && (string.IsNullOrWhiteSpace(report.FITR176) != true && (IsNumeric(report.FITR176) == false || AsNumeric(report.FITR176) < 12001 || AsNumeric(report.FITR176) > 999999 || Length(report.FITR176) != 6)));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.500136",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"BSB Number must be greater than 012000.",
                    Location = "/xbrli:xbrl/tns:FinancialInstitutionAccount/tns:FinancialInstitutionAccount.BankStateBranch.Number",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.500136" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR176", Value = report.FITR176 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.500136

            #region VR.ATO.GEN.500137

            /*  VR.ATO.GEN.500137
            Account Number and Account Name must be present.
    
            Legacy Rule Format:
            WHERE IN TUPLE(financialinstitutionaccount1.xx.xx:FinancialInstitutionAccount)
            IF pyid.xx.xx:FinancialInstitutionAccount.FinancialInstitutionAccount.Number = BLANK OR pyid.xx.xx:FinancialInstitutionAccount.FinancialInstitutionAccountName.Text = BLANK
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^FinancialInstitutionAccount <> NULL) AND (^FITR177 = BLANK OR ^FITR178 = BLANK)
    
            Data Elements:
    
            ^FITR177 = RP:FinancialInstitutionAccount:FinancialInstitutionAccount.FinancialInstitutionAccount.Number
    
            ^FITR178 = RP:FinancialInstitutionAccount:FinancialInstitutionAccount.FinancialInstitutionAccountName.Text
    
            ^FinancialInstitutionAccount = RP:FinancialInstitutionAccount
            */
            assertion = (report.FinancialInstitutionAccountCollectionExists != false && (string.IsNullOrWhiteSpace(report.FITR177) == true || string.IsNullOrWhiteSpace(report.FITR178) == true));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.500137",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Account Number and Account Name must be present.",
                    Location = "/xbrli:xbrl/tns:FinancialInstitutionAccount/tns:FinancialInstitutionAccount.FinancialInstitutionAccount.Number",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.500137" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR177", Value = report.FITR177 });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR178", Value = report.FITR178 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.500137

            #region VR.ATO.GEN.500138

            /*  VR.ATO.GEN.500138
            An individual's first name must begin with an alpha character in the range 'A to Z' or 'a to z'.
    
            Legacy Rule Format:
            WHERE IN TUPLE(personstructuredname3.xx.xx:PersonNameDetails)
            IF pyde.xx.xx:PersonNameDetails.GivenName.Text DOES NOT STARTWITH SET("A-Z","a-z")
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^PersonNameDetails <> NULL) AND (NOT StartsWithSet(^FITR114,'"A-Z","a-z"'))
    
            Data Elements:
    
            ^FITR114 = INT:PersonNameDetails:PersonNameDetails.GivenName.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
    
            ^PersonNameDetails = INT:PersonNameDetails
            */
            assertion = (report.PersonNameDetailsCollection_LGL_Contact_C_INTExists != false && !(IsMatch(report.FITR114, @"^([A-Z]|[a-z])", RegexOptions.IgnoreCase)));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.500138",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"An individual's first name must begin with an alpha character in the range 'A to Z' or 'a to z'.",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_INTOccurrenceIndex) + "/tns:PersonNameDetails.GivenName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.500138" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR114", Value = report.FITR114 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.500138

            #region VR.ATO.GEN.500139

            /*  VR.ATO.GEN.500139
            First name must be present.
    
            Legacy Rule Format:
            WHERE IN TUPLE(personstructuredname3.xx.xx:PersonNameDetails)
            IF pyde.xx.xx:PersonNameDetails.GivenName.Text = BLANK
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^PersonNameDetails <> NULL) AND (^FITR114 = BLANK)
    
            Data Elements:
    
            ^FITR114 = INT:PersonNameDetails:PersonNameDetails.GivenName.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
    
            ^PersonNameDetails = INT:PersonNameDetails
            */
            assertion = (report.PersonNameDetailsCollection_LGL_Contact_C_INTExists != false && string.IsNullOrWhiteSpace(report.FITR114) == true);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.500139",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"First name must be present.",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_INTOccurrenceIndex) + "/tns:PersonNameDetails.GivenName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.500139" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR114", Value = report.FITR114 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.500139

            #region VR.ATO.GEN.500139

            /*  VR.ATO.GEN.500139
            First name must be present.
    
            Legacy Rule Format:
            WHERE IN TUPLE(personstructuredname3.xx.xx:PersonNameDetails)
            IF pyde.xx.xx:PersonNameDetails.GivenName.Text = BLANK
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^PersonNameDetails <> NULL) AND (^FITR105 = BLANK)
    
            Data Elements:
    
            ^FITR105 = RP:PersonNameDetails:PersonNameDetails.GivenName.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
    
            ^PersonNameDetails = RP:PersonNameDetails
            */
            assertion = (report.PersonNameDetailsCollection_LGL_Contact_C_RPExists != false && string.IsNullOrWhiteSpace(report.FITR105) == true);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.500139",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"First name must be present.",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_RPOccurrenceIndex) + "/tns:PersonNameDetails.GivenName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.500139" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR105", Value = report.FITR105 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.500139

            #region VR.ATO.GEN.500141

            /*  VR.ATO.GEN.500141
            For an Australian address, state or territory must be one of: ACT, NSW, NT, QLD, SA, VIC, WA, TAS
    
            Legacy Rule Format:
            IF (pyde.xx.xx:AddressDetails.StateOrTerritory.Code <> NULLORBLANK)
            AND (pyde.xx.xx:AddressDetails.StateOrTerritory.Code <> SET ("ACT","NSW","NT","QLD","SA","VIC","WA","TAS"))
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^FITR10 <> BLANK) AND (NotInSet(^FITR10, '"ACT","NSW","NT","QLD","SA","VIC","WA","TAS"'))
    
            Data Elements:
    
            ^FITR10 = RP:AddressDetails:AddressDetails.StateOrTerritory.Code WHERE [AddressDetails.Usage.Code='POS'].[AddressDetails.Currency.Code='C']
            */
            assertion = (string.IsNullOrWhiteSpace(report.FITR10) != true && !(IsMatch(report.FITR10, @"^(ACT|NSW|NT|QLD|SA|VIC|WA|TAS)$", RegexOptions.IgnoreCase)));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.500141",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"For an Australian address, state or territory must be one of: ACT, NSW, NT, QLD, SA, VIC, WA, TAS",
                    Location = "/xbrli:xbrl/tns:AddressDetails" + OccurrenceIndex(report.AddressDetailsCollection_POS_C_RPOccurrenceIndex) + "/tns:AddressDetails.StateOrTerritory.Code",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.500141" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR10", Value = report.FITR10 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.500141

            #region VR.ATO.GEN.500142

            /*  VR.ATO.GEN.500142
            Family name must contain at least one alphabetical character
    
            Legacy Rule Format:
            WHERE IN TUPLE(personstructuredname3.xx.xx:PersonNameDetails)
            IF pyde.xx.xx:PersonNameDetails.FamilyName.Text DOES NOT CONTAIN SET("A-Z","a-z")
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^PersonNameDetails <> NULL) AND (NotContainsSet(^FITR112, '"A-Z","a-z"'))
    
            Data Elements:
    
            ^FITR112 = INT:PersonNameDetails:PersonNameDetails.FamilyName.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
    
            ^PersonNameDetails = INT:PersonNameDetails
            */
            assertion = (report.PersonNameDetailsCollection_LGL_Contact_C_INTExists != false && !(IsMatch(report.FITR112, @"[A-Z]|[a-z]", RegexOptions.IgnoreCase)));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.500142",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Family name must contain at least one alphabetical character.",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_INTOccurrenceIndex) + "/tns:PersonNameDetails.FamilyName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.500142" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR112", Value = report.FITR112 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.500142

            #region VR.ATO.GEN.500142

            /*  VR.ATO.GEN.500142
            Family name must contain at least one alphabetical character
    
            Legacy Rule Format:
            WHERE IN TUPLE(personstructuredname3.xx.xx:PersonNameDetails)
            IF pyde.xx.xx:PersonNameDetails.FamilyName.Text DOES NOT CONTAIN SET("A-Z","a-z")
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^PersonNameDetails <> NULL) AND (NotContainsSet(^FITR103, '"A-Z","a-z"'))
    
            Data Elements:
    
            ^FITR103 = RP:PersonNameDetails:PersonNameDetails.FamilyName.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
    
            ^PersonNameDetails = RP:PersonNameDetails
            */
            assertion = (report.PersonNameDetailsCollection_LGL_Contact_C_RPExists != false && !(IsMatch(report.FITR103, @"[A-Z]|[a-z]", RegexOptions.IgnoreCase)));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.500142",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Family name must contain at least one alphabetical character.",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_RPOccurrenceIndex) + "/tns:PersonNameDetails.FamilyName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.500142" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR103", Value = report.FITR103 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.500142

            #region VR.ATO.GEN.500143

            /*  VR.ATO.GEN.500143
            Given Name must contain at least one alphabetical character
    
            Legacy Rule Format:
            WHERE IN TUPLE(personstructuredname3.xx.xx:PersonNameDetails)
            IF pyde.xx.xx:PersonNameDetails.GivenName.Text DOES NOT CONTAIN SET("A-Z","a-z")
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^PersonNameDetails <> NULL) AND (NotContainsSet(^FITR114, '"A-Z","a-z"'))
    
            Data Elements:
    
            ^FITR114 = INT:PersonNameDetails:PersonNameDetails.GivenName.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
    
            ^PersonNameDetails = INT:PersonNameDetails
            */
            assertion = (report.PersonNameDetailsCollection_LGL_Contact_C_INTExists != false && !(IsMatch(report.FITR114, @"[A-Z]|[a-z]", RegexOptions.IgnoreCase)));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.500143",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Given Name must contain at least one alphabetical character.",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_INTOccurrenceIndex) + "/tns:PersonNameDetails.GivenName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.500143" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR114", Value = report.FITR114 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.500143

            #region VR.ATO.GEN.500143

            /*  VR.ATO.GEN.500143
            Given Name must contain at least one alphabetical character
    
            Legacy Rule Format:
            WHERE IN TUPLE(personstructuredname3.xx.xx:PersonNameDetails)
            IF pyde.xx.xx:PersonNameDetails.GivenName.Text DOES NOT CONTAIN SET("A-Z","a-z")
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (^PersonNameDetails <> NULL) AND (NotContainsSet(^FITR105, '"A-Z","a-z"'))
    
            Data Elements:
    
            ^FITR105 = RP:PersonNameDetails:PersonNameDetails.GivenName.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
    
            ^PersonNameDetails = RP:PersonNameDetails
            */
            assertion = (report.PersonNameDetailsCollection_LGL_Contact_C_RPExists != false && !(IsMatch(report.FITR105, @"[A-Z]|[a-z]", RegexOptions.IgnoreCase)));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.500143",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Given Name must contain at least one alphabetical character.",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_RPOccurrenceIndex) + "/tns:PersonNameDetails.GivenName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.500143" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR105", Value = report.FITR105 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.500143

            #region VR.ATO.GEN.500145

            /*  VR.ATO.GEN.500145
            Account number must contain the following: A to Z a to z or 0 to 9
    
            Legacy Rule Format:
            IF pyid.xx.xx:FinancialInstitutionAccount.FinancialInstitutionAccount.Number <> NULLORBLANK AND ANY CHARACTER OF pyid.xx.xx:FinancialInstitutionAccount.FinancialInstitutionAccount.Number <> SET("a-z","A-Z","0-9")
                RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            ^FITR177 <> BLANK AND NotCharacterInSet(^FITR177, '"a-z","A-Z","0-9"')
    
            Data Elements:
    
            ^FITR177 = RP:FinancialInstitutionAccount:FinancialInstitutionAccount.FinancialInstitutionAccount.Number
            */
            assertion = (string.IsNullOrWhiteSpace(report.FITR177) != true && !(IsMatch(report.FITR177, @"^[a-zA-Z0-9]*$", RegexOptions.IgnoreCase)));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.500145",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Account number contains invalid text",
                    LongDescription = @"Account number must contain the following: A to Z a to z or 0 to 9",
                    Location = "/xbrli:xbrl/tns:FinancialInstitutionAccount/tns:FinancialInstitutionAccount.FinancialInstitutionAccount.Number",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.500145" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR177", Value = report.FITR177 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.500145

            #region VR.ATO.GEN.500151

            /*  VR.ATO.GEN.500151
            First name cannot contain suffix ESQ, JNR, JP, MHA, MHR, MLA, MLC, MP, QC or SNR
    
            Legacy Rule Format:
            IF (pyde.xx.xx:PersonNameDetails.GivenName.Text = FOUND("ESQ","JNR","JP","MHA","MHR","MLA","MLC","MP","QC","SNR"))
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (FoundSet(^FITR114, '"ESQ","JNR","JP","MHA","MHR","MLA","MLC","MP","QC","SNR"'))
    
            Data Elements:
    
            ^FITR114 = INT:PersonNameDetails:PersonNameDetails.GivenName.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
            */
            assertion = IsMatch(report.FITR114, @"(^(ESQ|JNR|JP|MHA|MHR|MLA|MLC|MP|QC|SNR)$)|(^(.*)( ESQ | JNR | JP | MHA | MHR | MLA | MLC | MP | QC | SNR )(.*)$)|(^(ESQ |JNR |JP |MHA |MHR |MLA |MLC |MP |QC |SNR )(.*)$)|(^(.*)( ESQ| JNR| JP| MHA| MHR| MLA| MLC| MP| QC| SNR)$)", RegexOptions.IgnoreCase);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.010436",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"First name cannot contain suffix ESQ, JNR, JP, MHA, MHR, MLA, MLC, MP, QC or SNR",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_INTOccurrenceIndex) + "/tns:PersonNameDetails.GivenName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.500151" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR114", Value = report.FITR114 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.500151

            #region VR.ATO.GEN.500151

            /*  VR.ATO.GEN.500151
            First name cannot contain suffix ESQ, JNR, JP, MHA, MHR, MLA, MLC, MP, QC or SNR
    
            Legacy Rule Format:
            IF (pyde.xx.xx:PersonNameDetails.GivenName.Text = FOUND("ESQ","JNR","JP","MHA","MHR","MLA","MLC","MP","QC","SNR"))
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (FoundSet(^FITR105, '"ESQ","JNR","JP","MHA","MHR","MLA","MLC","MP","QC","SNR"'))
    
            Data Elements:
    
            ^FITR105 = RP:PersonNameDetails:PersonNameDetails.GivenName.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
            */
            assertion = IsMatch(report.FITR105, @"(^(ESQ|JNR|JP|MHA|MHR|MLA|MLC|MP|QC|SNR)$)|(^(.*)( ESQ | JNR | JP | MHA | MHR | MLA | MLC | MP | QC | SNR )(.*)$)|(^(ESQ |JNR |JP |MHA |MHR |MLA |MLC |MP |QC |SNR )(.*)$)|(^(.*)( ESQ| JNR| JP| MHA| MHR| MLA| MLC| MP| QC| SNR)$)", RegexOptions.IgnoreCase);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.010436",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"First name cannot contain suffix ESQ, JNR, JP, MHA, MHR, MLA, MLC, MP, QC or SNR",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_RPOccurrenceIndex) + "/tns:PersonNameDetails.GivenName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.500151" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR105", Value = report.FITR105 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.500151

            #region VR.ATO.GEN.500152

            /*  VR.ATO.GEN.500152
            Other given name cannot contain suffix ESQ, JNR, JP, MHA, MHR, MLA, MLC, MP, QC or SNR
    
            Legacy Rule Format:
            IF (pyde.xx.xx:PersonNameDetails.OtherGivenName.Text = FOUND("ESQ","JNR","JP","MHA","MHR","MLA","MLC","MP","QC","SNR"))
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (FoundSet(^FITR115, '"ESQ","JNR","JP","MHA","MHR","MLA","MLC","MP","QC","SNR"'))
    
            Data Elements:
    
            ^FITR115 = INT:PersonNameDetails:PersonNameDetails.OtherGivenName.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
            */
            assertion = IsMatch(report.FITR115, @"(^(ESQ|JNR|JP|MHA|MHR|MLA|MLC|MP|QC|SNR)$)|(^(.*)( ESQ | JNR | JP | MHA | MHR | MLA | MLC | MP | QC | SNR )(.*)$)|(^(ESQ |JNR |JP |MHA |MHR |MLA |MLC |MP |QC |SNR )(.*)$)|(^(.*)( ESQ| JNR| JP| MHA| MHR| MLA| MLC| MP| QC| SNR)$)", RegexOptions.IgnoreCase);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.010448",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Other given name cannot contain suffix ESQ, JNR, JP, MHA, MHR, MLA, MLC, MP, QC or SNR",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_INTOccurrenceIndex) + "/tns:PersonNameDetails.OtherGivenName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.500152" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR115", Value = GetValueOrEmpty(report.FITR115) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.500152

            #region VR.ATO.GEN.500152

            /*  VR.ATO.GEN.500152
            Other given name cannot contain suffix ESQ, JNR, JP, MHA, MHR, MLA, MLC, MP, QC or SNR
    
            Legacy Rule Format:
            IF (pyde.xx.xx:PersonNameDetails.OtherGivenName.Text = FOUND("ESQ","JNR","JP","MHA","MHR","MLA","MLC","MP","QC","SNR"))
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (FoundSet(^FITR106, '"ESQ","JNR","JP","MHA","MHR","MLA","MLC","MP","QC","SNR"'))
    
            Data Elements:
    
            ^FITR106 = RP:PersonNameDetails:PersonNameDetails.OtherGivenName.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
            */
            assertion = IsMatch(report.FITR106, @"(^(ESQ|JNR|JP|MHA|MHR|MLA|MLC|MP|QC|SNR)$)|(^(.*)( ESQ | JNR | JP | MHA | MHR | MLA | MLC | MP | QC | SNR )(.*)$)|(^(ESQ |JNR |JP |MHA |MHR |MLA |MLC |MP |QC |SNR )(.*)$)|(^(.*)( ESQ| JNR| JP| MHA| MHR| MLA| MLC| MP| QC| SNR)$)", RegexOptions.IgnoreCase);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.010448",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Other given name cannot contain suffix ESQ, JNR, JP, MHA, MHR, MLA, MLC, MP, QC or SNR",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_RPOccurrenceIndex) + "/tns:PersonNameDetails.OtherGivenName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.500152" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR106", Value = GetValueOrEmpty(report.FITR106) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.500152

            #region VR.ATO.GEN.500153

            /*  VR.ATO.GEN.500153
            Family name cannot contain suffix ESQ, JNR, JP, MHA, MHR, MLA, MLC, MP, QC or SNR
    
            Legacy Rule Format:
            IF (pyde.xx.xx:PersonNameDetails.FamilyName.Text = FOUND("ESQ","JNR","JP","MHA","MHR","MLA","MLC","MP","QC","SNR"))
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (FoundSet(^FITR112, '"ESQ","JNR","JP","MHA","MHR","MLA","MLC","MP","QC","SNR"'))
    
            Data Elements:
    
            ^FITR112 = INT:PersonNameDetails:PersonNameDetails.FamilyName.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
            */
            assertion = IsMatch(report.FITR112, @"(^(ESQ|JNR|JP|MHA|MHR|MLA|MLC|MP|QC|SNR)$)|(^(.*)( ESQ | JNR | JP | MHA | MHR | MLA | MLC | MP | QC | SNR )(.*)$)|(^(ESQ |JNR |JP |MHA |MHR |MLA |MLC |MP |QC |SNR )(.*)$)|(^(.*)( ESQ| JNR| JP| MHA| MHR| MLA| MLC| MP| QC| SNR)$)", RegexOptions.IgnoreCase);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.500153",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Family name cannot contain suffix ESQ, JNR, JP, MHA, MHR, MLA, MLC, MP, QC or SNR",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_INTOccurrenceIndex) + "/tns:PersonNameDetails.FamilyName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.500153" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR112", Value = report.FITR112 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.500153

            #region VR.ATO.GEN.500153

            /*  VR.ATO.GEN.500153
            Family name cannot contain suffix ESQ, JNR, JP, MHA, MHR, MLA, MLC, MP, QC or SNR
    
            Legacy Rule Format:
            IF (pyde.xx.xx:PersonNameDetails.FamilyName.Text = FOUND("ESQ","JNR","JP","MHA","MHR","MLA","MLC","MP","QC","SNR"))
               RETURN VALIDATION MESSAGE
            ENDIF

            Technical Business Rule Format:
            (FoundSet(^FITR103, '"ESQ","JNR","JP","MHA","MHR","MLA","MLC","MP","QC","SNR"'))
    
            Data Elements:
    
            ^FITR103 = RP:PersonNameDetails:PersonNameDetails.FamilyName.Text WHERE [PersonNameDetails.PersonNameType.Code='LGL'].[PersonNameDetails.Usage.Code='Contact'].[PersonNameDetails.Currency.Code='C']
            */
            assertion = IsMatch(report.FITR103, @"(^(ESQ|JNR|JP|MHA|MHR|MLA|MLC|MP|QC|SNR)$)|(^(.*)( ESQ | JNR | JP | MHA | MHR | MLA | MLC | MP | QC | SNR )(.*)$)|(^(ESQ |JNR |JP |MHA |MHR |MLA |MLC |MP |QC |SNR )(.*)$)|(^(.*)( ESQ| JNR| JP| MHA| MHR| MLA| MLC| MP| QC| SNR)$)", RegexOptions.IgnoreCase);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.GEN.500153",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Family name cannot contain suffix ESQ, JNR, JP, MHA, MHR, MLA, MLC, MP, QC or SNR",
                    Location = "/xbrli:xbrl/tns:PersonNameDetails" + OccurrenceIndex(report.PersonNameDetailsCollection_LGL_Contact_C_RPOccurrenceIndex) + "/tns:PersonNameDetails.FamilyName.Text",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.500153" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "FITR103", Value = report.FITR103 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.GEN.500153

            #region Repeating report.TransfereeCollection
            if (report.TransfereeCollection != null)
            {
                int itemIndex1 = 0;
                foreach (FITR2021.Transferee transferee in report.TransfereeCollection)
                {
                    itemIndex1++;

                    #region VR.ATO.FITR.434101

                    /*  VR.ATO.FITR.434101
                    Transferee Name, ABN, Amount and Market value must be present
    
                    Legacy Rule Format:
                    WHERE IN TUPLE(fitr.0008.lodge.req.xx.xx:Transferee)
                    IF [FITR85] = NULLORBLANK OR [FITR87] = 0 OR [FITR87] = NULL OR [FITR86] = NULLORBLANK OR [FITR88] = NULL 
                        RETURN VALIDATION MESSAGE
                    ENDIF
        
                    Technical Business Rule Format:
                    ^FITR85 = BLANK OR ^FITR87 = 0 OR ^FITR87 = NULL OR ^FITR86 = BLANK OR ^FITR88 = NULL
            
                    Data Elements:
            
                    ^FITR85 = RP:Transferee:OrganisationNameDetails:OrganisationNameDetails.OrganisationalName.Text WHERE [OrganisationNameDetails.OrganisationalNameType.Code='MN'].[OrganisationNameDetails.Currency.Code='C']
            
                    ^FITR87 = RP:Transferee:IncomeTax.Superannuation.TransferredAssessable.Amount
            
                    ^FITR86 = RP:Transferee:Identifiers.AustralianBusinessNumber.Identifier
            
                    ^FITR88 = RP:Transferee:IncomeTax.Superannuation.MarketValueTransfer.Amount
                    */
                    assertion = (string.IsNullOrWhiteSpace(transferee.FITR85) == true || transferee.FITR87 == 0 || transferee.FITR87 == null || string.IsNullOrWhiteSpace(transferee.FITR86) == true || transferee.FITR88 == null);
                    if (assertion)
                    {
                        processMessage = new ProcessMessageDocument()
                        {
                            Code = "CMN.ATO.FITR.434101",
                            Severity = ProcessMessageSeverity.Error,
                            Description = @"Transferee Name, ABN, Amount and Market value must be present",
                            LongDescription = @"If the fund or trust has transferred assessable taxable contributions to a life insurance company, registered organisation or pooled superannuation trust, then the details of each transferee must be present. This includes the Name, ABN, Amount and Market value of each transferee",
                            Location = "/xbrli:xbrl/tns:Transferee" + OccurrenceIndex(transferee.OccurrenceIndex) + "/tns:OrganisationNameDetails" + OccurrenceIndex(transferee.Transferee_OrganisationNameDetailsCollection_MN_C_RPOccurrenceIndex) + "/tns:OrganisationNameDetails.OrganisationalName.Text",
                            Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434101" } },
                        };
                        processMessage.Parameters.Add(new ProcessMessageParameter
                        { Name = "FITR85", Value = transferee.FITR85 });

                        processMessage.Parameters.Add(new ProcessMessageParameter
                        { Name = "FITR87", Value = GetValueOrEmpty(transferee.FITR87) });

                        processMessage.Parameters.Add(new ProcessMessageParameter
                        { Name = "FITR86", Value = transferee.FITR86 });

                        processMessage.Parameters.Add(new ProcessMessageParameter
                        { Name = "FITR88", Value = GetValueOrEmpty(transferee.FITR88) });

                        response.Add(processMessage);
                    }
                    #endregion // VR.ATO.FITR.434101

                    #region VR.ATO.FITR.434178

                    /*  VR.ATO.FITR.434178
                    Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
                    Legacy Rule Format:
                    IF [FITR87] <> NULL AND [FITR87] <> MONETARY(U,11,0)
                       RETURN VALIDATION MESSAGE
                    ENDIF
        
                    Technical Business Rule Format:
                    NotMonetary(^FITR87, 'U', 11, 0)
            
                    Data Elements:
            
                    ^FITR87 = RP:Transferee:IncomeTax.Superannuation.TransferredAssessable.Amount
                    */
                    assertion = NotMonetary(transferee.FITR87, @"U", 11, 0);
                    if (assertion)
                    {
                        processMessage = new ProcessMessageDocument()
                        {
                            Code = "CMN.ATO.GEN.400011",
                            Severity = ProcessMessageSeverity.Error,
                            Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                            LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                            Location = "/xbrli:xbrl/tns:Transferee" + OccurrenceIndex(transferee.OccurrenceIndex) + "/tns:IncomeTax.Superannuation.TransferredAssessable.Amount",
                            Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434178" } },
                        };
                        processMessage.Parameters.Add(new ProcessMessageParameter
                        { Name = "FITR87", Value = GetValueOrEmpty(transferee.FITR87) });

                        response.Add(processMessage);
                    }
                    #endregion // VR.ATO.FITR.434178

                    #region VR.ATO.FITR.434179

                    /*  VR.ATO.FITR.434179
                    Monetary field - max 11 characters, no sign, no decimal, no illegal characters
    
                    Legacy Rule Format:
                    IF [FITR88] <> NULL AND [FITR88] <> MONETARY(U,11,0)
                       RETURN VALIDATION MESSAGE
                    ENDIF
        
                    Technical Business Rule Format:
                    NotMonetary(^FITR88, 'U', 11, 0)
            
                    Data Elements:
            
                    ^FITR88 = RP:Transferee:IncomeTax.Superannuation.MarketValueTransfer.Amount
                    */
                    assertion = NotMonetary(transferee.FITR88, @"U", 11, 0);
                    if (assertion)
                    {
                        processMessage = new ProcessMessageDocument()
                        {
                            Code = "CMN.ATO.GEN.400011",
                            Severity = ProcessMessageSeverity.Error,
                            Description = @"Field incorrect format. Exceeds maximum digits or decimals and/or contains illegal characters",
                            LongDescription = @"Field is in incorrect format for a monetary field. Field either exceeds maximum digits or decimals and/or contains illegal characters and/or for a signed field does not have the sign as the first character",
                            Location = "/xbrli:xbrl/tns:Transferee" + OccurrenceIndex(transferee.OccurrenceIndex) + "/tns:IncomeTax.Superannuation.MarketValueTransfer.Amount",
                            Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434179" } },
                        };
                        processMessage.Parameters.Add(new ProcessMessageParameter
                        { Name = "FITR88", Value = GetValueOrEmpty(transferee.FITR88) });

                        response.Add(processMessage);
                    }
                    #endregion // VR.ATO.FITR.434179

                    #region VR.ATO.GEN.410038

                    /*  VR.ATO.GEN.410038
                    Non-individual name field cannot contain 'Pship', 'P'ship or 'P/Ship'
    
                    Legacy Rule Format:
                    IF (pyde.xx.xx:OrganisationNameDetails.OrganisationalName.Text STARTSWITH "T/A ") AND (pyde.xx.xx:OrganisationNameDetails.OrganisationalName.Text ENDSWITH SET(" Pship"," P'ship"," P/ship"))
                      RETURN VALIDATION MESSAGE
                    ENDIF
        
                    Technical Business Rule Format:
                    (StartsWith(^FITR85, 'T/A ')) AND (EndsWithSet(^FITR85, '" Pship"," P'ship"," P/ship"'))
            
                    Data Elements:
            
                    ^FITR85 = RP:Transferee:OrganisationNameDetails:OrganisationNameDetails.OrganisationalName.Text WHERE [OrganisationNameDetails.OrganisationalNameType.Code='MN'].[OrganisationNameDetails.Currency.Code='C']
                    */
                    assertion = (IsMatch(transferee.FITR85, @"^(T/A )", RegexOptions.IgnoreCase) && IsMatch(transferee.FITR85, @"( Pship| P'ship| P/ship)$", RegexOptions.IgnoreCase));
                    if (assertion)
                    {
                        processMessage = new ProcessMessageDocument()
                        {
                            Code = "CMN.ATO.GEN.410038",
                            Severity = ProcessMessageSeverity.Error,
                            Description = @"Non-individual name field cannot start with 'T/A' and end with 'Pship', 'P'ship or 'P/Ship'",
                            Location = "/xbrli:xbrl/tns:Transferee" + OccurrenceIndex(transferee.OccurrenceIndex) + "/tns:OrganisationNameDetails" + OccurrenceIndex(transferee.Transferee_OrganisationNameDetailsCollection_MN_C_RPOccurrenceIndex) + "/tns:OrganisationNameDetails.OrganisationalName.Text",
                            Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.410038" } },
                        };
                        processMessage.Parameters.Add(new ProcessMessageParameter
                        { Name = "FITR85", Value = transferee.FITR85 });

                        response.Add(processMessage);
                    }
                    #endregion // VR.ATO.GEN.410038

                    #region VR.ATO.GEN.410039

                    /*  VR.ATO.GEN.410039
                    Organisation name must contain at least one alpha or numeric character
    
                    Legacy Rule Format:
                    IF (pyde.xx.xx:OrganisationNameDetails.OrganisationalName.Text  <> NULLORBLANK) AND (pyde.xx.xx:OrganisationNameDetails.OrganisationalName.Text  DOES NOT CONTAIN SET("a-z", "A-Z", "0-9"))
                       RETURN VALIDATION MESSAGE
                    ENDIF
        
                    Technical Business Rule Format:
                    (^FITR85  <> BLANK) AND (NotContainsSet(^FITR85, '"a-z", "A-Z", "0-9"'))
            
                    Data Elements:
            
                    ^FITR85 = RP:Transferee:OrganisationNameDetails:OrganisationNameDetails.OrganisationalName.Text WHERE [OrganisationNameDetails.OrganisationalNameType.Code='MN'].[OrganisationNameDetails.Currency.Code='C']
                    */
                    assertion = (string.IsNullOrWhiteSpace(transferee.FITR85) != true && !(IsMatch(transferee.FITR85, @"[a-z]|[A-Z]|[0-9]", RegexOptions.IgnoreCase)));
                    if (assertion)
                    {
                        processMessage = new ProcessMessageDocument()
                        {
                            Code = "CMN.ATO.GEN.410039",
                            Severity = ProcessMessageSeverity.Error,
                            Description = @"Organisation name must contain at least one alpha or numeric character",
                            Location = "/xbrli:xbrl/tns:Transferee" + OccurrenceIndex(transferee.OccurrenceIndex) + "/tns:OrganisationNameDetails" + OccurrenceIndex(transferee.Transferee_OrganisationNameDetailsCollection_MN_C_RPOccurrenceIndex) + "/tns:OrganisationNameDetails.OrganisationalName.Text",
                            Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.410039" } },
                        };
                        processMessage.Parameters.Add(new ProcessMessageParameter
                        { Name = "FITR85", Value = transferee.FITR85 });

                        response.Add(processMessage);
                    }
                    #endregion // VR.ATO.GEN.410039

                    #region VR.ATO.GEN.428258

                    /*  VR.ATO.GEN.428258
                    Non Individual Name cannot contain space hyphen space combinations
    
                    Legacy Rule Format:
                    IF (pyde.xx.xx:OrganisationNameDetails.OrganisationalName.Text CONTAINS  " - " )
                       RETURN VALIDATION MESSAGE
                    ENDIF
        
                    Technical Business Rule Format:
                    Contains(^FITR85, ' - ')
            
                    Data Elements:
            
                    ^FITR85 = RP:Transferee:OrganisationNameDetails:OrganisationNameDetails.OrganisationalName.Text WHERE [OrganisationNameDetails.OrganisationalNameType.Code='MN'].[OrganisationNameDetails.Currency.Code='C']
                    */
                    assertion = (transferee.FITR85 == null ? false : transferee.FITR85.ToUpperInvariant().Contains(@" - "));
                    if (assertion)
                    {
                        processMessage = new ProcessMessageDocument()
                        {
                            Code = "CMN.ATO.GEN.000413",
                            Severity = ProcessMessageSeverity.Error,
                            Description = @"Non Individual Name cannot contain space hyphen space combinations",
                            Location = "/xbrli:xbrl/tns:Transferee" + OccurrenceIndex(transferee.OccurrenceIndex) + "/tns:OrganisationNameDetails" + OccurrenceIndex(transferee.Transferee_OrganisationNameDetailsCollection_MN_C_RPOccurrenceIndex) + "/tns:OrganisationNameDetails.OrganisationalName.Text",
                            Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.428258" } },
                        };
                        processMessage.Parameters.Add(new ProcessMessageParameter
                        { Name = "FITR85", Value = transferee.FITR85 });

                        response.Add(processMessage);
                    }
                    #endregion // VR.ATO.GEN.428258

                    #region VR.ATO.GEN.428259

                    /*  VR.ATO.GEN.428259
                    Non Individual Name cannot contain "P/L"
    
                    Legacy Rule Format:
                    IF (pyde.xx.xx:OrganisationNameDetails.OrganisationalName.Text CONTAINS "P/L")
                       RETURN VALIDATION MESSAGE
                    ENDIF
        
                    Technical Business Rule Format:
                    Contains(^FITR85, 'P/L')
            
                    Data Elements:
            
                    ^FITR85 = RP:Transferee:OrganisationNameDetails:OrganisationNameDetails.OrganisationalName.Text WHERE [OrganisationNameDetails.OrganisationalNameType.Code='MN'].[OrganisationNameDetails.Currency.Code='C']
                    */
                    assertion = (transferee.FITR85 == null ? false : transferee.FITR85.ToUpperInvariant().Contains(@"P/L"));
                    if (assertion)
                    {
                        processMessage = new ProcessMessageDocument()
                        {
                            Code = "CMN.ATO.GEN.000414",
                            Severity = ProcessMessageSeverity.Error,
                            Description = @"Non Individual Name cannot contain ""P/L""",
                            Location = "/xbrli:xbrl/tns:Transferee" + OccurrenceIndex(transferee.OccurrenceIndex) + "/tns:OrganisationNameDetails" + OccurrenceIndex(transferee.Transferee_OrganisationNameDetailsCollection_MN_C_RPOccurrenceIndex) + "/tns:OrganisationNameDetails.OrganisationalName.Text",
                            Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.428259" } },
                        };
                        processMessage.Parameters.Add(new ProcessMessageParameter
                        { Name = "FITR85", Value = transferee.FITR85 });

                        response.Add(processMessage);
                    }
                    #endregion // VR.ATO.GEN.428259

                    #region VR.ATO.GEN.428260

                    /*  VR.ATO.GEN.428260
                    Non Individual Name cannot contain repeated hyphen, apostrophe, or space characters
    
                    Legacy Rule Format:
                    IF (pyde.xx.xx:OrganisationNameDetails.OrganisationalName.Text CONTAINS SET("--","''","  ") )
                       RETURN VALIDATION MESSAGE
                    ENDIF
        
                    Technical Business Rule Format:
                    (ContainsSet(^FITR85, '"--","''","  "'))
            
                    Data Elements:
            
                    ^FITR85 = RP:Transferee:OrganisationNameDetails:OrganisationNameDetails.OrganisationalName.Text WHERE [OrganisationNameDetails.OrganisationalNameType.Code='MN'].[OrganisationNameDetails.Currency.Code='C']
                    */
                    assertion = IsMatch(transferee.FITR85, @"--|''|  ", RegexOptions.IgnoreCase);
                    if (assertion)
                    {
                        processMessage = new ProcessMessageDocument()
                        {
                            Code = "CMN.ATO.GEN.000416",
                            Severity = ProcessMessageSeverity.Error,
                            Description = @"Non Individual Name cannot contain repeated hyphen, apostrophe, or space characters",
                            Location = "/xbrli:xbrl/tns:Transferee" + OccurrenceIndex(transferee.OccurrenceIndex) + "/tns:OrganisationNameDetails" + OccurrenceIndex(transferee.Transferee_OrganisationNameDetailsCollection_MN_C_RPOccurrenceIndex) + "/tns:OrganisationNameDetails.OrganisationalName.Text",
                            Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.428260" } },
                        };
                        processMessage.Parameters.Add(new ProcessMessageParameter
                        { Name = "FITR85", Value = transferee.FITR85 });

                        response.Add(processMessage);
                    }
                    #endregion // VR.ATO.GEN.428260

                    #region VR.ATO.GEN.428261

                    /*  VR.ATO.GEN.428261
                    Non Individual Name cannot end with "T/A", "T/A P'ship", "T/A Pship", "T/A P/Ship" or "T/A Partnership"
    
                    Legacy Rule Format:
                    IF (pyde.xx.xx:OrganisationNameDetails.OrganisationalName.Text ENDSWITH SET(" T/A"," T/A P'ship"," T/A Pship"," T/A P/Ship"," T/A Partnership"))
                      RETURN VALIDATION MESSAGE
                    ENDIF
        
                    Technical Business Rule Format:
                    (EndsWithSet(^FITR85, '" T/A"," T/A P'ship"," T/A Pship"," T/A P/Ship"," T/A Partnership"'))
            
                    Data Elements:
            
                    ^FITR85 = RP:Transferee:OrganisationNameDetails:OrganisationNameDetails.OrganisationalName.Text WHERE [OrganisationNameDetails.OrganisationalNameType.Code='MN'].[OrganisationNameDetails.Currency.Code='C']
                    */
                    assertion = IsMatch(transferee.FITR85, @"( T/A| T/A P'ship| T/A Pship| T/A P/Ship| T/A Partnership)$", RegexOptions.IgnoreCase);
                    if (assertion)
                    {
                        processMessage = new ProcessMessageDocument()
                        {
                            Code = "CMN.ATO.GEN.000417",
                            Severity = ProcessMessageSeverity.Error,
                            Description = @"Non Individual Name cannot end with ""T/A"", ""T/A P'ship"", ""T/A Pship"", ""T/A P/Ship"" or ""T/A Partnership""",
                            Location = "/xbrli:xbrl/tns:Transferee" + OccurrenceIndex(transferee.OccurrenceIndex) + "/tns:OrganisationNameDetails" + OccurrenceIndex(transferee.Transferee_OrganisationNameDetailsCollection_MN_C_RPOccurrenceIndex) + "/tns:OrganisationNameDetails.OrganisationalName.Text",
                            Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.428261" } },
                        };
                        processMessage.Parameters.Add(new ProcessMessageParameter
                        { Name = "FITR85", Value = transferee.FITR85 });

                        response.Add(processMessage);
                    }
                    #endregion // VR.ATO.GEN.428261

                    #region VR.ATO.GEN.434223

                    /*  VR.ATO.GEN.434223
                    ABN must pass the ABN algorithm
    
                    Legacy Rule Format:
                    IF (pyid.xx.xx:Identifiers.AustralianBusinessNumber.Identifier <> NULLORBLANK) and (ABNALGORITHM(pyid.xx.xx:Identifiers.AustralianBusinessNumber.Identifier) = FALSE)
                    RETURN VALIDATION MESSAGE
                    ENDIF
        
                    Technical Business Rule Format:
                    (^FITR86 <> BLANK) and (FailsABNAlgorithm(^FITR86))
            
                    Data Elements:
            
                    ^FITR86 = RP:Transferee:Identifiers.AustralianBusinessNumber.Identifier
                    */
                    assertion = (string.IsNullOrWhiteSpace(transferee.FITR86) != true && FailsABNAlgorithm(transferee.FITR86));
                    if (assertion)
                    {
                        processMessage = new ProcessMessageDocument()
                        {
                            Code = "CMN.ATO.GEN.000477",
                            Severity = ProcessMessageSeverity.Error,
                            Description = @"ABN is invalid",
                            Location = "/xbrli:xbrl/tns:Transferee" + OccurrenceIndex(transferee.OccurrenceIndex) + "/tns:Identifiers.AustralianBusinessNumber.Identifier",
                            Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.434223" } },
                        };
                        processMessage.Parameters.Add(new ProcessMessageParameter
                        { Name = "FITR86", Value = transferee.FITR86 });

                        response.Add(processMessage);
                    }
                    #endregion // VR.ATO.GEN.434223

                    #region VR.ATO.GEN.500132

                    /*  VR.ATO.GEN.500132
                    Organisation name must be present
    
                    Legacy Rule Format:
                    WHERE IN TUPLE(organisationname2.xx.xx:OrganisationNameDetails)
                    IF pyde.xx.xx:OrganisationNameDetails.OrganisationalName.Text = BLANK
                       RETURN VALIDATION MESSAGE
                    ENDIF
        
                    Technical Business Rule Format:
                    (^OrganisationNameDetails <> NULL) AND (^FITR85 = BLANK)
            
                    Data Elements:
            
                    ^FITR85 = RP:Transferee:OrganisationNameDetails:OrganisationNameDetails.OrganisationalName.Text WHERE [OrganisationNameDetails.OrganisationalNameType.Code='MN'].[OrganisationNameDetails.Currency.Code='C']
            
                    ^OrganisationNameDetails = RP:Transferee:OrganisationNameDetails
                    */
                    assertion = (transferee.Transferee_OrganisationNameDetailsCollection_MN_C_RPExists != false && string.IsNullOrWhiteSpace(transferee.FITR85) == true);
                    if (assertion)
                    {
                        processMessage = new ProcessMessageDocument()
                        {
                            Code = "CMN.ATO.GEN.500132",
                            Severity = ProcessMessageSeverity.Error,
                            Description = @"Organisation name must be present.",
                            Location = "/xbrli:xbrl/tns:Transferee" + OccurrenceIndex(transferee.OccurrenceIndex) + "/tns:OrganisationNameDetails" + OccurrenceIndex(transferee.Transferee_OrganisationNameDetailsCollection_MN_C_RPOccurrenceIndex) + "/tns:OrganisationNameDetails.OrganisationalName.Text",
                            Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.500132" } },
                        };
                        processMessage.Parameters.Add(new ProcessMessageParameter
                        { Name = "FITR85", Value = transferee.FITR85 });

                        response.Add(processMessage);
                    }
                    #endregion // VR.ATO.GEN.500132
                }
            }

            #endregion // Foreach loop

            #region Repeating FITR204Collection
            if (report.FITR204Collection != null)
            {
                int itemIndex1 = 0;
                foreach (string FITR204 in report.FITR204Collection)
                {
                    itemIndex1++;

                    #region VR.ATO.FITR.434288

                    /*  VR.ATO.FITR.434288
                    Invalid exemption or rollover code
    
                    Legacy Rule Format:
                    IF [FITR204] <> NULL AND [FITR204] <> SET (DOMAIN(FITRCGTExemptionOrRolloverCodes))
                      RETURN VALIDATION MESSAGE
                    ENDIF
        
                    Technical Business Rule Format:
                    ^FITR204 <> NULL AND NotInDomain(^FITR204, 'Fitrcgtexemptionorrollovercodes')
            
                    Data Elements:
            
                    ^FITR204 = RP:CapitalGainsTax.ExemptionOrRolloverApplied.Code
                    */
                    assertion = (FITR204 != null && !(IsMatch(FITR204, @"^(A - Small business active asset reduction|B - Small business retirement exemption|C - Small business roll-over|D - Small business 15 year exemption|F - Scrip for scrip roll-over|L - Replacement asset roll-overs|M - Exchange of shares or units|N - Exchange of rights or options|O - Exchange of shares in one company for shares in another company|P - Exchange of units in a unit trust for shares in a company|Q - Trust restructure roll-over|S - Same asset roll-overs|U - Early stage investor|V - Venture capital investment|X - Other exemptions and roll-overs)$", RegexOptions.IgnoreCase)));
                    if (assertion)
                    {
                        processMessage = new ProcessMessageDocument()
                        {
                            Code = "CMN.ATO.FITR.434288",
                            Severity = ProcessMessageSeverity.Error,
                            Description = @"Invalid exemption or rollover code",
                            Location = "/xbrli:xbrl/tns:CapitalGainsTax.ExemptionOrRolloverApplied.Code[" + itemIndex1 + "]",
                            Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.FITR.434288" } },
                        };
                        processMessage.Parameters.Add(new ProcessMessageParameter
                        { Name = "FITR204", Value = GetValueOrEmpty(FITR204) });

                        response.Add(processMessage);
                    }
                    #endregion // VR.ATO.FITR.434288
                }
            }

            #endregion // Foreach loop

            foreach (ProcessMessageDocument currentProcessMessage in response)
            {
                if (currentProcessMessage.Parameters != null)
                {
                    foreach (ProcessMessageParameter currentParameter in currentProcessMessage.Parameters)
                    {
                        if (string.IsNullOrEmpty(currentParameter.Name))
                        {
                            currentParameter.Name = _emptyParameterValue;
                        }

                        if (currentParameter.Name.Length > _maxParameterNameLength)
                        {
                            currentParameter.Name = currentParameter.Name.Substring(0, _maxParameterNameLength - 1);
                        }

                        if (string.IsNullOrEmpty(currentParameter.Value))
                        {
                            currentParameter.Value = _emptyParameterValue;
                        }

                        if (currentParameter.Value.Length > _maxParameterValueLength)
                        {
                            currentParameter.Value = currentParameter.Value.Substring(0, _maxParameterValueLength - 1);
                        }
                    }
                }
            }

            return response;
        }


    }
}
